#ifdef __GNUG__
#pragma implementation
#endif

#include "Alert.h"

#include "Class.h"
#include "Env.h"
#include "ImageItem.h"
#include "Buttons.h"
#include "CheapText.h"
#include "StaticTView.h"
#include "StyledText.h"
#include "OrdColl.h"
#include "ObjArray.h"
#include "Window.h"
#include "Expander.h"
#include "BorderItems.h"

extern char *gProgname;

static u_short NoteBits[]= {
#   include "images/note.image"
};
SmartBitmap gNote(Point(64, 64), NoteBits);

static u_short CautionBits[]= {
#   include "images/caution.image"
};
SmartBitmap gCaution(Point(64, 64), CautionBits);

static u_short StopBits[]= {
#   include "images/stop.image"
};
SmartBitmap gStop(Point(64, 64), StopBits);

static u_short ErrorBits[]= {
#   include "images/error.image"
};
SmartBitmap gError(Point(64, 64), ErrorBits);

//---- Alert -------------------------------------------------------------------

NewMetaImpl(Alert,Dialog, (TP(text), TP(textview), TP(image), TP(buttons), TP(alerts)));

ObjArray *Alert::alerts;

Alert::Alert(AlertType at, byte *message, Bitmap *va_(bm), ...) : Dialog(gProgname)
{
    va_list ap;
    OrdCollection *ol;
    va_start(ap,va_(bm));

    if (message) {
	text= new CheapText(message);
	textview= new StaticTextView((View*)0,
				    Rectangle(gSysFont->Width('n')*40, cFit), text);
    
	if (va_(bm))
	    image= new ImageItem(va_(bm), 0);
    
	if (at == eAlertMessage)
	    image= textview;
	else
	    image= new HBox(gPoint10, eVObjVTop, image, textview, 0);
    
	ol= new OrdCollection;
	char *s;
	for (int i= 0; s= va_arg(ap, char*); i++)
	    ol->Add(new ActionButton(va_arg(ap, int), s, i == 0));   
	buttons= new HBox(20, (VObjAlign)(eVObjVBase|eVObjHEqual|eVObjHExpand), ol);
    }
    va_end(ap);
}

Alert::~Alert()
{
    SafeDelete(text);
}

VObject *Alert::DoMakeContent()
{
    return new Matte(new VBox(20, eVObjHLeft, image, buttons, 0));
}

int Alert::Show(char* va_(fmt), ...)
{
    int code;
    va_list ap;
    
    va_start(ap,va_(fmt));
    code= ShowV(va_(fmt), ap);
    va_end(ap);
    return code;
}

int Alert::ShowV(char *fmt, va_list ap)
{
    return ShowV(gSysFont, fmt, ap);
}

int Alert::ShowV(Font *fp, char *fmt, va_list ap)
{
    char *buf= strvprintf(fmt, ap);
    SafeDelete(text);
    text= new StyledText(fp, buf);
    SafeDelete(buf);
    textview->SetText(text);
    textview->SetExtent(Point(textview->GetExtent().x, cFit));
    return Dialog::ShowUnderMouse();
}

void Alert::InspectorId(char *buf, int sz)
{
    if (textview)
	textview->InspectorId(buf, sz);
    else
	Dialog::InspectorId(buf, sz);   
}

static Alert *GetAlert(AlertType at, char *yes, char *no)
{
    Alert *al= 0;

    if (Alert::alerts == 0)
	Alert::alerts= new ObjArray(10);	// !!
	
    if ((al= (Alert*) Alert::alerts->At(at)) == 0) {
	switch (at) {
	case eAlertNote:
	    al= new Alert(at, (byte*) "note", gNote, "Ok", cIdOk, 0);
	    break;
	case eAlertCaution:
	    if (yes == 0 || *yes == '\0') 
	    	yes = "Yes";
	    if (no && *no == '\0') 
	    	no = "No";
	    if (no)
		al= new Alert(at, (byte*) "Caution", gCaution,
					    yes,    cIdYes,
					    no,     cIdNo,
					    "Cancel", cIdCancel, 0);
	    else
		al= new Alert(at, (byte*) "Caution", gCaution,
					    yes,    cIdYes,
					    "Cancel", cIdCancel, 0);
	    break;
	case eAlertQuestion:
	    if (yes == 0 || *yes == '\0') 
	    	yes = "Yes";
	    if (no == 0 || *no == '\0') 
	    	no = "No";
	    al= new Alert(at, (byte*) "Question", gCaution,
					    yes,	cIdYes,
					    no, 	cIdNo, 0);
	    break;
	case eAlertStop:
	    if (yes == 0 || *yes == '\0') 
	    	yes = "Yes";
	    if (no == 0 || *no == '\0') 
	    	no = "No";
	    al= new Alert(at, (byte*) "Stop", gStop,
					 yes,    cIdYes,
					 no,     cIdNo,
					 "Cancel", cIdCancel, 0);
	    break;
	case eAlertError:
	    if (Env::GetValue("Application.DebugButtonInAlert", FALSE)) 
		al= new Alert(at, (byte*) "Error", gError,
					  "Ignore",     cIdIgnore,
					  "Dump Core",  cIdAbort,
					  "Debug",      cIdInspect, 0);
	    else
		al= new Alert(at, (byte*) "Error", gError,
					  "Ignore",     cIdIgnore,
					  "Dump Core",  cIdAbort, 0);
	    break;
	case eAlertSun:
	case eAlertMessage:
	    al= new Alert(at, (byte*) "Message", 0, "Ok", cIdOk, 0);
	    break;
	}
	if (al)
	    Alert::alerts->AtPut(at, al);
    }

    return al;
}

static void ClearAlert(AlertType at)
{
    Alert *al= 0;

    if (Alert::alerts == 0)
	return;
	
    if ((al= (Alert*) Alert::alerts->At(at)) != 0) 
    	{
	Alert::alerts->RemoveAt(at);
    	delete al;
	}
}

int ShowAlert(AlertType at, char* va_(fmt), ...)
{
    int code= 0;
    va_list ap;
    Alert *al;
    
    va_start(ap, va_(fmt));
    al= GetAlert(at, "", "");    
    if (al)
	code= al->ShowV(va_(fmt), ap);
    va_end(ap);
    return code;
}

int ShowAlert(AlertType at, Font *fp, char* va_(fmt), ...)
{
    int code= 0;
    va_list ap;
    Alert *al;
    
    va_start(ap, va_(fmt));
    al= GetAlert(at, "", "");    
    if (al)
	code= al->ShowV(fp, va_(fmt), ap);
    va_end(ap);
    return code;
}

int ShowAlert(char *yes, char *no, AlertType at, char* va_(fmt), ...)
{
    int code= 0;
    va_list ap;
    Alert *al;
    
    va_start(ap, va_(fmt));
    // delete alert if it already exists 
    ClearAlert(at);    
    // create a new with new yes/no titels
    al= GetAlert(at, yes, no);    
    if (al)
	code= al->ShowV(va_(fmt), ap);
    va_end(ap);

    ClearAlert(at);    
    
    return code;
}
