#ifdef __GNUG__
#pragma implementation
#endif

#include "Application.h"

#include "Document.h"
#include "Class.h"
#include "Storage.h"
#include "FileDialog.h"
#include "ObjectTable.h"
#include "ClassManager.h"
#include "ClipBoard.h"
#include "ProgEnv.h"
#include "Buttons.h"
#include "StreamConnection.h"
#include "WindowSystem.h"
#include "Env.h"
#include "ProgressDialog.h"
#include "Data.h"
#include "Alert_e.h"
#include "Error.h"
#include "Menu.h"
#include "ETString.h"
#include "Expander.h"
#include "MenuBar.h"
#include "BorderItems.h"
#include "Box.h"
#include "ImageItem.h"
#include "Filler.h"
#include "CLib.h"
#include "System.h"		// BaseName() is in this header

Application         *gApplication;
ClipBoard           *gClipBoard;
int                 gArgc;
char                **gArgv;

//---- Handling virtual memory exhaustion --------------------------------------

class MemoryWarnAlert : public Dialog {
public:
    static MemoryWarnAlert *memory_warn_alert;
public:
    MemoryWarnAlert();
    VObject *DoMakeContent();

    static void Init ();
    static void MallocFailedHook ();
};

MemoryWarnAlert *MemoryWarnAlert::memory_warn_alert;

MemoryWarnAlert::MemoryWarnAlert () : Dialog ("Out of Memory")
{
}

void
MemoryWarnAlert::Init ()
{
    memory_warn_alert = new MemoryWarnAlert;
    memory_warn_alert->MakeWindows();
    CLib::SetMallocFailedHook (MemoryWarnAlert::MallocFailedHook);
}

void
MemoryWarnAlert::MallocFailedHook ()
{
    memory_warn_alert->ShowAt(NULL, gPoint_1);
}

VObject *
MemoryWarnAlert::DoMakeContent ()
{
    VObject *icon;
    Bitmap *bm = gApplication->GetIcon();
    if ( bm )
		icon = (VObject *)new ImageItem (bm);
	else
		icon = (VObject *)new Filler (gPoint16);
    return
	new Matte (
	    new VBox (gPoint16, VObjAlign(eVObjHCenter|eVObjHExpand|eVObjVExpand),
		new HBox (gPoint16, eVObjVTop,
		    icon,
		    new VBox (gPoint0, eVObjHLeft,
			new TextItem ("The system ran out of virtual memory.",
					gSysFont->WithFace(eFaceBold)),
			new Filler (gPoint8),
			new TextItem ("Please shutdown other processes"),
			new TextItem ("or add swap space before continuing."),
			0
		    ),
		    0
		),
		new HBox (gPoint8, eVObjVBase,
		    new ActionButton (cIdOk, "Continue"),
//		    new ActionButton (cIdCancel, "Quit"),
		    0
		),		
		0
	    )
	);
}

//---- AppRpcConnection --------------------------------------------------------

class AppRpcConnection : public StreamConnection {
public:
    MetaDef(AppRpcConnection);
    AppRpcConnection();
    ~AppRpcConnection();
    void Dispatch(int from, char *req, char *buf, int len, char *&ret, int &retlen);
};
 
inline AppRpcConnection::~AppRpcConnection()
	{ SendTo(0, "remove", gProgname); }
inline  void AppRpcConnection::Dispatch(int from, char *req, char *buf, int len, char *&ret, int &retlen)
	{ gApplication->ExtCommand(from, req, buf, len, ret, retlen);
	  gWindowSystem->Update(); }
	  
NewMetaImpl0(AppRpcConnection,StreamConnection);

AppRpcConnection::AppRpcConnection() : StreamConnection(-1)
{
    char *retbuf = 0;
    int retlen = -1;
    int to;

    Open(gSystem->OpenConnection(SERVERNAME, SERVICENAME));
    if (Reconnect() == TRUE) // error situation (probably no appcomm)
    	return;
    Talk(0, "enroll", gProgname);
    Talk(1, "gettimeout", "", 0, &retbuf, &retlen);
    if ( retbuf ) {
	sscanf (retbuf, "%d", &to);
	StreamConnection::SetTimeout(to);
	delete retbuf;		// memory leak detected by Purify
    }
}

//---- ApplIntHandler ----------------------------------------------------------

class ApplIntHandler : public SysEvtHandler {
public:
    MetaDef(ApplIntHandler);
    ApplIntHandler() : SysEvtHandler(eSigInterrupt)
	{ }
    void Notify(SysEventCodes, int);
};

void ApplIntHandler::Notify(SysEventCodes, int)
{ 
    if(gQuitApp)
	gSystem->Exit(-1,FALSE);
    else 
	gSystem->ExitControl(); 
}
	
NewMetaImpl0(ApplIntHandler,SysEvtHandler);

//---- ApplErrorHandler --------------------------------------------------------

static void ApplErrorHandler(int level, bool, char *location, char *msg)
{
    gApplication->DoOnError(level, location, msg);
} 

//---- Application -------------------------------------------------------------

NewAbstractMetaImpl(Application,Manager, (T(lastWindowPos), TS(version),
	T(mainDocumentType), TP(fileDialog), TP(gApplication), TP(gClipBoard),
	TVP(gArgv,gArgc), TP(gClassManager)));

FileDialog *Application::fileDialog;

Application::Application(int ac, char **av, const Symbol &dt)
{
    if (gApplication)
	Error("Application", "only one Application !!");

    gApplication= this;
    version= "Version 3.0, 7/1/92, \251IFI & UBS-UBILAB";
    gArgc= ac;
    gArgv= av;
    mainDocumentType= dt;

    ETInit(gArgc, gArgv);
    SetName(ProgramName());
    WindowSystem::WSInit();
    
    SetErrorHandler(ApplErrorHandler);
    lastWindowPos= Point(150, 100);
    char *cp= Env::GetValue("Document.Position", (char*)0);
    if (cp) {
	IStream is(strlen(cp), cp);
	is >> lastWindowPos;
    }
}

Application::~Application()
{
    if (gApplication == this)
	gApplication= 0;

    SafeDelete(fileDialog);
}

char *Application::GetName()
{
    return gProgname;
}

FileDialog *Application::MakeFileDialog()
{
    return new FileDialog;
}

Manager *Application::NewManager(const Symbol &type)
{
    Manager *mp= DoMakeManager(type); // let user make his managers
    if (mp) {
	AddManager(mp);
	mp->Show();
    }
    return mp;
}

void Application::ParseCommandLine(int argc, char **argv)
{
    for (int i= 1; i < argc;) {
	if (argv[i][0] == '-') {
	    int n= DoParseOptions(argv[i], argv[i+1]);
	    if (n > 0)
		i+= n;
	    else
		i++;
	} else 
	    OpenDocument(argv[i++]);
    }
}

int Application::DoParseOptions(char *arg0, char*)
{
    if (arg0[0] == '-' && arg0[1] == 'E') {
	if (arg0[2] == 'e') 
	    gProgEnv->Start();
	if (arg0[2] == 'h') 
	    gProgEnv->ShowInHierarchy(this->IsA());
	return 1;
    }
    if(strcmp(arg0,"-s")==0) // session stuff
	return 2;
    return 0;
}

void Application::RemoveManager(Manager *mp)
{
    Manager::RemoveManager(mp);
    if (Size() <= 0 && !IsOpen())
	Quit();
}

const char *Application::ProgramName()
{
    return gSystem->BaseName(gProgname);
}

Point Application::GetInitialWindowPos()
{
    return GetNewDocumentPos();
}

void Application::About()
{
    ShowAlert(eAlertSun, "%s %s", ProgramName(), version);
}

Command *Application::DoMenuCommand(int cmd)
{
    switch(cmd) {
    case cNEW:
	NewManager(mainDocumentType);
	break;
    case cQUIT:
	Quit();
	break;
    case cABOUT:
	About();
	break;
    case cOPEN:
	OpenDocument(0);
	break;
    default:
	Manager::DoMenuCommand(cmd);
	break;
    }
    return gNoChanges;
}

void Application::DoSetupMenu(Menu *m)
{
    Manager::DoSetupMenu(m);
    m->EnableItems(cOPEN, cNEW, cQUIT, cABOUT, 0);
}

void Application::Control(int id, int part, void *val)
{
    SendDown(id, part, val);
}

void Application::ExtCommand(int from, char *rq, char *args, int len,
							char *&ret, int &retlen)
{
    char req[100], *cp;

    strcpy(req, rq);
    cp= strchr(req, ':');

    if (cp) {
	*cp++= 0;
	
	if (strcmp(req, "document") == 0) {
	    Iter next(MakeIterator());
	    Manager *first= (Manager*)next();
	    if (first)
		first->ExtCommand(from, cp, args, len, ret, retlen);
	} else if (strcmp(req, "application") == 0) {
	    if (strcmp(cp, "quit") == 0)
		Quit();
	    else if (strcmp(cp, "about") == 0)
		About();
	    else if (strcmp(cp, "new") == 0)
		NewManager(mainDocumentType);
	    else if (strcmp(cp, "open") == 0)
		OpenDocument(args);
	    else if (strcmp(cp, "inspect") == 0)
		Inspect();
	    else if (strcmp (cp, "settimeout") == 0) {
	        // inserted by bischi
	        int to;
	        sscanf(args,"%d",&to);
	        gRpc->SetTimeout(to);
	    }
	    ret= "ok";
	}
    }
}

void Application::Quit()
{
    if (Close())
	gSystem->ExitControl();
}

Manager *Application::DoMakeManager(Symbol type)
{
    return DoMakeDocuments(type);
}

bool Application::OpenDocument(char *name)
{
    bool rc= FALSE, dodelete= FALSE;
    Data *data;
    Manager *mp;

    if (name == 0) {
	if (fileDialog == 0)
	    fileDialog= MakeFileDialog();
	if (fileDialog->ShowInWindow(eFDRead, 0, this) != cIdOk)
	    return FALSE;
	name= fileDialog->FileName();
	data= fileDialog->GetData();
    } else {
	data= new FileData(name);
	dodelete= TRUE;
    }
      
    if (!CanOpen(data)) {
	ShowAlert(eAlertNote, "Cannot handle document @B%s@B (%s)\n",
				    data->FullName(), data->Type().AsString());
    } else {
	mp= DoMakeManager(data->Type()); // let user make his managers
	if (mp) {
	    AddManager(mp);
	    mp->Show();
	    Guard(mp,Document)->LoadData(data, TRUE);
	    gWindowSystem->Update();
	    rc= TRUE;
	}
    }
    if (dodelete)
	delete data;
    return rc;
}

bool Application::CanOpen(Data *data)
{
    if (mainDocumentType == cDocTypeUndef || mainDocumentType == data->Type())
	return TRUE;
    if (mainDocumentType == cDocTypeAscii && data->IsAscii())
	return TRUE;
    return FALSE;
}

void Application::MakeInitManager()
{
    if (Size() <= 0)
	NewManager(mainDocumentType);
}
void Application::LooseControl( void )
{
    gSystem->Control();
}
int Application::Run()
{
    gRpc= new AppRpcConnection;
    gClipBoard= new ClipBoard;
    gProgress= new ProgressImpl;
    gSystem->AddAsyncSignalHandler(new ApplIntHandler);

    MemoryWarnAlert::Init();

    ParseCommandLine(gArgc, gArgv);

    MakeInitManager();

    Open();

    LooseControl( ); // call gSystem->Control();

    SetErrorHandler(DefaultErrorHandler);

    SafeDelete(gClipBoard);
    SafeDelete(gRpc);
    SafeDelete(gProgress);

    return 0;
}

void Application::DoOnError(int level, char *location, char *msg)
{
    static bool inError= FALSE;
    
    if (level < Env::GetValue("System.IgnoreLevel", cError) || inError)
	return;
    inError= TRUE;
    
    char *type= "Warning";
    if (level >= cFatal)
	type= "Fatal";
    else if (level >= cSysError)
	type= "SysError";
    else if (level >= cError)
	type= "Error";
    
    fprintf(stderr, "%s: %s in <%s>: %s\n", ProgramName(), type, location, msg);
    if (level >= cError) {
	int r= ShowAlert(eAlertError, "%s: %s\nin @B%s@B: %s", ProgramName(),
						    type, location, msg);
	switch (r) {
	case cIdIgnore:
	    break;
	    
	case cIdAbort:
	    gSystem->Abort();
	    break;
	    
	case cIdInspect:
	    Debug();
	    break;
	}
    }
    inError= FALSE;
}

void Application::Debug()
{
    gRpc->SendTo (0, "attachto", form("%s %d", gProgPath, gSystem->GetPid()) ); 
    gSystem->Wait(15000);
    //ShowAlert(eAlertError, "%s: Debugger did not attach\n", ProgramName());
}

//---- obsolete ----------------------------------------------------------------

Document *Application::FindDocument(int id)
{
    return (Document*) Application::FindManager(id);
}

Document *Application::DoMakeDocuments(Symbol)
{
    return 0;
}

bool Application::CloseAllDocuments()
{
    return Close();
}

void Application::AddDocument(Document *dp)
{
    AddManager(dp);
}

void Application::RemoveDocument(Document *mp)
{
    RemoveManager(mp);
}

void Application::NewDocument(Symbol)
{
}

Point Application::GetNewDocumentPos()
{
    Point p= lastWindowPos;
    lastWindowPos+= Point(40, 30);
    return p;
}

//---- external calls of ET++PE ------------------------------------------------

extern "C" int ETPE_Enter(Object *op);

static int CallCount= 0;

int ETPE_Enter(Object *op)
{
    int level= CallCount++;
    Port *p= GrGetPort();

    if (op == 0)
	op= gApplication;
	
    if (!ObjectTable::PtrIsValid(op)) {
	fprintf(stderr, "ET++PE: 0x%x is not an instance of Object\n", (int)op);
	return 0;
    }

    op->Inspect();

    //---- nested event-loop    
    while (CallCount > level && !gQuitApp)
	gSystem->InnerLoop();
    
    if (gQuitApp)
	CallCount= 0;
	
    fprintf(stderr, "\nETPE: Exit\n");
    
    GrSetPort(p);
    
    return level;
}

int gCallCount(int d)
{
    CallCount+= d;
    return CallCount;
}
 
