#ifndef Application_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define Application_First

#include "Manager.h"
#include "Data.h"
#include "Main.h"

class Application;
class ClipBoard;
class FileDialog;
class Document;
class Manager;
class Menu;
class StreamConnection;

extern Application      *gApplication;
extern ClipBoard        *gClipBoard;
extern StreamConnection *gRpc;

//---- Application -------------------------------------------------------------

class Application: public Manager {
public:
    MetaDef(Application);
    Application(int ac, char** av, const Symbol &mainDocType= cDocTypeUndef);
    ~Application();
    
    //---- dialogs
    virtual FileDialog *MakeFileDialog();

    //---- application attributes
    char *GetName();
    Symbol GetMainDocumentType()
	{ return mainDocumentType; }

    //---- Command line arguments
    virtual int DoParseOptions(char *arg0, char *arg1);
    virtual void ParseCommandLine(int argc, char **argv);
    const char *ProgramName();

    //---- Manager related methods
    virtual Point GetInitialWindowPos();
    virtual Manager *DoMakeManager(Symbol type);
    virtual bool CanOpen(Data*);
    virtual Manager *NewManager(const Symbol &type);
    virtual void MakeInitManager();
    void RemoveManager(Manager *mp);

    //---- generic application commands
    virtual void About();
    virtual void Quit();
    virtual void DoOnError(int level, char *location, char *msg);
    virtual void Debug();
    
    //---- menu related methods
    Command *DoMenuCommand(int);
    void DoSetupMenu(Menu *);
    void Control(int id, int, void*);
    void ExtCommand(int client, char *req, char *args, int len,
							char *&ret, int &retlen);
    //---- run the Application
    int Run();
    virtual void LooseControl( void );

    //---- obsolete
    virtual Document *DoMakeDocuments(Symbol type);
    virtual void AddDocument(Document*);
    void NewDocument(Symbol type);
    Document *FindDocument(int id);
    void RemoveDocument(Document*);
    virtual bool OpenDocument(char *name);  // returns TRUE on success
    virtual bool CloseAllDocuments();       // returns TRUE on success
    virtual Point GetNewDocumentPos();

protected:
    static FileDialog *fileDialog;

    char *version;
    Symbol mainDocumentType;
    Point lastWindowPos;
};
 
 
//--- StdMain (Michael Scharf)
// This is compatible with et322. The macro might change in the future that maybe 
// multiple applications wihtin one executable are possible (like in Andres version
// of et++ on the mac). Currently it just initializes et++ and runs the application.
//
// Simple applications with one subclass of Application should call StdMain 
// with the classname as parameter.
//
// Example:
//    class Hello :public Application {
//    public:
//        Hello(int ac, char** av)
//           : Application(ac,av) 
//         {}
//    ...
//    };
//
//    StdMain(Hello);
 
#define StdMain(name) \
int main(int argc, char *argv[]) \
{ \
    ETInitAll(argc,argv); \
    return name(argc, argv).Run(); \
}

#endif

