#ifndef Bitmap_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define Bitmap_First

#include "Rectangle.h"
#include "Ink.h"

class DevBitmap;
class ColorMapper;

//---- Bitmap ----------------------------------------------------------------

class Bitmap : public Ink {
protected:
    ColorMapper *cmapper;
    DevBitmap *dbm;
    
public:
    MetaDef(Bitmap);

    Bitmap(Point sz, u_short depth, ColorMapper *cmapper);
    Bitmap(Point sz, u_short depth= 1);
    Bitmap(Point sz, u_short *data);
    Bitmap(DevBitmap *);
    Bitmap(const char *name);
    Bitmap(const char *xpmdata[]);
    ~Bitmap();

    ColorMapper *GetColorMapper()
	{ return cmapper; }
    DevBitmap *GetDevBitmap()
	{ return dbm; }
    void SetDevBitmap(DevBitmap*);

    void SetInk(Port *p);
    
    void SetPixel(u_int x, u_int y, u_long value);
    u_long GetPixel(u_int x, u_int y);

    void SetRGB(u_int x, u_int y, RGB *rgb);
    void GetRGB(u_int x, u_int y, RGB *result);

    u_long MaxPixelValue();
    void PixelToRGB(u_long pixel, RGB *result);
    bool RGB2Pixel(RGB *rgb, u_long *pixel);
    bool ChangeRGB(u_long pixel, RGB *rgb);

    void CopyPixel(Rectangle to, Bitmap *from, Point fromorigin);
    void ExpandDepth();

    int GetDepth();
    Point Size();
    int ShortsPerLine();
    int BytesPerLine();
    int Shorts()
	{ return ShortsPerLine() * Size().y; }
    int Bytes()
	{ return Shorts() * 2; }

    virtual void Show(Rectangle *r, Port *port);

    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);
    char *ReadXPM(char *data[]);
    Object *deepclone();
};

class SmartBitmap {
    Point size;
    u_short *bits;
    Bitmap *bm;
    const char *name;
public:
    SmartBitmap(const Point &s, u_short *b)
	{ size= s; bits= b; }
    SmartBitmap(const char *nm)
	{ name= nm; }
    SmartBitmap(const char *xpmdata[])
	{ bits= (u_short*) xpmdata; name= (const char*) xpmdata; }
    ~SmartBitmap()
	{ SafeDelete(bm); }
    Bitmap *MakeBitmap();
    operator Bitmap* ()
	{ return bm ? bm : MakeBitmap(); }
};

#endif

