#ifdef __GNUG__
#pragma implementation
#endif

#include "Box.h"

#include "Class.h"
#include "Collection.h"
#include "ETMath.h"
#include "Look.h"
#include "ETString.h"

struct rcinfo {
    int wd, ht, bs;
    int hfixed, vfixed;
};

//---- Box ---------------------------------------------------------------------

NewMetaImpl(Box,CompositeVObject, (T(colsrows), T(gap), TE(align)));

Box::Box(int id, Point cr, Point g, VObjAlign a)
    : colsrows(cr), gap(g), align(a), CompositeVObject(id, (SeqCollection*)0)
{
}

Box::Box(int id, Point cr, Point g, VObjAlign a, VObject *va_(vop), ...)
    : colsrows(cr), gap(g), align(a), CompositeVObject(id, (SeqCollection*)0)
{
    va_list ap;
    va_start(ap, va_(vop));
    SetItems(va_(vop), ap);
    va_end(ap);
}

Box::Box(int id, Point cr, Point g, VObjAlign a, SeqCollection *sc)
    : colsrows(cr), gap(g), align(a), CompositeVObject(id, sc)
{
}

Box::Box(int id, Point cr, Point g, VObjAlign a, va_list ap)
    : colsrows(cr), gap(g), align(a), CompositeVObject(id, ap)
{
}

void Box::SetAlign(VObjAlign a)
{
    align= a;
}

void Box::SetGap(Point g)
{
    gap= g;
}

void Box::SetColsRowsSize(Point cr)
{
    colsrows= cr;
}

Point Box::ColsRowsSize()
{
    int n= Size();
    if (colsrows.y <= 0 && colsrows.x <= 0) {
	int c= Math::Sqrt(n);
	return Point(c, (n-1)/c+1);
    }
    if (colsrows.y <= 0)
	return Point(colsrows.x, (n-1)/colsrows.x+1);
    if (colsrows.x <= 0)
	return Point((n-1)/colsrows.y+1, colsrows.y);
    return colsrows;
}

Metric Box::getMinSize(Point &rc, rcinfo *rci, bool natural)
{
    register int x, y;
    Metric m(-gap, 0);

    getGrid(rc, rci, natural);
    for (x= 0; x < rc.x; x++)
	m.extent.x+= rci[x].wd+gap.x;
    for (y= 0; y < rc.y; y++)
	m.extent.y+= rci[y].ht+gap.y;
    
    if (align & eVObjVBase)
	m.base= rci[0].bs;
    else if (Size() > 0)
	m.base= At(0)->Base();
    return m;
}

void Box::getGrid(Point &rc, rcinfo *rci, bool natural)
{
    Iter next(MakeIterator());
    register VObject *vop;
    register int x, y, hh, bb, maxwd= 0, maxht= 0, n;
    bool hexpand= align & eVObjHExpand,
	 vexpand= (align & eVObjVExpand) && !(align & eVObjVBase);
    
    for (y= 0; y < rc.y; y++) {
	hh= bb= 0;
	for (x= 0; x < rc.x; x++) {
	    if (vop= (VObject*) next()) {
		Metric m(vop->GetMinSize());
		if (natural) {
		    Point e(vop->GetExtent());
		    if (e.x > m.extent.x)
			m.extent.x= e.x;
		    if (e.y > m.extent.y)
			m.extent.y= e.y;
		}
		if (align & eVObjVBase) {
		    hh= Math::Max(hh, m.Base());
		    int eb= m.Height()-m.Base();
		    bb= Math::Max(bb, eb);
		} else {
		    hh= Math::Max(hh, m.extent.y);
		}
		rci[x].wd= Math::Max(rci[x].wd, m.Width());
		maxwd= Math::Max(maxwd, m.Width());
		if (!hexpand || vop->TestFlag(eVObjHFixed))
		    rci[x].hfixed++;
		if (!vexpand || vop->TestFlag(eVObjVFixed))
		    rci[y].vfixed++;
	    }
	    
	}
	rci[y].ht= hh+bb;
	rci[y].bs= hh;
	maxht= Math::Max(maxht, hh+bb);
    }
    
    for (x= 0; x < rc.x; x++) {
	if (rci[x].hfixed >= rc.y)
	    rci[x].hfixed= 1;
	else
	    rci[x].hfixed= 0;
    }
    for (y= 0; y < rc.y; y++) {
	if (rci[y].vfixed >= rc.x)
	    rci[y].vfixed= 1;
	else
	    rci[y].vfixed= 0;
    }
    /*
    for (x= 0; x < rc.x; x++) {
	if (rci[x].hfixed > 0)
	    rci[x].hfixed= 1;
	else
	    rci[x].hfixed= 0;
    }
    for (y= 0; y < rc.y; y++) {
	if (rci[y].vfixed > 0)
	    rci[y].vfixed= 1;
	else
	    rci[y].vfixed= 0;
    }
    */
    
    if ((align & eVObjHGapExpand) && rc.x > 1) {
	ResetFlag(eVObjHFixed);
    } else {
	for (n= x= 0; x < rc.x; x++)
	    if (rci[x].hfixed)
		n++;
	SetFlag(eVObjHFixed, n >= rc.x);
    }
    
    if ((align & eVObjVGapExpand) && rc.y > 1) {
	ResetFlag(eVObjVFixed);
    } else {
	for (n= y= 0; y < rc.y; y++)
	    if (rci[y].vfixed)
		n++;
	SetFlag(eVObjVFixed, n >= rc.y);
    }
    
    if (align & eVObjHEqual)
	for (x= 0; x < rc.x; x++)
	    rci[x].wd= maxwd;
    if (align & eVObjVEqual)
	for (y= 0; y < rc.y; y++)
	    rci[y].ht= maxht;
}

void Box::expand(Point &rc, rcinfo *rci, Point &g, Point &r)
{
    register int x, y;
    bool hagain, vagain;
    rcinfo *rci2= new rcinfo[Math::Max(rc.x, rc.y)];

    Point ne(GetExtent());
    Point oe(getMinSize(rc, rci, TRUE).extent);
    Point mine(getMinSize(rc, rci2, FALSE).extent);
    Point diff(ne-mine);
    
    g= gap;
    r= gPoint0;

    if (diff.x > 0) {
	if (align & eVObjHGapExpand) {
	    if (rc.x > 1) {
		g.x= diff.x/(rc.x-1);
		r.x= diff.x-g.x*(rc.x-1);
		g.x+= gap.x;
	    }
	}
    }

    if (diff.y > 0) {
	if (align & eVObjVGapExpand) {
	    if (rc.y > 1) {
		g.y= diff.y/(rc.y-1);
		r.y= diff.y-g.y*(rc.y-1);
		g.y+= gap.y;
	    }
	}
    }

    hagain= vagain= TRUE;
    
    for (int i= 0; i < 2; i++) {
	Point newe(ne), olde(oe);
	if (hagain && (align & eVObjHExpand)) {
	    for (x= 0; x < rc.x; x++) {
		if (rci[x].hfixed) {  // col not hfixed
		    newe.x-= rci[x].wd;
		    olde.x-= rci[x].wd;
		}
		if (x < rc.x-1) {
		    newe.x-= gap.x;
		    olde.x-= gap.x;
		}
	    }
	    
	    if (olde.x > 0) {
		int total= newe.x;
		hagain= FALSE;
		for (x= 0; x < rc.x; x++) {
		    if (!rci[x].hfixed) {
			rci[x].wd= (rci[x].wd*newe.x)/olde.x;
			total-= rci[x].wd;
		    }
		}
		for (x= 0; total > 0 && x < rc.x; x++) {
		    if (!rci[x].hfixed) {
			rci[x].wd++;
			total--;
		    }
		}
		for (x= 0; x < rc.x; x++) {
		    if (!rci[x].hfixed && rci[x].wd < rci2[x].wd) {
			hagain= TRUE;
			break;
		    }
		}
	    }
	}
    
	if (vagain && (align & eVObjVExpand)) {
	    for (y= 0; y < rc.y; y++) {
		if (rci[y].vfixed) {   // row not vfixed
		    newe.y-= rci[y].ht;
		    olde.y-= rci[y].ht;
		}
		if (y < rc.y-1) {
		    newe.y-= gap.y;
		    olde.y-= gap.y;
		}
	    }
	    if (olde.y > 0) {
		int total= newe.y;
		vagain= FALSE;
		for (y= 0; y < rc.y; y++) {
		    if (!rci[y].vfixed) {
			rci[y].ht= (rci[y].ht*newe.y)/olde.y;
			total-= rci[y].ht;
		    }
		}
		for (y= 0; total > 0 && y < rc.y; y++) {
		    if (!rci[y].vfixed) {
			rci[y].ht++;
			total--;
		    }
		}
		for (y= 0; y < rc.y; y++) {
		    if (!rci[y].vfixed && rci[y].ht < rci2[y].ht) {
			vagain= TRUE;
			break;
		    }
		}
	    }
	}
	
	if (hagain)
	    for (x= 0; x < rc.x; x++)
		rci[x].wd= rci2[x].wd;
	if (vagain)
	    for (y= 0; y < rc.y; y++)
		rci[y].ht= rci2[y].ht;
    }
    
    delete rci2;
}

Metric Box::GetMinSize()
{
    Point rc(ColsRowsSize());
    rcinfo *rci= new rcinfo[Math::Max(rc.x, rc.y)];
    Metric m(getMinSize(rc, rci, FALSE));
    delete rci;
    return m;
}

void Box::SetOrigin(Point at)
{
    Iter next(MakeIterator());
    register VObject *vop;
    register int x, y;
    Point g, r, rr, a, rc(ColsRowsSize());
    rcinfo *rci= new rcinfo[Math::Max(rc.x, rc.y)];
	
    expand(rc, rci, g, r);

    VObject::SetOrigin(at);
    
    a.y= at.y;
    rr.y= r.y;
    for (y= 0; y < rc.y; y++) {
	a.x= at.x;
	rr.x= r.x;
	for (x= 0; x < rc.x; x++) {
	    if (vop= (VObject*) next())
		vop->Align(a, Metric(rci[x].wd, rci[y].ht, rci[y].bs), align);
	    a.x+= rci[x].wd + g.x;
	    if (rr.x > 0) {
		a.x++;
		rr.x--;
	    }
	}
	a.y+= rci[y].ht + g.y;
	if (rr.y > 0) {
	    a.y++;
	    rr.y--;
	}
    }
    delete rci;
}

void Box::SetExtent(Point ee)
{
    Iter next(MakeIterator());
    register VObject *vop;
    register x, y;
    Point g, r, rc(ColsRowsSize());
    rcinfo *rci= new rcinfo[Math::Max(rc.x, rc.y)];
    bool hexpand= align & (eVObjHExpand|eVObjHEqual),
	 vexpand= (align & (eVObjVExpand|eVObjVEqual)) && !(align & eVObjVBase);

    Point e(getMinSize(rc, rci, FALSE).extent);
    if (!TestFlag(eVObjHFixed))
	e.x= ee.x;
    if (!TestFlag(eVObjVFixed) && !(align & eVObjVBase))
	e.y= ee.y;
    VObject::SetExtent(e);
    
    expand(rc, rci, g, r);
    
    for (y= 0; y < rc.y; y++) {
	for (x= 0; x < rc.x; x++) {
	    if (vop= (VObject*) next()) {
		Point ms(vop->GetMinSize().extent);
		e= vop->GetExtent();
		if (e.x < ms.x || e.y < ms.y) {
		    if (e.x < ms.x)
			e.x= ms.x;
		    if (e.y < ms.y)
			e.y= ms.y;
		}
		if (hexpand && !vop->TestFlag(eVObjHFixed))
		    e.x= rci[x].wd;
		if (vexpand && !vop->TestFlag(eVObjVFixed))
		    e.y= rci[y].ht;
		vop->SetExtent(e);
	    }
	}
    }

    delete rci;
}

OStream& Box::PrintOn(OStream &s)
{
    CompositeVObject::PrintOn(s);
    return s << colsrows SP << gap SP << align SP;
}

IStream& Box::ReadFrom(IStream &s)
{
    CompositeVObject::ReadFrom(s);
    return s >> colsrows >> gap >> Enum(align);
}

int Box::FindVObject(VObject *v)
{
    VObject *vop;
    Iter next(MakeIterator());
    
    for (int i= 0; vop= (VObject*) next(); i++)
	if (v == vop)
	    return i;
    return -1;
}

void Box::CalcResize(int where, Point d, rcinfo *minsize, rcinfo *normsize, rcinfo *result)
{
    Point rc(ColsRowsSize());
    int i;

    if (d.x) {  // vertical
	bool got= TRUE;
	int dd, gg= 0;
	if (d.x < 0) {
	    dd= -d.x;
	    while (dd > 0 && got) {
		got= FALSE;
		for (i= 0; i < where; i++) {
		    if (!normsize[i].hfixed && result[i].wd > minsize[i].wd) {
			result[i].wd--;
			dd--;
			gg++;
			got= TRUE;
			break;
		    }
		}
	    }
	    while (gg > 0) {
		for (i= where+1; i < rc.x; i++) {
		    if (!normsize[i].hfixed) {
			result[i].wd++;
			gg--;
			break;
		    }
		}
	    }
	} else {
	    dd= d.x;
	    while (dd > 0 && got) {
		got= FALSE;
		for (i= where+1; i < rc.x; i++) {
		    if (!normsize[i].hfixed && result[i].wd > minsize[i].wd) {
			result[i].wd--;
			dd--;
			gg++;
			got= TRUE;
			break;
		    }
		}
	    }
	    while (gg > 0) {
		for (i= 0; i < where; i++) {
		    if (!normsize[i].hfixed) {
			result[i].wd++;
			gg--;
			break;
		    }
		}
	    }
	}
    }

    if (d.y) { // hor
	bool got= TRUE;
	int dd, gg= 0;
	if (d.y < 0) {
	    dd= -d.y;
	    while (dd > 0 && got) {
		got= FALSE;
		for (i= 0; i < where; i++) {
		    if (!normsize[i].vfixed && result[i].ht > minsize[i].ht) {
			result[i].ht--;
			dd--;
			gg++;
			got= TRUE;
			break;
		    }
		}
	    }
	    while (gg > 0) {
		for (i= where+1; i < rc.y; i++) {
		    if (!normsize[i].vfixed) {
			result[i].ht++;
			gg--;
			break;
		    }
		}
	    }
	} else {
	    dd= d.y;
	    while (dd > 0 && got) {
		got= FALSE;
		for (i= where+1; i < rc.y; i++) {
		    if (!normsize[i].vfixed && result[i].ht > minsize[i].ht) {
			result[i].ht--;
			dd--;
			gg++;
			got= TRUE;
			break;
		    }
		}
	    }
	    while (gg > 0) {
		for (i= 0; i < where; i++) {
		    if (!normsize[i].vfixed) {
			result[i].ht++;
			gg--;
			break;
		    }
		}
	    }
	}
    }     
}

void Box::Resize(rcinfo *newsize)
{
    register VObject *vop;
    register int x, y;
    Iter next(MakeIterator());
    Point at(GetOrigin()), a, rc(ColsRowsSize());
    Rectangle inval;

    bool hexpand= align & (eVObjHExpand|eVObjHEqual),
	 vexpand= (align & (eVObjVExpand|eVObjVEqual)) && !(align & eVObjVBase);

    a.y= at.y;
    for (y= 0; y < rc.y; y++) {
	a.x= at.x;
	for (x= 0; x < rc.x; x++) {
	    if (vop= (VObject*) next()) {
		Rectangle oldr(vop->contentRect);
		Point e(vop->GetExtent());
		if (e.x == 0 || e.y == 0)
		    e= vop->GetMinSize().extent;
		if (hexpand && !vop->TestFlag(eVObjHFixed))
		    e.x= newsize[x].wd;
		if (vexpand && !vop->TestFlag(eVObjVFixed))
		    e.y= newsize[y].ht;
		vop->SetExtent(e);
		vop->Align(a, Metric(newsize[x].wd, newsize[y].ht, newsize[y].bs), align);
		if (vop->contentRect != oldr)
		    inval.Merge(vop->contentRect);
	    }
	    a.x+= newsize[x].wd + gap.x;
	}
	a.y+= newsize[y].ht + gap.y;
    }
    
    InvalidateRect(inval);
}

//---- SplitMover --------------------------------------------------------------

class SplitMover: public Command {
    Rectangle constrainRect;
    Point delta, rc;
    Box *sbox;
    VObject *where;
    rcinfo *minsize, *result, *normsize;
    int iwhere;
public:
    SplitMover(Box *sb, VObject *w);
    ~SplitMover();
    Command *TrackMouse(TrackPhase, Point, Point, Point);
    void TrackConstrain(Point ap, Point pp, Point *np);
    void TrackFeedback(Point, Point, bool on);
};

SplitMover::SplitMover(Box *sb, VObject *w)
{
    sbox= sb;
    constrainRect= sbox->contentRect;
    where= w;
    iwhere= sbox->FindVObject(where);
    rc= sbox->ColsRowsSize();
    int n= Math::Max(rc.x, rc.y);
    minsize= new rcinfo[n];
    sbox->getGrid(rc, minsize, FALSE);
    normsize= new rcinfo[n];
    sbox->getGrid(rc, normsize, TRUE);
}

SplitMover::~SplitMover()
{
    SafeDelete(minsize);
    SafeDelete(normsize);
    SafeDelete(result);
}

void SplitMover::TrackConstrain(Point ap, Point, Point *np)
{
    Rectangle r(where->contentRect);
    r.origin+= *np-ap;
    *np+= r.AmountToTranslateWithin(constrainRect);
}

Command *SplitMover::TrackMouse(TrackPhase tp, Point ap, Point, Point np)
{
    delta= np-ap;
    
    SafeDelete(result);
    int n= Math::Max(rc.x, rc.y);
    result= new rcinfo[n];
    memcpy(result, normsize, n*sizeof(struct rcinfo));
    sbox->CalcResize(iwhere, delta, minsize, normsize, result);
    
    if (tp == eTrackRelease) {
	sbox->Resize(result);
	return gNoChanges;
    }
    return this;
}
    
void SplitMover::TrackFeedback(Point, Point, bool)
{
    Point gap(sbox->GetGap()),
	  at(sbox->GetOrigin()),
	  e(sbox->GetExtent());

    if (delta.x) {  // vertical
	for (int x= 0; x < rc.x; x++) {
	    at.x+= result[x].wd;
	    GrLine(at, at+Point(0, e.y));
	    at.x+= gap.x;
	}
    }
	
    if (delta.y) { // hor
	for (int y= 0; y < rc.y; y++) {
	    at.y+= result[y].ht;
	    GrLine(at, at+Point(e.x, 0));
	    at.y+= gap.y;
	}
    }
}

//---- HBox --------------------------------------------------------------------

NewMetaImpl0(HBox,Box);

HBox::HBox(Point g, VObjAlign a) : Box(cIdNone, Point(0,1), g, a)
{
}

HBox::HBox(Point g, VObjAlign a, VObject *va_(vop), ...)
					    : Box(cIdNone, Point(0,1), g, a)
{
    va_list ap;
    va_start(ap, va_(vop));
    SetItems(va_(vop), ap);
    va_end(ap);
}

HBox::HBox(Point g, VObjAlign a, SeqCollection *c)
					    : Box(cIdNone, Point(0,1), g, a, c)
{
}

HBox::HBox(VObject *va_(vop), ...)
		    : Box(cIdNone, Point(0,1), gPoint2,
			    (VObjAlign)(eVObjVTop|eVObjHExpand|eVObjVExpand))
{
    va_list ap;
    va_start(ap, va_(vop));
    SetItems(va_(vop), ap);
    va_end(ap);
}

//---- VBox --------------------------------------------------------------------

NewMetaImpl0(VBox,Box);

VBox::VBox(Point g, VObjAlign a) : Box(cIdNone, Point(1,0), g, a)
{
}

VBox::VBox(Point g, VObjAlign a, VObject *va_(vop), ...)
					    : Box(cIdNone, Point(1,0), g, a)
{
    va_list ap;
    va_start(ap, va_(vop));
    SetItems(va_(vop), ap);
    va_end(ap);
}

VBox::VBox(Point g, VObjAlign a, SeqCollection *c)
					    : Box(cIdNone, Point(1,0), g, a, c)
{
}

VBox::VBox(VObject *va_(vop), ...)
		: Box(cIdNone, Point(1,0), gPoint2,
			    (VObjAlign)(eVObjHLeft|eVObjHExpand|eVObjVExpand))
{
    va_list ap;
    va_start(ap, va_(vop));
    SetItems(va_(vop), ap);
    va_end(ap);
}

//---- SplitHandle -------------------------------------------------------------

NewMetaImpl0(SplitHandle,VObject);

SplitHandle::SplitHandle()
{
}

Metric SplitHandle::GetMinSize()
{
    Metric m(gLook->SashLayout()->GetMinSize(this));
    Box *b= (Box*) GetContainer();
    if (b && b->IsKindOf(Box)) {
	Point rc(b->ColsRowsSize());
	if (rc.x > rc.y) {
	    SetFlag(eVObjHFixed);
	    m.extent.y*= 2;
	} else if (rc.x < rc.y) {
	    SetFlag(eVObjVFixed);
	    m.extent.x*= 2;
	}
    }
    return m;
}

void SplitHandle::Draw(Rectangle r)
{
    gLook->SashLayout()->Adorn(this, r, 0);
}

Command *SplitHandle::DoLeftButtonDownCommand(Point, Token, int)
{
    Box *sb= (Box*) GetContainer();
    if (sb && sb->IsKindOf(Box))
	return new SplitMover(sb, this);
    return gNoChanges;
}

