#ifndef Container_First
#ifdef __GNUG__
#pragma interface
#endif
#define Container_First

#include "Object.h"
#include "Iterator.h"

extern char *cWarningActiveIterator;
extern char *cWarningNullArgument;  

const int cContainerInitCap= 16;        

const bool cIterForward= TRUE,
	   cIterBackward= ! cIterForward;
	   
//---- Container --------------------------------------------------------

class Container : public Object {
friend class Iterator;
friend class IterIter;
protected:
    Container();
public:
    MetaDef(Container);
    ~Container();
    void InitNew();
    virtual int Capacity();
    bool assertclass(Class *);
    void InspectorId(char*, int);

    //---- comparison 
    bool IsEqual(Object *);
    u_long Hash();

    //---- access     
    int Size();
    bool IsEmpty();
    bool Contains(Object *op);
    bool ContainsPtr(Object *op);
    
    virtual Iterator *MakeIterator(bool dir= cIterForward);
    Iterator *MakeReversedIterator();
    virtual Object *AnyElement();       // return any element
    virtual Object *Find(Object *);     // returns 0 if not found, uses IsEqual
    virtual Object *FindPtr(Object *);  // returns 0 if not found, uses identity
    virtual int OccurrencesOf(Object *);    // uses IsEqual
    virtual int OccurrencesOfPtr(Object *); // uses identity

    //---- object i/o
    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);
    
protected:
    bool IsArgNull(char *where, Object *arg);   
    bool CheckForIterator(char *where);        
	       
    //---- iterators
    bool HasIterators();                                              
    void IteratorActivates(Iterator *which);
    void IteratorTerminates(Iterator *which);   
    void DetachIterators();               
    
    //---- robust iterators - place holder approach (old style)
    virtual void RemoveDeleted();
    void AnnounceRemove();
    bool HasDeletedObjects();
protected:
    int size;
    int nDeleted;       // for robust iterators (place holder approach)
    int iterCount;      // active iterators
private:
    Iterator iterHead;
public:
    void setCurrentContainer() { currentContainer= this; }; // for ForEach macro
    static Container *currentContainer;                     // for ForEach macro
};


//---- class Iter -------------------------------------------------------------

class Iter {
    Iterator *seq;
public:
    Iter(Container *coll, bool dir= cIterForward)
	{ seq= coll->MakeIterator(dir); }
    Iter(Iterator *it)
	{ seq= it; }
    ~Iter()
	{ delete seq; }
    Object *operator()()
	{ return (*seq)(); }
    void Reset(Container *coll, bool dir= cIterForward)
	{ delete seq; seq= coll->MakeIterator(dir); }
    void Reset()
	{ seq->Reset(); }
};

//---- macros -----------------------------------------------------------------

#define ForEach(type,proc)                          \
setCurrentContainer();                              \
Iter _NAME3_(type,proc,_next)(Container::currentContainer);     \
register type *_NAME3_(type,proc,_op);              \
while (_NAME3_(type,proc,_op)= (type *) _NAME3_(type,proc,_next)())     \
    _NAME3_(type,proc,_op)->proc

//---- Container inlines ------------------------------------------------------

inline int Container::Size()
{
    return size - nDeleted;
}

inline bool Container::IsEmpty()
{
    return (bool) (Size() <= 0); 
}

inline bool Container::Contains(Object *op)
{ 
    return Find(op) != 0; 
}

inline bool Container::ContainsPtr(Object *op)
{ 
    return FindPtr(op) != 0; 
}

inline Iterator *Container::MakeReversedIterator()
{
    return MakeIterator(FALSE);
}

//---- inlines (protected and private) ----

inline bool Container::IsArgNull(char *where, Object *arg) 
{ 
    return arg ? FALSE : (Warning(where, cWarningNullArgument), TRUE); 
}

inline bool Container::HasIterators()
{ 
    return (iterCount > 0);  
}

inline bool Container::CheckForIterator(char *where)
{ 
    return HasIterators() ? 
	    (Warning(where, cWarningActiveIterator), TRUE) : FALSE; 
}

inline void Container::AnnounceRemove() 
{ 
    nDeleted++; 
}

inline bool Container::HasDeletedObjects() 
{ 
    return (bool) (nDeleted > 0); 
}

#endif /* Container_First */

