#ifdef __GNUG__
#pragma implementation
#endif

#include "Dictionary.h"
#include "Class.h"

#include "DictionaryHashTable.h"

//---- Dictionary -------------------------------------------------------------

NewMetaImpl(Dictionary,Container, (TP(ht)));

Object *Dictionary::ignore;

Dictionary::Dictionary(int initCapacity) : Container()
{
    Init(initCapacity);
}

Dictionary::~Dictionary()
{
    SafeDelete(ht);
}

Object *Dictionary::Clone()
{
    Dictionary *clone;
    Object *key, *value;
   
    clone= (Dictionary *) New();
    DictionaryIterator next(this);
    while (key= next()) {
	value= next.GetValue();
	clone->PutAtKeyIfAbsent(value, key);
    }
    return clone;
}

void Dictionary::InitNew()
{
    Container::InitNew();
    Init(cContainerInitCap);
}

void Dictionary::FreeAll()
{
    DictionaryIterator next(this);
    Object *key, *value;
    
    CheckForIterator("FreeAll");
 
    while ( key= next() ) {
	key->FreeAll();
	value= next.GetValue();
	value->FreeAll();
	SafeDelete(key);
	SafeDelete(value);
    }
    Empty(cContainerInitCap);
}


int Dictionary::Capacity()
{
    return ht->Capacity();
}

void Dictionary::Empty(int newCapacity)
{
    CheckForIterator("Empty");
    
    ht->Empty(newCapacity);
    if ( IsObserved() )
	Changed();
}

void Dictionary::FreeValues()
{
    DictionaryIterator it(this);
    Object *value;
    
    CheckForIterator("FreeValues");
    
    while ( value= it.NextValue() ) 
	SafeDelete(value);
    Empty(cContainerInitCap); 
}

void Dictionary::FreeAllValues()
{
    DictionaryIterator it(this);
    Object *value;
    
    CheckForIterator("FreeAllValues");
    
    while ( value= it.NextValue() ) {
	value->FreeAll();
	SafeDelete(value);
    }
    Empty(cContainerInitCap); 
}

Object *Dictionary::PutAtKeyIfAbsent(Object *value, Object *key)
{
    Object *val;
    
    if ( IsArgNull("PutAtKeyIfAbsent(value)", value) || 
	 IsArgNull("PutAtKeyIfAbsent(key)", key) )
	return value;
    CheckForIterator("AddAtKey");
    
    val= (Object *) ht->AddAt(value, key, FALSE);
    if ( ! val && IsObserved() )
	Changed();
    return val;
}

Object *Dictionary::PutAtKey(Object *value, Object *key)
{
    Object *oldValue;
    
    if ( IsArgNull("PutAtKey(value)", value) || IsArgNull("PutAtKey(key)", key) )
	return value;
    CheckForIterator("AddAtKey");
    
    oldValue= (Object *) ht->AddAt(value, key, TRUE);
    if ( oldValue != value && IsObserved() )
	Changed();
    return oldValue;
}

Object *Dictionary::RemoveKey(Object *key, Object *&value)
{ 
    Object *removedKey;
	
    if ( IsArgNull("RemoveKey", key) )
	return 0;
    CheckForIterator("RemoveKey");

    removedKey= (Object *) ht->RemoveKey(key, (void **) &value, eEqValue);
    if ( removedKey && IsObserved() )
	Changed();
    return removedKey;
}

Iterator *Dictionary::MakeIterator(bool)
{
    return new DictionaryIterator(this);
}

Object *Dictionary::AtKey(Object *key)
{
    return (Object *) ht->AtKey(key, eEqValue);
}

OStream &Dictionary::PrintOn(OStream &os)
{
    Object *key;
    DictionaryIterator next(this);
    Container::PrintOn(os);
    
    os NL;
    while ( key= next() )
	os << key SP << next.GetValue() NL;
    return os NL;
}

IStream &Dictionary::ReadFrom(IStream &is)
{
    Object *key, *value;
    int nread, sz;
    
    Container::ReadFrom(is);
    nread= 0;
    sz= size; 
    size= 0;
    while ( nread < sz ) {
	is >> key >> value;
	nread++;
	PutAtKey(value, key);
    }
    return is;
}

void Dictionary::Init(int initCapacity)
{
    ht= new DictionaryHashTable(
	    initCapacity, 
	    &ContainerTypes::pEqual, &ContainerTypes::pHash, 
	    &size);
}

Dictionary::Dictionary(Dictionary *) : Container()
{
    ht= 0;
}

//---- DictionaryIterator ----------------------------------------------------------

DictionaryIterator::DictionaryIterator(Dictionary *adt)
{
    dt= adt;
    pix= 0; 
    ht= 0;     
}

DictionaryIterator::~DictionaryIterator()
{
    Terminate();    
}

void DictionaryIterator::Reset()
{
    Iterator::Reset();
    pix= 0;
}

Object *DictionaryIterator::operator()()
{
    return NextIndex();
}

Object *DictionaryIterator::NextValue()
{
    Object *key= NextIndex();

    if ( key )
	return (Object *) ht->ValueAt(pix);
    return 0; 
}
 
Object *DictionaryIterator::NextKey()
{
    return NextIndex();
}

Object *DictionaryIterator::GetValue()
{
    if ( IsActive() )
	return (Object *) ht->ValueAt(pix);
    return 0; 
}
 
Container *DictionaryIterator::GetContainer()
{
    return dt;
}

Object *DictionaryIterator::NextIndex()
{   
    Object *op;
     
    if ( ! Activated() ) {
	if ( Activate() ) {
	    ht= dt->ht;
	    pix= ht->GetInitialIndex();
	}
    }
    if ( ! Terminated() ) {
	op= (Object *) ht->AtNextIndex(pix);
	if ( ! op ) 
	    Terminate();
	return op;
    } 
    return 0;
};

//---- IdDictionary ------------------------------------------------------------

NewMetaImpl0(IdDictionary,Dictionary);

IdDictionary::IdDictionary(int initCapacity) : Dictionary(this)
{
    Init(initCapacity);
}

void IdDictionary::InitNew()
{
    Container::InitNew();
    Init(cContainerInitCap);
}

void IdDictionary::Init(int initCapacity)
{
    ht= new DictionaryHashTable(
	    initCapacity, 
	    &ContainerTypes::pIdentical, &ContainerTypes::pHashPtr, 
	    &size);    
}


