#ifndef Dictionary_First
#ifdef __GNUG__
#pragma interface
#endif
#define Dictionary_First

#include "Container.h"

class DictionaryHashTable;

//---- Dictionary -------------------------------------------------------------

class Dictionary : public Container {
friend class DictionaryIterator;
    static Object *ignore;  
public:
    MetaDef(Dictionary);
    Dictionary(int initCapacity= cContainerInitCap);
    ~Dictionary();
    Object *Clone();
    void InitNew();    
    void FreeAll();
    
    int Capacity();    
    virtual void Empty(int newCapacity= cContainerInitCap);    
    void FreeValues();      // delete all values and empty receiver
    void FreeAllValues();   // FreeAll values and empty receiver
    
    //---- manipulation           
    virtual Object *PutAtKeyIfAbsent(Object *value, Object *key);  
    virtual Object *PutAtKey(Object *value, Object *key);  
    virtual Object *RemoveKey(Object *key, Object *&value= ignore);
    
    //---- access
    Iterator *MakeIterator(bool dir= cIterForward);     // Iterator yields keys
    Object *AtKey(Object *key);             // return the value at key  

    //---- object i/o
    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);
protected:
    Dictionary(Dictionary *);
    void Init(int initCapacity);
    bool CheckKey(char *name, Object *op); 
protected:    
    class DictionaryHashTable *ht;
};

//---- DictionaryIterator -----------------------------------------------------

class DictionaryIterator : public Iterator {
public:
    DictionaryIterator(Dictionary *aDict);
    ~DictionaryIterator();
    
    void Reset(); 
    Object *operator()();    // return key
    Object *GetValue();  
    Object *NextValue();  
    Object *NextKey();  
protected:
    Container *GetContainer();
    Object *NextIndex();
protected:
    int pix;
    class DictionaryHashTable *ht;
    Dictionary *dt;
};

//---- IdDictionary ------------------------------------------------------------

class IdDictionary : public Dictionary {
public:
    MetaDef(IdDictionary);
    IdDictionary(int initCap= cContainerInitCap);
    void InitNew();
    
protected:
    void Init(int initCapacity);
};

//---- Dictionary inlines -----------------------------------------------------

inline bool Dictionary::CheckKey(char *name, Object *op) 
{   
    return IsArgNull(name, op) ? TRUE : (CheckForIterator(name), FALSE);
} 

#endif /* Dictionary_First */

