#ifdef __GNUG__
#pragma implementation
#endif

#include "DictionaryHashTable.h"
#include "Class.h"

//---- DictionaryHashTable -----------------------------------------------------

NewMetaImpl(DictionaryHashTable,HashTable, 
	    (TVP(keys,capacity), TVP(vals, capacity)));

DictionaryHashTable::DictionaryHashTable(
    int initCapacity, pEqualFunc eq, pHashFunc hash, int *size) : 
    HashTable(initCapacity, eq, hash, size)
{   
    keys= new Object*[capacity + 1];
    vals= new Object*[capacity + 1];       
}

DictionaryHashTable::~DictionaryHashTable()
{
    SafeDelete(keys);
    SafeDelete(vals);
}

void DictionaryHashTable::Empty(int newCapacity)
{
    SafeDelete(keys);
    SafeDelete(vals);

    CalcBounds(newCapacity);
    keys= new Object*[capacity + 1];
    vals= new Object*[capacity + 1];       
	
    *size= 0;
    
    if ( IsObserved() )
	Changed();
}

void *DictionaryHashTable::AddAt(void *value, void *key, bool replace)
{
    int slot;
    void *replaced;
    
    if ( SearchFirst(key, slot) )
	if ( replace ) {
	    replaced= ValueAt(slot);
	    SetValueAt(value, slot);
	    if ( IsObserved() )
		Changed();
	    return replaced;
	} else
	    return value;
    else {        
	SetKeyAt(key, slot);
	SetValueAt(value, slot);
	(*size)++;
	if ( *size >= highWatermark )
	    Rehash(NewCapacity(TRUE));
	if ( IsObserved() )
	    Changed();
	return 0; 
    }
}

void *DictionaryHashTable::AtKey(void *key, EqualType equal)
{
    int slot;
       
    if ( equal ? SearchFirst(key, slot) : SearchPtr(key, slot) )
	return ValueAt(slot);
    return 0;
}

void *DictionaryHashTable::FindKey(void *key, void **value, EqualType equal)
{
    int slot;  

    key= equal ? SearchFirst(key, slot) : SearchPtr(key, slot);      
    if ( key )
	*value= ValueAt(slot);
    else
	*value= 0;
    return key;
}

void *DictionaryHashTable::RemoveKey(void *key, void **value, EqualType equal)
{
    int slot;
    
    key= equal ? SearchFirst(key, slot) : SearchPtr(key, slot);
    if ( key ) {
	if ( value )
	    *value= ValueAt(slot);
	RemoveSlot(slot);
	if ( IsObserved() )
	    Changed();
    }
    return key;
}

//----- primitives --------

void *DictionaryHashTable::KeyAt(int i)
{
    return keys[i];
}

void DictionaryHashTable::SetKeyAt(void *key, int i)
{
    keys[i]= (Object *) key; 
}

void *DictionaryHashTable::ValueAt(int i)
{
    return vals[i];
}

void DictionaryHashTable::SetValueAt(void *value, int i)
{
    vals[i]= (Object *) value;
}

void DictionaryHashTable::NullAt(int i)
{
    keys[i]= 0;
    vals[i]= 0;
}

void DictionaryHashTable::Swap(int i, int j)
{
    void *key, *val;
	
    key= keys[i];
    val= vals[i];
    keys[i]= keys[j];
    vals[i]= vals[j];
    keys[j]= (Object *) key;
    vals[j]= (Object *) val;
}

void DictionaryHashTable::DoRehash(HashTable *ht)
{
    register int i;
    DictionaryHashTable *dt;
    
    dt= (DictionaryHashTable *) ht;
    for ( i= 0; i < capacity; i++ ) {
	if ( keys[i] )
	    dt->AddAt(vals[i], keys[i], FALSE);
    }
}

HashTable *DictionaryHashTable::MakeNew(int initCapacity)
{
    return new DictionaryHashTable(initCapacity, equal, hash);
}

void DictionaryHashTable::Assign(HashTable *ht)
{
    DictionaryHashTable *dht= (DictionaryHashTable *) ht;
    
    HashTable::Assign(ht);
    SafeDelete(keys);
    SafeDelete(vals);
    keys= dht->keys;
    vals= dht->vals;
    dht->keys= 0;
    dht->vals= 0;   
}

