#ifndef Iterator_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define Iterator_First

#include "Types.h"

class Object;
class Container;

//---- class IteratorBase ------------------------------------------------------

class IteratorBase {
#ifdef __GNUG__
    char *dummy;        
#endif
protected:
    IteratorBase();
public:  
    virtual ~IteratorBase();
    
    virtual Object *operator()();
    virtual void Reset();
};

//---- Iterator ----------------------------------------------------------------

class Iterator : public IteratorBase {
friend class Container;
friend class IterIter;
protected:
    Iterator();
public:
    ~Iterator();
    void Reset();
    
    //---- memory allocation
#if !defined(OS_WINDOWS)
    void *operator new(size_t sz);
    void operator delete(void *vp); 
#endif
protected:
    bool IsActive();
    bool Activate();                   
    void Terminate();
    bool Activated();
    bool Terminated();
    virtual Container *GetContainer();
    virtual void Detach();                                 
private:
    //---- only used by Container
    void ChainIn(Iterator *anchor);
    void ChainOut();
    void Init() { next= this; prev= this; };
private:
    bool activated, terminated;         
    Iterator *next, *prev;  // linking iterators - only used by Container
};

//---- Iterator0 ---------------------------------------------------------------

class Iterator0 : public Iterator {
public:
    Iterator0();
    Object *operator()();
    void Reset();
};

//---- Iterator inlines --------------------------------------------------------

inline bool Iterator::IsActive()
{
    return activated && ! terminated;
}

inline bool Iterator::Activated()
{
    return activated;
}

inline bool Iterator::Terminated()
{
    return terminated;
}

#endif /* Iterator_First */
 
