#ifndef ObjArray_First
#ifdef __GNUG__
#pragma interface
#endif
#define ObjArray_First

#include "SeqColl.h"

extern char *cOutOfBoundsError;
extern char *cMethodName;

//---- class ObjArray --------------------------------------------------

class ObjArray: public SeqCollection {
friend class ObjArrayIter;
friend class OrdCollection;
public:  
    MetaDef(ObjArray);

    //---- creation, destruction  
    ObjArray(int s= cContainerInitCap, int lowerBound= 0); 
    ~ObjArray();
    void InitNew();              
    void Expand (int);              // expand or shrink an array
    Object *Add(Object*);
    int Index(Object*);
    void FreeAll();
    
    //---- accessing
    Iterator *MakeIterator(bool dir= cIterForward); 
    int IndexOfPtr (Object*);     // return -1 if not found     
    int IndexOf (Object*);        //        "           "
    Object *At(int i);
    Object *AtPut(int i, Object *op);
    void AtPutAndExpand(int i, Object *op);// expand the array if necessary
    Object*& operator[](int i);
    Object *UncheckedAt(int i);
    Object *RemoveAt(int i);   
    Object *Remove(Object *a);
    Object *RemovePtr(Object *a);
    int LowerBound()
	{ return lb; }
    void Sort(int upto = cMaxInt);
    int BinarySearch(Object*, int upto = cMaxInt);
	// the ObjArray has to be sorted, -1 == not found !!

    //---- comparing   
    unsigned long Hash ();
    bool IsEqual (Object*);
    int  Compare (Object*);

    //---- activation/passivation
    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);
private:
    Object **cont;
    int  lb; // lower bound of the array
    bool BoundsOk(char *where, int at);       
};

//---- ObjArrayIter -----------------------------------------------------------

class ObjArrayIter: public Iterator {
public:
    ObjArrayIter(ObjArray *s);
    ~ObjArrayIter();
    
    void Reset();
    Object *operator()();
    virtual bool Filter(Object*);
private:
    int ce;
    ObjArray *cs;
};

//---- inlines ----------------------------------------------------------------
    
inline bool ObjArray::BoundsOk(char *where, int at)
{ 
    if ( at < lb || at-lb >= Size()) {
	Error(where, cOutOfBoundsError, at, Size(), (int)this);
	return FALSE;
    }
    else
	return TRUE;
}
	
inline Object*& ObjArray::operator[](int i)      
{ 
    if ( i < lb || i-lb >= size)
	Error(cMethodName, cOutOfBoundsError, i, size, this);
    return cont[i-lb]; 
}

inline Object *ObjArray::UncheckedAt(int i)
{ 
    return cont[i]; 
}

#endif /* ObjArray_h */

