#ifdef __GNUG__
#pragma implementation
#endif

#include "ObjList.h"
#include "Class.h"
	
//---- private imports --------

#include "IterIter.h"

#include "Error.h"
#include "Storage.h"

//---- ObjLink -----------------------------------------------------------------

NewMetaImpl(ObjLink,Object, (TP(next), TP(prev), TP(obj)));

ObjLink::ObjLink(Object *op, ObjLink *lnk)
{
    obj= op;
    next= lnk;
    prev= lnk->prev;
    prev->next= this;
    next->prev= this;
}

void ObjLink::FreeAll()
{
    if ( obj ) {
	obj->FreeAll();
	SafeDelete(obj);
    }
}

#if !defined(OS_WINDOWS)
//---- memory allocation ----

void *ObjLink::operator new(size_t sz)
{
    return Storage::ChunkAlloc(sz);
}

void ObjLink::operator delete(void *vp)
{
    ::delete(vp);
}
#endif

//---- ObjList ----------------------------------------------------------------

NewMetaImpl(ObjList,SeqCollection, (TP(head)));

ObjList::ObjList()
{
    head= new ObjLink;
}

ObjList::~ObjList()
{
    register ObjLink *lnk, *tlk;

    lnk= head->Next();
    while ( lnk != head ) {
	tlk= lnk;
	lnk= lnk->Next();
	delete tlk;
    }
    delete head;
}

void ObjList::InitNew()
{
    SeqCollection::InitNew();
    head= new ObjLink;
}

void ObjList::Empty(int)
{
    register ObjLink *ll, *l;
    
    l= head->next;
    while ( l != head ) {
	ll= l;
	l= l->next;
	delete ll;
    }
    head->next= head;
    head->prev= head;
    size= 0;
    if ( HasIterators() )
	ForEachIterDo(ObjListIter, OnEmpty());
}

void ObjList::FreeAll()
{
    register ObjLink *l, *ll;

    l= head->next;
    while ( l != head ) {
	ll= l;
	l= l->next;
	ll->FreeAll();
	delete ll;
    }
    head->next= head;
    head->prev= head;
    size= 0;
    if ( HasIterators() )
	ForEachIterDo(ObjListIter, OnEmpty());
}

//-------- accessing --------

Iterator *ObjList::MakeIterator(bool dir) 
{ 
    if ( dir == cIterForward )
	return new ObjListIter(this); 
    else
	return new RevObjListIter(this);
}

//-------- primitives --------

Object *ObjList::primAt(int idx)
{
    ObjLink *lnk;
    
    lnk= LinkAt(idx);
    return lnk->GetObject();
}

Object *ObjList::primRemoveAt(int idx)
{
    Object *removed;
    ObjLink *lnk;
    
    lnk= LinkAt(idx);
    removed= lnk->GetObject();
    RemoveLink(lnk);
    
    size--;
    if ( IsObserved() )
	Changed();    
    return removed;
}

Object *ObjList::primAddAt(Object *op, int idx)
{
    ObjLink *lnk;

    lnk= LinkAt(idx);
    InsertLink(op, lnk);
    return 0;
}

Object *ObjList::primPutAt(Object *op, int idx)
{
    Object *replaced;
    ObjLink *lnk;
    
    lnk= LinkAt(idx);
    replaced= lnk->GetObject();
    lnk->SetObject(op);
    if ( op == replaced && IsObserved() )
	Changed();
    return replaced;
}

Object *ObjList::primIndexOf(Object *op, int offset, int *idx)
{
    ObjLink *lnk;
    int i, edx;
    
    lnk= FindLink(op, i);
    if ( lnk ) {
	edx= i + offset;
	if ( edx >= 0 && edx < Size() ) {
	    lnk= LinkByOffset(lnk, offset);
	    if ( idx )
		*idx= edx;
	    return lnk->GetObject();
	}
    }
    if ( idx )
	*idx= -1;
    return 0;
}

Object *ObjList::primIndexOfPtr(Object *op, int offset, int *idx)
{
    ObjLink *lnk;
    int i, edx;
    
    lnk= FindLinkPtr(op, i);
    if ( lnk ) {
	edx= i + offset;
	if ( edx >= 0 && edx < Size() ) {
	    lnk= LinkByOffset(lnk, offset);
	    if ( idx )
		*idx= edx;
	    return lnk->GetObject();
	}
    }
    if ( idx )
	*idx= -1;
    return 0;
}

Object *ObjList::AddRelative(Object *op, EqualType eq, int offset, Object *to)
{
    ObjLink *lnk;
    int idx, edx;

    if ( eq )
	lnk= FindLink(to, idx);
    else
	lnk= FindLinkPtr(to, idx);
    if ( lnk ) {
	edx= idx + offset;
	if ( edx >= 0 && edx <= Size() ) {
	    lnk= LinkByOffset(lnk, offset);
	    InsertLink(op, lnk);
	    return 0;
	} else
	    return op;
    }
    return op;    
}

Object *ObjList::RemoveRelative(Object *op, EqualType eq, int offset)
{
    ObjLink *lnk;
    int idx, edx;
    Object *removed;

    if ( eq )
	lnk= FindLink(op, idx);
    else
	lnk= FindLinkPtr(op, idx);
    if ( lnk ) {
	edx= idx + offset;
	if ( edx >= 0 && edx < Size() ) {
	    lnk= LinkByOffset(lnk, offset);
	    removed= lnk->GetObject();
	    RemoveLink(lnk);
	    size--;
	    if ( IsObserved() )
		Changed();
	    return removed;
	} else
	  return 0;
    }
    return 0;
}

ObjLink *ObjList::NewLink(Object *op, ObjLink *prev)
{
    return new ObjLink(op, prev);
}

void ObjList::DeleteLink(ObjLink *lnk)
{
    delete lnk;
}

//-------- implementation --------

ObjLink *ObjList::FindLink(Object *op, int &idx)
{
    register ObjLink *lnk;
    
    idx= 0;
    lnk= head->next;
    while ( lnk != head ) {
	if ( lnk->GetObject()->IsEqual(op) ) 
	    return lnk;
	lnk= lnk->next;
	idx++;
    }
    return 0;
}

ObjLink *ObjList::FindLinkPtr(Object *op, int &idx)
{
    register ObjLink *lnk;
   
    idx= 0; 
    lnk= head->next;
    while ( lnk != head ) {
	if ( lnk->GetObject() == op ) 
	    return lnk;
	lnk= lnk->next;
	idx++;
    }
    return 0;
}

ObjLink *ObjList::LinkAt(int idx)
{
    register ObjLink *lnk;
    register int i;

    i= 0;
    lnk= head->Next();
    while ( i < idx && lnk != head ) {
	lnk= lnk->Next();
	i++;
    }
    return lnk;
}

ObjLink *ObjList::LinkByOffset(ObjLink *lnk, int offset)
{
    if ( offset >= 0 ) {
	while ( offset > 0 ) {
	    offset--;
	    lnk= lnk->Next();
	}
    } else {
	while ( offset < 0 ) {
	    offset++;
	    lnk= lnk->Prev();
	}
    }
    return lnk;
}

void ObjList::RemoveLink(ObjLink *lnk)
{
    if ( HasIterators() )
	ForEachIterDo(ObjListIter, OnRemove(lnk));
    lnk->prev->next= lnk->next;
    lnk->next->prev= lnk->prev;
    DeleteLink(lnk);
}

void ObjList::InsertLink(Object *op, ObjLink *prev)
{
    NewLink(op, prev);
    size++;
    if ( IsObserved() )
	Changed();
}

//---- ObjListIter ------------------------------------------------------------

ObjListIter::ObjListIter(ObjList *s)
{
    list= s;
    ce= 0;
}

ObjListIter::~ObjListIter()
{
    Terminate();
}

void ObjListIter::Reset()
{
    Iterator::Reset();
    ce= 0;
}

Object* ObjListIter::operator()()
{
    Object *op;

    if ( ! Activated() )
	if ( Activate() )
	    ce= list->head;
    if ( ! Terminated() ) { 
	ce= ce->Next();
	op= ce->GetObject();
	if ( op )
	    return op; 
	else {
	    Terminate();
	    return 0;
	}
    } else
	return 0;
}

Container *ObjListIter::GetContainer()
{
    return list;
}

void ObjListIter::Detach()
{
    list= 0;
    Iterator::Detach();
}

void ObjListIter::OnEmpty()
{
    if ( IsActive() )
	ce= list->head;
}

void ObjListIter::OnRemove(ObjLink *lnk)
{
    if ( IsActive() && lnk == ce )
	ce= ce->prev;
}

//---- RevObjListIter ---------------------------------------------------------

RevObjListIter::RevObjListIter(ObjList *s) : ObjListIter(s)
{ 
}

Object* RevObjListIter::operator()()
{
    Object *op;

    if ( ! Activated() )
	if ( Activate() )
	    ce= list->head;
    if ( ! Terminated() ) { 
	ce= ce->Prev();
	op= ce->GetObject();
	if ( op )
	    return op; 
	else {
	    Terminate();
	    return 0;
	}
    } else
	return 0;
}

void RevObjListIter::OnRemove(ObjLink *lnk)
{
    if ( IsActive() && lnk == ce )
	ce= ce->Next();
}

//------------------------------------------------------------------------------

