#ifndef Set_First
#ifdef __GNUG__
#pragma interface
#endif
#define Set_First

#include "Collection.h"

class HashTable;
class SetIter;

//---- Set ---------------------------------------------------------------------

class Set : public Collection {
friend class SetIter;
friend class RobustHashTable;
public:
    MetaDef(Set);
    Set(int initCapacity= cContainerInitCap);
    ~Set();
    void InitNew();
    void Empty(int newCapacity);
    void FreeAll();
    int Capacity();

    //---- manipulation
    Object *Add(Object *op);
    Object *Remove(Object *op);
    Object *RemovePtr(Object *op);
    
    //---- access 
    Iterator *MakeIterator(bool= cIterForward);    
    Object *Find(Object *op);
    Object *FindPtr(Object *op);
    int OccurrencesOf(Object *op);
    int OccurrencesOfPtr(Object *op);
    
    //---- statistics
    double AverageCollisions();  
protected:
    //---- initialization
    Set(Set *);
    virtual void Init(int initCapacity);

    //---- robust iterators
    void SaveItersPosition(int forPix);
    void ResetItersPosition();
protected:
    class HashTable *ht;
};

//---- SetIter ------------------------------------------------------------------

class SetIter : public Iterator {
friend class Set;
public:
    SetIter(Set *aSet);
    ~SetIter();
    
    Object *operator()();
    void Reset();
protected:
    Container *GetContainer();
    void SavePosition(int pix);
    void ResetPosition();
protected:
    Set *set;
    HashTable *ht;
    int pix;
    Object *saved;      
};

//---- IdSet -------------------------------------------------------------------

class IdSet : public Set {
public:
    MetaDef(IdSet);
    IdSet(int initCapacity= cContainerInitCap); 
    void InitNew();
protected:
    void Init(int aCapacity);
};

#endif /* Set_First */

