#include "Converter.h"

#include "Class.h"
#include "Data.h"
#include "Text.h"
#include "TypeMatcher.h"

//---- AsciiTypeMatcher -----------------------------------------------------

class AsciiTypeMatcher: public TypeMatcher {
public:
    AsciiTypeMatcher() : TypeMatcher(cDocTypeAscii, 0, 0, TRUE, 9999)
	{ }
    bool MatchContents(Data*, char*, int);
    bool MatchPathName(Data*, char*, int);
};

inline    bool AsciiTypeMatcher::MatchPathName(Data*, char*, int)
	{ return FALSE; }
	
static AsciiTypeMatcher asciimatcher;

bool AsciiTypeMatcher::MatchContents(Data*, char *p, int l)
{
    unsigned char *bp = (unsigned char *)p;
    for ( ; l > 0; l--, bp++)
	if ( ! Isascii(*bp) && ! Isprint(*bp) )
	    return FALSE;
    return TRUE;
}

//---- AsciiConverter ------------------------------------------------------

class AsciiConverter : public Converter {
public:
    MetaDef(AsciiConverter);
    AsciiConverter()
	{ }
    char *AsString();
    Object *Convert(Data *data, Class *want);
    bool CanConvert(Data *data, Class *want);    
    bool CanConvertTo(Object *op);
    bool ConvertTo(Data *data, Object *op);
};

inline    char *AsciiConverter::AsString()
	{ return "ASCII Text"; }
inline   bool AsciiConverter::CanConvert(Data *data, Class *want)
	{ return (data->Type() == cDocTypeAscii || data->IsAscii())
						&& want->isKindOf(Meta(Text)); }
inline  bool AsciiConverter::CanConvertTo(Object *op)
	{ return op->IsKindOf(Text); }  
		
NewMetaImpl0(AsciiConverter, Converter);

Object *AsciiConverter::Convert(Data *data, Class *want)
{
    Text *t= (Text*)want->Proto()->New();
    IStream from(data->GetStreamBufForReading(), 0, TRUE);
    t->ReadFromAsPureText(from, data->SizeHint());
    return t;
}

bool AsciiConverter::ConvertTo(Data *data, Object *op)
{
    OStream to(data->GetStreamBufForWriting(), TRUE);
    ((Text*)op)->PrintOnAsPureText(to);
    return TRUE;
}


void RefAsciiConverter()
{
    Class *cl= Meta(AsciiConverter);
    TypeMatcher *tm = &asciimatcher;
}
