#ifndef CheapText_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define CheapText_First

#include "Text.h"

//---- CheapText ------------------------------------------

class CheapText: public Text {
public:
    MetaDef(CheapText);
    CheapText(int s= 16, Font *fd= gSysFont);
    CheapText(byte *buf, int len= -1, Font *fd= gSysFont);
    ~CheapText();

    //---- editing
    void CopyInStr(byte *str,int strSize, int from, int to);
    void AddChar(int at, byte b);

    //---- accessing
    byte& operator[](int i);      
    int Size();
    int Search(class RegularExp *rex, int *nMatched, int start= 0, 
	     int range= cMaxInt, bool dir= cSearchForward);

    //---- iterators
    TextIter *MakeIterator(int from=0, int to= cMaxInt);

    //---- TextPainter contract
    byte *GetLineAccess(byte **buf, int from, int to);

    //---- input/output
    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);
    OStream& PrintOnAsPureText(OStream &s);
    IStream& ReadFromAsPureText(IStream&, long sizeHint= -1);

protected:
    void InitNew();
    void Terminate();
    void InsertBytes(byte *b, int size, int from, int to);
    void ReplaceRange(int from, int to, Text *src, int sfrom, int sto);
    Text *MakeScratchText(byte *buf, int size); 
    
    byte CharAt(int i);

private:                                                           
friend class CheapTextIter;
    int next;                               // next free slot
    int size;                               // size of allocated memory
    byte *cont;

    void Init(int, Font*);
    void Expand(int);    
    bool HighWaterMark(int n);
    bool LowWaterMark(); 

};

inline bool CheapText::HighWaterMark(int n)
{ 
    return (bool) (cont == NULL || next+n >= size); 
}

inline bool CheapText::LowWaterMark()                    
{ 
    return (bool) (next < size/5); 
}  

inline byte CheapText::CharAt(int i)
{ 
    return cont[i]; 
}

//---- class CheapTextIter -------------------------------------------------

class CheapTextIter: public TextIter {
    Font *font;
public:
    CheapTextIter(Text *s, int from= 0, int to= cMaxInt);
    int operator()(int *width, LineDesc* l);// returns width of the character too
    int Token(int *width, LineDesc* l = 0); // returns next token and its size
};

#endif
