#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma implementation
#endif

#include "CodeTextView.h"

#include "Class.h"
#include "VObjectText.h"
#include "RunArray.h"
#include "Env.h"
#include "ETMath.h"
#include "WindowSystem.h"

static char class_keyword[] = "class";
static char struct_keyword[] = "struct";

static char obrackets[] = "({[\"",
	    cbrackets[] = ")}]\"";
 
static Ink *commentInk= 0;
static Ink *functionInk= 0;
static Ink *classInk= 0;

//---- CodeTextView ------------------------------------------------------------

NewMetaImpl(CodeTextView,VObjectTextView, (TB(autoIndent), T(cursorPos)));

CodeTextView::CodeTextView(
    EvtHandler *eh, Rectangle r, VObjectText *t, 
    bool w, TextViewFlags f, Point b, 
    TViewAlign ta, int id
) : VObjectTextView(eh, r, t, w, f, b, ta, id) 
{
    SetStopChars("\n\r");
    autoIndent= Env::GetValue("CodeText.AutoIndent", TRUE);
    SetupStyles(t->GetFont());
    ExitCursorMode();
}   

CodeTextView::~CodeTextView()
{
}

Command *CodeTextView::DoKeyCommand(int ch, Token token)
{
    Command *cmd;

    ExitCursorMode();
    if ((ch == '\r' || ch == '\n') && autoIndent) { // auto indent
	int from, to, i;
	Text *t= GetText();
	GetSelection(&from, &to);
	scratchText->Empty();
	scratchText->Append('\n'); 
	for (i= StartLine(CharToLine(from)); i < from && i < t->End(); i++) {
	    ch= (*t)[i];
	    if (ch == ' ' || ch == '\t')
		scratchText->Append(ch);
	    else
		break;
	}
	cmd= InsertText(scratchText);
	scratchText->Empty();
	RevealSelection();
	return cmd;
    }
    return TextView::DoKeyCommand(ch, token);
}

void CodeTextView::SetFont(Font *fp)
{
    SetupStyles(fp);
    TextView::SetFont(fp);
}

void CodeTextView::SetupStyles(Font *fp)
{
    int declsize= fp->Size()
		  + Env::GetValue("CodeText.DeclarationSizeIncrement", 2);
    if (commentInk == 0) {
	if (gColor) {
	    commentInk= new_RGBColor(0, 0, 255);
	    Env::Bind(commentInk, "CodeText.CommentColor");
	    functionInk= new_RGBColor(0, 0, 0);
	    Env::Bind(functionInk, "CodeText.FunctionColor");
	    classInk= new_RGBColor(0, 0, 0);
	    Env::Bind(classInk, "CodeText.ClassColor");
	} else
	    commentInk= functionInk= classInk= gInkBlack; 
    }   
		  
    commentStyle= new_CharStyle(
	new_Font(fp->Fid(), fp->Size(), eFaceItalic), commentInk
    );
    functionStyle= new_CharStyle(
	new_Font(fp->Fid(), declsize, eFaceBold), functionInk
    );
    classDeclStyle= new_CharStyle(
	new_Font(fp->Fid(), declsize, eFaceBold), classInk
    );
    plainStyle= new_CharStyle(
	new_Font(fp->Fid(), fp->Size(), eFacePlain)
    );
}

void CodeTextView::SetAutoIndent(bool b)
{
    autoIndent= b;
}
    
bool CodeTextView::GetAutoIndent()
{
    return autoIndent;
}

void CodeTextView::ExitCursorMode()
{
    cursorPos= gPoint_1;
}

Command *CodeTextView::DoLeftButtonDownCommand(Point p, Token t, int cl)
{
    Point sp= p;
    ExitCursorMode();
    if (!Enabled())
	return gNoChanges;
    if (cl >= 2) {
	int line, cpos;
	char *br;
	Point pp;
	Text *text= GetText();

	p-= GetInnerOrigin();
	PointToPoint(p, &pp, &line, &cpos);
	
	if (cpos > 0 && (br= strchr(obrackets,(*text)[cpos-1]))) {
	    MatchBracketForward(cpos, *br, cbrackets[br-obrackets]);
	    return gNoChanges;
	}
	if (cpos < text->End() && (br= strchr(cbrackets,(*text)[cpos]))) {
	    MatchBracketBackward(cpos-1, obrackets[br-cbrackets], *br);
	    return gNoChanges;
	}
    }
    return TextView::DoLeftButtonDownCommand(sp, t, cl);
}

Command *CodeTextView::DoMenuCommand(int m)
{
    ExitCursorMode();
    return TextView::DoMenuCommand(m);
}
    
Command *CodeTextView::DoOtherEventCommand(Point p, Token t)
{
    ExitCursorMode();
    return TextView::DoOtherEventCommand(p, t);
}

void CodeTextView::MatchBracketForward(int from, int obracket, int cbracket)
{
    Text *text= GetText();
    int ch, stack= 0;

    for (int i= from; i < text->End(); i++) {
	ch= (*text)[i];
	if (ch == cbracket) {
	    if (stack-- == 0) 
		break;
	} else if (ch == obracket)
	    stack++;
    }    
    SetSelection(from, i, TRUE);
}

void CodeTextView::MatchBracketBackward(int from, int obracket, int cbracket)
{
    Text *text= GetText();
    int ch, stack= 0;

    for (int i= from; i >= 0; i--) {
	ch= (*text)[i];
	if (ch == obracket) { 
	    if (stack-- == 0) 
		break;
	} else if (ch == cbracket)
	    stack++;
    }
    SetSelection(i+1, from+1, TRUE);
}

Command *CodeTextView::DoCursorKeyCommand(EvtCursorDir cd, Token t)
{
    if (cd == eCrsLeft || cd == eCrsRight)
	ExitCursorMode();
    return TextView::DoCursorKeyCommand(cd, t);
}

int CodeTextView::CursorPos(int at, int line, EvtCursorDir d, Point p)
{    
    int charNo;
    Point basePoint;
    
    if (cursorPos == gPoint_1) 
	CharToPoint(at, &line, &cursorPos);

    if (d == eCrsDown)
	line= Math::Min(nLines-1, line+1);
    else
	line= Math::Max(0, line-1);        
    basePoint= LineToPoint(line, TRUE) + Point(cursorPos.x, 0);
    PointToPoint(basePoint, &p, &line, &charNo);
    return charNo;
}

PrettyPrinter *CodeTextView::MakePrettyPrinter(
    Text *t, CharStyle *cs, CharStyle *fs, CharStyle *cds, CharStyle *ps
) 
{
    return new PrettyPrinter(t, cs, fs, cds, ps);
}

void CodeTextView::FormatCode()
{
    if (!text->IsKindOf(StyledText)) 
	return;
    PrettyPrinter *pp= MakePrettyPrinter(
	text, commentStyle, functionStyle, classDeclStyle, plainStyle
    );
    if (pp)
	pp->Doit();
    SafeDelete(pp);
    needsReformat= FALSE;
}

void CodeTextView::SetDefaultStyle()
{
    if (!text->IsKindOf(StyledText)) 
	return;
    StyledText *stext= (StyledText*)text;
    RunArray *st= new RunArray();
    st->Insert(plainStyle, 0, 0, stext->Size());
    RunArray *tmpp= stext->SetCharStyles(st);
    delete tmpp;
}

void CodeTextView::DoObserve(int id, int part, void *what, Object *op)
{
    TextView::DoObserve(id, part, what, op);
    if (op == GetText()) {
        if (part == eTextDeleted || eTextReplaced) 
            InvalidateFormatting();
    }
}

void CodeTextView::InvalidateFormatting()
{
    needsReformat= TRUE;
}

bool CodeTextView::NeedsReformat()
{
    return needsReformat;
}

void CodeTextView::Reformat()
{
    FormatCode();
    if (NumberOfLines() > 500) // cheat
        ForceRedraw();
    else 
        RepairAll();
}

//---- CodeAnalyzer ----------------------------------------------------------

CodeAnalyzer::CodeAnalyzer(Text *t)
{
    text= t;
}

void CodeAnalyzer::Doit()
{
    bool inDefine;
    bool escape;
    bool inClass;
    int lastComment;
    int braceStack;
    int inString;
    int prevCh;
    int canBeClass;
    int canBeFunction;
    int c;

    inDefine= escape= inClass= FALSE;
    lastComment= braceStack= inString= 0;
    line = 1;
    prevCh= '\n'; 
    canBeClass= canBeFunction= -1;

    AutoTextIter next(text, 0, text->Size());
    Start();
    
    while ((c= next(0, 0)) != cEOT) {
	if (escape)
	{
	    if (c == '\n')
	    	line++;
	    escape=FALSE;
	}
	else
	    switch (c) {
	    case '#':
		if (prevCh == '\n')
		    inDefine= TRUE;
		break;
	    case '\\':
		escape= TRUE;
		break;
	    case '\n':
		inDefine= FALSE;
		line++;
		break;
	    case '*':
		if (prevCh == '/' && inString == 0) {
		    if (canBeFunction == -1 && canBeClass == -1)
			lastComment= next->GetPos(); 
		    FoundComment(&next);
		    prevCh= 0;
		    continue;
		}
		break;
	    case '/':
		if (prevCh == '/' && inString == 0) {
		    if (canBeFunction == -1 && canBeClass == -1)
			lastComment= next->GetPos(); 
		    FoundEndOfLineComment(&next);
		    prevCh= 0;
		    continue;
		}
		break;
	    case ';':
		canBeFunction= canBeClass= -1;
		break;
	    case '{':
		if (canBeFunction != -1) {
		    FoundFunctionOrMethod(canBeFunction, lastComment);
		    canBeFunction= -1;
		}
		if (canBeClass != -1) {
		    FoundClassDecl(canBeClass);
		    canBeClass= -1;
		}
		if (inString == 0)
		    braceStack++;
		break;
	    case '}':
		if (inString == 0)
		    braceStack--;
		break;
	    case '(':
		if (!inDefine && inString == 0) {
		    if (braceStack == 0 && canBeFunction == -1)
			canBeFunction= next->GetPos()-1;
		    braceStack++;
		}
		break;
	    case ')':
		if (!inDefine && inString == 0)
		    braceStack--;
		break;
	    case '\'':
	    case '\"':
		if (inString == 0) {
		    if (c == '\"')
			inString= '\"';
		    else
			inString= '\'';
		} else {
		    if ((inString == '\"' && c == '\"') || (inString == '\'' 
							 && c == '\''))
			inString= 0;
		}
		break;
	    default:
		if ((c == 'c' || c == 's') && !Isinword(prevCh)
		    && !inDefine && inString == 0 
		    && !inClass && canBeClass == -1) {
		    if (braceStack == 0)
			if (inClass= IsClassDecl(next->GetPos()-1))
			    next->SetPos(next->GetPos() - 1
				+ (c == 'c' ? strlen(class_keyword)
					    : strlen(struct_keyword)));
		}
		else if (inClass && Isinword(c)) {
		    inClass= FALSE;
		    canBeClass= next->GetPos();
		}
		break;
	}
	prevCh= c;
    }
    End();
}

void CodeAnalyzer::FoundComment(AutoTextIter *next)
{
    int c, prevCh= 0, l= line, start= (*next)->GetPos()-2;
   
    while ((c= (*next)(0, 0)) != cEOT) {
	if (c == '\n')
	    line++;
	if (c == '/' && prevCh == '*') {
	    Comment(l, start, (*next)->GetPos());
	    break;
	}
	prevCh= c;
    }
}

void CodeAnalyzer::FoundEndOfLineComment(AutoTextIter *next)
{
    int c, start= (*next)->GetPos()-2;

    while ((c= (*next)(0, 0)) != cEOT) 
	if (c == '\n') {
	    Comment(line, start, (*next)->GetPos()-1);
	    line++;
	    break;
	}
}

void CodeAnalyzer::FoundFunctionOrMethod(int at, int lastComment)
{
    byte c;
    int pos= at, len= 0;
    while (--pos >= lastComment) {
	c= (*text)[pos];
	if (!Isspace(c) && !strchr("[]*+-=%&><|:^%()~/!",c)) // overloaded operators
	    break;
    }
    while (pos >= lastComment) {
	c= (*text)[pos];
	if (!(Isinword(c) || c == ':' || c == '~' ))
	    break;
	if (pos-1 < 0)
	    break;
	pos--;
	len++;
    }
    
    if (len) {
	char buf[500];
	pos++;
	text->CopyInStr((byte*)buf, sizeof buf, pos, pos+len);
	char *p= strchr(buf, ':');
	if (!p)
	    Function(line, pos, pos+len, buf, 0);
	else {
	    *p= '\0';
	    char *pp= p+2;
	    Function(line, pos, pos+len, pp, buf);
	}
    }
}

bool CodeAnalyzer::IsClassDecl(int at)
{
    char *p;
    int c;
    AutoTextIter next(text, at, text->Size());

    if ((c= next(0, 0)) == 'c')
	p= &class_keyword[1];
    else
	p= &struct_keyword[1];

    while ((c= next(0, 0)) != cEOT && *p)
	if (c != *p++) 
	    return FALSE;

    if (c == cEOT || !Isspace(c))
	return FALSE;

    return TRUE;
}

void CodeAnalyzer::FoundClassDecl(int at)
{
    int end, c, start= at-1;
    AutoTextIter next(text, start, text->Size());

    while ((c= next(0, 0)) != cEOT) 
	if (!Isinword(c)) 
	    break;
    end= next->GetPos()-1;
    if (start != end) {
	static byte name[1000];
	text->CopyInStr(name, sizeof name, start, end);
	ClassDecl(line, start, end, (char*)name);
    }
}

void CodeAnalyzer::Start()
{
}

void CodeAnalyzer::End()
{
}

void CodeAnalyzer::Comment(int, int, int)
{
}

void CodeAnalyzer::ClassDecl(int, int, int, char *)
{
}
    
void CodeAnalyzer::Function(int, int, int, char *, char *)
{
}

//---- PrettyPrinter ---------------------------------------------------------

PrettyPrinter::PrettyPrinter(
    Text *t, CharStyle *cs, CharStyle *fs, CharStyle *cds, CharStyle *ps
) : CodeAnalyzer(t)
{
    stext= Guard(t, StyledText);
    plainStyle= ps;
    commentStyle= cs;
    classDeclStyle= cds;
    functionStyle= fs;
}

void PrettyPrinter::Start()
{
    st= new RunArray();
    st->Insert(plainStyle, 0, 0, stext->Size());
}

void PrettyPrinter::End()
{
    RunArray *tmpp= stext->SetCharStyles(st);
    delete tmpp;
}

void PrettyPrinter::Comment(int, int start, int end)
{
    st->Insert(commentStyle, start, end, end-start);
}
 
void PrettyPrinter::ClassDecl(int, int start, int end, char *)
{
    st->Insert(classDeclStyle, start, end, end-start);
}
    
void PrettyPrinter::Function(int, int start, int end, char *, char *)
{
    st->Insert(functionStyle, start, end, end-start);
}
