#ifndef _CodeTextView_h_
#define _CodeTextView_h_

#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma interface
#endif

#include "VObjectTView.h"
#include "TextFormatter.h"

//---- CodeTextView --------------------------------------------------------
// a TextView for editing programm text
// adds: auto indenting
//       pretty printing
//       highlighting matching brackets
//       (double clicking on a bracket selects the text enclosed by the brackets)

class PrettyPrinter;
class CharStyle;

class CodeTextView: public VObjectTextView {
public:
    MetaDef(CodeTextView);

    CodeTextView(
	EvtHandler *eh, Rectangle r, VObjectText *t, 
	bool wrap= TRUE,
	TextViewFlags= eTextViewDefault, 
	Point border= gBorder, 
	TViewAlign m= eTViewLeft,
	int id= cIdNone
    ); 
    ~CodeTextView();

    //---- formatting of source code
    virtual PrettyPrinter *MakePrettyPrinter(
	Text *t, CharStyle *cs, CharStyle *fs, CharStyle *cds, CharStyle *ps
    );
    void FormatCode();
    bool NeedsReformat();
    void Reformat();
    void InvalidateFormatting();
    void SetDefaultStyle();
    void SetFont(Font *fp);

    //---- controller methods 
    Command *DoLeftButtonDownCommand(Point, Token, int); // matching brackets
    Command *DoKeyCommand(int, Token);      // auto indent
    Command *DoMenuCommand(int);
    Command *DoOtherEventCommand(Point p, Token t);
    Command *DoCursorKeyCommand(EvtCursorDir, Token);    
    void DoObserve(int id, int part, void *what, Object *op);
    void SetAutoIndent(bool);
    bool GetAutoIndent();

protected:
    bool autoIndent;
    Point cursorPos;
    bool needsReformat;

    CharStyle *commentStyle;
    CharStyle *functionStyle;
    CharStyle *classDeclStyle;
    CharStyle *plainStyle;

protected:
    void MatchBracketForward(int from, int obracket, int cbracket);
    void MatchBracketBackward(int from, int obracket, int cbracket);
    void ExitCursorMode();
    int CursorPos(int ch, int line, EvtCursorDir d, Point);
    void SetupStyles(Font *fp);
    
};

//---- CodeAnalyzer --------------------------------------------------------

class CodeAnalyzer {
    int line;

protected:
    virtual void FoundComment(AutoTextIter *next);
    virtual void FoundEndOfLineComment(AutoTextIter *next);
    virtual void FoundFunctionOrMethod(int at, int lastComment);
    virtual void FoundClassDecl(int at);
    virtual bool IsClassDecl(int at);

    Text *text;

    virtual void Start();
    virtual void Comment(int line, int start, int end);
    virtual void ClassDecl(int line, int start, int end, char *name);
    virtual void Function(
	int line, int start, int end, char *name, char *classname
    );
    virtual void End();

public:
    CodeAnalyzer(Text *t);

    virtual void Doit();
};

//---- PrettyPrinter --------------------------------------------------------

class PrettyPrinter: public CodeAnalyzer {
protected:
    class StyledText *stext;
    class RunArray *st;
    CharStyle *commentStyle;
    CharStyle *functionStyle;
    CharStyle *classDeclStyle;
    CharStyle *plainStyle;

protected:
    void Start();
    void Comment(int line, int start, int end);
    void ClassDecl(int line, int start, int end, char *name);
    void Function(int line, int start, int end, char *name, char *classname);
    void End();

public:
    PrettyPrinter(
	Text *t, CharStyle *cs, CharStyle *fs, CharStyle *cds, CharStyle *ps
    );
};

#endif
