#ifdef __GNUG__
#pragma implementation
#endif

#include "CollView.h"
#include "Window.h"
#include "WindowSystem.h"
#include "ETString.h"
#include "OrdColl.h"
#include "CmdNo.h"
#include "Error.h"
#include "ETMath.h"
#include "Class.h"
#include "Menu.h"

extern bool testopen;
			
//---- CollectionView ----------------------------------------------------------

NewMetaImpl(CollectionView,View, (TP(coll), T(selection), T(gap), T(minExtent),
					    T(rows), T(cols), TP(defaultItem)));

CollectionView::CollectionView(EvtHandler *eh, SeqCollection *m, CollViewOptions o,
					    int r, int c) : View(eh, gRect0)
{
    incr_match = new byte [1024];
    incr_match[0] = 0;
    SetFlag(o);
    defaultItem= new VObject();
    rows= r;
    cols= c;
    sync_selection = FALSE;
    menu=0;
    SetCollection(m);
}

CollectionView::~CollectionView()
{
    if (coll) {
	coll->RemoveObserver(this);
	coll->FreeAll();
	SafeDelete(coll);
    }
    SafeDelete(xPos);
    SafeDelete(yPos);
    SafeDelete(defaultItem);
    SafeDelete(incr_match);
    SafeDelete(menu);
}

//---- init methods

void CollectionView::SetSelection(Rectangle newsel, bool sendcontrol)
{
    if (selection != newsel) {
	if (IsOpen()) {
	    InvalidateRect(ItemRect(selection));
	    InvalidateRect(ItemRect(newsel));
	}
	selection= newsel;
    }
    incr_match[0] = 0;
    if (sendcontrol)
	DoSelect(selection, WindowSystem::Clicks);
}

void CollectionView::SetSelectionWithClicks(Rectangle newsel, bool sendcontrol, int nrOfClicks)
{
    if (selection != newsel) {
	if (IsOpen()) {
	    InvalidateRect(ItemRect(selection));
	    InvalidateRect(ItemRect(newsel));
	}
	selection= newsel;
    }
    incr_match[0] = 0;
    if (sendcontrol)
	DoSelect(selection, nrOfClicks);
}

void CollectionView::SetColsRows(Point cr)
{
    cols= cr.x;
    rows= cr.y;
    Modified();
}

void CollectionView::DoSelect(Rectangle r, int clicks)
{
    if (r.IsNotEmpty() && (clicks == 1 || clicks == 2)) {
	int partcode= clicks >= 2 ? cPartCollDoubleSelect : cPartCollSelect;
	Control(GetId(), partcode, (void*) r.origin.y);
    }
}

void CollectionView::ClearSelection(bool)
{
    SetSelection(gRect0);
    Control(GetId(), cPartCollSelect, (void *)-1);
}

void CollectionView::SetCollection(SeqCollection *m, bool freeold)
{
    if (coll) {
	coll->RemoveObserver(this);
	{
	    coll->ForEach(VObject,ClearContainer)(this);
	}
	if (freeold) {
	    coll->FreeAll();
	    delete coll;
	}
    }
    coll= m;
    if (coll) {
	coll->assertclass(Meta(VObject));
	{
	    coll->ForEach(VObject,SetContainer)(this);
	}
	if (IsOpen()) {
	    coll->ForEach(VObject,Open)(TRUE);
	}
	coll->AddObserver(this);
    }
    selection= gRect0;
    incr_match[0] = 0;
    Modified();
    //if (IsOpen())
    //    Scroll(cPartScrollAbs, gPoint0, FALSE);
}

void CollectionView::SetDefaultItem(class VObject *d)
{ 
    if (defaultItem)
	delete defaultItem;
    defaultItem= d;
}

void CollectionView::SetMinExtent(Point e)
{
    if (minExtent != e) {
	minExtent= e;
	Modified();
    }
}

Metric CollectionView::GetMinSize()
{
    Update();
    return GetExtent();
}

void CollectionView::Open(bool m)
{
    View::Open(m);
    if (testopen) {
	if (m) {
	    if (coll) {
		coll->ForEach(VObject,Open)(TRUE);
	    }
	} else {
	    if (!IsOpen() && coll) {
		coll->ForEach(VObject,Open)(FALSE);
	    }
	}
    }
}

//---- layout

void CollectionView::ConstrainScroll(Point *p)
{
/*
    if (xPos) {
	bool outside;
	Point pp= PointToItem(*p, &outside);
	if (! outside)
	    p->y= ItemRect(0, pp.y).origin.y;
    }
*/
}

void CollectionView::SetOrigin(Point at)
{
    View::SetOrigin(at);
    if (coll && coll->Size() > 0) {
	register int x, y;
	at+= gap;
	for (x= 0; x < cols; x++)
	    for (y= 0; y < rows; y++)
		GetItem(x, y)->SetOrigin(Point(xPos[x]+at.x, yPos[y]+at.y));
    }
}

void CollectionView::Update()
{
    register VObject *gop;
    register int x, y, ww, hh, w, h;
    int xpos= 0, ypos= 0, sz;
    Point g;

    if (!TestFlag(eCVModified))
	return;
    ResetFlag(eCVModified);

    if (coll == 0 || coll->Size() <= 0) {
	ForceRedraw();
	SetExtent(gPoint1);
	ForceRedraw();
	return;
    }
    
    if (testopen) {
	{
	    coll->ForEach(VObject,SetContainer)(this);
	}
	if (IsOpen()) {
	    coll->ForEach(VObject,Open)(TRUE);
	}
    }
    
    sz= coll->Size();
    if (rows <= 0)
	SetFlag(eCVExpandRows);
    if (cols <= 0)
	SetFlag(eCVExpandCols);
    
    if (TestFlag(eCVExpandRows) && TestFlag(eCVExpandCols)) {
	cols= Math::Sqrt(sz);
	rows= (sz+cols-1) / cols;
    } else if (TestFlag(eCVExpandRows))
	rows= (sz+cols-1) / cols;
    else if (TestFlag(eCVExpandCols))
	cols= (sz+rows-1) / rows;

    yPos= (int*) Storage::ReAlloc(yPos, (rows+1) * sizeof(int));
    xPos= (int*) Storage::ReAlloc(xPos, (cols+1) * sizeof(int));

    xpos= ypos= 0;
    g= 2*gap;
    if (TestFlag(eCVGrid))
	g+= gPoint1;
    
    for (x= 0; x < cols; x++) {
	xPos[x]= xpos;
	ww= minExtent.x;
	for (y= 0; y < rows; y++) {
	    gop= GetItem(x, y);
	    gop->SetContainer(this);
	    if (!testopen) {
		if (IsOpen())
		    gop->Open();
	    }
	    gop->CalcExtent();
	    w= gop->Width();
	    ww= Math::Max(ww, w);
	}
	xpos+= ww+g.x;
    }
    xPos[x]= xpos;

    for (y= 0; y < rows; y++) {
	yPos[y]= ypos;
	hh= minExtent.y;
	for (x= 0; x < cols; x++) {
	    h= GetItem(x, y)->Height();
	    hh= Math::Max(hh, h);
	}
	ypos+= hh+g.y;
    }
    yPos[y]= ypos;

    for (x= 0; x < cols; x++)
	for (y= 0; y < rows; y++)
	    GetItem(x, y)->SetContentRect(ItemRect(x, y).Inset(gap), FALSE);

    if (TestFlag(eCVGrid)) {
	xpos--;
	ypos--;
    }
    
    ForceRedraw();
    SetExtent(Point(xpos, ypos));
    ForceRedraw();
}

//---- update

void CollectionView::DoObserve(int, int part, void *, Object *op)
{
    if (part == cPartSenderDied)
	coll= 0;
    else if (op == coll)
	Modified();
}

void CollectionView::Modified()
{
    SetFlag(eCVModified);
    if (IsOpen())
	Update();
}

//---- mapping

Point CollectionView::PointToItem(Point p, bool *outside)
{
    register int x, y;
    register bool out= FALSE;
    
    if (coll == 0 || coll->Size() <= 0) {
	if (outside)
	    *outside= TRUE;
	return gPoint0;
    }
    Update();
    p-= GetOrigin();
    if (p.x < xPos[0]) {
	out= TRUE;
	x= 0;
    } else if (p.x >= xPos[cols]) {
	out= TRUE;
	x= cols-1;
    } else {
	for (x= 0; x < cols; x++)
	    if (p.x >= xPos[x] && p.x < xPos[x+1])
		break;
    }
    if (p.y < yPos[0]) {
	out= TRUE;
	y= 0;
    } else if (p.y >= yPos[rows]) {
	out= TRUE;
	y= rows-1;
    } else {
	for (y= 0; y < rows; y++)
	    if (p.y >= yPos[y] && p.y < yPos[y+1])
		break;
    }
    if (outside)
	*outside= out;
    return Point(x, y);
}

Rectangle CollectionView::ItemRect(Rectangle r)
{
    Rectangle rr;

    if (r.IsEmpty())
	return gRect0;
    Update();
    rr.origin.x= xPos[r.origin.x];
    rr.origin.y= yPos[r.origin.y];
    rr.extent.x= xPos[r.origin.x+r.extent.x] - rr.origin.x;
    rr.extent.y= yPos[r.origin.y+r.extent.y] - rr.origin.y;
    if (TestFlag(eCVGrid))
	rr.extent-= gPoint1;
    return rr+GetOrigin();
}

Point CollectionView::ItemPos(VObject *g)
{
    VObject *gop;
    Iter next(coll);

    Update();
    for (int i= 0; gop= (VObject*) next(); i++)
	if (gop == g)
	    break;
    if (i >= coll->Size())
	return gPoint_1;
    return Point(i % cols, i / cols);
}

Rectangle CollectionView::ItemRect(int x, int y)
{
    Update();
    x= Math::Range(0, cols-1, x);
    y= Math::Range(0, rows-1, y);
    Rectangle r(xPos[x], yPos[y], xPos[x+1]-xPos[x], yPos[y+1]-yPos[y]);
    if (TestFlag(eCVGrid))
	r.extent-= gPoint1;
    r.origin+= GetOrigin();
    return r;
}

VObject *CollectionView::GetItem(int x, int y)
{
    VObject *vop= 0;
    int ix= y*cols+x;
    
    if (ix >= 0 && ix < coll->Size())
	vop= (VObject*)coll->At(ix);
    return vop ? vop : defaultItem;
}

//---- event handling

Command *CollectionView::DoLeftButtonDownCommand(Point lp, Token t, int clicks)
{
    if (coll && coll->Size() > 0) {
	SetSelection(gRect0);
	return new CellSelector(this, TestFlag(eCVDontStuckToBorder));
    }
    return View::DoLeftButtonDownCommand(lp, t, clicks);
}

Command *CollectionView::DoRightButtonDownCommand(Point lp, Token t, int clicks, Clipper *vf)
{
    if (vf) {
	menu= GetMenu();
	if(menu) {
	    Manager *m= GetManager();
	    EvtHandler *fh= 0;
	    if (m)
		fh= m->GetFirstHandler();
	    int cmdno= menu->Pulldown(lp, vf);
	    if (cmdno >= 0 && cmdno != cIdCollSizeEntry) {     // show menu in window
		vf->Focus();
		if (fh)
		    fh->PerformCommand(fh->DoMenuCommand(cmdno));
		else
		    PerformCommand(DoMenuCommand(cmdno));
	    }
	}
    }
    return gNoChanges;
}

Command *CollectionView::DoOtherEventCommand(Point lp, Token t)
{
    if (t.Code==eEvtHelpButton || t.Code==eEvtHelpKey) {
	 if( coll && coll->Size() > 0) {
	    Point itemp=PointToItem(lp);
	    VObject* item=GetItem(itemp.x,itemp.y);
	    if(item)
		return item->DoOtherEventCommand(lp, t); 
	 }
    }
    return View::DoOtherEventCommand(lp,t);
}

Menu *CollectionView::GetMenu()
{
    int colSize=coll==0?0:coll->Size();
    
    int selSize=selection!=gRect0;
    char *label=form("Total %d  Selected %d",colSize,selSize);

    if (menu == 0) {
	menu= new Menu(cIdNone, "CollViewPopup",FALSE,TRUE);
	
	Menu *oldMenu= VObject::GetMenu();
	if (oldMenu) 
	    menu->GetCollection()->AddAll((OrdCollection *)oldMenu->GetCollection()->DeepClone());
	menu->AppendItems(
	    "-",
	    "-",	    
	    label,cIdCollSizeEntry,
	    0);	
	menu->DisableAll();
    }
    else {
	menu->ReplaceItem(cIdCollSizeEntry,label);
    }	
    
    menu->EnableItem(cIdCollSizeEntry);
    return menu;
}

Command *
CollectionView::DoKeyCommand(int ch, Token)
{
    if ( ch == '\033' ) {		// Esc key
	ClearSelection();
	return gNoChanges;
    }

    int matchlen = strlen((char *)incr_match);
    bool is_alpha = FALSE;

    if ( ch == '\b' ) {			// Backspace key
	if ( matchlen > 0 )
	    incr_match[--matchlen] = 0;
	if ( matchlen == 0 ) {
	    ClearSelection();
	    return gNoChanges;
	}
    } else if ( ch == '\n' || ch == '\r' || ch == '\t' ) { // Return or Tab key
	if ( selection.IsNotEmpty() )
	    // simulate double click on Return and single click on Tab key
	    DoSelect(selection, /* clicks = */ (ch == '\t') ? 1 : 2);
	else
	{
	    Window *win = GetWindow();
	    if (win)
		win->Bell();
	}
	incr_match[0] = 0;
	return gNoChanges;
    } else {
	incr_match[matchlen++] = ch;
	incr_match[matchlen] = 0;
	is_alpha = TRUE;
    }

    Iter next(coll);
    VObject *item;
    while ( item = (VObject *) next() ) {
	byte *str = (byte *) item->AsString();
	if ( str && StrCmp(incr_match, str, matchlen, sortmap) == 0 )
	    break;
    }

    if ( item ) {
	Rectangle newsel (ItemPos(item), gPoint1);
	Rectangle itemrect (ItemRect(newsel));
	InvalidateRect (ItemRect(selection));
	InvalidateRect (itemrect);
	selection = newsel;
	itemrect.extent.x = 1;
	RevealAlign (itemrect, eVObjHLeft);
//	RevealRect (itemrect, gPoint0);
	if ( SyncSelection() )
	    DoSelect (selection, /* clicks = */ 1);
    } else {
	Window *win = GetWindow();
	if (win)
	    win->Bell();
	if ( is_alpha )
	    incr_match[matchlen - 1] = 0;
    }

    return gNoChanges;
}

Command *
CollectionView::DoCursorKeyCommand (EvtCursorDir d, Token t)
{
    Rectangle sel (GetSelection());
    if ( d == eCrsUp || d == eCrsDown ) {
	int size = coll ? coll->Size() : 0;
	if ( sel.IsEmpty() ) {
	    sel.origin.x = 0;
	    sel.origin.y = (d == eCrsUp) ? (size - 1) : 0;
	    sel.extent = gPoint1;
	} else
	    ( d == eCrsUp ) ? --sel.origin.y : ++sel.origin.y;

	if ( sel.origin.y >= 0 && sel.origin.y < size ) {
	    SetSelection (sel, FALSE);
	    Rectangle itemrect (ItemRect(sel));
	    itemrect.extent.x = 1;
	    RevealAlign (itemrect, eVObjHLeft);
//	    RevealRect (ItemRect(sel), gPoint0);
	    if ( SyncSelection() )
		DoSelect (sel, /* clicks = */ 1);
	}
	return gNoChanges;
    } else
	return View::DoCursorKeyCommand(d, t);
}

//---- drawing

void CollectionView::Draw(Rectangle r)
{
    register int x, y;
    register VObject *gop;
    Point p1, p2;

    if (coll == 0 || coll->Size() <= 0)
	return;
    p1= PointToItem(r.NW());
    p2= PointToItem(r.SE()+gPoint1);
    int i= 0;
    for (x= p1.x; x <= p2.x; x++) {
	for (y= p1.y; y <= p2.y; y++) {
	    i++;
	    gop= GetItem(x, y);
	    gop->DrawAll(gop->contentRect,
			selection.ContainsPoint(Point(x, y)) && gop->Enabled());
	}
    }
    if (TestFlag(eCVGrid)) {
	GrSetPenNormal();
	DrawGrid(p1, p2);
    }
} 

void CollectionView::DrawGrid(Point p1, Point p2)
{
    register int x, y;
    int xx= 0, yy= 0;
    Point o= GetOrigin();

    if (p2.x >= cols-1)
	xx= 1;
    if (p2.y >= rows-1)
	yy= 1;
    for (x= p1.x; x <= p2.x-xx; x++)
	GrLine(Point(xPos[x+1]-1, yPos[p1.y])+o, Point(xPos[x+1]-1, yPos[p2.y+1]-1)+o);
    for (y= p1.y; y <= p2.y-yy; y++)
	GrLine(Point(xPos[p1.x], yPos[y+1]-1)+o, Point(xPos[p2.x+1]-1, yPos[y+1]-1)+o);
}

// --- printing

Rectangle
CollectionView::NextPageBreak (int page, Rectangle r)
{
  bool outside;
  Point sw (r.SW());
  Point item = PointToItem (sw, &outside);
  if ( ! outside )
    {
      Rectangle item_rect = ItemRect (Rectangle (item.x, item.y, 1, 1));
      r.extent.y -= ((sw.y - item_rect.origin.y) + gap.y + 1);
    }
  return r;
}

//---- input/output

OStream& CollectionView::PrintOn(OStream &s)
{
    Object::PrintOn(s);
    return s << gap SP << minExtent SP << rows SP << cols SP << defaultItem SP << coll SP;
}

bool CollectionView::PrintOnWhenObserving(Object *from)
{
    return from != coll;
}

IStream& CollectionView::ReadFrom(IStream &s)
{
    SeqCollection *m;
    VObject *di= 0;

    Object::ReadFrom(s);
    s >> gap >> minExtent >> rows >> cols >> di >> m;
    SetCollection(m);
    SetDefaultItem(di);
    return s;
}

void CollectionView::CollectParts(Collection* col)
{
    View::CollectParts(col);
    if (coll && coll->Size() < 12) // hack
	col->Add(coll);
}

VObject *CollectionView::Detect(BoolFun find, void *arg)
{
    if (coll) {
	Iter next(coll);
	register VObject *v1, *v2;

	while (v1= (VObject*) next())
	    if (v2= v1->Detect(find, arg))
		return v2;
    }
    return 0;
}

//---- CellSelector ------------------------------------------------------------

CellSelector::CellSelector(CollectionView* v, bool dontstucktoborder)
{ 
    SetFlag(eCmdNoReplFeedback);
    lvp= v;
    last= 0;
    dontstuck= dontstucktoborder;
}

void CellSelector::TrackFeedback(Point, Point pp, bool on)
{
    if (on) {
	if (itemptr == last) {
	    if (last)
		last->DoTrackMouse((TrackPhase)eTrackMove, pp);
	} else {
	    if (last)
		last->DoTrackMouse((TrackPhase)eTrackRelease, pp);
	    if (itemptr)
		itemptr->DoTrackMouse((TrackPhase)eTrackPress, pp);
	}
	last= itemptr;
    }
}

Command *CellSelector::TrackMouse(TrackPhase atp, Point, Point, Point np)
{
    bool outside= FALSE;
    
    item= lvp->PointToItem(np, &outside);
    if (outside && dontstuck)
	itemptr= 0;
    else
	itemptr= lvp->GetItem(item.x, item.y);
	
    switch (atp) {
    case eTrackRelease:
	if (last)
	    last->DoTrackMouse((TrackPhase)eTrackRelease, np);
	if (itemptr && itemptr->Enabled())
	    lvp->SetSelection(Rectangle(item, gPoint1), TRUE);
	else
	    lvp->SetSelection(gRect0, TRUE);
    case eTrackExit:
	return gNoChanges;
    default:
	break;
    }
    return this;
}

//---- MultiSelCollView --------------------------------------------------------

MultiSelCollView::MultiSelCollView(EvtHandler *eh, SeqCollection *coll,
				   CollViewOptions options, int rows, int columns)
    : CollectionView(eh, coll, options, rows, columns)
{
}

Command *MultiSelCollView::DoLeftButtonDownCommand(Point lp, Token t, int clicks)
{
    if (coll && coll->Size() > 0)
	return new MultiCellSelector(this, (t.Flags & eFlgShiftKey), clicks);
    return View::DoLeftButtonDownCommand(lp, t, clicks);
}

Menu *MultiSelCollView::GetMenu()
{
    int colSize=coll==0?0:coll->Size();
    Collection *selColl=GetSelection();
    int selSize=selColl==0?0:selColl->Size();
    
    char *label=form("Total %d  Selected %d",colSize,selSize);

    if (menu == 0) {
	menu= new Menu(cIdNone, "CollViewPopup",FALSE,TRUE);
    
	Menu *oldMenu= VObject::GetMenu();
	if (oldMenu) 
    	    menu->GetCollection()->AddAll((OrdCollection *)oldMenu->GetCollection()->DeepClone());
	menu->AppendItems(
	    "-",
	    "-",	    
	    label,cIdCollSizeEntry,
	    0);	
    	menu->DisableAll();
    }
    else {
	menu->ReplaceItem(cIdCollSizeEntry,label);
    }	
    
    menu->EnableItem(cIdCollSizeEntry);
    return menu;
}

Command *
MultiSelCollView::DoKeyCommand(int ch, Token)
{
    if ( ch == '\033' ) {		// Esc key
	ClearSelection(TRUE, TRUE);
	return gNoChanges;
    }

    int matchlen = strlen((char *)incr_match);
    bool is_alpha = FALSE;

    if ( ch == '\b' ) {			// Backspace key
	if ( matchlen > 0 )
	    incr_match[--matchlen] = 0;
	if ( matchlen == 0 ) {
	    ClearSelection(TRUE, TRUE);
	    return gNoChanges;
	}
    } else if ( ch == '\n' || ch == '\r' || ch == '\t' ) { // Return or Tab key
	Collection *sel = GetSelection();
	bool has_selection = sel ? (sel->Size() > 0) : FALSE;
	if ( sel )
	  delete sel;
	if ( has_selection )
	    // simulate double click on Return and single click on Tab key
	    Control (GetId(), (ch == '\t') ? cPartCollSelect : cPartCollDoubleSelect, 0);
	else
	{
	    Window *win = GetWindow();
	    if (win)
		win->Bell();
	}
	incr_match[0] = 0;
	return gNoChanges;
    } else {
	incr_match[matchlen++] = ch;
	incr_match[matchlen] = 0;
	is_alpha = TRUE;
    }

    Iter next(coll);
    VObject *item;
    while ( item = (VObject *) next() ) {
	byte *str = (byte *) item->AsString();
	if ( str && StrCmp(incr_match, str, matchlen, sortmap) == 0 )
	    break;
    }

    if ( item ) {
	next.Reset();
	VObject *vop;
	while ( vop = (VObject *) next() )
	    vop->ResetFlag (eVObjHighlight);
	ForceRedraw();
	item->SetFlag (eVObjHighlight);
	Rectangle itemrect (item->ContentRect());
	itemrect.extent.x = 1;
	RevealAlign (itemrect, eVObjHLeft);
//	RevealRect (itemrect, gPoint0);
	if ( SyncSelection() )
	    Control (GetId(), cPartCollSelect, 0);
    } else {
	Window *win = GetWindow();
	if (win)
	    win->Bell();
	if ( is_alpha )
	  incr_match[matchlen - 1] = 0;
    }

    return gNoChanges;
}

Command *
MultiSelCollView::DoCursorKeyCommand (EvtCursorDir d, Token t)
{
    if ( d == eCrsUp || d == eCrsDown ) {
	int size = GetCollection() ? GetCollection()->Size() : 0;
	Iter next (GetCollection(), (d == eCrsUp) ? cIterForward : cIterBackward);
	VObject *item;
	int i = (d == eCrsUp) ? 0 : (size - 1);
	while ( item = (VObject *) next() ) {
	    if ( item->TestFlag(eVObjHighlight) )
		break;
	    ( d == eCrsUp ) ? ++i : --i;
	}
	if ( item )
	    ( d == eCrsUp ) ? --i : ++i;
	else
	    i = (d == eCrsUp) ? (size - 1) : 0;
	if ( i >= 0 && i < size ) {
	    ClearSelection(FALSE, FALSE);
	    item = (VObject *) GetCollection()->At(i);
	    item->SetFlag(eVObjHighlight);
	    ForceRedraw();
	    Rectangle itemrect (item->ContentRect());
	    itemrect.extent.x = 1;
	    RevealAlign (itemrect, eVObjHLeft);
//	    RevealRect (item->ContentRect(), gPoint0);
	    if ( SyncSelection() )
		Control (GetId(), cPartCollSelect, 0);
	}
	return gNoChanges;
    } else
	return View::DoCursorKeyCommand(d, t);
}

SeqCollection *MultiSelCollView::GetSelection()
{
    if (coll && coll->Size() > 0) {
	SeqCollection *list = new OrdCollection;
	Iter next(coll);
	VObject *vop;
	while ( vop = (VObject *) next() ) {
	    if ( vop->TestFlag(eVObjHighlight) )
		list->Add(vop);
	}
	return list;
    } else
	return NULL;
}

void MultiSelCollView::ClearSelection(bool redraw, bool send_control)
{
    if (coll && coll->Size() > 0) {
	Iter next(coll);
	VObject *vop;
	while ( vop = (VObject *) next() )
	    vop->ResetFlag(eVObjHighlight);
    	if (redraw)
	    ForceRedraw();
    }
    incr_match[0] = 0;
    if ( send_control )
	Control (GetId(), cPartCollSelect, 0);
}

void MultiSelCollView::SetSelection(SeqCollection *col, bool redraw)
{
    if (coll && coll->Size() > 0) {
	Iter next(coll);
	VObject *vop;
	while ( vop = (VObject *) next() ) {
	    if(col->FindPtr(vop))
		vop->SetFlag(eVObjHighlight);
	}
    	if (redraw)
	    ForceRedraw();
    }
    incr_match[0] = 0;
}

void 
MultiSelCollView::SelectAll (bool redraw)
{
    if (coll && coll->Size() > 0) {
	Iter next(coll);
	VObject *vop;
	while ( vop = (VObject *) next() ) 
	    vop->SetFlag(eVObjHighlight);
    	if (redraw)
	    ForceRedraw();
    }
    incr_match[0] = 0;
}


//---- MultiCellSelector ------------------------------------------------

// ---- Michael Scharf, Wed Aug 24 22:19:29 1994.
// who decided to put MultiSelCollView into CollView.[Ch]
// Changing this causes that almost everything gets recompiled,
// though only 2 or 3 files actually use MultiSelCollView!

MultiCellSelector::MultiCellSelector(CollectionView *cv, bool shift, int cl)
{
    view = cv;
    shift_key = shift;
    clicks = cl;
    first_track = first_item = TRUE;
    last = 0;
    view->incr_match[0] = 0;
    SetFlag(eCmdNoReplFeedback);
}

Command *MultiCellSelector::TrackMouse(TrackPhase atp, 
				       Point anchorP, 
				       Point prevP, Point newP)
{
    switch (atp) {
    case eTrackPress: {
	// clear old selection unless shift key is pressed
	if (!shift_key) {
	    Iter next(view->GetCollection());
	    VObject *vop;
	    while (vop = (VObject *) next())
		if (vop != itemptr && vop->TestFlag(eVObjHighlight)) {
		    vop->ResetFlag(eVObjHighlight);
		    vop->ForceRedraw();
		}
	}
    }
    case eTrackMove: {
	int ix,iy;
	bool outside= FALSE;

	// check for eCVDontStuckToBorder 
	// does anyone really ever want eCVDontStuckToBorder ??????
	view->PointToItem(newP, &outside);
	if (outside && view->TestFlag(eCVDontStuckToBorder))
	    break;

	Point maxI= view->PointToItem(Max(prevP,newP), &outside);
	Point minI= view->PointToItem(Min(prevP,newP), &outside);
	for(ix = minI.x; ix <= maxI.x; ix++) {
	    for(iy = minI.y; iy <= maxI.y; iy++) {
		itemptr= view->GetItem(ix, iy);
		if (itemptr) {
		    if (first_item) {
			first_item = FALSE;
			do_select = !(itemptr->TestFlag(eVObjHighlight) && shift_key);
		    }
		    if (itemptr->TestFlag(eVObjHighlight) != do_select) {
			itemptr->SetFlag(eVObjHighlight, do_select);
			itemptr->ForceRedraw();
		    }
		}
	    }
	}
	break;
    }
    case eTrackRelease:
	view->Control(view->GetId(), clicks >= 2 ? cPartCollDoubleSelect : cPartCollSelect, 0);
	break;
    case eTrackExit:
	return gNoChanges;
    }
    
    return this;
}
