#ifndef CollView_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define CollView_First

#include "View.h"

class SeqCollection;

//---- CollectionView options --------------------------------------------------

enum CollViewOptions {
    eCVGrid             = BIT(eViewLast+1),
    eCVDontStuckToBorder= BIT(eViewLast+2),
    eCVExpandRows       = BIT(eViewLast+3),
    eCVExpandCols       = BIT(eViewLast+4),
    eCVModified         = BIT(eViewLast+5),
    eCVDefault          = 0,
    eCVLast             = eViewLast + 5
};

const int 	cIdCollSizeEntry	= 	cIdFirstUser+1;

//---- CollectionView ----------------------------------------------------------

class MultiCellSelector;

class CollectionView: public View {
    friend class MultiCellSelector;
protected:
    SeqCollection *coll;
    Rectangle selection;
    Point gap, minExtent;
    int *xPos, *yPos, rows, cols;
    VObject *defaultItem;
    byte *incr_match;
    bool sync_selection;
    Menu *menu;

public:
    MetaDef(CollectionView);

    CollectionView (EvtHandler *, SeqCollection *, CollViewOptions = eCVDefault,
		    int rows = 0, int columns = 1);
    ~CollectionView();

    bool SyncSelection()
	{ return sync_selection; }
    void SetSyncSelection (bool s)
	{ sync_selection = s; }

    void SetCollection(SeqCollection*, bool freeold= TRUE);
    
    SeqCollection *GetCollection()
	{ return coll; }
    void SetDefaultItem(VObject *d);
    void SetColsRows(Point cr);

    //---- layout
    void SetMinExtent(Point);
    Metric GetMinSize();
    void SetOrigin(Point at);
    void SetGap(Point g)
	{ gap= g; }
    void Update();
    void Open(bool m);
	
    //---- drawing
    void Draw(Rectangle);
    virtual void DrawGrid(Point from, Point to);

    // --- printing
    Rectangle NextPageBreak (int page, Rectangle pr);
    
    //---- event handling
    Command *DoLeftButtonDownCommand(Point, Token, int);
    Command *DoRightButtonDownCommand (Point, Token, int, Clipper *);
    
    Command *DoKeyCommand(int, Token);
    Command *DoCursorKeyCommand (EvtCursorDir, Token);
    Command *DoOtherEventCommand(Point lp, Token t);
    virtual void DoSelect(Rectangle sel, int clicks);
    VObject *Detect(BoolFun find, void *arg);

    //---- updating (change propagation)
    void DoObserve(int, int, void*, Object*);
    void Modified();
    
    //---- selection
    Rectangle GetSelection()
	{ return selection; }
    void SetSelection(Rectangle, bool sendcontrol= FALSE);
    // This method dont use the value of WindowSystem::Clicks to pass to Control 
    // (if sendcontrol == TRUE) instead it uses the value of nrOfClicks
    void SetSelectionWithClicks(Rectangle, bool sendcontrol= FALSE, int nrOfClicks=1);
    void ClearSelection(bool redraw= TRUE);
	
    //---- mapping
    Point PointToItem(Point, bool *outside= 0);
    Rectangle ItemRect(int, int);
    VObject *GetItem(int x, int y);
    Point ItemPos(VObject *);
    Rectangle ItemRect(Rectangle r);
    void InvalidateItemRect(Rectangle r)
	{ InvalidateRect(ItemRect(r)); }
    void ConstrainScroll(Point*);

    //---- input/output
    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);
    bool PrintOnWhenObserving(Object*);
    
    //---- inspecting
    void CollectParts(Collection*);
    
    //----- popup menu
    Menu *GetMenu();
};

//---- CellSelector ------------------------------------------------------------

class CellSelector: public Command {
    CollectionView *lvp;
    VObject *itemptr, *last;
    Point item;
    bool dontstuck;
public:
    CellSelector(CollectionView*, bool dontstucktoborder);
    void TrackFeedback(Point, Point, bool);
    Command *TrackMouse(TrackPhase, Point, Point, Point);
};


//---- MultiSelCollView --------------------------------------------------------
/*
   A CollectionView with multiple selection.
   The current selection can be extended/shrinked by pressing the Shift key.
   Clients are notified of selection changes through
   "Control(id, cPartCollSelect, 0)". Selected items can then be obtained by
   iterating over the Collection and inspecting the items' "eVObjHighlight"
   flag.
*/

class MultiSelCollView : public CollectionView {
public:
    MultiSelCollView(EvtHandler*, SeqCollection*, CollViewOptions o= eCVDefault,
			int r= 0, int c= 1);
    //---- event handling
    Command *DoLeftButtonDownCommand (Point, Token, int);
    Menu *GetMenu();
    Command *DoKeyCommand (int, Token);
    Command *DoCursorKeyCommand (EvtCursorDir, Token);

    //---- selection
    SeqCollection *GetSelection();
    void SetSelection(SeqCollection *, bool);
    void ClearSelection(bool redraw= TRUE, bool send_control = FALSE);
    void SelectAll (bool redraw = TRUE);
};

//---- MultiCellSelector ------------------------------------------

class MultiCellSelector : public Command {
public:
    MultiCellSelector(CollectionView *, bool shift, int clicks);

protected:
    Command *TrackMouse(TrackPhase, Point, Point, Point);

    CollectionView *view;
    VObject *itemptr, *last;
    bool shift_key;		// Shift key pressed ?
    bool do_select;		// select or unselect items
    bool first_track;
    bool first_item;
    int clicks;
};

#endif
