#ifdef __GNUG__
#pragma implementation
#endif

#include "CommandProcessor.h"

#include "Class.h"
#include "Command.h"
#include "ObjArray.h"
#include "Menu.h"
#include "CmdNo.h"

//---- CommandProcessor --------------------------------------------------------

NewMetaImpl(CommandProcessor,RefCounted, (T(changeCount), T(cap)));

CommandProcessor::CommandProcessor()
{
    cap= changeCount= 0;
}

CommandProcessor::~CommandProcessor()
{
}

Command *CommandProcessor::GetAt(int)
{
    return 0;
}

void CommandProcessor::PerformCommand(Command* cmd)
{
    if (cmd == 0)
	return;

    switch (cmd->GetType()) {

    case eCmdTypeCantUndo:
	FinishFrom(0);
	changeCount+= cmd->Perform();
	Changed();
	cmd->Done(0);
	cmd->Finish();
	break;

    case eCmdTypeNoChange:
	CommandProcessor::PerformNormalCommand(cmd);
	break;

    case eCmdTypeNormal:
	FinishFrom(changeCount);
	PerformNormalCommand(cmd);
	break;
    }
}

bool
CommandProcessor::CanUndo()
{
    Command *cmd;
    if ((cmd= GetAt(changeCount-1)) && (cmd->GetType() != eCmdTypeCantUndo))
	return TRUE;
    return FALSE;
}

void
CommandProcessor::Undo()
{
    if(!CanUndo())
	return;
    Command *c;
    if (c= GetAt(changeCount-1)) {
	changeCount+= c->Perform();
	Changed();
    }
}

bool
CommandProcessor::CanRedo()
{
    Command *cmd;
    if (Size() >= 2) {
	if (cmd= GetAt(changeCount)) {
	    return TRUE;
	}
    }
    return FALSE;
}

void
CommandProcessor::Redo()
{
    if(!CanRedo())
	return;
    Command *c;
    if (c= GetAt(changeCount)) {
	changeCount+= c->Perform();
	Changed();
    }
}


void CommandProcessor::PerformNormalCommand(Command* cmd)
{
    if (cmd) {
	cmd->Perform();
	cmd->Done(0);
	cmd->Finish();
    }
}

void CommandProcessor::DoCreateMenu(Menu *menu)
{
    if (menu->GetId() == cEDITMENU) {
	menu->AppendItem("Undo@Z", cUNDO);
	if (Size() >= 2)
	    menu->AppendItem("Redo", cREDO);
    }
}

void CommandProcessor::DoSetupMenu(Menu *menu)
{
    if (menu->GetId() == cEDITMENU) {
	Command *cmd;
	
	if (cmd= GetAt(changeCount-1)) {
	    menu->ReplaceItem(cUNDO, cmd->GetUndoName());
	    if (cmd->GetType() != eCmdTypeCantUndo)
		menu->EnableItem(cUNDO);
	} else {
	    menu->ReplaceItem(cUNDO, "Undo");
	    menu->DisableItem(cUNDO);
	}
	
	if (Size() >= 2) {
	    if (cmd= GetAt(changeCount)) {
		menu->ReplaceItem(cREDO, cmd->GetUndoName());
		menu->EnableItem(cREDO);
	    } else {
		menu->ReplaceItem(cREDO, "Redo");
		menu->DisableItem(cREDO);
	    }
	}
    }
}

void CommandProcessor::DoMenuCommand(int cmdId)
{
    switch (cmdId) {
    case cUNDO:
	Undo();
	break;
    case cREDO:
	Redo();
	break;
    }
}

void CommandProcessor::Finish()
{
    FinishFrom(0);
    changeCount= 0;
    Changed();
}

void CommandProcessor::FinishFrom(int)
{
}

//---- SingleCommandProcessor --------------------------------------------------

NewMetaImpl(SingleCommandProcessor,CommandProcessor, (TP(lastCmd)));

SingleCommandProcessor::SingleCommandProcessor()
{
    lastCmd= 0;
    cap= 1;
}

Command *SingleCommandProcessor::GetAt(int)
{
    return lastCmd;
}

void SingleCommandProcessor::PerformNormalCommand(Command* cmd)
{
    if (lastCmd)
	lastCmd->Done(cmd);
    lastCmd= cmd;
    changeCount+= cmd->Perform();
    Changed();
}

void SingleCommandProcessor::FinishFrom(int)
{
    if (lastCmd) {
	lastCmd->Finish();
	lastCmd= 0;
    }
}

//---- UnboundedCommandProcessor -----------------------------------------------

NewMetaImpl(UnboundedCommandProcessor,CommandProcessor, (TP(cmdHistory)));

UnboundedCommandProcessor::UnboundedCommandProcessor()
{
    cmdHistory= new ObjArray;
    cap= 99999;
}

UnboundedCommandProcessor::~UnboundedCommandProcessor()
{
    SafeDelete(cmdHistory);
}

Command *UnboundedCommandProcessor::GetAt(int i)
{
    if (i >= 0 && i < cmdHistory->Size())
	return (Command*) cmdHistory->At(i);
    return 0;
}

void UnboundedCommandProcessor::PerformNormalCommand(Command* cmd)
{
    Command *lastcmd;
    if (lastcmd= GetAt(changeCount-1))
	lastcmd->Done(cmd);
    cmdHistory->AtPutAndExpand(changeCount, cmd);
    changeCount+= cmd->Perform();
    Changed();
}

void UnboundedCommandProcessor::FinishFrom(int n)
{
    for (int i= n; i < cmdHistory->Size(); i++) {
	Command *c= (Command*) cmdHistory->At(i);
	if (c) {
	    c->Finish();
	    cmdHistory->RemoveAt(i);
	}
    }
}

//---- BoundedCommandProcessor -------------------------------------------------

NewMetaImpl(BoundedCommandProcessor,CommandProcessor, (TVP(mem,cap), T(start), T(size)));

BoundedCommandProcessor::BoundedCommandProcessor(int s)
{
    cap= s;
    mem= new Command* [cap];
    size= start= 0;
}

BoundedCommandProcessor::~BoundedCommandProcessor()
{
    SafeDelete(mem);
}

Command *BoundedCommandProcessor::GetAt(int i)
{
    if (i >= start && i < start+size)
	return mem[i % cap];
    return 0;
}

void BoundedCommandProcessor::PerformNormalCommand(Command* cmd)
{
    Command *lastcmd;
    if (size >= cap) {      // overflow
	start++;
	GetAt(changeCount)->Finish();
	size--;
    }
    size++;
    if (lastcmd= GetAt(changeCount-1))
	lastcmd->Done(cmd);
    mem[changeCount % cap]= cmd;
    changeCount+= cmd->Perform();
    Changed();
}

void BoundedCommandProcessor::FinishFrom(int n)
{
    Command *c;
    
    n= (n == 0) ? start : changeCount;
    
    for (int i= n; c= GetAt(i); i++)
	c->Finish();
    size-= (i-changeCount);
}

