#ifdef __GNUG__
#pragma implementation
#endif

#include "Converter.h"

#include "Class.h"
#include "ETString.h"
#include "Data.h"
#include "CLib.h"
#include "OrdColl.h"
#include "ClassManager.h"

#include "CollView.h"
#include "TextItem.h"
#include "BorderItems.h"
#include "Box.h"
#include "Scroller.h"
#include "Expander.h"
#include "Dialog.h"
#include "Buttons.h"
#include "TypeMatcher.h"

//---- ConverterDialog ---------------------------------------------------------

class ConverterDialog : public Dialog {
public:
    MetaDef(ConverterDialog);
    ConverterDialog();
    ~ConverterDialog();
    Converter *ShowInWindow(SeqCollection *list);
    VObject *DoMakeContent();
    void DoSetup();
    void Control(int id, int p, void *v);
private:
    CollectionView *collView;
    SeqCollection *converterList;
};
	    
NewMetaImpl0(ConverterDialog,Dialog);

ConverterDialog::ConverterDialog() : Dialog("Converter")
{
}

ConverterDialog::~ConverterDialog()
{
    SafeDelete(collView);
}

Converter *ConverterDialog::ShowInWindow(SeqCollection *list)
{
    converterList= list;
    
    OrdCollection *col= new OrdCollection;
    Iter next(converterList);
    Converter *cv;
    
    while (cv= (Converter*) next())
	col->Add(new TextItem(cv->AsString()));
    
    collView= new CollectionView(this, col, eCVDontStuckToBorder);
    collView->SetMinExtent(Point(100, 0));
    
    Port *old= GrGetPort();
    Dialog::Show();
    if (old)
	GrSetPort(old);

    Rectangle r= collView->GetSelection();
    if (r.IsNotEmpty())
	return (Converter*) converterList->At(r.origin.y);
    return 0;
}

VObject *ConverterDialog::DoMakeContent()
{
    return
	new Matte(
	    new Expander(cIdNone, eVert, 20,
		new Scroller(collView, Point(150, cAutoSize)),
		new HBox(10, (VObjAlign)(eVObjVBase|eVObjHEqual|eVObjHExpand),
		    new ActionButton(cIdOk, "Convert", TRUE),
		    new ActionButton(cIdCancel, "Cancel"),
		    0
		),
		0
	    )
	);
}

void ConverterDialog::DoSetup()
{
    EnableItem(cIdOk, !collView->GetSelection().IsEmpty());
}

void ConverterDialog::Control(int id, int p, void *v)
{
    if (p == cPartCollDoubleSelect) {
	Dialog::Control(cIdOk, cPartAction, v);
    } else
	Dialog::Control(id, p, v);
}

//---- Converter ---------------------------------------------------------------

NewMetaImpl(Converter,Object, (TP(filters)));

OrdCollection *Converter::filters;

Converter::Converter()
{
}

char *Converter::AsString()
{
	return "Abstract Converter";
}

const char *Converter::GetExtension(Data*)
{
	return 0;
}

void Converter::CleanUp()
{
    if (filters) {
	filters->FreeAll();
	SafeDelete(filters);
    }
}

Converter *Converter::SelectConverter(OrdCollection *list)
{
    Converter *cv= 0;
    if (list->Size() > 1) {
	ConverterDialog *dia= new ConverterDialog;
	cv= dia->ShowInWindow(list);
	gSystem->AddCleanupObject(dia);
    } else if (list->Size() == 1)
	cv= (Converter*) list->At(0);
    return cv;
}

OrdCollection *Converter::CollectConverters()
{
    if (filters == 0) {
	Iter next(gClassManager->Iterator());
	Class *cla;
    
	filters= new OrdCollection;
	while (cla= (Class*)next())
	    if (cla->isKindOf(Meta(Converter)) && cla != Meta(Converter))
		filters->Add(cla->Proto()->New());
    }
    return filters;
}

Converter *Converter::FindConverter(Data *data, Class *want, OrdCollection *l)
{
    Iter next(CollectConverters());
    register Converter *filter;
    
    while (filter= (Converter*) next()) {
	if (filter->CanConvert(data, want)) {
	    if (l == 0)
		return filter;
	    if (l->FindPtr(filter) == 0)
		l->AddFirst(filter);
	}
    }
    if (l && l->Size() > 0)
	return (Converter*)l->At(0);
    return 0;
}

bool Converter::CanConvert(Data*, Class*)
{
    return FALSE;
}

Object *Converter::Convert(Data*, Class*)
{
    return 0;
}

bool Converter::CanConvertTo(Object*)
{
    return FALSE;
}

bool Converter::ConvertTo(Data*, Object*)
{
    return FALSE;
}

bool Converter::FindOutputConverter(Data *data, Object *op)
{
    if (op == 0)
	return FALSE;
    Iter next(CollectConverters());
    register Converter *filter, *theConverter= 0;
    OrdCollection *l= new OrdCollection;
    
    while (filter= (Converter*) next())
	if (filter->CanConvertTo(op)) {
	    if (l->FindPtr(filter) == 0)
		l->AddFirst(filter);
	}
    theConverter= SelectConverter(l);
    if (theConverter)
	return theConverter->ConvertTo(data, op);
    return FALSE;
}

