#ifdef __GNUG__
#pragma implementation
#endif

#include "Data.h"
#include "Class.h"
#include "OrdColl.h"
#include "FileBuf.h"
#include "MemBuf.h"
#include "ETString.h"
#include "System.h"
#include "Directory.h"
#include "Converter.h"
#include "TypeMatcher.h"
#include "Alert_e.h"
#include "Env.h"

const Symbol
	cDocCreatorUndef("__UNKNOWN__"),
	cDocTypeUndef("UNDEF"),
	cDocTypeAscii("ASCII"),
	cDocTypeET("ET++OIO"),
	cDocTypeCPlusPlus("C++"),
	cDocTypeCHeader("CHEADER"),
	cDocTypeDirectory("DIRECTORY"),
	cDocTypeExec("EXEC"),
	cDocTypeOFile("OFILE");
	
extern SmartBitmap gFileIcon;  // bitmap images for file icons

 
//used by:
//-flag: determines if attribute is already set (eg. modtime)
//-bits: sets respective (binary) attribute (eg. exec) 
enum {		
    eDataType,
    eDataCreator,
    eDataSize,
    eDataTime,
    eDataId,
    eDataExec,
    eDataValid,
    eDataWrite,
    eDataRead,
    eDataDeepType,
    eDataNew,
    eDataIsAscii
};

static OrdCollection *gConverters;
static OrdCollection *gTypes;

void PrepareTypes()
{
    SafeDelete(gTypes);
    gTypes= new OrdCollection;
}

bool EndTypes(Data *data)
{
    bool b = FALSE;
    if (data && gTypes)
	b = data->CheckTypes(gTypes);
    SafeDelete(gTypes);
    return b;
}

//---- Data --------------------------------------------------------------------

NewMetaImpl(Data,RefCounted, (TP(name), T(creator), T(type), T(bits), T(flags),
	T(uniqueid), T(sizehint), T(modtime), TP(classtype), TP(typematcher)));

Data::Data(char *n)
{
    name= strsave(n);
    classtype= 0;
    type= cDocTypeUndef;
    creator= cDocCreatorUndef;
    typematcher= 0;
    bits= flags= 0;
    uniqueid= 0;
    sizehint= -1;
    modtime= 0;
}

Data::~Data()
{
    SafeDelete(name);
}

#define SETBIT2(bi,b) ((bi)|= BIT(b))
#define CLRBIT2(bi,b) ((bi)&= ~BIT(b))
#define TESTBIT2(bi,b) (bool) (((bi) & BIT(b)) != 0)

void Data::FlushStatCache()
{
    bits=flags=0;
}
 
void Data::UpdateBit(u_long b, bool val)
{
    if (val)
	SETBIT2(bits, b);
    else
	CLRBIT2(bits, b);
    SETBIT2(flags, b);
}

Class *Data::GetClassType()
{
    UpdateFields(eDataDeepType);
    return classtype;
}

bool Data::IsWritable()
{
    UpdateFields(eDataWrite);
    return TESTBIT2(bits, eDataWrite);
}
  
bool Data::IsReadable()
{
    UpdateFields(eDataRead);
    return TESTBIT2(bits, eDataRead);
}
  
bool Data::IsExecutable()
{
    UpdateFields(eDataExec);
    return TESTBIT2(bits, eDataExec);
}

bool Data::IsETFormat()
{
    if (classtype)
	return TRUE;
    UpdateFields(eDataDeepType);
    if (type == cDocTypeET)
	return TRUE;
    if (typematcher && typematcher->GetType() == cDocTypeET)
	return TRUE;
    return FALSE;
}

bool Data::IsAscii()
{
    UpdateFields(eDataIsAscii);
    return TESTBIT2(bits, eDataIsAscii);
}
 
void Data::SetIsAscii(bool isAscii)
{
    UpdateBit(eDataIsAscii, isAscii);
}
 
ELineSeparatorStyle Data::GetLineSeparatorStyleOrg()
{
    AbstractMethod("GetLineSeparatorStyleOrg");
    return cLineSeparatorStyleUnknown;
}
 
ELineSeparatorStyle Data::GetLineSeparatorStyleNew()
{
    AbstractMethod("GetLineSeparatorStyleNew");
    return cLineSeparatorStyleUnknown;
}
 
void Data::SetLineSeparatorStyleNew(ELineSeparatorStyle lss)
{ 
    AbstractMethod("SetLineSeparatorStyleNew");
}

bool Data::IsValid()
{
    UpdateFields(eDataValid);
    return TESTBIT2(bits, eDataValid);
}

bool Data::IsUntitled()
{
    UpdateFields(eDataNew);
    return TESTBIT2(bits, eDataNew);
}

bool Data::HasContents()
{
    UpdateFields(eDataType);
    return type == cDocTypeDirectory;
}

u_long Data::ModifiedAt()
{
    UpdateFields(eDataTime);
    return modtime;
}

long Data::SizeHint()
{
    UpdateFields(eDataSize);
    return sizehint;
}

Symbol Data::Type()
{
    UpdateFields(eDataDeepType);
    return type;
}

void Data::SetType(const Symbol &t)
{
    type= t;
    SETBIT2(flags, eDataDeepType);
}

Symbol Data::Creator()
{
    UpdateFields(eDataCreator);
    return creator;
}

void Data::SetCreator(const Symbol &crea)
{
    creator= crea;
    SETBIT2(flags, eDataCreator);
}

char *Data::FullName()
{
    return name;
}

char *Data::ShortName()
{
    return FullName();
}

u_long Data::UniqueId()
{
    UpdateFields(eDataId);
    return uniqueid;
}

StreamBuf *Data::GetStreamBuf(bool forwriting)
{
    AbstractMethod("GetStreamBuf");
    return 0;
}

StreamBuf *Data::GetStreamBufForReading()
{
    StreamBuf *sb= GetStreamBuf(FALSE);
    if (sb)
	sb->seek(0);
    return sb;
}

StreamBuf *Data::GetStreamBufForWriting()
{
    return GetStreamBuf(TRUE);
}

bool Data::SaveObject(Object*, bool)
{
    AbstractMethod("SaveObject");
    return FALSE;
}

Object *Data::AsObject(Class *want)
{
    if (gTypes) {   // AsObject used in test mode (ala CanPaste)
	if (want)
	    gTypes->AddFirst(want);
	return 0;
    }
    // AsObject used in normal mode (ala PasteData)
    return MakeObject(want);
}

Object *Data::MakeObject(Class *want)
{
    Converter *converter= Converter::FindConverter(this, want);
    if (converter && converter->CanConvert(this, want))
	return converter->Convert(this, want);
    return 0;
}

bool Data::CanConvert(Class *want)
{
    if (gTypes) {
	if (want)
	    gTypes->AddFirst(want);
	return FALSE;
    }
    return Converter::FindConverter(this, want) != 0;
}

bool Data::CheckTypes(OrdCollection *typelist)
{
    if (typelist) {
	Iter next(typelist);
	Class *cl;
	
	SafeDelete(gConverters);
	gConverters= new OrdCollection;
	
	while (cl= (Class*) next())
	    Converter::FindConverter(this, cl, gConverters);

	if (gConverters->Size() <= 0)
	    SafeDelete(gConverters);
    }
    bool result=FALSE;
    if (gConverters && gConverters->Size() > 0)
	result= TRUE;
    SafeDelete(gConverters);
    return result;
}

Bitmap *Data::GetBitmap()
{
    UpdateFields(eDataType);
    if (typematcher)
	return typematcher->GetBitmap();
    return gFileIcon;
}

SeqCollection *Data::GetContents()
{
    return 0;
}

Data *Data::GetParent()
{
    return 0;
}

bool Data::HasParent()
{
    Data *d= GetParent();
    bool b= (d != 0);
    SafeDelete(d);
    return b;
}

OStream& Data::PrintOn(OStream &s)
{
    RefCounted::PrintOn(s);
    s.PrintString(name);
    return s << bits SP << flags SP << creator SP << type SP
	    << sizehint SP << modtime SP << uniqueid SP
	    << (long)(classtype) SP
	    << (long)(typematcher) SP;
}

IStream& Data::ReadFrom(IStream &s) 
{
    long ct, tm;
    RefCounted::ReadFrom(s);
    SafeDelete(name);
    s.ReadString(&name);
    s >> bits >> flags >> creator >> type >> sizehint >> modtime >> uniqueid
							    >> ct >> tm;
    classtype= (Class*) ct;
    typematcher= (TypeMatcher*) tm;
    return s;
}

u_long Data::UpdateFields(u_long f)
{
    if (!TESTBIT2(flags, f)) {
	switch (f) {
	case eDataType:
	case eDataDeepType:
	case eDataCreator:
	case eDataIsAscii:
	    typematcher= TypeMatcher::MatchByContents(this);
	    if (typematcher && type == cDocTypeUndef)
		type= typematcher->GetType();
	    SETBIT2(flags, eDataType);
	    SETBIT2(flags, eDataDeepType);
	    SETBIT2(flags, eDataCreator);
	    if (typematcher) UpdateBit(eDataIsAscii, typematcher->ascii);
	    break;
	}
    }
    return 0;
}

//---- FileData -------------------------------------------------------------

NewMetaImpl(FileData, Data, (TP(converter), TB(lookDeep)));

FileData::FileData(char *nm, bool deep) : Data()
{
    lookDeep= deep;	//why???
    lookDeep= FALSE;	//why???
    SetName(nm);
    converter= 0;
    fLineSeparatorStyleOrg= cLineSeparatorStyleNotSet;
    fLineSeparatorStyleNew= cLineSeparatorStyleNotSet;
}

FileData::~FileData()
{
    converter= 0;
}

Object *FileData::deepclone()
{
    FileData*fd= new FileData(name, lookDeep);
    SafeDelete(fd->name);
    fd->name= strsave(name);
    return fd;
}

void FileData::SetName(char *nm)
{
    flags= 0;
    SafeDelete(name);
    if (nm) {
	if (!gSystem->IsAbsolutePath(nm)) {
	    char *wd= gSystem->WorkingDirectory();
	    if (strcmp(nm, ".") == 0)
		name= strsave(wd);
	    else {
		//ruedi:why the following
		//if really necessary, use another test to find out if 
		//at root (cf. GetParent()
		//old:
		//if (wd && strcmp(wd, path_separator)) 
		//    name= strsave(gSystem->ConcatPathNames(wd, nm,0));
		//else
		//    name= strsave(gSystem->ConcatPathNames(" ", nm, 0)); 
		//new:
		name= strsave(gSystem->ConcatPathNames(wd, nm,0));
	    }
	} else if(nm[0]) {
	    name= strsave(nm);
	} else {
	    static int untitledCnt=0;
	    if(untitledCnt>0)
		name= strsave(form("Untitled-%d",untitledCnt));
	    else
		name= strsave("Untitled");
	    untitledCnt++;
	}
    }
    else
	name= strsave("Untitled");
}

char *FileData::ShortName()
{
    return gSystem->BaseName(FullName());
}

char *FileData::GetLoadDir()
{
    return gSystem->DirName(FullName());
}

Converter *FileData::GetConverter()
{
    return converter;
}

//delayed determination schema:
//DetermineLineSeparatorStyle() is only called when: needed for 
//writing and LineSeparatorStyleNew has not been explicitly set!
ELineSeparatorStyle FileData::DetermineLineSeparatorStyle()
{
    const int cTestSize= 512; //max. bytes considered for decision
    char buf[cTestSize+1];
    
    StreamBuf *sb= new Filebuf(FullName(), input); //it MUST be Filebuf
    if (sb) sb->seek(0);
    IStream is(sb, 0, TRUE);
    int effSize= is.read(buf, cTestSize);
    buf[effSize]= '\0';
    if (effSize<=0) {	//empty file or no file
	return cLineSeparatorStyleUnknown;
    }
    
    int countLF= 0;	//unix
    int countCRLF= 0;	//dos
    int countCR= 0;	//mac
    bool lastCharIsCR= FALSE;
    int i;
    for (i=0 ; i<effSize; i++) {
	if (buf[i]=='\r'){
	    countCR++;
	    lastCharIsCR= TRUE;
	} else if (buf[i]=='\n') {
	    countLF++;
	    if (lastCharIsCR) {
		countCR--;countLF--;
		countCRLF++;
	    }
	    lastCharIsCR= FALSE;
	} else {
	    lastCharIsCR= FALSE;
	}
    }
    if (countLF>=countCRLF && countLF>=countCR)	//UNIX has first priority!
	return cLineSeparatorStyleUnix;
    else if (countCRLF>=countLF && countCRLF>=countCR)
	return cLineSeparatorStyleDos;
    else 
	return cLineSeparatorStyleMac;
}
	
ELineSeparatorStyle FileData::GetLineSeparatorStyleOrg()
{
    if (fLineSeparatorStyleOrg==cLineSeparatorStyleNotSet) {
	fLineSeparatorStyleOrg= DetermineLineSeparatorStyle();
    }
    return fLineSeparatorStyleOrg;
}
ELineSeparatorStyle FileData::GetLineSeparatorStyleNew()
{
    if (fLineSeparatorStyleNew==cLineSeparatorStyleNotSet ||
	fLineSeparatorStyleNew==cLineSeparatorStyleAsIs) {
	fLineSeparatorStyleNew= GetLineSeparatorStyleOrg();
	if (fLineSeparatorStyleNew==cLineSeparatorStyleUnknown) {
	    fLineSeparatorStyleNew= (ELineSeparatorStyle)Env::GetValue("FileData.LineSeparatorStyle", cLineSeparatorStyleUnix);
        }
    }
    return fLineSeparatorStyleNew;
}
void FileData::SetLineSeparatorStyleNew(ELineSeparatorStyle lss)
{ 
    fLineSeparatorStyleNew= lss; 
}
 
u_long FileData::UpdateFields(u_long f)
{
    if (!TESTBIT2(flags, f)) {
    
	switch (f) {
	case eDataId:
	case eDataSize:
	case eDataTime:
	case eDataExec:
	case eDataValid:
	    {
		u_long ff = 0;
		bool valid= gSystem->GetPathInfo(FullName(), &uniqueid,
					    (u_long*)&sizehint, &ff, &modtime) == 0;
		SETBIT2(flags, eDataId);
		SETBIT2(flags, eDataSize);
		SETBIT2(flags, eDataTime);
		UpdateBit(eDataExec, (u_int)(ff & ePiExec));
		UpdateBit(eDataValid, valid);
	    }
	    break;
	
	case eDataType:
	    if (lookDeep) {
		Data::UpdateFields(f);
	    } else {
		if (!typematcher) typematcher= TypeMatcher::MatchByName(this);
		if (typematcher && type == cDocTypeUndef)
		    type= typematcher->GetType();
		SETBIT2(flags, eDataType);
		if (type == cDocTypeDirectory)
		    SETBIT2(flags, eDataDeepType);   // no need to look deeper
	    }
	    break;

	case eDataDeepType:
	case eDataCreator:
	case eDataIsAscii:
	    Data::UpdateFields(f);
	    break;

	case eDataRead:
	    UpdateBit(eDataRead, !gSystem->AccessPathName(FullName(), 4));
	    break;
	    
	case eDataWrite:
	    {
		char *nm= FullName();
		if ( gSystem->AccessPathName(nm, 0) ) {
		    char buf[1024];
		    strcpy(buf, nm);
		    nm = gSystem->DirName(buf);
		}
		UpdateBit(eDataWrite, !gSystem->AccessPathName(nm, 2));
	    }
	    break;

	case eDataNew:
	    UpdateBit(eDataNew, !gSystem->IsAbsolutePath(FullName()));
	    break;

	}
    }
    return 0;
}

Object *FileData::MakeObject(Class *want)
{
    if (converter == 0) {
	if (gConverters == 0 || gConverters->Size() <= 0) {
	    if (gConverters == 0)
		gConverters= new OrdCollection;
	    Converter::FindConverter(this, want, gConverters);
	}
	converter= Converter::SelectConverter(gConverters);
    }
    SafeDelete(gConverters);
    if (converter && converter->CanConvert(this, want))
	return converter->Convert(this, want);
    return 0;
}

static bool _isFirstTime= TRUE;	//dirty!?!
 
StreamBuf *FileData::GetStreamBuf(bool forwriting)
{
    //problem: IsAscii() calls (indirectly) GetStreamBuf()
    //ie: we'll get a nasty Head-Recursion!
    //solution: as usu: stack to terminate recursion 
    //(since we only have depth one-> static var is sufficient)
    //sorry, I'don't like this solution too much; better solutions wellcome!
    
    if (!TESTBIT2(flags,eDataIsAscii)) {
	if (_isFirstTime) {
	    _isFirstTime= FALSE;
	    IsAscii();
	    _isFirstTime= TRUE;
	} else {
	    return new Filebuf(FullName(), forwriting ? output : input);
	}
    }
    
    if (IsAscii()) {
	if (forwriting) {
	    ELineSeparatorStyle lss= GetLineSeparatorStyleNew();
	    return new FilebufAsciiEncoder(FullName(), lss);
	} else { 
	    return new FilebufAsciiDecoder(FullName());
	}
    } else {
	return new Filebuf(FullName(), forwriting ? output : input);
    }
}

SeqCollection *FileData::GetContents()
{
    Dir dir(FullName());
    char *nm;
    OrdCollection *files= new OrdCollection;
    
    for (int i= 0; nm= dir(); i++)
	if (strcmp(nm, "."))
	    files->Add(new FileData(nm));
    
    return files;
}

Data *FileData::GetParent()
//ruedi: old test if at root of file-directory-hierarchy
//    if ( strcmp(fname, path_separator) == 0 ) return 0; //old
//this is unfortunate for DOS/MAC where volume-letters may also be used,
//therefore, use the following: (copied from aw)
{
    char *parent= gSystem->DirName(FullName());
    if (parent == 0)
	    return 0;
    if (strcmp(FullName(), parent) == 0)	// <<--
	    return 0;              
    return new FileData(parent, lookDeep);
}

extern Converter *gEtConverter;

bool FileData::SaveObject(Object *op, bool saveas)
{
    if (converter == 0) {
	Converter::CollectConverters();
	converter= gEtConverter;
    }
    if (! IsWritable()) {
	ShowAlert(eAlertNote, "File @B%s@B is not writable\n%s",
						Name(), gSystem->GetErrorStr());
	return FALSE;
    }
    if (converter) {
	return converter->ConvertTo(this, op);
    }
    return FALSE;
}

 
OStream& FileData::PrintOn(OStream &s)
{
    Data::PrintOn(s);
    return s << lookDeep SP << (long)converter SP;
}

IStream& FileData::ReadFrom(IStream &s)
{
    long cv;
    Data::ReadFrom(s);
    s >> Bool(lookDeep) >> cv;
    converter= (Converter*) cv;
    return s;
}

//---- StreamBufData --------------------------------------------------------

NewMetaImpl0(StreamBufData,Data);

StreamBufData::StreamBufData (StreamBuf *s, const Symbol &t)
    : Data("streambufdata")
{
    sb= s;
    type= t;
    SETBIT2(flags, eDataType);
    SETBIT2(flags, eDataDeepType);
    if (sb)
	sizehint= sb->tell();
    else
	sizehint= -1;
}

StreamBuf *StreamBufData::GetStreamBuf(bool)
{
    return sb;
}

//---- ObjectData --------------------------------------------------------------

NewMetaImpl(ObjectData,Data, (TP(object)));

extern TypeMatcher *gEtMatcher;

ObjectData::ObjectData(Object *op) : Data()
{
    object= op;
    typematcher= gEtMatcher;
    type= cDocTypeET;
    if (object)
	classtype= object->IsA();
    SETBIT2(flags, eDataType);
    SETBIT2(flags, eDataDeepType);
    SETBIT2(flags, eDataCreator);
    UpdateBit(eDataIsAscii, FALSE);
}

ObjectData::~ObjectData()
{
    SafeDelete(object);
}

Object *ObjectData::MakeObject(Class *want)
{
    if (GetClassType()->isKindOf(want))
	return object->DeepClone();
    Converter *converter= Converter::FindConverter(this, want);
    if (converter)
	return converter->Convert(this, want);
    return 0;
}

bool ObjectData::CanConvert(Class *want)
{
    if (gTypes)
	return Data::CanConvert(want);
    if (GetClassType()->isKindOf(want))
	return TRUE;
    if (Converter::FindConverter(this, want))
	return TRUE;
    return FALSE;
}

OStream& ObjectData::PrintOn(OStream &s)
{
    Data::PrintOn(s);
    return s << object;
}

IStream& ObjectData::ReadFrom(IStream &s)
{
    Data::ReadFrom(s);
    return s >> object;
}

