#ifndef Data_First  
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define Data_First

#include "RefCounted.h"
#include "Symbol.h"
#include "FileBuf.h"

class StreamBuf;
class MemBuf;
class Bitmap;
class SeqCollection;
class Converter;
class TypeMatcher;

extern const Symbol
	cDocTypeUndef,      // "UNDEF" document type
	cDocTypeAscii,      // document type for ordinary ascii-files
	cDocTypeET,         // ET++ file format
	cDocTypeCPlusPlus,
	cDocTypeCHeader,
	cDocTypeDirectory,  // System directory
	cDocTypeExec,
	cDocTypeOFile;      // .o file

extern const Symbol
	cDocCreatorUndef;  // Creator of document unknown

//---- Data --------------------------------------------------------------------

class Data : public RefCounted {
protected:
    u_long flags, bits, modtime, uniqueid;
    long sizehint;
    char *name;
    Symbol creator, type;
    Class *classtype;
    TypeMatcher *typematcher;
public:
    MetaDef(Data);
    Data(char *name= 0);
    ~Data();
    
    // Data internaly caches the stat (modtime etc)
    // Calling FlushStatCache forces the data to get the stuff
    // from disk instead its internal cache.
    void FlushStatCache();
 
    void UpdateBit(u_long b, bool val);

    virtual char *FullName();
    virtual char *ShortName();
    char *Name()
	{ return FullName(); }

    Class *GetClassType();
    void SetClassType(Class *cl)
	{ classtype= cl; }

    bool IsUntitled();
    Symbol Type();
    void SetType(const Symbol &t);
    Symbol Creator();
    void SetCreator(const Symbol &crea);
    virtual StreamBuf *GetStreamBuf(bool forwriting);
    StreamBuf *GetStreamBufForReading();
    StreamBuf *GetStreamBufForWriting();
    virtual bool SaveObject(Object *op, bool saveas);

    virtual Bitmap *GetBitmap();
    virtual SeqCollection *GetContents();
    virtual bool HasContents();
    virtual Data *GetParent();
    bool HasParent();			//added by SJH 24.6.92

    bool IsWritable();
    bool IsReadable();
    bool IsExecutable();
    bool IsValid();
    
    bool IsDirectory()
	{ return HasContents(); }
    bool IsETFormat();
    bool IsAscii();
    void SetIsAscii(bool isAscii);
    virtual ELineSeparatorStyle GetLineSeparatorStyleOrg();
    virtual ELineSeparatorStyle GetLineSeparatorStyleNew();
    virtual void SetLineSeparatorStyleNew(ELineSeparatorStyle lss);

    u_long UniqueId();
    long SizeHint();

    u_long ModifiedAt();

    virtual Object *AsObject(Class *want);
    virtual bool CanConvert(Class *want);
    virtual bool CheckTypes(OrdCollection *typelist);

    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);

protected:
    virtual u_long UpdateFields(u_long flags);
    virtual Object *MakeObject(Class *want);
};

//---- FileData ----------------------------------------------------------------
class FileData : public Data {
protected:
    bool lookDeep;
    Converter *converter;

    void WhatType(bool deep= FALSE);
private:
    bool fIsAscii;
    ELineSeparatorStyle fLineSeparatorStyleOrg;	//style we read
    ELineSeparatorStyle fLineSeparatorStyleNew;	//style we'll write in
    ELineSeparatorStyle DetermineLineSeparatorStyle();
 
public:
    MetaDef(FileData);
    FileData(char *nm, bool lookDeep= TRUE);
		// lookDeep indicates that the contents of the file is
		// considered for determining a file type
    ~FileData();

    Object *deepclone();	// optimization

    char *ShortName();
    char *GetLoadDir();
    char *GetBaseName()
	{ return ShortName(); }
    char *PathName()
	{ return FullName(); }
    void SetName(char *nm);
    Converter *GetConverter();
    ELineSeparatorStyle GetLineSeparatorStyleOrg();
    ELineSeparatorStyle GetLineSeparatorStyleNew();
    void SetLineSeparatorStyleNew(ELineSeparatorStyle lss);

    bool SaveObject(Object *op, bool saveas= FALSE);
    StreamBuf *GetStreamBuf(bool forwriting);

    SeqCollection *GetContents();
    Data *GetParent();

    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);

protected:
    u_long UpdateFields(u_long flags);
    Object *MakeObject(Class *want);
};

//---- StreamBufData -----------------------------------------------------------

class StreamBufData : public Data {
public:
    MetaDef(StreamBufData);
    StreamBufData(StreamBuf *sb= 0, const Symbol &type= cDocTypeUndef);
    StreamBuf *GetStreamBuf(bool forwriting);
protected:
    StreamBuf *sb;
};

//---- ObjectData --------------------------------------------------------------

class ObjectData : public Data {
protected:
    Object *object;
public:
    MetaDef(ObjectData);
    ObjectData(Object *op);
    ~ObjectData();
    bool CanConvert(Class *want);    
    Object *MakeObject(Class *want);
    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);
};

#endif
