#ifndef _Document_h_
#define _Document_h_

#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma interface
#endif

#include "Manager.h"
#include "Data.h"

enum FileDialogType {
    eFDTypeRead,
    eFDTypeWrite,
    eFDTypeImport
};

class Data;
class FileDialog;
class VObject;

class Document : public Manager {
protected:
    virtual void Store();
public:
    MetaDef(Document);

    Document(const Symbol &documentType= cDocTypeUndef);
    ~Document();
    
    //---- dialogs
    virtual FileDialog *MakeFileDialog(FileDialogType); 

    //---- document attributes
    int UniqueId();
    Symbol GetDocumentType()
	{ return docType; }
    bool IsUntitled();
    bool IsConverted();
    char *GetBaseName();
    void SetData(Data *fd);
    Data *GetData()
	{ return docData; }

    //---- change management
    virtual CommandProcessor *MakeCmdProcessor();
    virtual void Finish(); //here: GetCmdP()->Finish()
    virtual bool Modified();
    virtual bool SavedChanges();

    //---- misc
    void CollectParts(Collection*);
    void ExtCommand(int from, char *req, char *args, int len,
						    char *&ret, int &retlen);
    //---- menues
    virtual MenuBar *DoMakeMenuBar();
    Menu *MakeMenu(int menuId);
    void DoSetupMenu(Menu*);
    Command *DoMenuCommand(int);
    
    //---- loading
    virtual bool CanLoad(Data *data);
    virtual void AboutToLoad(Data *data);
    virtual bool LoadData(Data *data, bool unique);

    //---- generic menu commands
    virtual bool Close();       // return TRUE if OK
    virtual void Revert();
    virtual bool Save();        // return TRUE if OK
    virtual bool SaveAs();      // return TRUE if OK

    //---- input/output
    virtual bool DoFileIsAlreadyOpen(Document *shown, char *name);
    virtual void MakeBackup(char *loadDir, char *docName);
    virtual bool DoReadData(Data *data);        // return TRUE if OK
    virtual bool DoRead(IStream&, Data *data);  // return TRUE if OK
    virtual bool DoWrite(OStream&, Data *data); // return TRUE if OK
    
    //---- obsolete 
    virtual bool Load(char *filename= 0, bool unique= TRUE);

private:
    Symbol docType;
    Data *docData;
};

#endif
