#ifdef __GNUG__
#pragma implementation
#endif

#include "Env.h"

#include "Class.h"
#include "ETString.h"
#include "System.h"
#include "OrdColl.h"
#include "StreamBuf.h"
#include "CLib.h"

#ifdef OS_LINUX
extern "C" char *cuserid (char *);
#endif

static char etrc_filename[] = "ETRC";
static char rect_format[] = "%d:%d:%d:%d";

//---- EnvParser ---------------------------------------------------------------

class EnvParser {
    FILE *ifp;
public:
    EnvParser(FILE *f)
	{ ifp= f; }
    virtual void KeyValue(char*, char*, char*)
	{ }
    virtual void Char(int)
	{ }
    void Parse();
};

void EnvParser::Parse()
{
    enum { eStart = 0,
           eComment,
           eName,
           eWs_value,
           eValue,
           eType,
           eStart_value,
           eStart_type
           } state;
           
    char name[200], type[200], value[200];
    int c, ni, ti, vi;
    ni= ti= vi= 0;
    state= eStart;

    while ((c= fgetc(ifp)) != EOF) {
	if (c == '\n') {
	    state= eStart;
	    if (ni > 0) {
		name[ni++]= 0;
		type[ti++]= 0;
		value[vi++]= 0;
		ni= ti= vi= 0;
		KeyValue(name, value, type);
	    }
	    Char(c);
	    continue;
	}
	switch (state) {
	case eStart:             // start of line 
	    switch (c) {
	    case ' ':
	    case '\t':
		break;
	    case '#':
		state= eComment;
		break;
	    default:
		state= eName;
		break;
	    }
	    break;

	case eComment:             // comment
	    break;

	case eName:                // name
	    switch (c) {
	    case ' ':
	    case '\t':
	    case ':':
		state= eWs_value;
		break;
	    case '(':
		state= eStart_type;
		break;
	    default:
		break;
	    }
	    break;
	    
	case eWs_value:             // ws before value
	    if (c != ' ' && c != '\t')
		state= eValue;
	    break;

	case eValue:             // value
	    break;

	case eType:             // type
	    if (c == ')')
		state= eStart_value;
	    break;

	case eStart_value:             // optional ':'
	    state= (c == ':') ? eWs_value : eValue;
	    break;

	case eStart_type:
	    state= (c == ')') ? eStart_value : eType;
	    break;

	}
	switch (state) {
	case eName:
	    name[ni++]= c;
	    break;
	case eValue:
	    value[vi++]= c;
	    break;
	case eType:
	    type[ti++]= c;
	    break;
	}
	if (state != eValue)
	    Char(c);
    }
}

//---- ReadEnvParser -----------------------------------------------------------

class ReadEnvParser: public EnvParser {
    EnvLevel level;
public:
    ReadEnvParser(FILE *f, EnvLevel l) : EnvParser(f)
	{ level= l; }
    void KeyValue(char *name, char *value, char *type)
	{ Env::SetValue(name, value, level, type); }
};

//---- WriteEnvParser ----------------------------------------------------------

class WriteEnvParser: public EnvParser {
    FILE *ofp;
public:
    WriteEnvParser(FILE *f, FILE *of) : EnvParser(f)
	{ ofp= of;  }
    void KeyValue(char *name, char *value, char *type);
    void Char(int c)
	{ fputc(c, ofp); }
};

void WriteEnvParser::KeyValue(char *name, char *value, char*)
{
    EnvRec *er= Env::Lookup(name);
    if (er && er->changed) {
	er->changed= FALSE;
	if (er->op)
	    er->Write(er->op);
	if (er->op2)
	    er->Write2(er->op);
	fprintf(ofp, "%s", er->value);
    } else
	fprintf(ofp, "%s", value);
}

//---- EnvRec ------------------------------------------------------------------

NewMetaImpl(EnvRec,Object, (T(name), TS(value), TS(type), TE(level), TB(changed)));

EnvRec::EnvRec(Symbol n, char *v, char *t, EnvLevel l)
{
    name= n;
    value= strsave(v);
    level= l;
    if (t)
	type= strsave(t);
    else
	type= 0;
    changed= (l == eEnvChange);
    op= 0;
}

EnvRec::~EnvRec()
{
    SafeDelete(value);
    SafeDelete(type);
}

void EnvRec::Read(Object *op)
{
    IStream is(strlen(value), value);
    op->ReadFrom(is);
}

Object *EnvRec::Read2()
{
    Object *op= 0;
    IStream is(strlen(value), value);
    is >> op;
    return op;
}

void EnvRec::Write(Object *op)
{
    char *cp= new char[1000];
    OStream os(1000, cp);
    op->PrintOn(os);
    ChangeValue(cp, 0, eEnvChange);
    delete cp;
}

void EnvRec::Write2(Object*)
{
}

void EnvRec::ChangeValue(char *v, char*, EnvLevel l)
{
    if (l != eEnvChange && level == l) {
	if (gDebug)
	    fprintf(stderr,
	"EnvRec::ChangeValue: duplicate entry <%s=%s> for level %d; ignored\n",
						name.AsString(), v, l);
	return;
    }
    if (strcmp(value, v) != 0) {
	if (l == eEnvChange)
	    changed= TRUE;
	else {
	    changed= FALSE;
	    level= l;
	}
	strreplace(&value, v);
	if (changed && op)
	    Read(op);
    }
}

void EnvRec::InspectorId(char *buf, int bufSize)
{
    strn0cpy(buf, name.AsString(), bufSize);
}

int EnvRec::Compare(Object *op)
{
    return StrCmp((byte*)name.AsString(),
		  (byte*)Guard(op,EnvRec)->name.AsString(), -1, sortmap);
}

//---- EnvRec ------------------------------------------------------------------

NewMetaImpl(Env,Object, (TP(table), TB(terminated)));

static Env gEnv;

OrdCollection *Env::table;
bool Env::terminated;

Env::Env()
{
}

Env::~Env()
{
    terminated= TRUE;
    if (table) {
	table->FreeAll();
	SafeDelete(table);
    }
}

void Env::Init()
{
    if (table)
	return;
    
    if(!gSystem) 
	return;
	
    table= new OrdCollection;
    
    if (gEtDir == 0)
	gEtDir= CLib::Getenv("ET_DIR");

    char *et_rc   = form("%s/%s", gEtDir, etrc_filename);
    char *home_rc = form("%s/.%s", gSystem->HomeDirectory(), etrc_filename);
    u_long et_id, home_id, dummy;

    et_id = home_id = 0;
    gSystem->GetPathInfo(et_rc, &et_id, &dummy, &dummy, &dummy);
    gSystem->GetPathInfo(home_rc, &home_id, &dummy, &dummy, &dummy);

    if (et_id > 0)
	ReadFile(et_rc, eEnvGlobal);

    if (home_id > 0 && home_id != et_id)
	ReadFile(home_rc, eEnvUser);

    char **env = CLib::Environ();
    if (env) {
	for (int i= 0; env[i]; i++)
	    SetValue(env[i], eEnvEnvironment);
    }
}

EnvRec *Env::Lookup(char *n)
{
    if (table == 0)
	Env::Init();
    char *name = SymbolTable::Lookup(n);
    if (name) {
	Iter next(table);
	EnvRec *er;
	while (er= (EnvRec*) next())
	    if (er->name.Id() == name)
		return er;
    }
    return 0;
}

//---- Loading/Storing ---------------------------------------------------------

void Env::ReadFile(char *fname, EnvLevel level)
{
    FILE *ifp;
    if (ifp= fopen(fname, "r")) {
	ReadEnvParser rp(ifp, level);
	rp.Parse();
	fclose(ifp);
    }
}

void Env::Save()
{
    FILE *ifp, *ofp;

    char etrc_file[1024];
    char etrc_new_file[1024];
    char tmp_file[100];

    sprintf (etrc_file, "%s/.%s", gSystem->HomeDirectory(), etrc_filename);
    strcpy (etrc_new_file, etrc_file);
    strcat (etrc_new_file, ".new");
    sprintf (tmp_file, "/tmp/%s_%s", etrc_filename, cuserid(0));

    ofp = fopen (etrc_new_file, "w");
    if ( ofp ) {
	ifp = fopen (etrc_file, "r");
	if ( ifp == 0 ) {     // try to create file
	    ifp = fopen (etrc_file, "w");
	    if ( ifp )
		fclose (ifp);
	    ifp = fopen (etrc_file, "r");
	    if (ifp == 0)
		ifp = fopen (tmp_file, "rw");
	}
	if ( ifp ) {
	    WriteEnvParser wp(ifp, ofp);
	    wp.Parse();
	    Iter next(table);
	    EnvRec *er;
	    while (er = (EnvRec*) next())
		if (er->changed) {
		    er->changed= FALSE;
		    fprintf (ofp, "%s: %s\n", er->name.AsString(), er->value);
		}
	    fclose (ifp);
	    fclose (ofp);
	    CLib::Unlink (etrc_file);
	    CLib::Rename (etrc_new_file, etrc_file);
	} else {
	    fclose (ofp);
	    CLib::Unlink (etrc_new_file);
	    CLib::Unlink (tmp_file);
	}
    }
}

//---- GetValue ----------------------------------------------------------------

char *Env::getvalue(char *name)
{
    if (terminated)
	return 0;
    EnvRec *er= Lookup(form("%s.%s", gProgname, name));
    if (er == 0)
	er= Lookup(form("*.%s", name));
    if (er == 0)
	er= Lookup(name);
    if (er == 0)
	return 0;
    return er->value;
}

int Env::GetValue(char *name, int dflt)
{
    char *cp= Env::getvalue(name);
    if (cp) {
	int n;
	if (sscanf(cp, "%d", &n) == 1)
	    return n;
	if (strcmp(cp, "TRUE") == 0 || strcmp(cp, "ON") == 0
						    || strcmp(cp, "YES") == 0)
	    return 1;
	if (strcmp(cp, "FALSE") == 0 || strcmp(cp, "OFF") == 0
						    || strcmp(cp, "NO") == 0)
	    return FALSE;
    }
    return dflt;
}

char *Env::GetValue(char *name, char *dflt)
{
    char *cp= Env::getvalue(name);
    if (cp)
	return cp;
    return dflt;
}

Rectangle Env::GetValue(char *name, Rectangle dflt)
{
    char *cp= Env::getvalue(name);
    if (cp) {
	Rectangle r;
	if (sscanf(cp, rect_format,
			&r.origin.x, &r.origin.y, &r.extent.x, &r.extent.y) == 4)
	    return r;
    }
    return dflt;
}

//---- SetValue ----------------------------------------------------------------

void Env::SetValue(char *name, char *value, EnvLevel level, char *type)
{
    if (table == 0)
	Env::Init();
    
    EnvRec *er= Lookup(name);
    if (er)
	er->ChangeValue(value, type, level);
    else
	table->Add(new EnvRec(name, value, type, level));
}

void Env::SetValue(char *name, EnvLevel level)
{
    char *buf= strsave(name), *cp;
    if (cp= strchr(buf, '=')) {
	*cp++= 0;
	SetValue(buf, cp, level);
    } else
	SetValue(name, "1", level);
    delete buf;
}

void Env::SetValue(char *name, int value)
{
    SetValue(name, form("%d", value));
}

void Env::SetValue(char *name, double value)
{
    SetValue(name, form("%f", value));
}

void Env::SetBoolValue(char *name, bool value)
{
    SetValue(name, value ? "TRUE" : "FALSE");
}

void Env::SetValue(char *name, Rectangle r)
{
    SetValue(name, form(rect_format, r.origin.x, r.origin.y, r.extent.x, r.extent.y));
}

//---- Bind --------------------------------------------------------------------

bool Env::Bind(Object *&op, char *name)
{
    EnvRec *er= Lookup(form("%s.%s", gProgname, name));
    if (er == 0)
	er= Lookup(form("*.%s", name));
    if (er == 0)
	er= Lookup(name);
    
    if (er) {
	er->op= op;
	er->Read(op);
	return TRUE;
    }
    return FALSE;
}
