#ifdef __GNUG__
#pragma implementation
#endif

#include "FileDialog.h"

#include "Class.h"
#include "PopupItem.h"
#include "ImageItem.h"
#include "BorderItems.h"
#include "Buttons.h"
#include "CollView.h"
#include "OrdColl.h"
#include "Data.h"
#include "Fields.h"
#include "Menu.h"
#include "MenuBar.h"
#include "Alert_e.h"
#include "Scroller.h"
#include "Document.h"
#include "Application.h"
#include "Error.h"
#include "ETString.h"
#include "CLib.h"
#include "Env.h"
#include "MemBuf.h"

static const int
	cListMinWidth     = 250,
	cListMinHeight    = 200,
	cConfListMinWidth = cListMinWidth + 100;

static const char the_history_file_name[] = "~/.FileDialogHistory";
static char history_file_name [1024];

//---- FileDialog --------------------------------------------------------------

NewMetaImpl(FileDialog, Dialog, (TP(eti), TP(title), TP(initDir), TE(flags),
	TP(collview), TP(pathList), TP(doc), TP(fileData), TP(pathPopup),
	TP(current)));

SeqCollection *FileDialog::fileList = 0;
SeqCollection *FileDialog::dirList = 0;
bool FileDialog::show_hidden_files = TRUE;
char *FileDialog::filter_str;

FileDialog::FileDialog (char *title_str) : Dialog (title_str, eWinBlock)
{
    title = new TextItem (title_str, gSysFont, gPoint0);
    doc = 0;
    fileData = 0;
    initDir = 0;
    pathList = 0;
    current = 0;
    confDiag = new ConfigureDiag();
}

FileDialog::~FileDialog()
{
    SafeUnref(fileData);
    SafeDelete(initDir);
    SafeDelete(collview);
    if (pathList) {
	pathList->FreeAll();
	SafeDelete(pathList);
    }
    SafeUnref(current);
    SafeDelete(confDiag);
}

char *FileDialog::FileName()
{
    return fileData->FullName();
}

int FileDialog::ShowInWindow(FileDialogFlags f, Clipper *fp, EvtHandler *eh, char *p)
{
    flags= f;
    doc= eh;

    if (p == 0) {
	if (f == eFDRead)
	    p= "Open File Named:";
	else if (f == eFDWrite)
	    p= "Save in File Named:";
	else
	    p= "Import File Named:";
    }

    title->SetString(p);

    // remember initial working directory
    strreplace(&initDir, gSystem->WorkingDirectory());

    return Dialog::ShowAt(0, gPoint_1);
}

void FileDialog::UpdateList()
{
    GrShowWaitCursor wc;

    // SPEED UP !
    gSystem->StatCache(TRUE);

    if ( filter_field )
	filter_str = filter_field->GetString();
    else
	filter_str = 0;

    SafeUnref(current);
    current = new FileData(".");
    current->Ref();

    OrdCollection *files = new OrdCollection;
    Iter next (current->GetContents());
    Data *data;
    VObject *vop;
    while ( data = (Data *) next() )
	if ( vop = MakeFileItem(data) )
	    files->Add(vop);
    files->Sort();
    collview->SetCollection(files);

    if ( eti )
	eti->SetString("");
    gSystem->StatCache(FALSE);
}

VObject *
FileDialog::MakeFileItem (Data *data)
{
    char *file_name = data->ShortName();
    bool show_it = TRUE;

    // always show ".." (to go up 1 directory level)
    if ( file_name[0] == '.' && file_name[1] == '.' && file_name[2] == '\0' )
	show_it = TRUE;
    else {
	// check for hidden files (files starting with a '.')
	if ( ! show_hidden_files )
	    show_it = (file_name[0] != '.');
	// pass file name through pattern filter
	if ( show_it ) {
	    if ( filter_str != 0 && filter_str[0] != 0
		 && !(filter_str[0] == '*' && filter_str[1] == '\0') )
		show_it = Glob::Match (filter_str, file_name, FALSE);
	}
    }

    if ( show_it )
	return new FileItem (data);
    else
	return 0;
}

VObject *FileDialog::DoMakeContent()
{
    pathPopup = new PopupButton (cIdFDPath, cIdNone,
				 new Menu ("Directory", FALSE, FALSE));

    collview = new CollectionView (this, new OrdCollection, eCVDontStuckToBorder);
    collview->SetMinExtent (Point (cListMinWidth, 0));
    collview->SetId (cIdFDFileList);
    collview->SetSyncSelection (TRUE);

    return
	new Matte (gPoint16,
	    new VBox (gPoint10, VObjAlign (eVObjHLeft | eVObjHExpand | eVObjVExpand),
		new HBox (gPoint4, VObjAlign (eVObjVBase | eVObjHExpand),
		    new TextItem ("Directory:", gSysFont, gPoint0),
		    pathPopup,
		    0
		),
		scroller = new Scroller (collview, Point (cListMinWidth, cListMinHeight)),
		new Box (cIdNone, Point(2,0), gPoint2, VObjAlign (eVObjHLeft | eVObjVBase | eVObjHExpand | eVObjVExpand),
		    title,
		    new TextItem ("Filter:", gSysFont, gPoint0),
		    eti = new TextField (cIdFDName, 20),
		    filter_field = new TextField (cIdFDFilter, 5),
		    0
		),
		MakeButtons(),
		0
	    )
	);
}

VObject *FileDialog::MakeButtons()
{
    return
	new HBox (Point(20), VObjAlign (eVObjVBase | eVObjHEqual | eVObjHExpand),
	    openButton = new ActionButton (cIdOk, flags == eFDWrite ? "Save" : "Open", TRUE),
	    new ActionButton (cIdCancel, "Cancel"),
	    new ActionButton (cIdFDUpdate, "Update"),
	    0
	);
}

void FileDialog::DoSetDefaults()
{
    show_hidden_files = Env::GetValue("FileDialog.HiddenFiles", TRUE);
    UpdatePath();
    UpdateList();
}

void FileDialog::DoSetup()
{
    bool have_path = (eti->GetTextSize() > 0);
    bool is_dir = FALSE;
    bool is_valid = FALSE;
    if ( have_path ) {
	char pathname[1024];
	eti->GetString(pathname, sizeof(pathname));
	if ( ! gSystem->ExpandPathName(pathname, sizeof(pathname)) ) {
	    FileData fd(pathname, FALSE);
	    is_valid = fd.IsValid();
	    if ( is_valid )
		is_dir = fd.IsDirectory();
	}
    }
    if ( flags == eFDWrite ) {
	openButton->SetLabel( is_dir ? "Open" : "Save", TRUE );
	openButton->Enable(have_path, TRUE);
    } else {
	openButton->Enable(is_valid, TRUE);
    }
}

void FileDialog::DoRestore()
{
    gSystem->ChangeDirectory(initDir);
}

bool FileDialog::AboutToDismiss()
{
    bool last_value=Env::GetValue("FileDialog.HiddenFiles",TRUE);
    if(last_value!=show_hidden_files) {
	Env::SetBoolValue("FileDialog.HiddenFiles", show_hidden_files);
	Env::Save();
    }
    return Dialog::AboutToDismiss();
}

void FileDialog::Control(int id, int part, void *val)
{
    VObject *vop;

    switch (id) {
    case cIdFDUpdate:
	UpdateList();
	return;

    case cIdFDFileList:
	switch (part) {
	case cPartCollSelect:
	    if ( int(val) >= 0 ) {
		vop = (VObject *) collview->GetCollection()->At( (int) val );
		char *fname= vop->AsString();
		if (fname && strlen(fname) > 0)
		    eti->SetString(fname);
	    } else
		eti->SetString("");
	    break;
	case cPartCollDoubleSelect:
	    if (OpenOrChangeDir()) {
		UpdateHistory(fileData->FullName());
		Dialog::Control(cIdOk, cPartAction, val);
		return;
	    }
	    break;
	}
	return;

    case cIdOk:
	if (!OpenOrChangeDir())
	    return;
	UpdateHistory(fileData->FullName());
	break;

    case cIdFDPath:
	{
	    Data *d= (Data*) pathList->At((int)val - cIdFDComponent);
	    if (d)
		eti->SetString(d->FullName());
	    OpenOrChangeDir();
	}
	return;
    }
    Dialog::Control(id, part, val);
}

void FileDialog::InputKbd(Token &t)
{
    if ( /* t.IsKey() && !(t.Flags & eFlgMetaKey) && */ scroller->ContainsPoint(t.Pos) ) {
	Command *cmd = scroller->Input(t.Pos, t, GetWindow());
	if ( cmd )
	    return;
    }
    Manager::InputKbd(t);
}

bool FileDialog::ChangeDirectory()
{
    if (fileData->IsDirectory()) {
        if(!strcmp(fileData->FullName(),gSystem->WorkingDirectory())) {// current dir is equal to wanted dir->do nothing 
								    // (happens if dir is selected from history)
	    eti->SetString("",TRUE);
	    return TRUE;
	}
	
	if (!fileData->IsReadable()) {
	    ShowAlert(eAlertNote, "Cannot open directory @B%s@B\n%s",
			fileData->FullName(), gSystem->GetErrorStr());
	    return TRUE;
	}
	if (!gSystem->ChangeDirectory(fileData->FullName()))
	    ShowAlert(eAlertNote, "Cannot change directory to @B%s@B", fileData->FullName());
	else {
	    UpdatePath();
	    UpdateList();
	}
	return TRUE;
    }
    return FALSE;
}

bool FileDialog::OpenOrChangeDir()
{
    char pathname[1024];
    strcpy( pathname, eti->GetString() );

    strreplace(&symbolic_path_name, pathname);

    if (gSystem->ExpandPathName(pathname, sizeof(pathname))) {
	ShowAlert(eAlertNote, "%s:\n%s", pathname, gSystem->GetErrorStr());
	return FALSE;
    }

    SafeUnref(fileData);
    fileData= new FileData(pathname, TRUE); // deep
    fileData->Ref();

    char *fname= fileData->FullName();
    char *shortname= fileData->ShortName();

    if (!gSystem->IsAbsolutePath(pathname))
	strreplace(&symbolic_path_name, fname);
    
    if (ChangeDirectory())
	return FALSE;

    if (flags == eFDWrite) {
	if (!fileData->IsWritable()) {
	    ShowAlert(eAlertNote, "File @B%s@B is not writable\n%s",
				shortname, gSystem->GetErrorStr());
	    return FALSE;
	}
	if (!gSystem->AccessPathName(fname, 0)) {
	    if (ShowAlert(eAlertCaution, "File @B%s@B exists !\nOverwrite ?",
			shortname) != cIdYes)
		return FALSE;
	}
    } else {
	if (!fileData->IsReadable()) {
	    ShowAlert(eAlertNote, "Cannot open file @B%s@B for reading\n%s",
					shortname, gSystem->GetErrorStr());
	    return FALSE;
	}
	if (WrongType()) {
	    ShowAlert(eAlertNote, "File @B%s@B has wrong type (@B%s@B)",
					shortname, fileData->Type().AsString());
	    return FALSE;
	}
    }
    return TRUE;
}

bool FileDialog::WrongType()
{
    if (doc)
    {
	if (doc->IsKindOf(Document))
	{
	    return ! ((Document*)doc)->CanLoad(fileData);
	}
	else if (doc->IsKindOf(View))
	{
	    return ! ((View*)doc)->CanPaste(fileData);
	}
	else
	    return ! gApplication->CanOpen(fileData);
    }
    return FALSE;
}

void FileDialog::UpdatePath()
{
    Data *here;
    if (pathList) {
	pathList->FreeAll();
	SafeDelete(pathList);
    }
    pathList= new OrdCollection;
    for (here= new FileData("."); here; here= here->GetParent())
	pathList->AddFirst(here);

    Iter next(pathList);
    Data *dp;
    SeqCollection *newpath= new OrdCollection;
    for (int i= 0; dp= (Data*)next(); i++)
	newpath->Add(new MenuButtonItem(cIdFDComponent+i, dp->ShortName()));

    pathPopup->SetCollection(newpath);
    pathPopup->SetValue(cIdFDComponent+i-1);
}

MenuBar *FileDialog::DoMakeMenuBar()
{
    MenuBar *mb = new MenuBar();
    mb->AddMenu(new Menu("Files"));
    mb->AddMenu(new Menu("Directories"));
    Menu *m= new Menu("Options");
    m->AppendItems(
	"Edit Files",		cIdFDConfFiles,
	"Edit Directories",	cIdFDConfDirs,
	"-",
	"Hidden Files",		cIdFDHidden,
	"Create Directory",	cIdFDMakeDir,
	0
    );
    mb->AddMenu(m);
    return mb;
}

char *
TruncateIfTooLong(char *name)
{
    if (!name || !*name || strlen(name) < 80)
	return name;
    static char	buf[100];
    int	orig_len = strlen(name);
    char *p = name;
    int len = orig_len - 77;
    p += len;
    sprintf(buf,"...%s", p);
    return buf;
}

void FileDialog::DoSetupMenu(Menu* menu)
{
    int i;
    TextItem *ti;

    LoadLists();

    if ( strcmp(menu->GetName(), "Files") == 0 ) {
	menu->RemoveAllItems();
	Iter next(fileList);
	for (i= 0; ti= (TextItem*)next(); i++) {
	    menu->AppendItem(TruncateIfTooLong(ti->AsString()), cIdFDLoadFile+i);
	    if (ti->GetFont()->Face() == eFaceBold)
		menu->CheckItem(cIdFDLoadFile+i, TRUE);
	}
	if ( i == 0 )
	    menu->AppendItem("--none--", cIdNone);
    } else if ( strcmp(menu->GetName(), "Directories") == 0 ) {
	menu->RemoveAllItems();
	Iter next(dirList);
	for (i= 0; ti= (TextItem*)next(); i++) {
	    menu->AppendItem(TruncateIfTooLong(ti->AsString()),cIdFDChangeDir+i);
	    if (ti->GetFont()->Face() == eFaceBold)
		menu->CheckItem(cIdFDChangeDir+i, TRUE);
	}
	if ( i == 0 )
	    menu->AppendItem("--none--", cIdNone);
    } else if ( strcmp(menu->GetName(), "Options") == 0 ) {
	menu->EnableItems(cIdFDConfFiles, cIdFDConfDirs, cIdFDHidden, 0);
	menu->CheckItem(cIdFDHidden, show_hidden_files);
	char *dir = NULL;
	char pathname[1024];
	eti->GetString(pathname, sizeof(pathname));
	if (gSystem->ExpandPathName(pathname, sizeof(pathname)) == FALSE)
	    if (gSystem->AccessPathName(pathname, 0))
	    	dir = pathname;
	menu->ReplaceItem(cIdFDMakeDir, form("Create Directory %s", dir ? dir : ""));
	if (dir && *dir)
	    menu->EnableItem(cIdFDMakeDir);
    }
}

Command *FileDialog::DoMenuCommand(int cmd)
{
    switch(cmd) {
    case cIdFDMakeDir:
	CreateDirectory(eti->GetString());
        return gNoChanges;
    case cIdFDConfFiles:
	confDiag->SetFileMode(TRUE);
	confDiag->SetName("Edit Files");
	confDiag->ShowUnderMouse();
        return gNoChanges;
    case cIdFDConfDirs:
	confDiag->SetFileMode(FALSE);
	confDiag->SetName("Edit Directories");
	confDiag->ShowUnderMouse();
        return gNoChanges;
    case cIdFDHidden:
	show_hidden_files = !show_hidden_files;
	UpdateList();
        return gNoChanges;
    }
    if (cmd >= cIdFDChangeDir && cmd < (cIdFDChangeDir + cFDMaxHistory)) {
	eti->SetString(dirList->At(cmd-cIdFDChangeDir)->AsString());
	Control(cIdOk, cPartAction, 0);
    }
    else if (cmd >= cIdFDLoadFile && cmd < (cIdFDLoadFile + cFDMaxHistory)) {
	eti->SetString(fileList->At(cmd-cIdFDLoadFile)->AsString());
	Control(cIdOk, cPartAction, 0);
    }
    return gNoChanges;
}

void FileDialog::CreateDirectory(char *dir)
{
    if (gSystem->CreateDirectory(dir, 0775))
	ShowAlert(eAlertNote, "Cannot create directory @B%s@B\n%s",
			gSystem->BaseName(dir), gSystem->GetError());
    else
	UpdateList();
}

void FileDialog::UpdateHistory(char *path)
{
    char file_name[1024];
    TextItem *ti;
    Object *op;
    char *ptr;
    bool have_directory = TRUE;

    LoadLists();

    strcpy(file_name, path);
    ptr = file_name;
    FileData fd(path);
    if ( !fd.IsDirectory() ) {
	ti= new TextItem(file_name, gFixedFont, gPoint0);
	if (op = fileList->Find(ti)) {
	    fileList->RemovePtr(op);
	    delete ti;
	    ti = (TextItem *) op;
	}
	while (fileList->Size() >= cFDMaxHistory)
	    RemoveLastNotMarked(fileList);
	fileList->AddFirst(ti);
	ptr = gSystem->DirName(file_name);
	if (strcmp(ptr, ".") == 0)
	    have_directory = FALSE;
    }

    if (have_directory) {
	ti= new TextItem(ptr, gFixedFont, gPoint0);
	if (op = dirList->Find(ti)) {
	    dirList->RemovePtr(op);
	    delete ti;
	    ti = (TextItem *) op;
	}
	while (dirList->Size() >= cFDMaxHistory)
	    RemoveLastNotMarked(dirList);
	dirList->AddFirst(ti);
    }

    StoreLists();
}

void FileDialog::RemoveLastNotMarked(SeqCollection *list)
{
    Object *tempLastObj;
    TextItem *ti;
    bool fixedMenuItem= TRUE;

    tempLastObj= list->Last();
    do
    {
	ti= (TextItem*)tempLastObj;
	if (ti->GetFont()->Face() != eFaceBold)
	    fixedMenuItem= FALSE;
	else
	    tempLastObj= list->Before(tempLastObj);

	// if all menuitems are marked remove the last one
	if (list->First() == tempLastObj) {
	    fixedMenuItem = FALSE;
	    tempLastObj= list->Last();
	}
    }
    while (fixedMenuItem);
    gSystem->AddCleanupObject(list->Remove(tempLastObj));
}

void FileDialog::StoreLists()
{
  if ( history_file_name[0] == '\0' )
    {
      strcpy (history_file_name, the_history_file_name);
      gSystem->ExpandPathName (history_file_name, sizeof (history_file_name));
    }

  if ( fileList == 0 )
    fileList = new OrdCollection;
  if ( dirList == 0 )
    dirList = new OrdCollection;

  MemBuf mb;
  OStream os (&mb);

  os << fileList NL << dirList NL;

  int fd = CLib::Creat (history_file_name, 0644);
  if ( fd == -1 )
    {
      ::SysError ("FileDialog", "cannot open %s for writing", the_history_file_name);
      return;
    }

  if ( -1 == CLib::Write (fd, mb.Base(), mb.tell()) )
    {
      ::SysError ("FileDialog", "cannot write to %s", the_history_file_name);
      CLib::Close (fd);
      CLib::Unlink (history_file_name);
      return;
    }
  CLib::Close (fd);
}

void FileDialog::LoadLists()
{
  if ( history_file_name[0] == '\0' )
    {
      strcpy (history_file_name, the_history_file_name);
      gSystem->ExpandPathName (history_file_name, sizeof (history_file_name));
    }

  FileData fd (history_file_name);

  if ( fd.IsValid() )
    {
      if ( fd.IsReadable() )
	{
	  if ( fd.SizeHint() > 40 )
	    {
	      OrdCollection * tmp;
	      IStream is (history_file_name);

	      if ( is.good() )
		{
		  tmp = 0;
		  is >> tmp;
		  if ( tmp != 0 )
		    {
		      if ( fileList != 0 )
			{
			  fileList->FreeAll();
			  delete fileList;
			}
		      fileList = tmp;
		    }
		}

	      if ( is.good() )
		{
		  tmp = 0;
		  is >> tmp;
		  if ( tmp != 0 )
		    {
		      if ( dirList != 0 )
			{
			  dirList->FreeAll();
			  delete dirList;
			}
		      dirList = tmp;
		    }
		}
	    }
	  else
	    {
	      ::Warning ("FileDialog", "file %s seems to be corrupted", the_history_file_name);
	    }
	}
      else
	{
	  ::SysError ("FileDialog", "cannot read %s", the_history_file_name);
	}
    }

  if ( fileList == 0 )
    fileList = new OrdCollection;
  if ( dirList == 0 )
    dirList = new OrdCollection;
}

//---- FileItem ----------------------------------------------------------------

NewMetaImpl0(FileItem,HBox);

FileItem::FileItem(char *name, bool shallow)
    : HBox( (VObject *)0, (VObject *)0 )
{
    data= new FileData(name, !shallow); // shallow
    data->Ref();
    Init();
}

FileItem::FileItem(Data *d)
    : HBox( (VObject *)0, (VObject *)0 )
{
    data= d;
    if (data)
	data->Ref();
    Init();
}

void FileItem::Init()
{
    SetGap( gPoint2 );
    SetAlign( VObjAlign(eVObjHExpand|eVObjVCenter) );
    Add( new ImageItem(data->GetBitmap(), /* base= */ 0, /* freeold= */ FALSE) );
    Add( new TextItem(data->ShortName(), gFixedFont, gPoint0) );
}

FileItem::~FileItem()
{
    SafeUnref(data);
}

char *FileItem::AsString()
{
    return data->ShortName();
}

int FileItem::Compare(Object *op)
{
    FileItem *other= Guard(op,FileItem);
    if ( strcmp(data->ShortName(), "..") == 0 )
	return -1;
    if ( strcmp(other->GetData()->ShortName(), "..") == 0 )
	return 1;
    return StrCmp( (byte*)data->ShortName(), (byte*)other->GetData()->ShortName(),
			-1, sortmap );
}

//---- ConfigureDiag -----------------------------------------------------

NewMetaImpl0(ConfigureDiag,Dialog);

ConfigureDiag::ConfigureDiag() : Dialog("Configure Dialog")
{
}

VObject *ConfigureDiag::DoMakeContent()
{
    collView= new CollectionView(this, NULL, eCVDontStuckToBorder);
    collView->SetMinExtent(Point(cConfListMinWidth, 0));
    collView->SetId(cIdFDConfigList);

    VObject *layout= new Matte(
	new VBox(Point(20), VObjAlign(eVObjVBase|eVObjHEqual|eVObjHExpand),
	    new TextItem("Click to make the item persistent"),
	    new HBox(Point(20), VObjAlign(eVObjVBase|eVObjHEqual|eVObjHExpand),
		new VBox(Point(20), VObjAlign(eVObjVBase|eVObjHEqual|eVObjHExpand),
		    new Scroller(collView, Point(cConfListMinWidth, cListMinHeight)),
		    0
		),
	       	0
	    ),
	    new HBox(Point(20), VObjAlign(eVObjVBase|eVObjHEqual|eVObjHExpand),
		new ActionButton(cIdOk, "Done", TRUE),
		toggle_pers_button= new ActionButton(cIdFDTogglePers, "Toggle Persistency"),
		remove_button= new ActionButton(cIdFDRemoveItem, "Remove"),
		0
	    ),
	    0
	)
    );

    return layout;
}

void ConfigureDiag::DoSetDefaults()
{
    SeqCollection *list;
    list = fileDirMode ? FileDialog::GetFileList() : FileDialog::GetDirList();
    collView->SetCollection(list, /* bool free_old = */ FALSE);
}

void ConfigureDiag::DoStore()
{
    FileDialog::StoreLists();
}

void ConfigureDiag::DoSetup()
{
    bool hasSelection = FALSE;
    Rectangle sel = collView->GetSelection();
    if ( sel.IsNotEmpty() ) {
	if ( sel.origin.y < collView->GetCollection()->Size() )
	    hasSelection = TRUE;
    }
    toggle_pers_button->Enable (hasSelection, TRUE);
    remove_button->Enable (hasSelection, TRUE);
}

void ConfigureDiag::Control(int id, int part, void *val)
{
    if ( id == cIdFDTogglePers
	|| (id == cIdFDConfigList && part == cPartCollDoubleSelect) )
    {
	Rectangle sel = collView->GetSelection();
	if ( sel.IsNotEmpty() ) {
	    int at = sel.origin.y;
	    if ( at < collView->GetCollection()->Size() ) {
		TextItem *ti = (TextItem *) collView->GetCollection()->At(at);
		Font *ft = ti->GetFont();
		GrFace face = (ft->Face() == eFacePlain) ? eFaceBold : eFacePlain;
		ti->SetFont(ft->WithFace(face));
		ti->ForceRedraw();
	    }
	}
	DoSetup();
    } else if ( id == cIdFDRemoveItem ) {
	Rectangle sel = collView->GetSelection();
	if ( sel.IsNotEmpty() ) {
	    int at = sel.origin.y;
	    if ( at < collView->GetCollection()->Size() ) {
		Object *item = collView->GetCollection()->At(at);
		collView->GetCollection()->RemovePtr(item);
		if ( at >= collView->GetCollection()->Size() )
		    collView->ClearSelection(FALSE);
		collView->ForceRedraw();
	    }
	}
	DoSetup();
    } else
	Dialog::Control(id, part, val);
}
