#ifndef FileDialog_First
#define FileDialog_First

#ifdef __GNUG__
#  pragma interface
#endif
 
#include "Dialog.h"
#include "Box.h"

class TextField;
class TextItem;
class CollectionView;
class OrdCollection;
class Menu;
class PopupButton;
class ImageItem;
class TextItem;
class ActionButton;
class Scroller;
class Data;
class ConfigureDiag;

const int cFDMaxHistory = 15;

const int cIdFDName       = cIdFirstUser + 0,
	  cIdFDFilter     = cIdFirstUser + 1,
	  cIdFDFileList   = cIdFirstUser + 2,
	  cIdFDUpdate     = cIdFirstUser + 3,
	  cIdFDPath       = cIdFirstUser + 4,
	  cIdFDHidden     = cIdFirstUser + 5,
	  cIdFDComponent  = cIdFirstUser + 6;

const int cIdFDConfFiles  = cIdFirstUser + 100,
          cIdFDConfDirs   = cIdFirstUser + 101,
          cIdFDConfigList = cIdFirstUser + 102,
          cIdFDTogglePers = cIdFirstUser + 103,
          cIdFDRemoveItem = cIdFirstUser + 104,
          cIdFDMakeDir    = cIdFirstUser + 105,
          cIdFDChangeDir  = cIdFirstUser + 106,
          cIdFDLoadFile   = cIdFDChangeDir + cFDMaxHistory;

const int cIdFDLast = cIdFDLoadFile + cFDMaxHistory;

//---- FileDialog --------------------------------------------------------

enum FileDialogFlags {
    eFDRead,
    eFDWrite,
    eFDImport
};

class FileDialog : public Dialog {
public:
    MetaDef(FileDialog);
    
    FileDialog(char *title= "File Dialog");
    ~FileDialog();
    
    int ShowInWindow(FileDialogFlags, Clipper*, EvtHandler *eh= 0, char *msg= 0);
    char *FileName();
    char *SymbolicFileName()
	{ return symbolic_path_name; }
    Data *GetFileData()
	{ return fileData; }
    Data *GetData()
	{ return fileData; }
    virtual bool WrongType();
    virtual bool ChangeDirectory();

protected:
    bool AboutToDismiss();
    void InputKbd(Token &t);
    VObject *DoMakeContent();
    void DoSetup();
    void DoRestore();
    void DoSetDefaults();
    void Control(int id, int, void *v);
    int ItemHeight();
    virtual VObject *MakeButtons();
    virtual void UpdateList();
    virtual void UpdatePath();
    virtual bool OpenOrChangeDir();
    virtual VObject *MakeFileItem(Data *);

    MenuBar* DoMakeMenuBar();
    void DoSetupMenu(Menu* menu);
    Command* DoMenuCommand(int cmd);

    void CreateDirectory(char *dir);
    static void RemoveLastNotMarked(SeqCollection *);

public:
    static void UpdateHistory(char *pn);
    static void StoreLists();
    static void LoadLists();

    static SeqCollection *GetFileList()
	{ return fileList; }
    static SeqCollection *GetDirList()
	{ return dirList; }

protected:
    TextField *eti;
    TextField *filter_field;
    TextItem *title;
    char *initDir;
    FileDialogFlags flags;
    CollectionView *collview;
    OrdCollection *pathList;
    EvtHandler *doc;
    Data *fileData;
    ActionButton *openButton;
    PopupButton *pathPopup;
    Scroller *scroller;
    Data *current;
    ConfigureDiag *confDiag;
    char *symbolic_path_name;	// non-expanded shell variables

    static SeqCollection *fileList;
    static SeqCollection *dirList;

    static bool show_hidden_files;
    static char *filter_str;
};

//---- FileItem --------------------------------------------------------------

class FileItem: public HBox {
public:
    MetaDef(FileItem);
    FileItem(char *name, bool shallow= TRUE);
    FileItem(Data *);
    ~FileItem();

    Data *GetData()
	{ return data; }
    char *AsString();
    int Compare(Object *);

protected:
    void Init();

    Data *data;
};

//---- ConfigureDiag -----------------------------------------------------

class ConfigureDiag : public Dialog {
    CollectionView	*collView;
    Scroller		*scroller;
    ActionButton	*toggle_pers_button;
    ActionButton	*remove_button;
    bool 		fileDirMode; 
public:
    MetaDef(ConfigureDiag);
    ConfigureDiag();
    void SetFileMode(bool mode) { fileDirMode = mode; }
    void Control(int, int, void *);
    VObject *DoMakeContent();
    void DoSetDefaults();
    void DoStore();
    void DoSetup();
};

extern char *TruncateIfTooLong(char *);
#endif
