#ifdef __GNUG__
#  pragma implementation
#endif

#include "FindDialog.h"

#include "Class.h"
#include "RegularExp.h"
#include "BorderItems.h"
#include "OneOfCluster.h"
#include "ManyOfCluster.h"
#include "Fields.h"
#include "Buttons.h"
#include "Form.h"
#include "Box.h"
#include "Alert_e.h"
#include "FindChange_e.h"
#include "TextView.h"
#include "TextCmd.h"
#include "ETMath.h"
#include "MacroCommand.h"

//---- tags ----------------------------------------------------------------

const int
  cIdFindText		= cIdFirstUser + 1,
  cIdChangeText		= cIdFirstUser + 2,

  cIdDirection		= cIdFirstUser + 3,
  cIdForward		= cIdDirection,
  cIdBackward		= cIdDirection + 1,

  cIdOptions		= cIdFirstUser + 6,
  cIdIgnoreCase		= cIdOptions,
  cIdMatchWholeWord	= cIdOptions + 1,

  cIdChangeAllScope	= cIdFirstUser + 9,
  cIdChangeAll		= cIdChangeAllScope,
  cIdChangeSel		= cIdChangeAllScope + 1,

  cIdDoNext		= cIdFirstUser + 12,
  cIdDoPrev		= cIdFirstUser + 13,
  cIdDoChange		= cIdFirstUser + 14,
  cIdDoChangeAndFind	= cIdFirstUser + 15,
  cIdDoChangeAll	= cIdFirstUser + 16;

//---- FindDialog --------------------------------------------------------------

NewMetaImpl (FindDialog, Dialog,
	     (TP(rex), TP(tvp), TP(ei1), TP(ei1), TP(msg_field),
	      TP(modecl), TP(optionscl), TP(scopecl),
	      TP(next), TP(change), TP(changeAndFind), TP(changeAll)));

FindDialog::FindDialog (char * title) : Dialog (title, eWinFixed)
{
  tvp = 0;
  rex = new RegularExp;
}

FindDialog::~FindDialog()
{
  SafeDelete (rex);
}

void
FindDialog::SetTextViewFocus (TextView * tv)
{
  if ( tvp != 0 )
    tvp->RemoveObserver (this);
  tvp = tv;
  if ( tvp != 0 )
    tvp->AddObserver (this);
  if ( ei1 != 0 )
    DoSetup();
}

void
FindDialog::DoObserve (int id, int part, void * val, Object * op)
{
  if ( op == tvp && part == cPartSenderDied )
    {
      tvp = 0;
      if ( ei1 != 0 )
	DoSetup();
    }
}

VObject *
FindDialog::DoMakeContent()
{
    VObject * layout =
	new Matte (gPoint8,
	    new VBox (gPoint8, eVObjHExpand,
		new Form (
		    "Find:",   ei1 = new TextField (cIdFindText, 20),
		    "Change:", ei2 = new TextField (cIdChangeText, 20),
		    0
		),
		new HBox (gPoint8, VObjAlign(eVObjVTop | eVObjHGapExpand),
		    new BorderItem ("Direction",
			modecl = new OneOfCluster (cIdDirection,
			    "Forward",
			    "Backward",
			    0
			)
		    ),
		    new BorderItem ("Options",
			optionscl = new ManyOfCluster (cIdOptions,
			    "Ignore Case",
			    "Match Whole Word",
			    0
			)
		    ),
		    new BorderItem ("Change All Scope",
			scopecl = new OneOfCluster (cIdChangeAllScope,
			    "All of Document",
			    "Selection Only",
			    0
			)
		    ),
		    0
		),
		msg_field = new TextField (cIdNone, 30),
		new HBox (gPoint8, VObjAlign(eVObjVBase|eVObjHExpand),
		    next = new ActionButton (cIdDoNext,           "Next", TRUE),
		    change = new ActionButton (cIdDoChange,       "Change"),
		    changeAndFind = new ActionButton (cIdDoChangeAndFind,       "Change & Find"),
		    changeAll = new ActionButton (cIdDoChangeAll, "Change All"),
		    new ActionButton (cIdCancel, "Close"),
		    0
		),
		0
	    )
	);

    msg_field->SetEditable (FALSE);
    msg_field->SetFlag (eVObjVFixed);
    optionscl->SetItemOn (cIdIgnoreCase);

    return layout;
}

void
FindDialog::DoSetDefaults()
{
  // copy selection in associated TextView as default value
  if ( tvp != 0 && tvp->HasSelection() )
    {
      int from, to;
      Text * txt = tvp->GetText();
      byte ch;
      byte buf [512];
      int i, j;

      // copy the first line of the current selection
      // and quote regex meta characters
      tvp->GetSelection (&from, &to);
      for ( i = from, j = 0; i < to && j < (sizeof (buf) - 1); i++, j++ )
	{
	  ch = (*txt)[i];
	  if ( ch == '\n' )
	    break;
	  if ( strchr (".*+?[]^$\\", ch) != 0 )
	    {
	      if ( j < (sizeof (buf) - 2) )
		{
		  buf[j++] = '\\';
		}
	      else
		break;
	    }
	  buf[j] = ch;
	}
      buf[j] = 0;
      ei1->SetString ((char *) buf, TRUE);
    }
}

void
FindDialog::DoSetup()
{
  bool find_enable = (tvp != 0 && ei1->GetTextSize() > 0);
  bool change_enable = (find_enable && !tvp->TestFlag (eTextViewReadOnly));
  next->Enable (find_enable);
  change->Enable (change_enable);
  changeAndFind->Enable (change_enable);
  changeAll->Enable (change_enable);
}

void
FindDialog::Control (int id, int part, void * val)
{
  switch (id)
    {
    case cIdDoNext:
      Find (ei1->GetString(), modecl->GetCurrentItem() == cIdForward);
      break;

    case cIdDoChange:
      DoChange (ei1->GetString(), ei2->GetString());
      break;

    case cIdDoChangeAndFind:
      DoChange (ei1->GetString(), ei2->GetString());
      Find (ei1->GetString(), modecl->GetCurrentItem() == cIdForward);
      break;

    case cIdDoChangeAll:
      DoChangeAll (ei1->GetString(), ei2->GetString());
      break;

    default:
      Dialog::Control (id, part, val);
      break;
    }
}

void
FindDialog::ShowMessage (char * str, bool bell)
{
  if ( msg_field )
    msg_field->SetString (str, TRUE);
  Window *win;
  if ( bell && (win = GetWindow()) )
    win->Bell();
}

void FindDialog::RevealSelection()
{
    if(!tvp)
	return;
    Rectangle rect= tvp->SelectionRect().Expand(Point(8,0));
    tvp->RevealRectAt(rect,-6);
}

void
FindDialog::SendDown (int id, int part, void * val)
{
  if ( part == cPartSelectionChanged )
    if ( ((Object *)val)->IsKindOf (TextView) )
      FindChange::SetFocus ((TextView *) val);
}

bool
FindDialog::Compile (char * pattern)
{
  char searchpattern [1000];
  if ( optionscl->IsItemOn (cIdMatchWholeWord) )
    strn0cpy (searchpattern,
	      form ((char *) rex->MatchWordPattern(), pattern),
	      sizeof (searchpattern));
  else
    strcpy (searchpattern, pattern);
  rex->Reset (searchpattern, optionscl->IsItemOn (cIdIgnoreCase));
  if ( rex->GetExprState() )
    {
      ShowMessage (rex->GetExprState());
      return FALSE;
    }
  return TRUE;
}

bool
FindDialog::Find (char * pattern, bool forward)
{
  if ( tvp == 0 )
    return FALSE;

  if ( pattern == 0 || pattern [0] == 0 )
    {
      ShowMessage ("", TRUE);
      return FALSE;
    }

  int from, to;
  tvp->GetSelection (&from, &to);
  bool found = DoFind (pattern, forward);
  if ( ! found )
    tvp->SetSelection (from, to);
  RevealSelection();
  return found;
}

void
FindDialog::FindAgain()
{
  Find (ei1->GetString(), modecl->GetCurrentItem() == cIdForward);
}

char *
FindDialog::GetSearchPattern()
{
  return ei1->GetString();
}

bool
FindDialog::DoFind (char * pattern, bool forward, bool msg)
{
  if ( ! Compile (pattern) )
    return FALSE;

  bool found = tvp->SelectRegExpr (rex, forward);
  if ( ! found )
    {	// wrap around
      if ( forward )
	tvp->Home();
      else
	tvp->Bottom();
      found = tvp->SelectRegExpr (rex, forward);
    }

  if ( msg )
    ShowMessage (found ? "" : "Not found", !found);

  return found;
}

void
FindDialog::DoChange (char * pattern, char * change)
{
  int orig_from, orig_to;

  if ( tvp == 0 )
    return;

  tvp->GetSelection (&orig_from, &orig_to);
  bool forward = (modecl->GetCurrentItem() == cIdForward);
  if ( forward )
    tvp->SetSelection (orig_from, orig_from);
  else
    tvp->SetSelection (orig_to, orig_to);

  if ( ! DoFind (pattern, forward) )
    tvp->SetSelection (orig_from, orig_to);
  else
    {
      int from, to;
      tvp->GetSelection (&from, &to);
      if ( from == orig_from && to == orig_to )
	{
	  Command * cmd = tvp->InsertString ((byte *) change);
	  cmd->SetName ("Change");
	  tvp->PerformCommand (cmd);
	  if ( ! forward )
	    tvp->SetSelection (from, from);
	  tvp->SetSelection (orig_from, orig_from + strlen (change));
	}
    }
  RevealSelection();
}

void
FindDialog::DoChangeAll (char * pattern, char * change)
{
  if ( tvp == 0 )
    return;

  int from, to;
  int scope_from, scope_to;
  bool change_all;

  tvp->GetSelection (&from, &to);
  scope_from = from;
  scope_to = to;
  change_all = (scopecl->GetCurrentItem() == cIdChangeAll);
  if ( change_all )
    {
      scope_from = 0;
      scope_to = tvp->GetText()->End();
    }

  int nChanges = ChangeAll (scope_from, scope_to, pattern, change, from, to);
  if ( nChanges >= 0 )
    {
      if ( nChanges > 0 )
	{
	  ShowMessage (form ("%d occurrence%s changed",
			     nChanges, (nChanges == 1) ? "" : "s"),
		       FALSE);
	}
      else
	{
	  ShowMessage ("Not found");
	}
      tvp->SetSelection (from, to);
      RevealSelection();
    }
}

int
FindDialog::ChangeAll (int from, int to, char * pattern, char * change,
		       int & orig_from, int & orig_to)
{
  GrShowWaitCursor wc;
  Command * cmd;
  MacroCommand * macro = new GrMacroCommand ("Change All");
  int sfrom, sto, nChanges;
  int next_to, len_incr;
  int changeLen = strlen (change);

  if ( ! Compile (pattern) )
    return -1;
  tvp->SetSelection (from, from, FALSE);
  for ( nChanges = 0; ; nChanges++ )
    {
      if ( ! tvp->SelectRegExpr (rex, TRUE) )
	break;
      tvp->GetSelection (&sfrom, &sto);
      if ( sfrom < from || sfrom > to || sto < from || sto > to )
	break;
      cmd = tvp->InsertString ((byte *) change);
      cmd->SetName ("Change");
      macro->Add (cmd);
      next_to = sfrom + changeLen;
      len_incr = changeLen - (sto - sfrom);
      if ( orig_from > sfrom )
	{
	  if ( orig_from < sto )
	    orig_from = Math::Min (orig_from, next_to);
	  else
	    orig_from += len_incr;
	}
      if ( orig_to > sfrom )
	{
	  if ( orig_to < sto )
	    orig_to = Math::Min (orig_to, next_to);
	  else
	    orig_to += len_incr;
	}
      to += len_incr;
      tvp->SetSelection (next_to, next_to, FALSE);
    }

  if ( macro->Size() == 0 )
    {
      delete macro;
    }
  else
    {
      tvp->PerformCommand (macro);
    }

  return nChanges;
}
