#ifdef __GNUG__
#pragma implementation
#endif

#include "MotifLook.h"

#include "Env.h"
#include "WindowSystem.h"
#include "Expander.h"
#include "Buttons.h"
#include "Layout.h"
#include "ETMath.h"

extern int gDoubleClickTime;
extern void WindowSystem_ReadEnvVars (void);
 
//---- Motif -------------------------------------------------------------------

MotifLook *gMotif;

Look *MakeMotifLook()
{
    return new MotifLook;
}

//---- MotifMenuBorderLayout ---------------------------------------------------

class MotifMenuBorderLayout : public Layout {
public:
    MotifMenuBorderLayout()
	{ }
    void SetOrigin(VObject*, Point);
    void SetExtent(VObject*, Point);
    Metric GetMinSize(VObject *acv);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

Metric MotifMenuBorderLayout::GetMinSize(VObject *acv)
{
    Point e= acv->At(0)->GetMinSize().extent + gPoint4;
    if (acv->Size() > 1) {
	Metric tm= acv->At(1)->GetMinSize();
	e.x= Math::Max(e.x, tm.extent.x);
	e.y+= tm.extent.y + 1;
    }
    return e;
}

void MotifMenuBorderLayout::SetExtent(VObject *acv, Point e)
{
    e-= gPoint4;
    if (acv->Size() > 1) {
	acv->At(1)->CalcExtent();
	e.y-= acv->At(1)->Height() + 1;
    }
    acv->At(0)->SetExtent(e);
}

void MotifMenuBorderLayout::SetOrigin(VObject *acv, Point at)
{
    at+= gPoint2;
    if (acv->Size() > 1) {
	Point a(at);
	a.x+= (acv->At(0)->Width()-acv->At(1)->Width())/2;
	acv->At(1)->SetOrigin(a);
	at.y+= acv->At(1)->Height() + 1;
    }
    acv->At(0)->SetOrigin(at);
}

void MotifMenuBorderLayout::Adorn(VObject *acv, Rectangle &r, int)
{
    GrPaintRect(r, gMotif->backgroundColor);
    if (acv->Size() > 1) {
	int titleht= acv->At(1)->Height();
	BevelLine(Point(2,titleht+1), Point(r.extent.x-2,titleht+1), TRUE,
						gMotif->dark, gMotif->light);
	acv->At(1)->DrawAll(r);
    }
    acv->At(0)->DrawAll(r);
    BevelRect(acv->contentRect, 2, TRUE, gMotif->dark, gMotif->light);
}

//---- MotifMenuLineLayout -----------------------------------------------------

class MotifMenuLineLayout : public Layout {
public:
    MotifMenuLineLayout()
	{ }
    Metric GetMinSize(VObject*);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

inline    Metric MotifMenuLineLayout::GetMinSize(VObject*)
	//{ return gPoint2; }		//ruedi
	{ return Metric(gPoint2); }	//ruedi
	
void MotifMenuLineLayout::Adorn(VObject *acv, Rectangle&, int)
{
    BevelLine(acv->contentRect.NW(), acv->contentRect.NE(), FALSE,
			gMotif->dark, gMotif->light);
}

//---- MotifScrollButtonLayout ------------------------------------------------------

class MotifScrollButtonLayout : public Layout {
public:
    MotifScrollButtonLayout()
	{ }
    Metric GetMinSize(VObject*);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

inline   Metric MotifScrollButtonLayout::GetMinSize(VObject*)
	{ return gMotif->scrollBarThickness; }
	
void MotifScrollButtonLayout::Adorn(VObject*, Rectangle &r, int code)
{
    FillArrow(r.Inset(3), TESTBIT(code, 4) ? gHighlightColor
					   : gMotif->backgroundColor, code & 0x03);
    BevelArrow(r.Inset(3), 2, code & 0x03, !TESTBIT(code, 4), gMotif->dark,
							     gMotif->light);
}

//---- MotifToggleButtonLayout ------------------------------------------------------

class MotifToggleButtonLayout : public Layout {
    bool toggle;
public:
    MotifToggleButtonLayout(bool t)
	{ toggle= t; }
    Metric GetMinSize(VObject *acv);
    void SetOrigin(VObject *acv, Point at);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

Metric MotifToggleButtonLayout::GetMinSize(VObject *acv)
{
    Metric m(17, 17, 13);
    if (acv->Size() > 0)
	m= HCat(m, acv->At(0)->GetMinSize(), 1);
    return m;
}

void MotifToggleButtonLayout::SetOrigin(VObject *acv, Point at)
{
    if (acv->Size() > 0) {
	at.x+= 17+1;
	acv->At(0)->SetOrigin(at);
    }
}

void MotifToggleButtonLayout::Adorn(VObject *acv, Rectangle &r, int code)
{
    Rectangle rr(acv->GetOrigin()+2, Point(17));

    //GrPaintRect(acv->contentRect, gMotif->backgroundColor);
    if (acv->Size() > 0)
	acv->At(0)->DrawAll(r);
    
    if (toggle) {
	if (TESTBIT(code, 3)) 
	    GrPaintRect(rr.Inset(3), gHighlightColor);
	else if (!TESTBIT(code, 1))
	    GrPaintRect(rr.Inset(3), gMotif->backgroundColor);
	else
	    GrPaintRect(rr.Inset(3), gMotif->recessed);
	BevelRect(rr.Inset(1), 2, !(TESTBIT(code, 1) || TESTBIT(code, 3)), gMotif->dark,
								 gMotif->light);
    } else {
	if (TESTBIT(code, 3)) 
	    FillDiamond(rr.Inset(2), gHighlightColor);
	else if (TESTBIT(code, 1))
	    FillDiamond(rr.Inset(2), gMotif->recessed);
	else if (TESTBIT(code, 3))
	    FillDiamond(rr.Inset(2), gHighlightColor);
	BevelDiamond(rr, 2, !(TESTBIT(code, 1) || TESTBIT(code, 3)), gMotif->dark,
								   gMotif->light);
    }
}

//---- MotifMenuItemLayout --------------------------------------------------------

class MotifMenuItemLayout : public Layout {
public:
    MotifMenuItemLayout()
	{ }
    Metric GetMinSize(VObject *acv);
    void SetOrigin(VObject *acv, Point at);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

Metric MotifMenuItemLayout::GetMinSize(VObject *acv)
{
    Metric m;
    if (acv->Size() > 0) {
	m= acv->At(0)->GetMinSize();
	if (acv->Size() > 1)
	    m= HCat(m, acv->At(1)->GetMinSize());
	m.extent.x+= 2+15+0+5+11;
    }
    return m;
}

void MotifMenuItemLayout::SetOrigin(VObject *acv, Point at)
{
    if (acv->Size() > 1)
	acv->At(1)->Align(at, acv->GetExtent(), eVObjHRight);
    at.x+= 2+15+0;
    acv->At(0)->SetOrigin(at);
}

void MotifMenuItemLayout::Adorn(VObject *acv, Rectangle &r, int code)
{
    Rectangle cr(acv->contentRect), rr(cr);
    rr.origin.x+= 2;
    rr.origin.y+= (cr.extent.y-15)/2;
    rr.extent= 15;

    if (TESTBIT(code, 3)) {
	if (GrHasColor())
	    GrPaintRect(cr, gHighlightColor);
    } else
	GrPaintRect(cr, gMotif->backgroundColor);
    
    acv->Draw(r);

    if (TESTBIT(code, 2)) {
	switch (code & 3) {
	case 1:
	    BevelDiamond(rr, 2, TRUE, gMotif->dark, gMotif->light);
	    break;
	case 2:
	    BevelRect(rr.Inset(1), 2, TRUE, gMotif->dark, gMotif->light);
	    break;
	}
    }
    
    if (TESTBIT(code, 4)) {
	Rectangle rr(Point(11));
	rr.origin= cr.NE()-Point(15,-4);
	BevelArrow(rr, 2, 3, TRUE, gMotif->dark, gMotif->light);
    }

    if (TESTBIT(code, 3)) {
	if (GrHasColor())
	    BevelRect(cr, 2, TRUE, gMotif->dark, gMotif->light);
	else
	    GrPaintRect(cr, gHighlightColor);
    }
}

//---- MotifScrollBarLayout ----------------------------------------------------

class MotifScrollBarLayout : public Layout {
public:
    MotifScrollBarLayout()
	{ }
    void Adorn(VObject *acv, Rectangle &r, int code);
};

void MotifScrollBarLayout::Adorn(VObject *acv, Rectangle &r, int)
{
    if (GrHasColor())
	GrPaintRect(r.Inset(2), gMotif->recessed);
    BevelRect(r, 2, FALSE, gMotif->dark, gMotif->light);
    ((Expander*)acv)->Expander::Draw(r);
}

//---- MotifActionButtonLayout -------------------------------------------------

class MotifActionButtonLayout : public Layout {
    bool dflt;
public:
    MotifActionButtonLayout(bool d)
	{ dflt= d; }
    void Adorn(VObject *acv, Rectangle &r, int code);
    Metric GetMinSize(VObject *acv);
};

Metric MotifActionButtonLayout::GetMinSize(VObject *acv)
{
    if (dflt)
	return acv->At(0)->GetMinSize().Expand(gPoint6);
    return acv->At(0)->GetMinSize().Expand(gPoint2);
}

void MotifActionButtonLayout::Adorn(VObject *acv, Rectangle &r, int code)
{
    Rectangle rr(r);
    VObject *vop= acv->At(0);
    
    if (dflt)
	rr= rr.Inset(4);
    
    GrPaintRect(acv->contentRect, gMotif->backgroundColor);

    if (TESTBIT(code, 3)) {
	if (GrHasColor())
	    GrPaintRect(rr, gHighlightColor);
    } else if (TESTBIT(code, 1))
	GrPaintRect(rr, gMotif->recessed);

    vop->DrawAll(r);

    if (TESTBIT(code, 3) && !GrHasColor())
	GrPaintRect(rr, gHighlightColor);

    BevelRect(rr, 2, !(TESTBIT(code, 1)||TESTBIT(code, 3)), gMotif->dark, gMotif->light);
    if (dflt)
	BevelRect(r, 1, FALSE, gMotif->dark, gMotif->light);
}

//---- MotifPopupButtonLayout --------------------------------------------------

class MotifPopupButtonLayout : public Layout {
public:
    MotifPopupButtonLayout()
	{ }
    void Adorn(VObject *acv, Rectangle &r, int code);
};

void MotifPopupButtonLayout::Adorn(VObject *acv, Rectangle&, int)
{
    BevelRect(acv->contentRect, 2, TRUE, gMotif->dark, gMotif->light);
    Rectangle r(acv->Width()-18, (acv->Height()-6)/2, 12, 6);
    r.origin+= acv->GetOrigin();
    BevelRect(r, 2, TRUE, gMotif->dark, gMotif->light);
}

//---- MotifThumbLayout --------------------------------------------------------

class MotifThumbLayout : public Layout {
public:
    MotifThumbLayout()
	{ }
    Metric GetMinSize(VObject*);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

inline    Metric MotifThumbLayout::GetMinSize(VObject*)
	{ return gMotif->scrollBarThickness; }
	
void MotifThumbLayout::Adorn(VObject*, Rectangle &r, int)
{
    GrPaintRect(r.Inset(2), gMotif->backgroundColor);
    BevelRect(r.Inset(2), 2, TRUE, gMotif->dark, gMotif->light);
}

//---- MotifFieldBorderLayout --------------------------------------------------

class MotifFieldBorderLayout : public Layout {
    bool noborder;
public:
    MotifFieldBorderLayout(bool nb)
	{ noborder= nb; }
    void Adorn(VObject *acv, Rectangle &r, int code);
    long GetValue(VObject*, int);
};
 
inline    long MotifFieldBorderLayout::GetValue(VObject*, int)
	{ return noborder ? 2 : 4; }
	
void MotifFieldBorderLayout::Adorn(VObject *acv, Rectangle&, int code)
{
    if (!noborder)
	BevelRect(acv->contentRect.Inset(gPoint2), 2, FALSE, gMotif->dark, gMotif->light);
    if (code)
	GrGetPort()->StrokeRect(gHighlightColor, 2, acv->contentRect);
}

//---- MotifMenuBarLayout ------------------------------------------------------

class MotifMenuBarLayout : public Layout {
public:
    MotifMenuBarLayout()
	{ }
    Metric GetMinSize(VObject*);
    void SetOrigin(VObject *acv, Point at);
    void SetExtent(VObject *acv, Point e);
    void Adorn(VObject *acv, Rectangle&, int);
};
 
inline    void MotifMenuBarLayout::SetOrigin(VObject *acv, Point at)
	{ acv->At(0)->SetOrigin(at+gPoint3); }
inline    void MotifMenuBarLayout::SetExtent(VObject *acv, Point e)
	{ acv->At(0)->SetExtent(e-6); }
inline    void MotifMenuBarLayout::Adorn(VObject *acv, Rectangle&, int)
	{ BevelRect(acv->contentRect, 2, TRUE, gMotif->dark, gMotif->light); }

Metric MotifMenuBarLayout::GetMinSize(VObject *acv)
{
    Metric m(acv->At(0)->GetMinSize());
    m.extent+= 2*gPoint3;
    m.base+= 3;
    return m;
}

//---- MotifSashLayout ---------------------------------------------------------

class MotifSashLayout : public Layout {
public:
    MotifSashLayout()
	{ }
    Metric GetMinSize(VObject*);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

inline    Metric MotifSashLayout::GetMinSize(VObject*)
	//{ return gPoint6; }		//ruedi
	{ return Metric(gPoint6); }	//ruedi
	
void MotifSashLayout::Adorn(VObject *acv, Rectangle&, int)
{
    Point p1, p2;
    Rectangle rr(acv->contentRect);
    if (rr.extent.x > rr.extent.y) { // horizontal
	p1= rr.W();
	p2= rr.E();
	p1.y--;
	p2.y--;
	rr.origin.x+= rr.extent.x - 19 - 1;
	rr.extent.x= 19;
    } else {    // vertical
	p1= rr.N();
	p2= rr.S();
	p1.x--;
	p2.x--;
	rr.origin.y+= rr.extent.y - 19 - 1;
	rr.extent.y= 19;
    }
    BevelLine(p1, p2, FALSE, gMotif->dark, gMotif->light);
    BevelRect(rr, 2, TRUE, gMotif->dark, gMotif->light);
    GrPaintRect(rr.Inset(2), gMotif->backgroundColor);
}

//---- MotifProgressBarLayout --------------------------------------------------

class MotifProgressBarLayout : public Layout {
public:
    MotifProgressBarLayout()
	{ }
    Metric GetMinSize(VObject*);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

inline    Metric MotifProgressBarLayout::GetMinSize(VObject*)
	{ return Metric(200, 20); }
	
void MotifProgressBarLayout::Adorn(VObject *acv, Rectangle&, int val)
{
    Rectangle rr(acv->contentRect.Inset(2));
    val= Math::Min(val, rr.extent.x);
    GrPaintRect(rr, gMotif->recessed);
    BevelRect(acv->contentRect, 2, FALSE, gMotif->dark, gMotif->light);
    rr.extent.x= val;
    GrPaintRect(rr, gMotif->backgroundColor);
    BevelRect(rr, 2, TRUE, gMotif->dark, gMotif->light);
}

//---- MotifScaleLayout --------------------------------------------------------

class MotifScaleLayout : public Layout {
public:
    MotifScaleLayout()
	{ }
    Metric GetMinSize(VObject *acv);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

inline    Metric MotifScaleLayout::GetMinSize(VObject *acv)
	{ return acv->At(0)->GetMinSize(); }
	
void MotifScaleLayout::Adorn(VObject *acv, Rectangle&, int)
{
    Rectangle cr= acv->contentRect;
    GrPaintRect(cr.Inset(2), gMotif->recessed);
    BevelRect(cr, 2, FALSE, gMotif->dark, gMotif->light);
}

//---- MotifScaleThumbLayout ---------------------------------------------------

class MotifScaleThumbLayout : public Layout {
    Direction dir;
public:
    MotifScaleThumbLayout(Direction d)
	{ dir= d; }
    Metric GetMinSize(VObject *acv);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

Metric MotifScaleThumbLayout::GetMinSize(VObject*)
{
    if (dir == eHor)
	return Metric(30, 20);
    return Metric(20, 30);
}

void MotifScaleThumbLayout::Adorn(VObject *acv, Rectangle&, int)
{
    Rectangle cr= acv->contentRect.Inset(2);
    GrPaintRect(cr, gMotif->backgroundColor);
    cr.extent[bool(dir == eVert)] /= 2;
    BevelRect(cr, 2, TRUE, gMotif->dark, gMotif->light);
    cr.origin[bool(dir == eVert)] += cr.extent[bool(dir == eVert)];
    BevelRect(cr, 2, TRUE, gMotif->dark, gMotif->light);
}

//---- MotifLayout --------------------------------------------------------

class MotifGroupLayout : public Layout {
    Point border;
    int cBorderTitleGap;
public:
    MotifGroupLayout()
	{ border= gPoint6; cBorderTitleGap= 4; }
    void SetOrigin(VObject*, Point);
    void SetExtent(VObject*, Point);
    Metric GetMinSize(VObject *acv);
    void Adorn(VObject *acv, Rectangle &r, int code);
};

Metric MotifGroupLayout::GetMinSize(VObject *acv)
{
    VObject *interior= acv->At(0);
    Metric m= interior->GetMinSize();
    
    if (acv->Size() > 1) {
	VObject *titleBar= acv->At(1);
	Metric tm= titleBar->GetMinSize();
	m.extent.x= Math::Max(m.extent.x, tm.extent.x+2*cBorderTitleGap);
	int d= tm.extent.y - border.y;
	if (d > 0) {
	    m.extent.y+= d;
	    m.base+= d;
	}
    }
    return m.Expand(border);
}

void MotifGroupLayout::SetExtent(VObject *acv, Point e)
{
    if (acv->Size() > 1) {
	VObject *titleBar= acv->At(1);
	titleBar->CalcExtent();
	int d= titleBar->Height() - border.y;
	if (d > 0)
	    e.y-= d;
    }
    acv->At(0)->SetExtent(e-2*border);
}

void MotifGroupLayout::SetOrigin(VObject *acv, Point at)
{
    if (acv->Size() > 1) {
	VObject *titleBar= acv->At(1);
	int th= titleBar->Height();
	int tgap= border.x+cBorderTitleGap;
	titleBar->Align(Point(at.x+tgap, at.y), Metric(acv->Width()-2*tgap, th),
				(VObjAlign)((eVObjHLeft & eVObjH) | eVObjVTop));
	int d= th - border.y;
	if (d > 0)
	    at.y+= d;
    }
    acv->At(0)->SetOrigin(at+border);
}

void MotifGroupLayout::Adorn(VObject *acv, Rectangle &r, int)
{
    if (! acv->At(0)->contentRect.ContainsRect(r)) {
	if (acv->Size() > 1) {
	    VObject *titleBar= acv->At(1);
	    int h= titleBar->Height()/2-1;
	    Rectangle rr(acv->contentRect);
	    rr.origin.y+= h;
	    rr.extent.y-= h;

	    int lpos= titleBar->GetOrigin().x;
	    int rpos= titleBar->contentRect.NE().x;
	    GrPaintLine(gMotif->dark, 1, eDefaultCap, Point(lpos, rr.origin.y), rr.NW());
	    GrPaintLine(gMotif->dark, 1, eDefaultCap, rr.NW(), rr.SW());
	    GrPaintLine(gMotif->light, 1, eDefaultCap, rr.SW(), rr.SE());
	    GrPaintLine(gMotif->light, 1, eDefaultCap, rr.SE(), rr.NE());
	    GrPaintLine(gMotif->dark, 1, eDefaultCap, rr.NE(), Point(rpos, rr.origin.y));
	    rr= rr.Inset(1);
	    GrPaintLine(gMotif->light, 1, eDefaultCap, Point(lpos, rr.origin.y), rr.NW());
	    GrPaintLine(gMotif->light, 1, eDefaultCap, rr.NW(), rr.SW());
	    GrPaintLine(gMotif->dark, 1, eDefaultCap, rr.SW(), rr.SE());
	    GrPaintLine(gMotif->dark, 1, eDefaultCap, rr.SE(), rr.NE());
	    GrPaintLine(gMotif->light, 1, eDefaultCap, rr.NE(), Point(rpos, rr.origin.y));
	} else {
	    BevelRect(acv->contentRect, 1, FALSE, gMotif->dark, gMotif->light);
	    BevelRect(acv->contentRect.Inset(1), 1, TRUE, gMotif->dark, gMotif->light);
	}
    }
}

//---- MotifFrameLayout --------------------------------------------------------

class MotifFrameLayout : public Layout {
public:
    MotifFrameLayout()
	{ }
    void Adorn(VObject *acv, Rectangle &r, int code);
    long GetValue(VObject*, int code);
};
 
inline    long MotifFrameLayout::GetValue(VObject*, int code)
	{ return code ? 6 : 2; }
	
void MotifFrameLayout::Adorn(VObject*, Rectangle &r, int)
{
    BevelRect(r.Expand(gPoint2), 2, FALSE, gMotif->dark, gMotif->light);
}

//---- Motif -------------------------------------------------------------------

MotifLook::MotifLook()
{
    gMotif= this;

    scrollBarThickness = Env::GetValue("ScrollBar.Thickness", 20);
    
    gCaretColor= new_RGBColor(255, 0, 0); // default is red
    Env::Bind(gCaretColor, "TextView.CaretColor");    
    
    gClipperBgColor = new_RGBColor(255,255,255,0); // default is white
    Env::Bind(gClipperBgColor, "WindowSystem.ViewBackgroundColor");

    if (gDepth == 1) {
	light= ePatGrey50;
	dark= gInkBlack;
	recessed= ePatGrey50;
    } else if (gDepth == 2) {
	backgroundColor= new_Grey(0.6);
	light= gInkWhite;
	dark= gInkBlack;
	recessed= new_Grey(0.6);
	disableInk= new_Grey(0.3);
    } else {
	backgroundColor= new_Grey(0.85);
	Env::Bind(backgroundColor, "WindowSystem.WindowBackgroundColor");
	light= new_Grey(0.95);
	dark= new_Grey(0.15);
	recessed= new_Grey(0.65);
	disableInk= new_Grey(0.7);
	Env::Bind(disableInk, "WindowSystem.DisableColor");
    }

    popUpMenuLayout= new MotifMenuBorderLayout;
    pullDownMenuLayout= new MotifMenuBorderLayout;
    menuLineLayout= new MotifMenuLineLayout;
    menuItemLayout= new MotifMenuItemLayout;
    scrollButtonLayout= new MotifScrollButtonLayout;
    radioButtonLayout= new MotifToggleButtonLayout(FALSE);
    toggleButtonLayout= new MotifToggleButtonLayout(TRUE);
    scrollBarLayout= new MotifScrollBarLayout;
    actionButtonLayout= new MotifActionButtonLayout(FALSE);
    defaultButtonLayout= new MotifActionButtonLayout(TRUE);
    popupButtonLayout= new MotifPopupButtonLayout;
    sliderLayout= new Layout;
    thumbLayout= new MotifThumbLayout;
    fieldBorderLayout= new MotifFieldBorderLayout(FALSE);
    fieldLayout= new MotifFieldBorderLayout(TRUE);
    menuBarLayout= new MotifMenuBarLayout;
    sashLayout= new MotifSashLayout;
    progressBarLayout= new MotifProgressBarLayout;
    scaleLayout= new MotifScaleLayout;
    scaleThumbLayout[eHor]= new MotifScaleThumbLayout(eHor);
    scaleThumbLayout[eVert]= new MotifScaleThumbLayout(eVert);
    groupLayout= new MotifGroupLayout;
    frameLayout= new MotifFrameLayout;
}

MotifLook::~MotifLook()
{
}

void MotifLook::ReReadEnvValues()
{
    scrollBarThickness = Env::GetValue("ScrollBar.Thickness", 20);
    Env::Bind(backgroundColor, "WindowSystem.WindowBackgroundColor");
    Env::Bind(disableInk, "WindowSystem.DisableColor");
    Env::Bind(gHighlightColor, "WindowSystem.HighlightColor");
    Env::Bind(gClipperBgColor, "WindowSystem.ViewBackgroundColor");
    Env::Bind(gCaretColor, "TextView.CaretColor");    
    gDoubleClickTime = Env::GetValue ("WindowSystem.DoubleClickTime", 500);
    WindowSystem_ReadEnvVars();
}


void MotifLook::DrawHighlight(Rectangle &r)
{
    GrPaintRect(r.Inset(2), gHighlightColor);
    BevelRect(r, 2, TRUE, dark, light);
}

//---- Utilities ---------------------------------------------------------------

void
BevelArrow (const Rectangle &r, int w, int dir, bool raised, Ink *C1, Ink *C2)
{
    Ink *c1, *c2;
    if (!raised) {
	c1= C1;
	c2= C2;
    } else {
	c1= C2;
	c2= C1;
    }
    switch (dir) {
    case 0: // up
	GrPaintLine(c2, 1, eDefaultCap, r.N(), r.SE());
	GrPaintLine(c2, 1, eDefaultCap, r.SE(), r.SW());
	GrPaintLine(c1, 1, eDefaultCap, r.SW(), r.N());
	break;
    case 1: // down
	GrPaintLine(c1, 1, eDefaultCap, r.NW(), r.NE());
	GrPaintLine(c2, 1, eDefaultCap, r.NE(), r.S());
	GrPaintLine(c1, 1, eDefaultCap, r.S(), r.NW());
	break;
    case 2: // left
	GrPaintLine(c1, 1, eDefaultCap, r.W(), r.NE());
	GrPaintLine(c2, 1, eDefaultCap, r.NE(), r.SE());
	GrPaintLine(c2, 1, eDefaultCap, r.SE(), r.W());
	break;
    case 3: // right
	GrPaintLine(c1, 1, eDefaultCap, r.NW(), r.E());
	GrPaintLine(c2, 1, eDefaultCap, r.E(), r.SW());
	GrPaintLine(c1, 1, eDefaultCap, r.SW(), r.NW());
	break;
    }
    
    if (w >= 2)
	BevelArrow(r.Inset(1), w-1, dir, raised, C1, C2);
}

void
BevelRect (const Rectangle &r, int w, bool raised, Ink *C1, Ink *C2)
{
    Ink *c1, *c2;
    if (!raised) {
	c1= C1;
	c2= C2;
    } else {
	c1= C2;
	c2= C1;
    }
    GrPaintLine(c1, 1, eDefaultCap, r.SW(), r.NW());
    GrPaintLine(c1, 1, eDefaultCap, r.NW(), r.NE());
    GrPaintLine(c2, 1, eDefaultCap, r.SW(), r.SE());
    GrPaintLine(c2, 1, eDefaultCap, r.SE(), r.NE());
    if (w >= 2)
	BevelRect(r.Inset(1), w-1, raised, C1, C2);
}

void
BevelDiamond (const Rectangle &rr, int w, bool raised, Ink *C1, Ink *C2)
{
    Rectangle r(rr);
    if ((r.extent.x & 1) == 0)
	r.extent.x--;
    if ((r.extent.y & 1) == 0)
	r.extent.y--;
    Ink *c1, *c2;
    if (!raised) {
	c1= C1;
	c2= C2;
    } else {
	c1= C2;
	c2= C1;
    }
    GrPaintLine(c2, 1, eDefaultCap, r.W(), r.S());
    GrPaintLine(c2, 1, eDefaultCap, r.S(), r.E());
    GrPaintLine(c1, 1, eDefaultCap, r.W(), r.N());
    GrPaintLine(c1, 1, eDefaultCap, r.N(), r.E());
    if (w >= 2)
	BevelDiamond(rr.Inset(1), w-1, raised, C1, C2);
}

void
BevelLine (Point p1, Point p2, bool raised, Ink *C1, Ink *C2)
{
    Ink *c1, *c2;
    if (!raised) {
	c1= C1;
	c2= C2;
    } else {
	c1= C2;
	c2= C1;
    }
    GrPaintLine(c1, 1, eDefaultCap, p1, p2);
    GrPaintLine(c2, 1, eDefaultCap, p1+gPoint1, p2+gPoint1);
}

void
FillDiamond (const Rectangle &rr, Ink *fill)
{
    Point p[4];
    Rectangle r(rr);
    if ((r.extent.x & 1) == 0)
	r.extent.x--;
    if ((r.extent.y & 1) == 0)
	r.extent.y--;
    p[0]= r.N();
    p[1]= r.E();
    p[2]= r.S();
    p[3]= r.W();
    GrPaintPolygon(gPoint0, fill, p, 4, ePolyDefault);
    GrPaintLine(fill, 1, eDefaultCap, r.N(), r.E());
    GrPaintLine(fill, 1, eDefaultCap, r.E(), r.S());
}

void
FillArrow (const Rectangle &rr, Ink *fill, int dir)
{
    Point p[3];
    Rectangle r(rr);

    switch (dir) {
    case 0: // up
	p[0]= r.N();
	p[1]= r.SE();
	p[2]= r.SW();
	break;
    case 1: // down
	p[0]= r.NW();
	p[1]= r.NE();
	p[2]= r.S();
	break;
    case 2: // left
	p[0]= r.W();
	p[1]= r.NE();
	p[2]= r.SE();
	break;
    case 3: // right
	p[0]= r.NW();
	p[1]= r.E();
	p[2]= r.SW();
	break;
    }

    GrPaintPolygon(gPoint0, fill, p, 3, ePolyDefault);
    GrPaintLine(fill, 1, eDefaultCap, p[0], p[1]);
    GrPaintLine(fill, 1, eDefaultCap, p[1], p[2]);
    GrPaintLine(fill, 1, eDefaultCap, p[2], p[0]);
}
