#ifdef __GNUG__
#pragma implementation
#endif

#include "Class.h"
#include "MacroCommand.h"
#include "OrdColl.h"

//---- MacroCommand -----------------------------------------------

NewMetaImpl(MacroCommand,Command,(TP(commands),TP(lastCmd)));


MacroCommand::MacroCommand(char*name):Command(name,eCmdTypeNoChange)
{
    commands=0;
    lastCmd=0;
}

MacroCommand::~MacroCommand()
{
    // destructor is called in Command::Finish():
    // all our commands will commit suicide now ;-)
    if(commands) {
	{commands->ForEach(Command,Finish)();}
	SafeDelete(commands);
    } else if(lastCmd) {
	lastCmd->Finish();
    }
}
void
MacroCommand::Add(Command*cmd)
{
    if(cmd==0 || cmd==gNoChanges) // that's safe :-)
	return;
    switch (cmd->GetType()) {
    case eCmdTypeCantUndo:
	// if one can't undo -> the whole macro can't undo!
	SetType(eCmdTypeCantUndo);
	break;
    case eCmdTypeNoChange:
	// perform it now and finish...
	cmd->Perform();
	cmd->Done(0);
	cmd->Finish();
	// don't store it -> return!
	return;
    case eCmdTypeNormal:
	// let's become a eCmdTypeNormal except if we are a eCmdTypeCantUndo
	if(GetType()==eCmdTypeNoChange)
	    SetType(eCmdTypeNormal);
	break;
    }
    // be economic: only allocate commands if more then 1 command...
    if (lastCmd) {
	if(commands==0) {
	    commands=new OrdCollection;
	    commands->Add(lastCmd);
	}
	lastCmd->Done(cmd);
    }
    lastCmd= cmd;
    // let's perform it now (let's be a little bit CommandProcessor)
    cmd->Perform();
    if(commands)
	commands->Add(cmd);
    
}
int
MacroCommand::Size()
{
    if(commands)
    	return commands->Size();
    if(lastCmd)
    	return 1;
    return 0;
}

void
MacroCommand::Cancel()
{
    if(commands) {
	// undo in reverse order!
	Iter next(commands,cIterBackward);
	Command *cmd;
	while(cmd=(Command*)next())
	    cmd->Perform();
    	{commands->ForEach(Command,Finish)();} // this kills the commands!
    	SafeDelete(commands);
    } else if(lastCmd) {
    	lastCmd->Finish();
    }
    commands=0;
    lastCmd=0;
}

void
MacroCommand::Done(Command *newCommand)
{
    // forward newCommand to the last in this macro...
    if(lastCmd)
	lastCmd->Done(newCommand);
}


void
MacroCommand::Commit()
{
    // done in the destructor via Finish...
}

void
MacroCommand::UndoIt()
{
    if(commands) {
	// undo in reverse order!
	Iter next(commands,cIterBackward);
	Command *cmd;
	while(cmd=(Command*)next())
	    cmd->UndoIt();
    } else if(lastCmd) {
	lastCmd->UndoIt();
    }
}

void
MacroCommand::DoIt()
{
    // is done when added...
    if(!commands && lastCmd) {
    	// if there is only one Command, let's get its name...
    	SetName(lastCmd->GetName());
    }
}

void
MacroCommand::RedoIt()
{
    if(commands) {
	commands->ForEach(Command,RedoIt)();
    }else if (lastCmd) {
	lastCmd->RedoIt();
    }
}

//---- GrMacroCommand -----------------------------------------------
#include "WindowSystem.h"

NewMetaImpl(GrMacroCommand,MacroCommand,(TB(grUpdate)));


GrMacroCommand::GrMacroCommand(char*name):MacroCommand(name)
{
    grUpdate=FALSE;
}

GrMacroCommand::~GrMacroCommand()
{
    EndDelayUpdate();
}
void
GrMacroCommand::Add(Command*cmd)
{
    BeginDelayUpdate();
    MacroCommand::Add(cmd);
}

void
GrMacroCommand::UndoIt()
{
    gWindowSystem->BeginDelayUpdate();
    MacroCommand::UndoIt();
    gWindowSystem->EndDelayUpdate();
}

void
GrMacroCommand::DoIt()
{
    MacroCommand::DoIt();
    EndDelayUpdate();
}

void
GrMacroCommand::RedoIt()
{
    gWindowSystem->BeginDelayUpdate();
    MacroCommand::RedoIt();
    gWindowSystem->EndDelayUpdate();
}


void GrMacroCommand::BeginDelayUpdate()
{
    if(grUpdate)
    	return;
    gWindowSystem->BeginDelayUpdate();
    grUpdate=TRUE;
}

void GrMacroCommand::EndDelayUpdate()
{
    if(!grUpdate)
    	return;
    gWindowSystem->EndDelayUpdate();
    grUpdate=FALSE;
}
