//-*-Mode: C++;-*-
#ifndef Manager_First
#define Manager_First

#ifdef __GNUG__
#  pragma interface
#endif

#include "Window.h"

class Menu;
class CommandProcessor;
class Command;
class MenuBar;
class Button;
class OrdCollection;
class TextViewOverlay;
class Bitmap;

enum MgrDismiss {
    eMgrClose,
    eMgrHide,
    eMgrIconize
};

//---- Manager -----------------------------------------------------------------

class Manager : public EvtHandler {
public:
    MetaDef(Manager);

    Manager(char *title= 0, MgrDismiss md= eMgrHide, WindowType wt= eWinStandard);
    Manager(Manager*, char *title= 0, MgrDismiss md= eMgrHide, WindowType wt= eWinStandard);
    ~Manager();

    //---- misc
    void Init(Manager*, char *name, MgrDismiss oc, WindowType wt);
    virtual void SetName(const char *name);
    char *GetName();
    bool IsEqual(Object*);
    bool IsOpen();
    void CollectParts(Collection*);
    void InspectorId(char *buf, int bufSize);

    //---- change management
    CommandProcessor *GetCmdP();
    virtual CommandProcessor *MakeCmdProcessor();
    void ClearCmdProcessor();
    void SetCmdProcessor(CommandProcessor *c);

    //---- managers
    virtual void Open();
    void OpenAt(VObject *fp, Point p, bool resize= FALSE, bool block= FALSE);
    virtual bool Close();       // return TRUE if OK
    virtual void Show();
    virtual int Run();
    void Hide();
    void HideAndRemember();
    void Unhide();
    virtual bool AboutToDismiss();
    bool Dismiss();
    virtual void Iconize();

    Iterator *MakeIterator();
    int Size();
    virtual void AddManager(Manager*);
    virtual void RemoveManager(Manager*);
    virtual Manager *FindManager(int id);
    void SetNextManager(Manager *mp);
    EvtHandler *GetNextHandler();
    virtual int UniqueId();

    //---- events
    void InputKbd(Token &t);
    bool SetFirstHandler(EvtHandler *eh);
    EvtHandler *GetFirstHandler();
    Button *GetDefaultButton();
    void TabFields(Token &t);
    void ExtCommand(int from, char *req, char *args, int len, char *&ret, int &retlen);
    void Control(int id, int part, void *vp);
    void SendDown(int id, int part, void *val);
    TextViewOverlay *GetTextViewOverlay();

    //---- windows
    virtual Window *DoMakeWindows();
    virtual VObject *DoMakeContent();
    virtual Bitmap *DoMakeIcon();
    virtual void MakeWindows();
    virtual Point GetInitialPos(Point e);
    virtual Point GetInitialWindowPos();
    virtual Point GetInitialWindowSize();
    virtual void SetWindowTitle(char *title);
    Window *GetWindow();
    Bitmap *GetIcon();
    void SetOnDismiss(MgrDismiss md);

    //---- menues
    virtual MenuBar *DoMakeMenuBar();
    virtual Menu *MakeMenu(int menuId);
    void DoSetupMenu(Menu*);
    Command *DoMenuCommand(int);
    void PerformCommand(Command*);
    Menu *GetMenu();
    MenuBar *GetMenuBar();

protected:
    WindowType wtype;
    WindowFlags wflags;
    Manager *nextManager;
    OrdCollection *managers;
    Menu *menu;
    MenuBar *menubar;
    Window *window;
    Bitmap *icon;
    EvtHandler *firstHandler;
    CommandProcessor *cmdp;
    MgrDismiss dismiss;
    char *name;
    TextViewOverlay *tvo;
    bool isopen, wasopen;
};

inline char * Manager::GetName()
  { return name; }

inline bool Manager::IsOpen()
  { return isopen; }

inline void Manager::SetNextManager (Manager * mgr)
  { nextManager = mgr; }

inline EvtHandler * Manager::GetFirstHandler()
  { return firstHandler; }

inline Window * Manager::GetWindow()
  { return window; }

inline void Manager::SetOnDismiss (MgrDismiss d)
  { dismiss = d; }

inline MenuBar * Manager::GetMenuBar()
  { return menubar; }

#endif
