#ifdef __GNUG__
#pragma implementation
#endif
 
#include "EtPeManager.h" 

#include "ET++.h" 

#include "ProgEnv.h"
#include "ObjectTable.h"
#include "PathLookup.h"
#include "ObjList.h"
#include "OrdColl.h"
#include "System.h"
#include "ImageItem.h"
#include "Window.h"
#include "Buttons.h"
#include "BorderItems.h"
#include "Env.h"

#include "Inspector.h"
#include "ClassTree.h"
#include "ClassManager.h"
#include "OStruBrowser.h"
#include "SrcBrowser.h"
#include "EtPeCmdNo.h"
#include "MethodAccessor.h"
#include "Reference.h"

static u_short IconBits[]= {
#   include "images/et++pe_icon.image"
};
static u_short SrcBits[]= {
#   include "images/et++pe_src.image"
};
static u_short InspBits[]= {
#   include "images/et++pe_insp.image"
};
static u_short OStruBits[]= {
#   include "images/et++pe_ostru.image"
};
static u_short HierarchyBits[]= {
#   include "images/et++pe_hier.image"
};

static SmartBitmap IconBm(Point(59, 52), IconBits),
		   SrcBm(64, SrcBits),
		   InspBm(64, InspBits),
		   OStruBm(64, OStruBits),
		   HierarchyBm(64, HierarchyBits);

EtPeManager *gEtPeManager;

//---- EtPeManager -----------------------------------------------------------------

NewMetaImpl0(EtPeManager, Dialog);

PathLookup *EtPeManager::universe;

EtPeManager::EtPeManager() : Dialog("ET++PE", eWinFixed)
{
    wtype= eWinStandard;
    SetOnDismiss(eMgrIconize);
    if (Env::GetValue("ProgEnv.UseMapFiles", TRUE))
	methodAccessor= new PeMapMethodAccessor();
    else
	methodAccessor= new PeScanMethodAccessor();
}

EtPeManager::~EtPeManager()
{
    SafeDelete(methodAccessor);
}
 
VObject *EtPeManager::DoMakeContent()
{
    return
	new HExpander(gPoint4,
	    new ActionButton(cIdNewInsp,     new ImageItem(InspBm)),
	    new ActionButton(cIdNewOBrowser, new ImageItem(OStruBm)),
	    new ActionButton(cIdNewBrowser,  new ImageItem(SrcBm)),
	    new ActionButton(cIdNewHier,     new ImageItem(HierarchyBm)),
	    new ActionButton(12345, "Hide", TRUE),
	    0
	);
}
   
void EtPeManager::Control(int id, int what, void *val)
{
    switch (id) {
    case cIdNewInsp:
	PeInspector::ShowObject(this, gApplication, FALSE);
	break;
    case cIdNewHier: 
	PeHierarchyBrowser::ShowClass(this, gClassManager->Find("Object"));
	break;
    case cIdNewOBrowser: 
	PeObjectBrowser::ShowObject(this, gApplication);
	break;
    case cIdNewBrowser:
	PeSourceBrowser::ShowClass(this, gApplication->IsA(), TRUE);
	break;
    case 12345:
	Hide();
	return;
    }
    
    switch (what) {
    
    case cPeNewInsp:
	PeInspector::ShowObject(this, gApplication, FALSE);
	break;
    case cPeShowInspector:
	PeInspector::ShowObject(this, (Object*)val, FALSE);
	break;
    case cPeSpawnInsp:
	PeInspector::Spawn(this, (Ref*)val, FALSE);
	break;

    case cPeNewHier: 
	PeHierarchyBrowser::ShowClass(this, gClassManager->Find("Object"));
	break;
    case cPeSpawnHier:
	PeHierarchyBrowser::Spawn(this, (Class*)val);
	break;
    case cPeShowHierarchy:
	PeHierarchyBrowser::ShowClass(this, (Class*)val);
	break;
	
    case cPeNewOBrowser: 
	PeObjectBrowser::ShowObject(this, gApplication);
	break;
    case cPeSpawnOBrowser:
	PeObjectBrowser::Spawn(this, (Object*)val);
	break;
    case cPeShowOBrowser:
	PeObjectBrowser::ShowObject(this, (Object*)val);
	break;
	
    case cPeSpawnBrowser:
	PeSourceBrowser::Spawn(this, (Class*)val, TRUE);
	break;
    case cPeShowBrowser:
	PeSourceBrowser::ShowClass(this, (Class*)val, TRUE);
	break;
    }   
    if (id < cIdFirstUser) 
	Dialog::Control(id, what, val);
}

Bitmap *EtPeManager::DoMakeIcon()
{
    return IconBm;
}

PathLookup *EtPeManager::Universe()
{
    if (universe == 0) {
	char *srcPath= Env::GetValue("ProgEnv.SrcPath", ".");
	universe= new PathLookup(srcPath);
	universe->Add(gSystem->ConcatPathNames(gEtDir, "src", 0));
	universe->Add(gSystem->ConcatPathNames(gEtDir, "src", "PROGENV", 0));
	universe->Add(gSystem->ConcatPathNames(gEtDir, "src", "PRINTERS", 0));
	universe->Add(gSystem->ConcatPathNames(gEtDir, "src", "SUNWINDOW", 0));
	universe->Add(gSystem->ConcatPathNames(gEtDir, "src", "XSERVER", 0));
	universe->Add(gSystem->ConcatPathNames(gEtDir, "src", "UNIX", 0));
	universe->Add(gSystem->ConcatPathNames(gEtDir, "src", "STREAMFILTERS", 0));
	universe->Add(gSystem->ConcatPathNames(gEtDir, "src", "CONTAINER", 0));
	universe->Add(gSystem->ConcatPathNames(gEtDir, "src", "CONVERTERS", 0));
	universe->Add(gSystem->ConcatPathNames(gEtDir, "src", "LOOKS", 0));
    }  
    return universe;  
}

bool EtPeManager::FileOfClass(Class *cl, char *fname, bool decl)
{
    const char *path, *base, *bp;
    fname[0] = '\0';
    if (decl)
	base= cl->GetDeclFileName();
    else
	base= cl->GetImplFileName();
    path= "" /* cl->GetCompDir() */;
    if (!gSystem->IsAbsolutePath((char*)base))
	strcpy(fname,gSystem->ConcatPathNames(path, " ", 0));
    if (base[0] == '.' && gSystem->IsAbsolutePath((char*)&base[1])) {
	bp= &base[2];
	strcat (fname,bp);
    }
    else {
	bp= base;
	strcat (fname,base);
    }
    if(gSystem->AccessPathName(fname, 4) != 0) {
	PathLookup *pl= gEtPeManager->Universe();
	if (! pl->Lookup(bp, fname))
	    return FALSE;
    }
    return TRUE;
}


void EtPeManager::CollectParts(Collection*)
{
    // empty: suppress display of ETPE objects
}

Collection *EtPeManager::LoadMethodsOfClass(Class *cl)
{
    Collection *methods= cl->GetMethods();
    if (methods == 0 && methodAccessor) {
	if (universe == 0)
	    universe= Universe();
	methods= methodAccessor->FindMethods(cl, universe);
	if (methods && methods->IsKindOf(OrdCollection))
	    Guard(methods, OrdCollection)->Sort();
	cl->SetMethods(methods);
    }
    return methods;
}

bool EtPeManager::SameMethods(char *m1, char *m2)
{
    return (strcmp(m1, m2) == 0); 
}

bool EtPeManager::IsDestructor(char *m1)
{
    return m1[0] == '~';
}

Command *EtPeManager::DoMenuCommand(int cmd)
{
    if (cmd < cUSERCMD)
	return Dialog::DoMenuCommand(cmd);
    return gNoChanges;
}

//---- MethodReference ---------------------------------------------------------

NewMetaImpl(PeMethodReference, ByteArray, (TP(clp), T(line), TB(isPublic)));

PeMethodReference::PeMethodReference(int l, Class *c, char *method, bool ip)
						  : ByteArray((byte*)method)
{
    clp= c;
    line= l;
    isPublic= ip;
}

bool PeMethodReference::IsEqual(Object *op)
{
    if (op == 0 || !op->IsKindOf(PeMethodReference))
	return FALSE;
    PeMethodReference *mrp= (PeMethodReference *)op;
    return (strcmp((char*)Str(), (char*)mrp->Str()) == 0) && (clp == mrp->clp);
}

