#ifndef Port_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define Port_First

#include "System.h"
#include "Point.h"
#include "Font.h"
#include "Bitmap.h"
#include "Token.h"

class ImageCache;
class Picture;

extern const int MaxTextBatchCnt;

enum GrHints {
    eHintBatch,
    eHintUnbatch,
    eHintPostScript,
    eHintPic,
    eHintCacheStart,
    eHintCacheEnd,
    eHintWinErrorMsg,
    eHintWinStatusMsg,
    eHintWinBusy,
    eHintWinSetName,
    eHintEPSFStart,
    eHintEPSFEnd
};

enum GrLineCap {
    eDefaultCap,
    eStartArrow,
    eEndArrow
};

enum GrPolyType {
    ePolyDefault,
    ePolyBezier
};

enum GrCursor {
    eCrsNone, 
    eCrsArrow,
    eCrsBoldArrow,
    eCrsCross,
    eCrsBoldCross,
    eCrsUpArrow,
    eCrsDownArrow,
    eCrsRightArrow,
    eCrsLeftArrow,
    eCrsHourglass,
    eCrsPen,
    eCrsMoveHand,
    eCrsHand,
    eCrsCrossHair,
    eCrsMoveStretch,
    eCrsIBeam,
    eCrsLeftRightArrow,
    eCrsUpDownArrow,
    eCrsHelp
};

class GrState {
public:
    Port *port;
    Rectangle clip;
    Point origin;
    float scalex, scaley;
public:
    GrState(Port *port= 0);
    ~GrState();
};

//---- Port --------------------------------------------------------------------

class Port : public SysEvtHandler {
friend ImageCache;
friend GrState;
friend Bitmap;
friend Ink;
protected:
    static Port *port;

    static Port *tbport;
    static Ink *tbink;
    static Point tbpos, tblastpos;
    static Rectangle tbbbox;

    /* graphic state */
    Point origin;
    Rectangle cliprect;
    Ink *fillink;
    
    /* pen state */
    Point penpos;
    short pensize;
    Ink *penink;
    GrLineCap pencap;
    
    /* text state */
    Point textpos;
    Ink *textink;
    Font *textfont;
    
    /* other */
    Ink *lastink;
    bool ignore;
    bool hascolor;
    int depth;
    
public:
    bool scale;
    float scalex, scaley;
    Font *sfont, *lastfdp;
	  
protected:
    Port();
    
    bool Visible(Ink *ink, const Rectangle &r);
    void InstallInk(Ink *ink)
	{ lastink= ink; ink->SetInk(this); }

    //---- device dependent primitives ------------------------------------------
    virtual void DevClip(Rectangle*);
    virtual bool DevScale(float sx, float sy);
    virtual void DevSetPattern(Bitmap*);
    virtual void DevSetColor(RGBColor*);
    virtual void DevFreeColor(RGBColor*);
    virtual void DevSetOther(int);
    virtual void DevSetGrey(float);
    virtual void DevStrokeLine(int, Rectangle*, GrLineCap, Point, Point);
    virtual void DevStrokeRect(int, Rectangle*);
    virtual void DevFillRect(Rectangle*);
    virtual void DevStrokeOval(int, Rectangle*);
    virtual void DevFillOval(Rectangle*);
    virtual void DevStrokeRRect(int, Rectangle*, Point);
    virtual void DevFillRRect(Rectangle*, Point);
    virtual void DevStrokeWedge(int, GrLineCap, Rectangle*, int, int);
    virtual void DevFillWedge(Rectangle*, int, int);
    virtual void DevStrokePolygon(Rectangle*, Point*, int, GrPolyType t, int, GrLineCap);
    virtual void DevFillPolygon(Rectangle*, Point*, int, GrPolyType t);
    virtual void DevShowBitmap(Rectangle*, Bitmap*);
    virtual void DevShowPicture(Rectangle *r, Picture *pic);
    virtual void DevShowChar(Font*, Point, byte, Rectangle*);
    virtual void DevShowTextBatch(Rectangle*, Point);
    virtual void DevGiveHint(int code, int len, void *vp);
    virtual GrCursor SetCursor(GrCursor c);
    virtual GrCursor SetWaitCursor(unsigned int d= 400, GrCursor c= eCrsHourglass);
    virtual GrCursor GetCursor();
    virtual bool DevImageCacheBegin(ImageCache *bb, Rectangle *r);
    virtual void DevImageCacheEnd(ImageCache *bb);

    virtual void DevStrokeRect2(int, Rectangle*);
    virtual void DevStrokeRRect2(int, Rectangle*, Point);
    virtual void DevFillRRect2(Rectangle*, Point);
    virtual void DevStrokeOval2(int, Rectangle*);
    virtual void DevFillOval2(Rectangle*);
    virtual void DevStrokeLine2(int, Rectangle*, GrLineCap, Point, Point);
    virtual void DevStrokeWedge2(int, GrLineCap, Rectangle*, int, int);
    virtual void DevFillWedge2(Rectangle*, int, int);
    virtual void DevStrokePolygon2(Rectangle*, Point*, int, GrPolyType, int, GrLineCap);
    virtual void DevFillPolygon2(Rectangle*, Point*, int, GrPolyType);
    virtual Point DevDrawArrow(int, Point, Point);

public:

    MetaDef(Port);
    virtual ~Port();
    void PortInit();

    bool HasColor()
	{ return hascolor; }
    int GetDepth()
	{ return depth; }

    //---- clipping ------------------------------------------------------------
    void Clip(const Rectangle &r);
    void ClipFurther(const Rectangle &r);
    Rectangle GetCliprect()
	{ return cliprect; }
    void Translate(const Point &p);
    Point GetOrigin()
	{ return origin; }
    void Scale(float x, float y);
    float GetXScale()
	{ return scalex; }
    float GetYScale()
	{ return scaley; }
    void InitClip();

    //---- Fill ----------------------------------------------------------------
    void SetNormal()
	{ fillink= gInkBlack; }
    void SetInk(Ink *p)
	{ fillink= p; }
    void SetColor(RGBColor*);
    void FreeColor(RGBColor *cp);
    void FillRect(Ink*, const Rectangle &r);
    void FillOval(Ink*, const Rectangle &r);
    void FillRRect(Ink*, const Rectangle &r, const Point &corner);
    void FillWedge(Ink*, const Rectangle &r, int, int);
    void FillPolygon(const Point &at, Ink*, Point*, int, GrPolyType);

    //---- Pen and Lines -------------------------------------------------------
    void SetPenNormal();
    void SetPenInk(Ink *p)
	{ penink= p; }
    Ink *GetPenInk()
	{ return penink; }
    void SetLineCap(GrLineCap lc)
	{ pencap= lc; }
    void SetPenSize(int w)
	{ pensize= w; }
    void Moveto(const Point &p)
	{ penpos= p; }
    void Moveby(const Point &p)
	{ penpos+= p; }

    void StrokeLine(Ink*, int, GrLineCap, Point, Point);
    void Line(const Point &p1, const Point &p2)
	{ StrokeLine(penink, pensize, pencap, p1, p2); }
    void PaintLine(Ink *ink, GrLineCap cap, const Point &p1, const Point &p2)
	{ StrokeLine(ink, pensize, cap, p1, p2); }
    void Lineto(const Point &p)
	{ StrokeLine(penink, pensize, pencap, penpos, p); penpos= p; }

    void StrokeRect(Ink*, int, const Rectangle &r);
    void StrokeOval(Ink*, int, const Rectangle &);
    void StrokeRRect(Ink*, int, const Rectangle &r, const Point &corner);
    void StrokeWedge(Ink*, int, GrLineCap, const Rectangle &r, int, int);
    void StrokePolygon(const Point &at, Ink*, Point*, int, GrPolyType, int, GrLineCap);

    //---- Text ----------------------------------------------------------------
    void SetTextNormal();
    void SetFamily(GrFont);
    void SetFont(Font *fd);
    Font *GetFont()
	{ return textfont; }
    void SetSize(int);
    void SetFace(GrFace);
    void SetTextInk(Ink *p)
	{ textink= p; }
    Ink *GetTextInk()
	{ return textink; }

    Point GetTextPos()   
	{ return textpos; }
    void TextMoveto(const Point &p)
	{ textpos= p; }
    void TextMoveby(const Point &p)
	{ textpos+= p; }
    void TextHMoveby(int h)
	{ textpos.x+= h; }
    void TextVMoveby(int v)
	{ textpos.y+= v; }
	
    int ShowChar(Font *fdp, Ink *ink, const Point &pos, byte c);
    void DrawChar(byte c)
	{ textpos.x+= ShowChar(textfont, textink, textpos, c); }

    int ShowString(Font*, Ink*, const Point &pos, byte*, int);
    void DrawString(byte *text, int l)
	{ textpos.x+= ShowString(textfont, textink, textpos, text, l); }

    void flushtext();
    void FlushMyText()
	{ if (tbport == this) flushtext(); }
    static void FlushAnyText()
	{ if (tbport) tbport->flushtext(); }


    //---- misc ----------------------------------------------------------------
    void ShowBitmap(Ink*, const Rectangle &r, Bitmap*);
    void ShowPicture(const Rectangle &r, Picture*);
    void GiveHint(int code, int len, void *vp);

    //---- Port syntactic sugar ------------------------------------------------
    friend void GrSetPort(Port *newport)
	{ Port::port= newport; }
    friend Port *GrGetPort()
	{ return Port::port; }

    //---- clipping ------------------------------------------------------------
    friend void GrInitClip()
	{ Port::port->InitClip(); }
    friend void GrSetClip(const Rectangle &r)
	{ Port::port->Clip(r); }
    friend void GrClipFurther(const Rectangle &r)
	{ Port::port->ClipFurther(r); }
    friend Rectangle GrGetClip()
	{ return Port::port->GetCliprect(); }
    friend void GrTranslate(const Point &o)
	{ Port::port->Translate(o); }
    friend Point GrGetOrigin()
	{ return Port::port->GetOrigin(); }
    friend void GrScale(float x, float y)
	{ Port::port->Scale(x, y); }
    friend float GrGetXScale()
	{ return Port::port->GetXScale(); }
    friend float GrGetYScale()
	{ return Port::port->GetYScale(); }

    //---- fill ----------------------------------------------------------------
    friend void GrSetInk(Ink *p)
	{ Port::port->SetInk(p); }
    
    //---- pens ----------------------------------------------------------------
    friend void GrSetPenSize(int w)
	{ Port::port->SetPenSize(w); }
    friend void GrSetLineCap(GrLineCap lc)
	{ Port::port->SetLineCap(lc); }
    friend void GrSetPenInk(Ink *p)
	{ Port::port->SetPenInk(p); }
    friend Ink *GrGetPenInk()
	{ return Port::port->GetPenInk(); }
    friend void GrSetPenNormal()
	{ Port::port->SetPenNormal(); }
    
    //---- lines ---------------------------------------------------------------
    friend void GrMoveto(const Point &p)
	{ Port::port->Moveto(p); }
    friend void GrLineto(const Point &p)
	{ Port::port->Lineto(p); }
    friend void GrLine(const Point &p1, const Point &p2)
	{ Port::port->StrokeLine(Port::port->penink, Port::port->pensize, Port::port->pencap, p1, p2); }
    friend void GrPaintLine(Ink *ink, int psz, GrLineCap cap, const Point &p1, const Point &p2)
	{ Port::port->StrokeLine(ink, psz, cap, p1, p2); }
    friend void GrStrokeLine(Ink *ink, GrLineCap cap, const Point &p1, const Point &p2)
	{ Port::port->StrokeLine(ink, Port::port->pensize, cap, p1, p2); }
    friend void GrInvertLine(const Point &p1, const Point &p2)
	{ Port::port->StrokeLine(gInkXor, Port::port->pensize, Port::port->pencap, p1, p2); }
    friend void GrEraseLine(const Point &p1, const Point &p2)
	{ Port::port->StrokeLine(gInkWhite, Port::port->pensize, Port::port->pencap, p1, p2); }

    //---- rectangle -----------------------------------------------------------
    friend void GrFillRect(const Rectangle &r)
	{ Port::port->FillRect(Port::port->fillink, r); }
    friend void GrStrokeRect(const Rectangle &r)
	{ Port::port->StrokeRect(Port::port->penink, Port::port->pensize, r); }
    friend void GrEraseRect(const Rectangle &r)
	{ Port::port->FillRect(gInkWhite, r); }
    friend void GrInvertRect(const Rectangle &r)
	{ Port::port->FillRect(gInkXor, r); }
    friend void GrPaintRect(const Rectangle &r, Ink *ink)
	{ Port::port->FillRect(ink, r); }
    
    //---- round rectangles ----------------------------------------------------
    friend void GrStrokeRoundRect(const Rectangle &r, const Point &dia)
	{ Port::port->StrokeRRect(Port::port->penink, Port::port->pensize, r, dia); }
    friend void GrFillRoundRect(const Rectangle &r, const Point &dia)
	{ Port::port->FillRRect(Port::port->fillink, r, dia); }
    friend void GrEraseRoundRect(const Rectangle &r, const Point &dia)
	{ Port::port->FillRRect(gInkWhite, r, dia); }
    friend void GrInvertRoundRect(const Rectangle &r, const Point &dia)
	{ Port::port->FillRRect(gInkXor, r, dia); }
    friend void GrPaintRoundRect(const Rectangle &r, const Point &dia, Ink *ink)
	{ Port::port->FillRRect(ink, r, dia); }
    
    //----- ovals --------------------------------------------------------------
    friend void GrStrokeOval(const Rectangle &r)
	{ Port::port->StrokeOval(Port::port->penink, Port::port->pensize, r); }
    friend void GrFillOval(const Rectangle &r)
	{ Port::port->FillOval(Port::port->fillink, r); }
    friend void GrEraseOval(const Rectangle &r)
	{ Port::port->FillOval(gInkWhite, r); }
    friend void GrInvertOval(const Rectangle &r)
	{ Port::port->FillOval(gInkXor, r); }
    friend void GrPaintOval(const Rectangle &r, Ink *ink)
	{ Port::port->FillOval(ink, r); }
    
    //----- wedges -------------------------------------------------------------
    friend void GrStrokeWedge(const Rectangle &r, int s, int e)
	{ Port::port->StrokeWedge(Port::port->penink, Port::port->pensize, Port::port->pencap, r, s, e); }
    friend void GrFillWedge(const Rectangle &r, int s, int e)
	{ Port::port->FillWedge(Port::port->fillink, r, s, e); }
    friend void GrEraseWedge(const Rectangle &r, int s, int e)
	{ Port::port->FillWedge(gInkWhite, r, s, e); }
    friend void GrInvertWedge(const Rectangle &r, int s, int e)
	{ Port::port->FillWedge(gInkXor, r, s, e); }
    friend void GrPaintWedge(const Rectangle &r, int s, int e, Ink *ink)
	{ Port::port->FillWedge(ink, r, s, e); }
    
    //---- bitmaps -------------------------------------------------------------
    friend void GrShowBitMap(const Rectangle &r, Bitmap *bmp)
	{ Port::port->ShowBitmap(Port::port->fillink, r, bmp); }
    friend void GrPaintBitMap(const Rectangle &r, Bitmap *bmp, Ink *ink)
	{ Port::port->ShowBitmap(ink, r, bmp); }
    friend void GrInvertBitMap(const Rectangle &r, Bitmap *bmp)
	{ Port::port->ShowBitmap(gInkXor, r, bmp); }
    
    //--- Polygon --------------------------------------------------------------
    friend void GrStrokePolygon(const Point &at, Point *pts, int npts, GrPolyType t)
	{ Port::port->StrokePolygon(at, Port::port->penink, pts, npts, t, Port::port->pensize, Port::port->pencap); }
    friend void GrFillPolygon(const Point &at, Point *pts, int npts, GrPolyType t)
	{ Port::port->FillPolygon(at, Port::port->fillink, pts, npts, t); }
    friend void GrErasePolygon(const Point &at, Point *pts, int npts, GrPolyType t)
	{ Port::port->FillPolygon(at, gInkWhite, pts, npts, t); }
    friend void GrInvertPolygon(const Point &at, Point *pts, int npts, GrPolyType t)
	{ Port::port->FillPolygon(at, gInkXor, pts, npts, t); }
    friend void GrPaintPolygon(const Point &at, Ink *ink, Point *pts, int npts, GrPolyType t)
	{ Port::port->FillPolygon(at, ink, pts, npts, t); }

    //---- text ----------------------------------------------------------------
    friend void GrSetFont(Font *fd)
	{ Port::port->SetFont(fd); }
    friend Font *GrGetFont()
	{ return Port::port->GetFont(); }
    friend void GrSetFamily(GrFont f)
	{ Port::port->SetFamily(f); }
    friend void GrSetSize(int s)
	{ Port::port->SetSize(s); }
    friend void GrSetFace(GrFace f)
	{ Port::port->SetFace(f); }
    friend void GrSetTextInk(Ink *tink)
	{ Port::port->SetTextInk(tink); }
    friend void GrSetTextNormal()
	{ Port::port->SetTextNormal(); }
    friend Ink *GrGetTextInk()
	{ return Port::port->GetTextInk(); }
    friend void GrTextMoveto(Point p)
	{ Port::port->TextMoveto(p); }
    friend void GrTextAdvance(int h)
	{ Port::port->TextHMoveby(h); }
    friend Point GrGetTextPos()
	{ return Port::port->GetTextPos(); }

    friend int GrShowChar(Font *fdp, Ink *ink, const Point &pos, byte c)
	{ return Port::port->ShowChar(fdp, ink, pos, c); }
    friend void GrDrawChar(byte c)
	{ Port::port->DrawChar(c); }
    friend int GrShowString(Font *fd, Ink *ink, const Point &pos, byte *text, int l= -1)
	{ return Port::port->ShowString(fd, ink, pos, text, l); }
    friend void GrDrawString(byte *s, int l= -1)
	{ Port::port->DrawString(s, l); }
    
    //---- pictures ------------------------------------------------------------
    friend void GrShowPicture(const Rectangle &r, Picture *pic)
	{ Port::port->ShowPicture(r, pic); }
    
    //---- cursor --------------------------------------------------------------
    friend GrCursor GrSetCursor(GrCursor c)
	{ return Port::port->SetCursor(c); }
    friend GrCursor GrGetCursor()
	{ return Port::port->GetCursor(); }
    friend GrCursor GrSetWaitCursor(unsigned int d= 400, GrCursor c= eCrsHourglass)
	{ return Port::port->SetWaitCursor(d, c); }

    //---- misc ----------------------------------------------------------------
    friend void GrGiveHint(int n, int l= 0, void *v= 0)
	{ Port::port->GiveHint(n, l, v); }
    friend bool GrHasColor()
	{ return Port::port->HasColor(); }
};

//---- ImageCache --------------------------------------------------------------

class ImageCache {
public:
    Bitmap *b;
    Rectangle vr, r;
    bool valid;
public:
    ImageCache()
	{ b= 0; valid= FALSE; }
    ~ImageCache()
	{ SafeDelete(b); }
    bool Open(const Rectangle &r)
	{ return Port::port->DevImageCacheBegin(this, (Rectangle*)&r); }
    void Close()
	{ Port::port->DevImageCacheEnd(this); }
    void Invalidate()
	{ valid= FALSE; }
};

#endif
