#ifdef __GNUG__
#pragma implementation
#endif

#include "ProgressDialog.h"

#include "Class.h"
#include "Dialog.h"
#include "Box.h"
#include "TextItem.h"
#include "WindowSystem.h"
#include "Window.h"
#include "WindowPort.h"
#include "Buttons.h"
#include "Look.h"
#include "BorderItems.h"

//---- ProgressBar -----------------------------------------------

class ProgressBar : public VObject {
public:
    int val;
public:
    MetaDef(ProgressBar);
    ProgressBar();
    Metric GetMinSize();
    void Draw(Rectangle);
    void SetVal(int);
};

ProgressBar::ProgressBar() : VObject(cIdNone)
{
    contentRect.extent = Point(20);
}

Metric ProgressBar::GetMinSize()
{
    return gLook->ProgressBarLayout()->GetMinSize(this);
}

void ProgressBar::Draw(Rectangle r)
{
    gLook->ProgressBarLayout()->Adorn(this, r, val);
}

void ProgressBar::SetVal(int v)
{
    if(v!=val)
    	Changed();
    val= v;
    ForceRedraw();
    if (GetWindow()) {
	WindowPort *wp= (WindowPort*) GetWindow()->GetPortDesc();
	wp->DevUpdate();
    }
}
NewMetaImpl(ProgressBar,VObject,(T(val)));

//---- ProgressDialog --------------------------------------------

const int cIdStop   = cIdFirstUser + 100;

class ProgressDialog : public Dialog {
    VObject *box;
    ProgressBar *bar;
    TextItem *ti;
    VObject *stop;
    bool stoppable;
public:
    bool quit;

public:
    MetaDef(ProgressDialog);
    ProgressDialog(ProgressBar *, TextItem *, bool stoppable);

    VObject *DoMakeContent();
    void SetStop(bool);
    void Show();
    void Control(int id, int part, void *val);
};

NewMetaImpl(ProgressDialog, Dialog,(TP(box),TP(bar),TP(ti),TP(stop),
	TB(stoppable),TB(quit)));

ProgressDialog::ProgressDialog (ProgressBar *b, TextItem *t, bool s)
    : Dialog ("Progress", eWinFixed)
{
    bar= b;
    ti= t;
    stoppable= s;
}

VObject *ProgressDialog::DoMakeContent()
{
    stop= new ActionButton(cIdStop, "Stop");
    stop->SetFlag(eVObjHFixed);

    return
	new Matte(
	    box= new VBox(Point(10,10), VObjAlign(eVObjHCenter|eVObjHExpand),
		ti,
		bar,
		stoppable ? stop : NULL,
		NULL
	    )
	);
}

void ProgressDialog::SetStop(bool s)
{
    if (stoppable != s) {
	stoppable = s;
	if (box && stop) {
	    if (stoppable)
		box->Add(stop);
	    else
		box->Remove(stop);
	}
    }
}

void ProgressDialog::Show()
{
    quit= FALSE;
    bar->val= 0;
    ShowAt(NULL, gPoint_1);
}

void ProgressDialog::Control(int id, int part, void *val)
{
    if (id == cIdStop)
	quit= TRUE;
    else if (id != cCLOSE)
	Dialog::Control(id, part, val);
}

//---- ProgressImpl ------------------------------------------------------------
NewMetaImpl(ProgressImpl,Progress,(TB(stoppable),TP(ti),TP(pb),TP(bar)));

ProgressImpl::ProgressImpl()
{
    stoppable= TRUE;
    ti= new TextItem("");
}

ProgressImpl::~ProgressImpl()
{
    if (pb == 0)
	SafeDelete(ti);
    SafeDelete(pb);
}

void ProgressImpl::SetUp()
{
    gWindowSystem->SetWait();
}

void ProgressImpl::SetStop(bool b)
{
    stoppable= b;
    if (pb)
	pb->SetStop(b);
}

void ProgressImpl::SetMessage(char *message)
{
    ti->SetString (message, /* redraw = */ TRUE);
}

bool ProgressImpl::SetVal()
{
    if ( (pb == 0 || ! pb->IsOpen()) && curr > max )
	return FALSE;

    if (bar == 0)
	bar= new ProgressBar();

    int v= (curr * bar->contentRect.extent.x) / max;

    if (v != bar->val) {
	if (pb == 0)
	    pb= new ProgressDialog(bar, ti, stoppable);
	bar->SetVal(v);
	if (! pb->IsOpen())
	    pb->Show();

	((WindowPort*) pb->GetWindow()->GetPortDesc())->Poll(0);
	if (pb->quit) {
	    pb->quit = FALSE;
	    Abort(FALSE);
	    return TRUE;
	}
    }
    return FALSE;
}

void ProgressImpl::Abort (bool skiptoend)
{
    gWindowSystem->ResetWait();
    if (pb && pb->IsOpen()) {
	if (skiptoend) {
	    if ( fwd ) {
		for (int i= bar->val; i < bar->contentRect.extent.x; i++)
		    bar->SetVal(i);
	    } else {
		for (int i= bar->val; i>0; i--)
		    bar->SetVal(i);
	    }
	}
	pb->Close();
    }
}

//---- ProgressFilter ----------------------------------------------------------

ProgressFilter::ProgressFilter() : StreamBuf()
{
}

int ProgressFilter::Underflow(u_char*, int size)
{
    size= sb->salloc(&base, BUFSIZE);
    if (size == EOF)
	return EOF;
    if (gProgress->Inc(size))
	return EOF;
    return size;
}

int ProgressFilter::Overflow(u_char *bp, int size)
{
    if (gProgress->Inc(size))
	return EOF;
    return size= sb->sputn((char*)bp, size);
}
