#ifdef __GNUG__
#pragma implementation
#endif

#include "Scroller.h"

#include "ETMath.h"
#include "Class.h"
#include "ScrollBar.h"
#include "Clipper.h"
#include "View.h"
#include "Look.h"
#include "SeqColl.h"

//---- Scroller ----------------------------------------------------------------

NewMetaImpl(Scroller,CompositeVObject,
	(TAP(sb), TP(vf), TP(vop), TE(controls), TE(visibleControls), T(ms)));

Scroller::Scroller(VObject *v, Point e, int id, ScrollDir w)
		: CompositeVObject(id, (SeqCollection*) 0)
{
    controls= w;
    vop= v;
    ms= e;
    visibleControls= eScrollNone;
}

Scroller::Scroller(VObject *v, Ink *bg, Point e, int id, ScrollDir w)
		: CompositeVObject(id, (SeqCollection*) 0)
{
    controls= w;
    vop= v;
    ms= e;
    visibleControls= eScrollNone;
    SetBgInk(bg);
}

Clipper *Scroller::MakeClipper()
{
    return new Clipper(vop, ms, GetId());
}

ScrollBar *Scroller::MakeScrollBar(int id)
{
    return new ScrollBar(cIdNone, (Direction) Math::Even(id));
}

Clipper *Scroller::GetClipper()
{
    if (vf == 0)
	Add(vf = MakeClipper());
    return vf;
}

ScrollBar *Scroller::GetScrollBar(int i)
{
    if ((sb[i] == 0) && (controls & BIT(i)))
	Add(sb[i]= MakeScrollBar(i));
    return sb[i];
}

Command *Scroller::Input(Point lp, Token &t, Clipper *vf)
{
    if (t.IsKey() || t.IsFunctionKey() || t.IsCursorKey())
	return GetClipper()->Input(lp, t, vf);
    return CompositeVObject::Input(lp, t, vf);
}

Command *Scroller::DoOtherFunctionKeyCommand(int key, Token t)
{
    Clipper *clipper = GetClipper();
    if ( clipper ) {
	switch ( key ) {
	case eEvtKey_PageUp:
	    clipper->Scroll(cPartScrollPage, Point(0,-1));
	    return gNoChanges;
	case eEvtKey_PageDown:
	    clipper->Scroll(cPartScrollPage, Point(0,1));
	    return gNoChanges;
	case eEvtKey_Home:
	    clipper->Scroll(cPartScrollAbs, gPoint0);
	    return gNoChanges;
	case eEvtKey_End:
	    clipper->Scroll(cPartScrollAbs, Point(0,cMaxInt));
	    return gNoChanges;
	}
    }
    return CompositeVObject::DoOtherFunctionKeyCommand(key, t);
}

void Scroller::Control(int id, int part, void *val)
{
    switch (part) {
    case cPartScrollPos:
	BubbleUpdate(*((Point*)val));
	break;
	
    case cPartViewSize:
	ViewSizeChanged();
	BubbleUpdate(GetClipper()->GetRelOrigin());
	break;
	
    case cPartScrollPage:
    case cPartScrollRel:
    case cPartScrollStep:
    case cPartScrollAbs:
    case cPartScrollHAbs:
    case cPartScrollVAbs:
	GetClipper()->Scroll(part, *((Point*)val));
	break;
	
    default:
	CompositeVObject::Control(id, part, val);
	break;
    }
}

void Scroller::BubbleUpdate(Point bs)
{
    for (int i= 0; i < 4; i++)
	if (visibleControls & BIT(i))
	    GetScrollBar(i)->BubbleUpdate(bs);
}

bool Scroller::NeedControls(Point e, Point vs)
{
    ScrollDir lastVisibleControls= visibleControls;
    Point fgap((int)gLook->FrameLayout()->GetValue(this, 1)),
	  border(2 * (int)gLook->FrameLayout()->GetValue(this, 0));
    e-= border;
    
    visibleControls= (ScrollDir)(controls & ~eScrollHideScrolls);
    if (controls & eScrollHideScrolls) {
	// check whether we need horizontal scrollbars
	// if yes, then the Clipper shrinks vertically
	if (vs.x > e.x) {
	    if (controls & eScrollDown)
		e.y -= GetScrollBar(1)->GetMinSize().extent.y + fgap.y;
	    if (controls & eScrollUp)
		e.y -= GetScrollBar(3)->GetMinSize().extent.y + fgap.y;
	}

	// check whether we need vertical scrollbars
	// if yes, then the Clipper shrinks horizontally
	if (vs.y > e.y) {
	    if (controls & eScrollRight)
		e.x -= GetScrollBar(0)->GetMinSize().extent.x + fgap.x;
	    if (controls & eScrollLeft)
		e.x -= GetScrollBar(2)->GetMinSize().extent.x + fgap.x;
	}
	else
	    visibleControls= (ScrollDir)(visibleControls&~(eScrollLeft | eScrollRight));

	// once again we must check whether we need horizontal scrollbars
	// because the Clipper's width might have changed through the
	// vertical scrollbar check
	if (vs.x <= e.x)
	    visibleControls= (ScrollDir)(visibleControls&~(eScrollUp | eScrollDown));
    }
    return visibleControls != lastVisibleControls;
}

void Scroller::ViewSizeChanged()
{
    Point newviewsize(GetClipper()->GetViewSize());

    if (NeedControls(GetExtent(), newviewsize)) {
	SetExtent(GetExtent());
	SetOrigin(GetOrigin());
	ForceRedraw();
    }

    for (int i= 0; i < 4; i++)
	if (visibleControls & BIT(i))
	    GetScrollBar(i)->ViewSizeChanged(newviewsize);
}

void Scroller::Open(bool mode)
{
    SetFlag(eVObjOpen, mode);
    GetClipper()->Open(mode);
    for (int i= 0; i < 4; i++)
	if (visibleControls & BIT(i))
	    GetScrollBar(i)->Open(mode);
}

void Scroller::Draw(Rectangle r)
{
    CompositeVObject::Draw(r);
    gLook->FrameLayout()->Adorn(this, GetClipper()->contentRect, 0);
}

void Scroller::SetBgInk(Ink *bg)
{
    GetClipper()->SetBgInk(bg);
}

Ink *Scroller::GetBgInk()
{
    return GetClipper()->GetBgInk();
}

Metric Scroller::GetMinSize()
{
    Point fgap((int)gLook->FrameLayout()->GetValue(this, 1)),
	  border(2 * (int)gLook->FrameLayout()->GetValue(this, 0)),
	  sbSizes,
	  clipperSize(GetClipper()->GetMinSize().extent + border);

    for (int i= 0; i < 4; i++) {
	ScrollBar *sb= GetScrollBar(i);
	if (sb) {
	    Point ee(sb->GetMinSize().extent + fgap);
	    if (Math::Even(i)) {      // right & left scrollbar
		sbSizes.x+= ee.x;
		clipperSize.y= Math::Max(clipperSize.y, ee.y);
	    } else {            // up & down scrollbar
		sbSizes.y+= ee.y;
		clipperSize.x= Math::Max(clipperSize.x, ee.x);
	    }
	}
    }
    return clipperSize + sbSizes;
}

void Scroller::SetExtent(Point e)
{
    Point fgap((int)gLook->FrameLayout()->GetValue(this, 1)),
	  border(2 * (int)gLook->FrameLayout()->GetValue(this, 0)),
	  viewExtent(GetClipper()->GetViewSize());
    
    VObject::SetExtent(e);
    
    NeedControls(e, viewExtent);

    if (visibleControls & eScrollRight)
	e.x= e.x - GetScrollBar(0)->GetMinSize().extent.x - fgap.x;

    if (visibleControls & eScrollDown)
	e.y= e.y - GetScrollBar(1)->GetMinSize().extent.y - fgap.y;
    
    if (visibleControls & eScrollLeft)
	e.x= e.x - GetScrollBar(2)->GetMinSize().extent.x - fgap.x;

    if (visibleControls & eScrollUp)
	e.y= e.y - GetScrollBar(3)->GetMinSize().extent.y - fgap.y;
    
    GetClipper()->SetExtent(e-border);
    Point vs(GetClipper()->GetExtent());

    for (int i= 0; i < 4; i++) {
	ScrollBar *sb= GetScrollBar(i);
	if (sb) {
	    if (visibleControls & BIT(i)) {
		Point size(sb->GetMinSize().extent);
		if (Math::Even(i))
		    size.y= e.y;
		else
		    size.x= e.x;
		if (!sb->IsOpen())
		    sb->Open(TRUE);
		sb->Init(size, vs, viewExtent);
	    } else {
		if (sb->IsOpen())
		    sb->Open(FALSE);
	    }
	}
    }
}

void Scroller::SetOrigin(Point at)
{
    Point fgap((int)gLook->FrameLayout()->GetValue(this, 1)),
	  border((int)gLook->FrameLayout()->GetValue(this, 0)),
	  left, middle(GetClipper()->GetExtent()+2*border);
    
    VObject::SetOrigin(at);
   
    if (visibleControls & eScrollLeft)
	left.x= GetScrollBar(2)->Width() + fgap.x;
    if (visibleControls & eScrollUp)
	left.y= GetScrollBar(3)->Height() + fgap.y;
	
    GetClipper()->SetOrigin(at+left+border);
    
    if (visibleControls & eScrollRight)
	GetScrollBar(0)->SetOrigin(at+left+Point(middle.x+fgap.x, 0));
    if (visibleControls & eScrollDown)
	GetScrollBar(1)->SetOrigin(at+left+Point(0, middle.y+fgap.y));
    if (visibleControls & eScrollLeft)
	GetScrollBar(2)->SetOrigin(at+Point(0, left.y));
    if (visibleControls & eScrollUp)
	GetScrollBar(3)->SetOrigin(at+Point(left.x, 0));        
}

OStream& Scroller::PrintOn(OStream &s)
{
    VObject::PrintOn(s);
    return s << vop SP << controls SP << ms SP;
}

IStream& Scroller::ReadFrom(IStream &s)
{
    FreeAll();
    sb[0]= sb[1]= sb[2]= sb[3]= 0;
    vf= 0;
    visibleControls= eScrollNone;
    
    VObject::ReadFrom(s);
    return s >> vop >> Enum(controls) >> ms;
}

