#include "StringGridView.h"

#include "Class.h"

#include "ETMath.h"
#include "Window.h"
#include "Look.h"
#include "Manager.h"
#include "TextView.h"
#include "TextPainter.h"
#include "BorderItems.h"

//---- CellEditField -----------------------------------------------------------

NewMetaImpl(CellEditField,EditField, (TP(fStringGridView), T(fEditCell)));

CellEditField::CellEditField(StringGridView *sgv, long minw, Font *f)
    : fStringGridView(sgv), fEditCell(gPoint_1), EditField(cIdNone, f->Width('w') * minw, f)
{
    SetFlag(eEditFieldNoBorder);
    SetContainer(fStringGridView);
}

char *CellEditField::GetStringForEditing()
{
    char *s= (char*)fStringGridView->GetString(fEditCell, FALSE);
    if (s)
	return s;
    return "";
}

bool CellEditField::ValidateString(const char *s)
{
    bool b= TRUE;
    const char *olds= fStringGridView->GetString(fEditCell, FALSE);
    if (olds == 0 || strcmp(s, (char*) olds) != 0)
	b= fStringGridView->SetString(fEditCell, s);
    if (b) {
	fEditCell= gPoint_1;
    }
    return b;
}

void CellEditField::SetEditCell(const Point &ec, const Rectangle &cr, Font *f)
{
    Manager *m= GetManager();
    if (m && ec != fEditCell) {
	fEditCell= ec;
	SetContentRect(cr, TRUE);
	SetFont(f);
	m->SetFirstHandler(this);
	UpdateEvent();
    }
}

Command *CellEditField::Input(Point lp, Token &t, Clipper *vf)
{
    if (t.IsKey() && t.Code == '\t') {
	fStringGridView->TabField(fEditCell, t);
	return gNoChanges;
    }
    return EditField::Input(lp, t, vf);
}

Command *CellEditField::InsertString(const char *s)
{
//XXX Michael: Adobt to T5 et++!!!!
//  TextViewOverlay *tv= GetTextViewOverlay();
//  if (tv)
//      return tv->GetOverlayTextView()->InsertString(s);
    return gNoChanges;
}

//---- StringGridView ----------------------------------------------------------------

NewMetaImpl(StringGridView,GridView, (TP(fEditField)));

StringGridView::StringGridView(EvtHandler *eh, bool c)
    : GridView(eh, c)
{
    fEditField= new CellEditField(this, 10);
    SetModified();
}

StringGridView::~StringGridView()
{
    SafeDelete(fEditField);
}

TextViewOverlay *StringGridView::GetTextViewOverlay()
{
    if (fEditField)
	return fEditField->GetTextViewOverlay();
    return 0;
}

const char *StringGridView::GetString(const Point&, bool)
{
    return 0;
}

bool StringGridView::SetString(const Point&, const char*)
{
    return FALSE;
}

Metric StringGridView::CellSize(const Point &cell)
{
    int b= (int)gLook->FieldBorderLayout(TRUE)->GetValue(this, 0);
    return StaticTextView::MeasureText((char*)GetString(cell, TRUE),
	GetCellAttributes(cell)->fFont, 0, 1).Expand(cBorder+Point(b));
}

void StringGridView::DrawCell(const Point &cell, const Rectangle &r,
					    const Rectangle &clip, bool highlight)
{
    if (fEditField && fEditField->fEditCell == cell) {
	fEditField->DrawAll(clip);
    } else {
	if (GrHasColor() && highlight)
	    GrPaintRect(r, gHighlightColor);
	const char *s= GetString(cell, TRUE);
	if (s) {
	    Rectangle rrr(r);
	    CellAttributes *ca= GetCellAttributes(cell);
	    int b= (int)gLook->FieldBorderLayout(TRUE)->GetValue(this, 0);
//XXX Michael: Adobt to T5 et++!!!!
//          TextPainter::DrawBoxedText(rrr.Inset(cBorder+Point(b)), s,
//                      ca->fFont, ca->fInk, FALSE, ca->fHAlign, ca->fVAlign);
	}
	if (!GrHasColor() && highlight)
	    GrPaintRect(r, gHighlightColor);
    }
}

bool StringGridView::SetEditCell(const Point &cell)
{
    if (fEditField == 0 || cell != fEditField->fEditCell) {
	Manager *m= GetManager();
	if (m->SetFirstHandler(this)) {
	    if (IsEditable(cell)) {
		if (!GetSelection().ContainsPoint(cell))
		    SetSelection(cell);
		fEditField->SetEditCell(cell, GetCellRect(cell), GetCellAttributes(cell)->fFont);
		return TRUE;
	    }
	    return FALSE;
	}
	return FALSE;
    }
    return TRUE;
}

Point StringGridView::CalculateTab(const Point &curr, const Region &reg, bool fwd)
{
    Rectangle r(reg.BBox());
    Point p(curr);

    for (;;) {
	if (fwd) {
	    if (p.y < r.origin.y+r.extent.y-1) {
		p.y++;
	    } else {
		if (p.x < r.origin.x+r.extent.x-1) {
		    p.x++;
		    p.y= r.origin.y;
		} else
		    p= r.origin;
	    }
	} else {
	    if (p.y > r.origin.y) {
		p.y--;
	    } else {
		if (p.x > r.origin.x) {
		    p.x--;
		    p.y= r.origin.y+r.extent.y-1;
		} else
		    p= r.SE();
	    }
	}
	if (p == curr)
	    break;
	if (reg.ContainsPoint(p) && IsEditable(p))
	    break;
    }
    return p;
}

void StringGridView::TabField(const Point &cell, Token &t)
{
    Region sel(GetSelection());
    bool fwd= !t.IsShift();
    if (sel.BBox().origin == cell && sel.BBox().extent == gPoint1)
	SetEditCell(CalculateTab(cell, Rectangle(gPoint0, GetColsRows()), fwd));
    else
	SetEditCell(CalculateTab(cell, sel, fwd));
}

Command *StringGridView::DispatchEvents(Point lp, Token &t, Clipper *vf)
{
    Manager *m= GetManager();

    if (m) {
	Command *currCmd= 0;
	Region sel(GetSelection());
	if (t.IsLeftClick() && !t.IsShift()) {
	    if (m->SetFirstHandler(this)) {
		Point cell;
		PosToCell(lp, &cell);
		if (sel.ContainsPoint(cell) && SetEditCell(cell)) {
		    if (currCmd= fEditField->Input(lp, t, vf))
			return currCmd;
		}
	    }
	}
	if (t.IsKey() && sel.IsNotEmpty()) {
	    Point p(CalculateTab(sel.BBox().origin, sel));
	    //if (!IsEditable(p))
	    //    p= CalculateTab(p, sel);
	    if (SetEditCell(p)) {
		if (t.Code != '\t')
		    if (currCmd= fEditField->Input(lp, t, vf))
			return currCmd;
	    }
	}
    }
    return View::DispatchEvents(lp, t, vf);
}

void StringGridView::SetExtent(Point e)
{
    GridView::SetExtent(e);
    if (fEditField)
	fEditField->SetExtent(GetCellRect(fEditField->fEditCell).extent);
}

void StringGridView::SetOrigin(Point at)
{
    GridView::SetOrigin(at);
    if (fEditField)
	fEditField->SetOrigin(GetCellRect(fEditField->fEditCell).origin);
}

