#ifdef __GNUG__
#pragma implementation
#endif

#include "Class.h"
#include "ClassManager.h"
#include "VObject.h"
#include "Alert_e.h"
#include "Filler.h"
#include "Picture.h"
#include "PictureItem.h"
#include "ImageItem.h"
#include "Data.h"
#include "System.h"
#include "PathLookup.h"
#include "FileDialog.h"
#include "Window.h"
#include "ETString.h"
#include "CLib.h"
#include "Env.h"

#include "TextMarks.h"

static PathLookup *docPath= 0;
static FileDialog *fileDialog= 0;

//------ VObjectMark -----------------------------------------------------------

NewMetaImpl(VObjectMark, VisualMark, (TP(vop)));

VObjectMark::VObjectMark(VObject *v) 
{   
    vop= v; 
}

VObjectMark::~VObjectMark() 
{
    SafeDelete(vop);
}

void VObjectMark::CalcExtent()
{
    vop->CalcExtent();
}

Point VObjectMark::GetExtent()
{
    return vop->GetExtent();
}
    
int VObjectMark::Base()
{
    return vop->Base();
}
    
void VObjectMark::SetContainer(VObject *cont)
{
    vop->SetContainer(cont);
    if (cont && cont->IsOpen())
	vop->Open(); 
}

void VObjectMark::Draw(Point at, Rectangle clip, Rectangle, bool border)
{
    vop->SetOrigin(at);
    vop->DrawAll(clip, FALSE);
    if (border) {
	GrSetPenInk(ePatGrey50);
	GrStrokeRect(vop->ContentRect());
    }
}

void VObjectMark::SendDown(int what, int p, void *arg)
{
    vop->SendDown(what, p, arg);
}

bool VObjectMark::WantsInput(Point lp)
{
    return vop->ContainsPoint(lp);
}

class Command *VObjectMark::Input(Point lp, Token &t, class Clipper *cl)
{
    return vop->Input(lp, t, cl);
}

void VObjectMark::FreeAll()
{
    vop->FreeAll();
    SafeDelete(vop); 
}

OStream &VObjectMark::PrintOn(OStream &s) 
{
    VisualMark::PrintOn(s);
    return s << vop->ClassName() SP << vop->GetExtent() SP << vop SP << '\007';
}

IStream &VObjectMark::ReadFrom(IStream &s) 
{
    VisualMark::ReadFrom(s);
    
    if (IsA()->LoadVersion() < 3.1)
	return s >> vop;
	
    char c, className[1024];
    Point extent;
    
    s >> className >> extent;
    
    if (gClassManager->Find(className))
	s >> vop;
    else
	vop= new Filler(extent, ePatGrey50);

    while (s.get(c))
	if (c == '\007')
	    break;
    return s;
}

void VObjectMark::CollectParts(Collection *col)
{
    VisualMark::CollectParts(col);
    col->Add(vop);
}

//------ ReferenceMark --------------------------------------------------------

NewMetaImpl(ReferenceMark, VisualMark, (TP(vop),TS(filename),TB(verify),T(sizeHint)));

ReferenceMark::ReferenceMark(char *fn, Point sh) 
{   
    filename= 0;
    verify=TRUE;
    strreplace(&filename, fn);
    vop= 0; 
    sizeHint= sh;
}

ReferenceMark::~ReferenceMark() 
{
    SafeDelete(filename);
}

VObject *ReferenceMark::GetVObject()
{
    Load();
    if (vop)
	return vop;
    return 0;
}

Point ReferenceMark::GetExtent()
{
    if (sizeHint != gPoint_1) 
	return sizeHint;
    Load();
    if (vop) {
	return vop->GetExtent();
    }
    return 0;
}
    
int ReferenceMark::Base()
{
    if (sizeHint != gPoint_1) 
	return sizeHint.y;
    Load();
    if (vop)
	return vop->Base();
    return 0;
}

void ReferenceMark::SetContainer(VObject *vop)    
{
    container= vop;
}

void ReferenceMark::Draw(Point at, Rectangle clip, Rectangle, bool border)
{
    Load();
    if (vop) {
	vop->SetOrigin(at);
	vop->DrawAll(clip, FALSE);
	if (border) {
	    GrSetPenInk(ePatGrey50);
	    GrStrokeRect(vop->ContentRect());
	}
    }
}

void ReferenceMark::SendDown(int what, int p, void *arg)
{
    if (vop)
	vop->SendDown(what, p, arg);
}

bool ReferenceMark::WantsInput(Point lp)
{
    if (vop)
	 return vop->ContainsPoint(lp);
    return FALSE;
}

class Command *ReferenceMark::Input(Point lp, Token &t, class Clipper *cl)
{
    if (vop)
	return vop->Input(lp, t, cl);
    return 0;
}

void ReferenceMark::FreeAll()
{
    if (vop)
	vop->FreeAll();
    SafeDelete(vop); 
}

OStream &ReferenceMark::PrintOn(OStream &s) 
{
    VisualMark::PrintOn(s);  
    s << sizeHint;  
    return s.PrintString(filename);
}

IStream &ReferenceMark::ReadFrom(IStream &s) 
{
    VisualMark::ReadFrom(s);
    SafeDelete(filename);
    vop= 0;
    s >> sizeHint;
    s.ReadString(&filename);
    VerifyFileName();
    return s;
}

void ReferenceMark::CollectParts(Collection *col)
{
    VisualMark::CollectParts(col);
    if (vop)
	col->Add(vop);
}

char *ReferenceMark::FileName()
{
    return filename;
}    

void ReferenceMark::SetFileName(char *fn)
{
    strreplace(&filename, fn);
}

void ReferenceMark::SetExtentHint(Point p)
{
    sizeHint=p;
}

void ReferenceMark::Load()
{
    if (vop)    
       return;
    if(verify) {
	VerifyFileName();
	verify=FALSE;
    }
    FileData *fd= new FileData(filename,FALSE); //don't look deep
    if (!fd->IsReadable()) {
	vop= MakePlaceHolder();
    } else {
	vop=Convert(fd);
	if(vop==0) {
	    SafeDelete(fd);
	    fd= new FileData(filename,TRUE); // look deep
	    vop=Convert(fd);
	}
    }
    
    if (vop == 0) {
	//ShowAlert(eAlertNote, "Cannot convert @B%s@B (%s)\n", fd->Name(), fd->Type().AsString());
	printf("%s??%s??\n",fd->FullName(),fd->Type().AsString());
	vop= MakePlaceHolder();
    }   
    if (vop) {
       vop->SetContainer(container);
	if (container && container->IsOpen())
	    vop->Open(); 
    }
    vop->CalcExtent();
//    if (sizeHint != gPoint_1)
//	vop->SetExtent(sizeHint);
    SafeDelete(fd);
}

VObject* ReferenceMark::Convert(Data*fd)
{
    Object *op;
    if ((op= fd->AsObject(Meta(Picture))))
	return new PictureItem((Picture*)op);
    
    if ((op= fd->AsObject(Meta(Bitmap))))
	return new ImageItem((Bitmap*)op);

    if ((op= fd->AsObject(Meta(VObject))))
	return (VObject*)op;
    return 0;
}

VObject* ReferenceMark::MakePlaceHolder()
{
    Point p(100);
    if (sizeHint != gPoint_1)
	p=sizeHint;
   
    return new Filler(p,ePatGrey25);
}

void ReferenceMark::SetupDocPath()
{
    if (docPath != 0)
	return;
    docPath= new PathLookup(Env::GetValue("CodeText.PicturePath", "."));
    char *etdir= CLib::Getenv("ET_DIR");
    if (etdir)
	docPath->Add(gSystem->ConcatPathNames(etdir,"doc",0));

}

void ReferenceMark::VerifyFileName()
{
    FileData *fd= new FileData(filename);
    if (!fd->IsReadable()) {
	SetupDocPath();
	char docPathName[1024];
	    SafeDelete(fd);
	if (!docPath->Lookup(gSystem->BaseName(filename), docPathName)) {
	    char *newName= AskUser();
	    if (newName == 0) 
		return;
		SetFileName(newName);
	    } else {
		SetFileName(docPathName);
	    }
    }   
    SafeDelete(fd);
}

char *ReferenceMark::AskUser()
{
    if (fileDialog == 0)
	fileDialog= new FileDialog();
    char *title= form("Show the Location of %s",gSystem->BaseName(filename));
    int ret= fileDialog->ShowInWindow(eFDRead, gWindow, 0,  title);
    if (ret == cIdOk) 
	    return fileDialog->FileName();
    return 0;
}
