#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma implementation
#endif
    
#include "TextView.h"

#include "TextCmd.h"
#include "CheapText.h"
#include "GapText.h"
#include "CmdNo.h"
#include "ETString.h"
#include "Window.h"
#include "Menu.h"
#include "Error.h"
#include "RegularExp.h"
#include "ClipBoard.h"
#include "TextFormatter.h"
#include "Env.h"
#include "Data.h"
#include "ETMath.h"
#include "Class.h"

#include "SortedOList.h"
#include "TextItem.h"
#include "MenuBar.h"

const int cCaretHt = 4,
	  cCaretWd = 7;

static u_short CaretBits[]= {
#   include "images/Caret.image"
};

SmartBitmap CaretImage(Point(cCaretWd, cCaretHt), CaretBits);

static int InputMap[] = {
0,       /* 0x00 */
0,       /* 0x01 */
0,       /* 0x02 */
0,       /* 0x03 */
0,       /* 0x04 */
0,       /* 0x05 */
0,       /* 0x06 */
0,       /* 0x07 */
0,       /* 0x08 */
'\t',    /* 0x09 */
'\r',    /* 0x0a */
0,       /* 0x0b */
0x0c,    /* 0x0c */
'\n',    /* 0x0d */
0,       /* 0x0e */
0,       /* 0x0f */
0,       /* 0x10 */
0,       /* 0x11 */
0,       /* 0x12 */
0,       /* 0x13 */
0,       /* 0x14 */
0,       /* 0x15 */
0,       /* 0x16 */
0,       /* 0x17 */
0,       /* 0x18 */
0,       /* 0x19 */
0,       /* 0x1a */
0,       /* 0x1b */
0,       /* 0x1c */
0,       /* 0x1d */
0,       /* 0x1e */
0,       /* 0x1f */
};

static int *inputmap= InputMap;

static Ink *caretColor;

void SwapSelPoints(SelPoint &p1, SelPoint &p2)
{
    SelPoint tmp;

    tmp= p1; p1= p2; p2= tmp;
}

//----- TextView Methods -------------------------------------------------------
 
NewMetaImpl(TextView, StaticTextView, (TB(updateSelection), TB(inTextSelector), 
    TB(enabled), TP(stopChars), TP(typing), TP(scratchText), 
    T(start.ch), T(start.line), T(start.viewp), T(end.ch), T(end.line), 
    T(end.viewp), TP(tvf)));

Controler *gTextViewControler; // Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
 
TextView::TextView(
    EvtHandler *eh, Rectangle r, Text *t, 
    bool w, TextViewFlags f, Point b, 
    TViewAlign ta, int id
) : StaticTextView(eh, r, t, w, f, b, ta, id) 
{
    Init();
}

void TextView::InitNew()
{
    View::InitNew();
    inTextSelector= FALSE;
}  

void TextView::Init()
{
    SetControler(gTextViewControler); // Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
    PrivSetSelection(0, 0, FALSE);
    tvf=0;
    typing= 0;
    stopChars= 0;
    selectionMark= 0;
    updateSelection= TRUE;
    text->AddObserver(this);
    scratchText= text->MakeScratchText(0, cMaxBatchedIns);
    active= TRUE;
    inTextSelector= FALSE;
    enabled= FALSE;
    if (Env::GetValue("TextView.DragAndDrop", TRUE))
	SetFlag(eTextViewDragAndDrop);
    SetFlag(eVObjKbdFocus);
}

TextView::~TextView()
{
    if (text)
	text->RemoveObserver(this);
    SafeUnref(tvf);
    SafeDelete(stopChars);
    SafeDelete(scratchText);
}

void TextView::InvalidateSelection()
{
    if (AnySelection())
	InvalidateRange(start.line, start.viewp, end.line, end.viewp);
}

void TextView::Draw(Rectangle r)
{
    if (GrHasColor() && CanShowSelection() && AnySelection())
	Invert(start.line, start.viewp, end.line, end.viewp);
    StaticTextView::Draw(r);
    if (CanShowSelection() && AnySelection()) {
	DrawCaret(start.viewp, start.line, On);
	if (!GrHasColor())
	    Invert(start.line, start.viewp, end.line, end.viewp);
    }
}
    
void TextView::DrawCaret(Point p, int line, HighlightState)
{
    if (! inTextSelector && AnySelection() && Caret()) {
	if (caretColor == 0) {
	    if (GrHasColor()) {
		caretColor= new_RGBColor(255, 0, 0);
		Env::Bind(caretColor, "TextView.CaretColor");
	    } else
		caretColor= gInkXor;
	}
	if(gCaretColor && GrHasColor()) 
	    caretColor=gCaretColor;
	   
	int bh= BaseHeight(line);
	p+= GetInnerOrigin();
	p.x--;
	Rectangle r(p.x-cCaretWd/2, p.y+bh-1, cCaretWd, cCaretHt);
	GrPaintBitMap(r, CaretImage, caretColor);
	GrPaintLine(caretColor, 0, eDefaultCap, p, Point(p.x, r.origin.y));
    }
}

void TextView::SetKeyMap(int *map)
{
    inputmap= map;
}

//---- highlight the characters in the given range -----------------------

void TextView::Invert(int from, Point fp, int to, Point tp)
{
    if (AnySelection() && !Caret()) {
	Rectangle r;
	Ink *c= gHighlightColor;
	
	if ((from == to && fp.x > tp.x) || from > to) { // normalize range
	    SwapRange(from, to);
	    Swap(fp, tp);   
	}
    
	if (from == to) {
	    r= Rectangle(LineToPoint(from)+Point (fp.x,0),
			   Point(tp.x-fp.x, LineHeight(from)));
	    GrPaintRect(r+GetInnerOrigin(), c);
	    return;
	}

	r= Rectangle(LineToPoint(from)+ Point(fp.x,0),
		      Point(GetInnerExtent().x-fp.x, LineHeight(from)));
	GrPaintRect(r+GetInnerOrigin(), c);

	r= Rectangle(LineToPoint(from + 1), Point(GetInnerExtent().x,
		      LineToPoint(to).y - LineToPoint(from + 1).y));
	GrPaintRect(r+GetInnerOrigin(), c);

	r= Rectangle(LineToPoint(to), Point(tp.x, LineHeight(to)));
	GrPaintRect(r+GetInnerOrigin(), c);
    }
}

void TextView::RepairAll()
{
    StaticTextView::RepairAll();
    if (CanShowSelection())
	SetSelection(start.ch, end.ch, FALSE);
}
 
void TextView::ClearSelection(bool redraw)
{
    if (AnySelection()) {
	InvalidateSelection();
	start.ch= end.ch= -1;
	if (redraw)
	    UpdateEvent();
    }
}

void TextView::ShowSelection(bool redraw, bool show)
{
    if (enabled != show) {
	enabled= show;
	InvalidateSelection();
	if (redraw)
	    UpdateEvent();
    }
}

//--- Added by Michael Scharf 4.12.1995
//--- begin
void TextView::RevealRectAt(Rectangle rect, int at)
{
    if(at<0) {
	rect.extent.y=Getclipper()->Height()/(-at);
	RevealRect(rect,rect.extent);
    } else {
	if (at >= 2) {
	    int clipperHeight = Getclipper()->Height();
	    if(clipperHeight > rect.extent.y) {
		rect.origin.y -= (clipperHeight - rect.extent.y)/at;
	    }
	    rect.extent.y = clipperHeight;
	}
	RevealAlign(rect, VObjAlign(eVObjHRight|eVObjVTop));
    }
}

void TextView::RevealLine (int line, bool do_select)
{
    line = Math::Max (line - 1, 0);
    line = LogicalToRealLine (line);
    line = Math::Min (line, NumberOfLines() - 1);
    line = Math::Max (line, 0);
    int start_line = StartLine (line);
    int end_line = EndLine (line);
    if ( do_select )
	SetSelection (start_line, end_line);
    Rectangle r = BoundingRect (start_line, end_line);
    r.origin.x = 0;
    r.origin.y -= r.extent.y;
    r.extent.x = 1;
    r.extent.y *= 3;
    RevealRect (r, r.extent);
}
//--- end

bool TextView::KbdFocus(bool in)
{
    ShowSelection(FALSE, in);
    return TRUE;
}

void TextView::Enable(bool enable, bool redraw)
{
    StaticTextView::Enable(enable, redraw);
    if (! enable)
	HideSelection(redraw);
} 

void TextView::PrivSetSelection(SelPoint s, SelPoint e, bool redraw)
{
    InvalidateSelection();          // invalidate old selection
    start= s;
    end= e;
    InvalidateSelection();          // invalidate new selection
    if (redraw)
	UpdateEvent();
    Send(GetId(), cPartSelectionChanged, this);
}

void TextView::PrivSetSelection(int s, int e, bool redraw)
{
    SelPoint sp, ep;

    if (text == 0)
	return;
	
    int end= text->End();
    sp.ch= Math::Range(0, end, s);
    ep.ch= Math::Range(0, end, e);
    sp.ch= CharToPoint(sp.ch, &sp.line, &sp.viewp);
    if (sp.ch != ep.ch)
	ep.ch= CharToPoint(ep.ch, &ep.line, &ep.viewp);
    else
	ep= sp;
    PrivSetSelection(sp, ep, redraw);
}

void TextView::SetSelection(int s, int e, bool redraw)
{
    DoneTyping();
    PrivSetSelection(s, e, FALSE);
    ShowSelection(redraw);
}

void TextView::SelectionAsString(byte *buf, int ma)
{
    GetText()->CopyInStr(buf, ma, start.ch, end.ch);
}

Text *TextView::SelectionAsText()
{
    return text->Save(start.ch, end.ch);
}

void TextView::SetDragAndDrop(bool on)
{
    SetFlag(eTextViewDragAndDrop, on);
}

Rectangle TextView::BoundingRect(int from, int to)
{
    Point sp, ep;
    int sl, el;
    
    if(from == start.ch) {
	sp= start.viewp;
	sl= start.line;
    } else 
	CharToPoint(from, &sl, &sp);
    if (to == end.ch) {
	ep= end.viewp;
	el= end.line;
    } else
	CharToPoint(to, &el, &ep);

    Point rp(ep.x+1, ep.y + LineHeight(el));
    Rectangle r= NormRect(sp, rp);
    r.origin+= GetInnerOrigin();
    return r;
}

void TextView::RevealSelection()
{
    if (AnySelection()) {
	Rectangle r= SelectionRect().Expand(Point(8,0));
	RevealRect(r, r.extent);
    }
}
     
Rectangle TextView::SelectionRect()
{
    if (AnySelection()) 
	return BoundingRect(start.ch, end.ch);
    return gRect0;
}

Command *TextView::DoLeftButtonDownCommand(Point lp, Token t, int clicks)
{
    SelPoint oldStart, oldEnd;
    TextRangeFP rf= CharacterRange;
	
    oldStart= start;
    oldEnd= end;

    if (! Enabled())
	return View::DoLeftButtonDownCommand(lp, t, clicks);
    
    if (clicks >= 3)
	rf= ParagraphRange;
    else if (clicks >= 2) 
	rf= WordRange;

    if (t.Flags == eFlgShiftKey) 
	return new ExtendRangeSelector(this, rf);  
    
    if (clicks == 1 && Writeable()) {
	if (!Caret()) {
	if (TestFlag(eTextViewDragAndDrop)) {
	    SelPoint n;
	    PointToPoint(lp-GetInnerOrigin(), &n.viewp, &n.line, &n.ch);
	    if (n.ch >= oldStart.ch && n.ch < oldEnd.ch)
		return new DragAndDropSelector(this, oldStart.ch, oldEnd.ch, t.Flags == eFlgCntlKey);
	}     
	if (t.Flags == eFlgCntlKey) 
	    return new DragAndDropSelector(this, oldStart.ch, oldEnd.ch, TRUE); 
	}
	if (t.Flags == (eFlgShiftKey|eFlgCntlKey))
	    return new QuickPasteSelector(this, oldStart.ch, oldEnd.ch);
    }
    if (t.Flags == 0)
	return new RangeSelector(this, rf);
    return View::DoLeftButtonDownCommand(lp, t, clicks);
}

void TextView::SetStopChars(char *stops)
{
    strreplace(&stopChars, stops);
}

Command *
TextView::DoKeyCommand (int ch, Token token)
{
  if ( ch == 0x7F || token.Flags == eFlgCntlKey )
    {
      Text * text = GetText();
      int length = text->End();
      bool delete_text = FALSE;

      switch ( ch )
	{
	  case 'a':
	    return DoCursorKeyCommand (eCrsLeft, token);

	  case 'e':
	    return DoCursorKeyCommand (eCrsRight, token);

	  case 'd':	// delete selection or character at cursor
	  case 0x7F:	// Delete key
	    if ( start.ch == end.ch && end.ch < length )
	      end.ch++;
	    // Is this a DOS CRLF?
	    if(end.ch<length && end.ch >0 && (*text)[end.ch-1]=='\r' && (*text)[end.ch]=='\n')
		end.ch++;
	    delete_text = TRUE;
	    break;

	  case 'k':	// delete from end of selection to end of line
	    start.ch = end.ch;
	    while ( end.ch < length && (*text)[end.ch] != '\n' )
	      end.ch++;
	    delete_text = TRUE;
	    break;

	  case '\b':	// delete selection or previous word
	    if ( start.ch == end.ch )
	      {
		while ( start.ch > 0 && !Isinword ((*text)[start.ch - 1]) )
		  start.ch--;
		while ( start.ch > 0 && Isinword ((*text)[start.ch - 1]) )
		  start.ch--;
	      }
	    delete_text = TRUE;
	    break;
	}

      if ( delete_text && start.ch != end.ch )
	{
	  if ( Writeable() )
	    {
	      SetSelection (start.ch, end.ch, /* bool redraw = */ FALSE);
	      return new CutCopyCommand (this, cCUT);
	    }
	  else
	    {
	      gWindow->Bell();
	    }
	}

      return gNoChanges;
    }

  bool newCmd;
  bool delSelection;       // delete the selection only

  if ( ! Writeable() )
    {
      gWindow->Bell();
      return gNoChanges;
    }

  if ( ch == '\r' )
    ch = '\n';

  newCmd = (typing == 0); // start new typing sequence?
  if ( newCmd )
    typing = MakeTypingCommand (cTYPEING, "Typing");

  delSelection = (newCmd && !Caret());
  if ( ch != '\b' || delSelection )
    {     
      if ( ch != '\b' && ch != '\0' )
	{
	  typing->AddChar();
	  scratchText->Append (ch);
	}
      Paste (scratchText);
      scratchText->Empty();
    }
  else
    {	// backspace
      // Is this a DOS CRLF?
      if(end.ch >= 2 && (*text)[end.ch-1]=='\n' && (*text)[end.ch-2]=='\r') {
	  typing->DelChar();
	  DelChar (1);
      }
      typing->DelChar();
      DelChar (1);
    }
  RevealSelection();

  return newCmd ? typing : gNoChanges;
}

Command *
TextView::DoCursorKeyCommand (EvtCursorDir cd, Token t)
{
  Text * text = GetText();
  int length = text->End();
  int charNo = -1;
  int line;
  Point p;

  if ( ! Enabled() )
    return gNoChanges;

  switch ( cd )
    {
      case eCrsLeft:
	charNo = start.ch;
	if ( t.Flags == eFlgCntlKey )		// beginning of line
	  {
	    while ( charNo > 0 && (*text)[charNo - 1] != '\n' )
	      charNo--;
	  }
	else if ( t.Flags == eFlgShiftKey )	// move back one word
	  {
	    while ( charNo > 0 && !Isinword ((*text)[charNo - 1]) )
	      charNo--;
	    while ( charNo > 0 && Isinword ((*text)[charNo - 1]) )
	      charNo--;
	  }
	else if ( t.Flags == 0 )		// one character to the left
	  charNo--;
	break;

      case eCrsRight:
	charNo = end.ch;
	if ( t.Flags == eFlgCntlKey )		// end of line
	  {
	    while ( charNo < length && (*text)[charNo] != '\n' )
	      charNo++;
	  }
	else if ( t.Flags == eFlgShiftKey )	// move forward one word
	  {
	    while ( charNo < length && !Isinword ((*text)[charNo]) )
	      charNo++;
	    while ( charNo < length && Isinword ((*text)[charNo]) )
	      charNo++;
	  }
	else if ( t.Flags == 0 )		// one character to the right
	  charNo++;
	  // Is this a DOS CRLF?
	  if(charNo < length && charNo > 0 && (*text)[charNo-1]=='\r' && (*text)[charNo]=='\n' )
	    charNo++;
	break;

      case eCrsUp:
	if ( t.Flags == eFlgCntlKey )		// beginning of document
	  {
	    charNo = 0;
	  }
	else if ( t.Flags == eFlgShiftKey )	// top of page
	  {
	    Clipper * cl = Getclipper();
	    int clh = cl->Height();
	    p = Point(1,1) + cl->GetRelOrigin() + cl->GetShift();
	    line = PointToLine(p);
	    if ( line == start.line )		// already at top of page
	      {					// go up one page
		p.y -= clh - clh/10;
		p.y = Math::Max (p.y, 0);
		line = PointToLine(p);
	      }
	    charNo = StartLine (line);
	  }
	else if ( t.Flags == 0 )		// up one line
	  {
	    charNo = CursorPos (start.ch, start.line, cd, p);
	  }
	break;

      case eCrsDown:
	if ( t.Flags == eFlgCntlKey )		// end of document
	  {
	    charNo = length;
	  }
	else if ( t.Flags == eFlgShiftKey )	// bottom of page
	  {
	    Clipper * cl = Getclipper();
	    int clh = cl->Height();
	    p = Point(1, clh - 1) + cl->GetRelOrigin() + cl->GetShift();
	    p.y = Math::Min (p.y, Height());
	    line = PointToLine(p) - 1;
	    if ( line < Lines() && line == end.line )
	      {					// already at bottom of page
		p.y += clh - clh/10;		// go down one page
		p.y = Math::Min (p.y, Height());
		line = PointToLine(p);
	      }
	    line = Math::Min (line, Lines() - 1);
	    charNo = StartLine (line);
	  }
	else if ( t.Flags == 0 )		// down one line
	  {
	    charNo = CursorPos (end.ch, end.line, cd, p);
	  }
	break;
    }

  if ( charNo >= 0 )
    {
      PrivSetSelection (charNo, charNo);
      DoneTyping();
      RevealSelection();
    }

  return gNoChanges;
}

int TextView::CursorPos(int at, int line, EvtCursorDir d, Point p)
{
    int charNo;
    Point basePoint, screenPoint;

    CharToPoint(at, &line, &screenPoint);
    if (d == eCrsDown)
	line= Math::Min(nLines-1, line+1);
    else
	line= Math::Max(0, line-1);        
    basePoint= LineToPoint(line, TRUE) + Point(screenPoint.x, 0);
    PointToPoint(basePoint, &p, &line, &charNo);
    return charNo;
}

GrCursor TextView::GetCursor(Point lp)
{
    if (TestFlag(eTextViewDragAndDrop)) {
	SelPoint n;
	PointToPoint(lp-GetInnerOrigin(), &n.viewp, &n.line, &n.ch);
	if (n.ch >= start.ch && n.ch < end.ch) 
	    return eCrsBoldArrow;
    }
    return eCrsIBeam;
}

void TextView::DoneTyping()
{
    typing= 0;
    if(text)
	text->ResetCurrentCharStyle();
}

void TextView::TypingDeleted(TypingCommand *t)
{
    if (t == typing)
	typing= 0;
}

TypingCommand *TextView::MakeTypingCommand(int cmdNo, char *name)
{
    return new TypingCommand(this, cmdNo, name);
}

bool TextView::DeleteRequest(int,int)
{
    return !IsReadOnly();
}

void TextView::Cut()
{
    if (!Writeable())
	return; 
    if (DeleteRequest(start.ch, end.ch) && AnySelection()) {
	updateSelection= FALSE;
	text->Cut(start.ch, end.ch);
	updateSelection= TRUE;
	PrivSetSelection(start.ch, start.ch, FALSE);
    }
}

void TextView::Copy(Text *t)
{
    text->Copy(t, start.ch, end.ch);
}

void TextView::Paste(Text *t)
{ 
    if (!Writeable())
	return;
    if (AnySelection() && DeleteRequest(start.ch, end.ch)) {
	updateSelection= FALSE;
	text->Paste(t, start.ch, end.ch);
	updateSelection= TRUE;
	PrivSetSelection(start.ch+t->End(), start.ch+t->End(), FALSE);
    }
}

Command *TextView::InsertText(Text *t)
{
    if (!Writeable())
	return gNoChanges;

    bool newCmd= (typing == 0); // start new typing sequence?

    if (newCmd)
	typing= MakeTypingCommand(cTYPEING, "Typing");

    typing->AddChar(t->End());
    Paste(t);
    if (newCmd)
	return typing;
    return gNoChanges;
}

Command *TextView::InsertString(byte *str, int len)
{
    Text *t= text->MakeScratchText(str, len);
    Command *cmd= InsertText(t);
    SafeDelete(t);
    return cmd;
}

void TextView::DelChar(int n)
{
    int newStart= Math::Max(0, start.ch-n);
    if (newStart == start.ch)
	return;
    if (newStart == 0 && start.ch == 0)
	return;
    if (DeleteRequest(newStart, start.ch)) {
	updateSelection= FALSE;   
	text->Cut(newStart, start.ch);
	updateSelection= TRUE;
	PrivSetSelection(newStart, newStart, FALSE);       
    }
}

Text *TextView::SetText(Text *t, bool scroll)
{
    Text *old= text;
    SetFlag(eTextViewModified);
    ForceRedraw();
    old->RemoveObserver(this);
    PrivSetSelection(0, 0, FALSE); // reset old selection
    text= t;    
    SafeDelete(scratchText);
    scratchText= text->MakeScratchText(0, cMaxBatchedIns);
    text->AddObserver(this);
    RepairAll();
    if (scroll == cRevealTop)
	Scroll(cPartScrollAbs, gPoint0, FALSE);
    Send(GetId(), cPartReplacedText, 0);
    PrivSetSelection(0, 0, FALSE); // set new selection
    typing= 0;
    return old;
}       

void TextView::SetString(byte *str, int len)
{
    ForceRedraw();
    PrivSetSelection(0, 0, FALSE);
    text->ReplaceWithStr(str, len);
    RepairAll();
    Send(GetId(), cPartReplacedText, 0);
    DoneTyping();
    PrivSetSelection(0, 0, FALSE);
    typing= 0;
}

void TextView::SetReadOnly(bool m)
{
    SetFlag(eTextViewReadOnly, m);
}

bool TextView::IsReadOnly()
{ 
    return TestFlag(eTextViewReadOnly); 
}

void TextView::DoObserve(int, int part, void *what, Object *op)
{
    if (op == text && part == cPartSenderDied) {
	text= 0;
	return;    
    }
	
    if (op == text) {
	TextChanges tc= *(TextChanges*)what;
	switch (part) {
	case eTextWillDelete:
	case eTextWillReplace:
	    if(selectionMark)
		Error("DoObserve","Should not occur!");
	    if(AnySelection() && updateSelection) {
		// let's put a mark into this View to keep track of the 
		// selection...
		selectionMark=new Mark(start.ch,end.ch-start.ch);
		GetText()->AddMark(selectionMark);
	    }
	    return;
	case eTextChangedRange:
	    SetFlag(eTextViewModified);
	    marks->RangeChanged(tc.from, tc.to - tc.from);
	    Repair(tc.from, tc.to, FALSE);
//	    if (updateSelection && AnySelection())
//		PrivSetSelection(tc.from, tc.to, TRUE);
	    break;

	case eTextDeleted:
	case eTextReplaced:
	    SetFlag(eTextViewModified);
	    marks->Replace(tc.from, tc.to, tc.size);
	    Repair(tc.from, tc.from+tc.size, FALSE);
	    if (selectionMark) {
		// did the selection move?
		if(start.ch!=selectionMark->Pos() || end.ch!=selectionMark->End())
			PrivSetSelection(selectionMark->Pos(),selectionMark->End(),TRUE);

		// cleanup
		GetText()->RemoveMark(selectionMark);
		SafeDelete(selectionMark);
	    }
	    break;
	default:
	    return;
	}
	if (tc.pagination == cPaginationChange) 
	    ForceRedraw();
	Send(GetId(), cPartChangedText, this);    // notify observers of the textview   
    }
}

bool TextView::PrintOnWhenObserving(Object *from)
{
    return GetText() != from;
}

IStream& TextView::ReadFrom(IStream &s)
{
    StaticTextView::ReadFrom(s);
    enabled= TRUE;
    GetText()->AddObserver(this);
    scratchText= text->MakeScratchText(0, cMaxBatchedIns);
    return s;
}

void TextView::NormSelection()
{
    if (start.ch > end.ch) 
	SwapSelPoints(start, end);
}

//---- menu related commands -----------------------------------------------

bool TextView::HasSelection()
{
    return AnySelection() && !Caret();
}

Command *TextView::PasteData(Data *data)
{
    if (!IsReadOnly())
    {
	Text *t;
	if ( (t= (Text*) data->AsObject(GetText()->IsA())) ||
	     (t= (Text*) data->AsObject(Meta(CheapText)))  ||
	     (t= (Text*) data->AsObject(Meta(GapText))) )
	    return new PasteCommand(this, t);
    }
    return gNoChanges;
}

Menu *TextView::MakeMenu(int menuId)
{
    int font, size= gSysFont->Size();
    GrFont fid= (GrFont)gSysFont->Fid();
    char *fontname;
    Menu *m = NULL;
    
    switch (menuId) {
    case cFONTMENU:
	m= new Menu(cFONTMENU, "Fonts", FALSE, TRUE, 0, 1, new SortedObjList);
	for (font= 0; fontname= gFontManager->IdToName((GrFont)font); font++)
	    m->AppendItem(fontname, cFIRSTFONT+font);
	break;
		
    case cSTYLEMENU:
	m= new Menu(cSTYLEMENU, "Styles", FALSE);
	m->Append(new MenuButtonItem(cFIRSTFACE+eFacePlain, 
	    new TextItem("Plain", new_Font(fid, size, eFacePlain))));
	m->Append(new MenuButtonItem(cFIRSTFACE+eFaceBold,
	    new TextItem("Bold", new_Font(fid, size, eFaceBold)), "B"));
	m->Append(new MenuButtonItem(cFIRSTFACE+eFaceItalic,
	    new TextItem("Italic", new_Font(fid, size, eFaceItalic)), "I"));
	m->Append(new MenuButtonItem(cFIRSTFACE+eFaceUnderline,
	    new TextItem("Underline", new_Font(fid, size, eFaceUnderline)), "U"));
	m->Append(new MenuButtonItem(cFIRSTFACE+eFaceOutline,
	    new TextItem("Outline", new_Font(fid, size, eFaceOutline))));
	m->Append(new MenuButtonItem(cFIRSTFACE+eFaceShadow,
	    new TextItem("Shadow", new_Font(fid, size, eFaceShadow))));
	break;
	
    case cSIZEMENU:
	{
	    m= new Menu(cSIZEMENU, "Sizes", FALSE);
	    for (int sz= 9; sz <= 24; sz++)
		m->AppendItem(form("%d  Points", sz), cFIRSTSIZE+sz); 
	}
	break;
	
    case cFORMATMENU:
	m= new Menu(cFORMATMENU, "Format", FALSE);
	m->Append(new MenuButtonItem(cFIRSTADJUST,   "Align Left"));
	m->Append(new MenuButtonItem(cFIRSTADJUST+1, "Align Right"));
	m->Append(new MenuButtonItem(cFIRSTADJUST+2, "Align Center"));
	m->Append(new MenuButtonItem(cFIRSTADJUST+3, "Justify"));
	m->AppendItems("-", 0);
	m->Append(new MenuButtonItem(cFIRSTSPACING+0, "Single Spacing"));
	m->Append(new MenuButtonItem(cFIRSTSPACING+18, "1-1/2 Spacing"));
	m->Append(new MenuButtonItem(cFIRSTSPACING+24, "Double Spacing"));
	break;
	
    }
    return m;
}

void TextView::DoSetupMenu(Menu *m)
{    
    View::DoSetupMenu(m);
    switch (m->GetId()) {
    case cEDITMENU:
	SetupEditMenu(m);
	break;
    case cSIZEMENU:
	SetupSizeMenu(m);
	break;
    case cFONTMENU:
	SetupFontMenu(m);
	break;
    case cFORMATMENU:
	SetupFormatMenu(m);
	break;
    case cSTYLEMENU:
	SetupCharStyleMenu(m);
	break;
    case cVIEWMENU:
	SetupViewMenu(m);
	break;
    }
}

void TextView::SetupEditMenu(Menu *m)
{
    if (!IsReadOnly() && !Caret()) 
	m->EnableItems(cCUT, cIMPORT, 0);
    if (AnySelection())
	m->EnableItems(cSELECTALL, cFIND, 0);
}

void TextView::SetupSizeMenu(Menu *m)
{
    Font *fp= GetCharStyle()->GetFont();
    m->CheckItem(cFIRSTSIZE+fp->Size());
}

void TextView::SetupFontMenu(Menu *m)
{
    Font *fp= GetCharStyle()->GetFont();
    m->CheckItem(gFontManager->NameToId(fp->Name())+cFIRSTFONT);
}

void TextView::SetupViewMenu(Menu *m)
{
    m->EnableItem(cSHOWINVIS);
    m->CheckItem(cSHOWINVIS, GetShowInvis(), cItemStateManyOf);
}

void TextView::SetupCharStyleMenu(Menu *m)
{
    GrFace face= GetCharStyle()->GetFont()->Face();
    m->CheckItem(cFIRSTFACE+eFacePlain, face == eFacePlain);
    m->CheckItem(cFIRSTFACE+eFaceBold, (face&eFaceBold) == eFaceBold, cItemStateManyOf);
    m->CheckItem(cFIRSTFACE+eFaceItalic, (face&eFaceItalic) == eFaceItalic, cItemStateManyOf);
    m->CheckItem(cFIRSTFACE+eFaceUnderline, (face&eFaceUnderline) == eFaceUnderline, cItemStateManyOf);
    m->CheckItem(cFIRSTFACE+eFaceOutline, (face&eFaceOutline) == eFaceOutline, cItemStateManyOf);
    m->CheckItem(cFIRSTFACE+eFaceShadow, (face&eFaceShadow) == eFaceShadow, cItemStateManyOf);        
}
      
void TextView::SetupFormatMenu(Menu *m)
{
    int from, to;
    GetSelection(&from, &to);
    ParaStyle *ps= text->GetParaStyle(from);
    m->CheckItem(ps->GetProperty(eTxtPAlign)+cFIRSTADJUST);
    m->CheckItem(ps->GetProperty(eTxtPSpacing)+cFIRSTSPACING);
} 

Command *SetParaProp(TextView *view, int cmd, TxtParaProp p, int value)
{
    ParaDesc pd;
    pd.SetProperty(p, value);
    return 
	new ParaStyleCommand(
	    view,
	    cmd,
	    "Change Formatting",
	    p,
	    pd
	);
}

Command *TextView::DoMenuCommand(int cmd)
{
    if (cmd >= cFIRSTSIZE && cmd <= cLASTSIZE) 
	return new CharStyleCommand(
		    this, 
		    cmd, 
		    "Change Size", 
		    eTxtPSize, 
		    CharStyleSpec(
			    eFontDefault, 
			    eFacePlain,
			    (GrFace)(cmd-cFIRSTSIZE)
		    )
	);

    if (cmd >= cFIRSTFONT && cmd <= cLASTFONT) 
	return new CharStyleCommand(
		    this, 
		    cmd, 
		    "Change Font", 
		    eTxTPFont, 
		    CharStyleSpec(
			    (GrFont)(cmd-cFIRSTFONT),
			    eFacePlain,
			    0
		    )
	);
	
    if (cmd >= cFIRSTFACE && cmd <= cLASTFACE) {
	return new CharStyleCommand(
		    this, 
		    cmd, 
		    "Change Face",
		    eTxTPFace, 
		    CharStyleSpec(
			    eFontDefault,
			    (GrFace)(cmd-cFIRSTFACE),
			    0
		    )
	);
    }

    if (cmd >= cFIRSTSPACING && cmd <= cLASTSPACING) {
	ParaDesc pd;
	pd.SetProperty(eTxtPSpacing, cmd-cFIRSTSPACING);
	return 
	    new ParaStyleCommand(
		this,
		cmd,
		"Change Formatting",
		eTxtPSpacing,
		pd
	    );
    }

    /*
    if (cmd >= cFIRSTCOLOR && cmd <= cLASTCOLOR) 
	return new CharStyleCommand(
		    this, 
		    cmd, 
		    "Change Color",
		    eTxtPInk, 
		    CharStyleSpec(
			    eFontDefault,
			    eFacePlain,
			    0,
			    palette[cmd-cFIRSTCOLOR]
		    )
	);
    */
    
    if (cmd >= cFIRSTADJUST && cmd <= cLASTADJUST) 
	return SetParaProp(this, cmd, eTxtPAlign, cmd-cFIRSTADJUST); 

    switch (cmd) {
	case cCUT:
	case cCOPY:
	    if (HasSelection())
	    {
		gClipBoard->SelectionToClipboard(SelectionAsText(), FALSE);
		return new CutCopyCommand(this, cmd);
	    }
	    break;

	case cSELECTALL:
	    SelectAll();
	    break;

	case cSHOWINVIS:
	    ShowInvisibles(!GetShowInvis());
	    break;

	default:
	    return View::DoMenuCommand(cmd);
    }

    return gNoChanges;
}

bool TextView::SelectRegExpr(RegularExp *rex, bool dir)
{
    int selStart, selEnd, pos, matched;

    GetSelection(&selStart, &selEnd);
    if (dir == cSearchForward)
	pos= GetText()->Search(rex, &matched, selEnd);
    else
	pos= GetText()->Search(rex, &matched, Math::Max(0, selStart-1), cMaxInt, FALSE);

    if (pos != -1 && matched > 0) {
	PrivSetSelection(pos, pos+matched);
	DoneTyping();
	return TRUE;
    }
    return FALSE;
}

void TextView::Home()
{
    PrivSetSelection(0, 0);
    DoneTyping();
}

void TextView::Bottom()
{
    PrivSetSelection(cMaxInt, cMaxInt);
    DoneTyping();
}

void TextView::SelectAll(bool redraw)
{
    PrivSetSelection(0, text->End(), redraw);
    DoneTyping();
}

CharStyle *TextView::GetCharStyle()
{
    int from, to;
    GetSelection(&from, &to);
    if (text->GetCurrentCharStyle())
	return text->GetCurrentCharStyle();
    else if (text->IsParaStart(from))
	from++;
    if (from == to)
	from--;
    return text->GetCharStyle(Math::Max(0, from));
}
 
//--- TextViewProxyFactory support, see comments in SmartTextView.h
static TextViewProxyFactory* gTextViewProxyFactory;
 
TextViewProxyFactory *
TextView::MakeTextViewProxyFactory()
{
    if(gTextViewProxyFactory==0)
	AssignRef(&gTextViewProxyFactory,new TextViewProxyFactory);
    return gTextViewProxyFactory;
}
 
TextViewProxyFactory *
TextView::GetTextViewProxyFactory()
{
    if(!tvf)
	SetTextViewProxyFactory(MakeTextViewProxyFactory());
    return tvf;
 
}
 
void
TextView::SetTextViewProxyFactory(TextViewProxyFactory*ntvf)
{
    AssignRef(&tvf,ntvf);
}
 
TextViewProxy*
TextView::MakeProxy()
{
    return GetTextViewProxyFactory()->MakeProxy(this);
}


