#ifdef __GNUG__
#pragma implementation
#endif

#include "TypeMatcher.h"

#include "Class.h"
#include "ETString.h"
#include "Data.h"
#include "CLib.h"
#include "OrdColl.h"
#include "Bitmap.h"
#include "System.h"

const char *cMagic= "@!OIO";
const int cContentsSize= 1024;

//---- bitmap images for file icons --------------------------------------------

/*
#include "images/directory.xpm3"
#include "images/exec.xpm3"
#include "images/ascii.xpm3"
#include "images/file.xpm3"
#include "images/image.xpm3"

SmartBitmap
	    gDirectoryIcon(directory),
	    gExecIcon(exec),
	    gAsciiIcon(ascii),
	    gFileIcon(file),
	    gImageIcon(image);
*/

static u_short DirectoryBits[]= {
#   include "images/directory.image"
};
static u_short ExecBits[]= {
#   include "images/exec.image"
};
static u_short AsciiBits[]= {
#   include "images/ascii.image"
};
static u_short FileBits[]= {
#   include "images/file.image"
};

SmartBitmap
	    gDirectoryIcon(gPoint16, DirectoryBits),
	    gExecIcon(gPoint16, ExecBits),
	    gAsciiIcon(gPoint16, AsciiBits),
	    gFileIcon(gPoint16, FileBits),
	    gImageIcon(gPoint16, AsciiBits);

//---- TypeMatcher -------------------------------------------------------------

NewMetaImpl(TypeMatcher,Object, (TP(type), TS(extension), TS(tag), TB(ascii),
							    TP(matchers)));

OrdCollection *TypeMatcher::matchers;
bool TypeMatcher::sorted;

void TypeMatcher::AddMatcher(TypeMatcher *ftm)
{
    if (ftm) {
	if (matchers == 0)
	    matchers= new OrdCollection;
	matchers->Add(ftm);
	sorted= FALSE;
    }
}
    
TypeMatcher::TypeMatcher(const Symbol &t, char *ext, char *tg, bool ia, int prio)
{
    type= (Symbol*) &t;
    extension= ext ? strsave(ext) : 0;  
    tag= tg ? strsave(tg) : 0;
    ascii= ia;
    priority= prio;
    AddMatcher(this);
}

TypeMatcher::~TypeMatcher()
{
    SafeDelete(extension);
    SafeDelete(tag);
    if (matchers) {
	matchers->RemovePtr(this);
	if (matchers->Size() == 0)
	    SafeDelete(matchers);
    }
}

Bitmap *TypeMatcher::GetBitmap()
{
    if (ascii)
	return gAsciiIcon;
    return gFileIcon;
}

bool TypeMatcher::MatchPathName(Data *data, char *pathname, int)
{
    u_long flags= 0;

    if (gSystem->GetPathInfo(pathname, 0, 0, &flags, 0) == 0)
      if (flags & ePiDir)
        return FALSE;

    char *s= strrchr(pathname, '.');
    if (s)
      {
      s++;
      if (*s)
	return MatchExtension(data, s);
      }
    return FALSE;
}

bool TypeMatcher::MatchExtension(Data*, char *ext)
{
    return extension && (*ext == *extension) && (strcmp(ext, extension) == 0);
}

bool TypeMatcher::MatchNameAndContents(Data *data, char*, char *buf, int len)
{
    return MatchContents(data, buf, len);
}

bool TypeMatcher::MatchContents(Data*, char *bp, int len)
{
    if (tag) {
	int taglen= strlen(tag);
	if (len >= taglen && *bp == *tag && strncmp(bp, tag, taglen) == 0)
	    return TRUE;
    }
    return FALSE;
}

TypeMatcher *TypeMatcher::MatchByName(Data *data)
{
    if (!sorted) {
	matchers->Sort();
	sorted= TRUE;
    }
    char *bp= data->FullName();
    int len= strlen(bp);
    Iter next(matchers);
    TypeMatcher *m;
    while (m= (TypeMatcher*) next())
	if (m->MatchPathName(data, bp, len))
	    break;
    return m;
}

TypeMatcher *TypeMatcher::MatchByContents(Data *data)
{
    char buf[cContentsSize+1];
    IStream is(data->GetStreamBufForReading(), 0, TRUE);
    int l= is.read(buf, cContentsSize);
    buf[l]= '\0';
    if (!sorted) {
	matchers->Sort();
	sorted= TRUE;
    }
    Iter next(matchers);
    TypeMatcher *m;
    while (m= (TypeMatcher*) next())
	if (m->MatchNameAndContents(data, data->FullName(), buf, l))
	    break;
    return m;
}

int TypeMatcher::Compare(Object *op)
{
    return priority - ((TypeMatcher*)op)->priority;
}

//---- CFileTypeMatcher --------------------------------------------------------

class CFileTypeMatcher: public TypeMatcher {
public:
    CFileTypeMatcher() : TypeMatcher(cDocTypeCPlusPlus, 0, 0, TRUE)
	{ }
    bool MatchExtension(Data *data, char *ext);
    bool MatchNameAndContents(Data *data, char *path, char *buf, int len);
    Bitmap *GetBitmap();
};
 
inline    Bitmap *CFileTypeMatcher::GetBitmap()
	{ return gAsciiIcon; }
	
static CFileTypeMatcher cmatcher;

bool CFileTypeMatcher::MatchExtension(Data*, char *ext)
{
    if (ext[1] == '\0' && strchr("cChH", ext[0]))
	return TRUE;
    if (strcmp(ext, "cp") == 0)
	return TRUE;
    if (strcmp(ext, "cxx") == 0)
	return TRUE;
    return FALSE;
}

bool CFileTypeMatcher::MatchNameAndContents(Data *data, char *pathname, char*, int)
{
    return MatchPathName(data, pathname, strlen(pathname));
}

//---- OFileTypeMatcher --------------------------------------------------------

class OFileTypeMatcher: public TypeMatcher {
public:
    OFileTypeMatcher() : TypeMatcher(cDocTypeOFile, 0, 0, FALSE)
	{ }
    bool MatchExtension(Data *data, char *ext);
    bool MatchNameAndContents(Data *data, char *path, char *buf, int len);
    Bitmap *GetBitmap();
};
 
inline    Bitmap *OFileTypeMatcher::GetBitmap()
	{ return gFileIcon; }
	
static OFileTypeMatcher omatcher;

bool OFileTypeMatcher::MatchExtension(Data*, char *ext)
{
     return (ext[0] == 'o' && ext[1] == '\0');
}

bool OFileTypeMatcher::MatchNameAndContents(Data *data, char *pathname, char*, int)
{
    return MatchPathName(data, pathname, strlen(pathname));
}

//---- DirectoryMatcher --------------------------------------------------------

class DirectoryMatcher: public TypeMatcher {
public:
    DirectoryMatcher() : TypeMatcher(cDocTypeDirectory, 0, 0, FALSE, 50)
	{ }
    bool MatchPathName(Data*, char *pathname, int len);
    Bitmap *GetBitmap();
};
 
inline  Bitmap *DirectoryMatcher::GetBitmap()
	{ return gDirectoryIcon; }
	
static DirectoryMatcher directorymatcher;

bool DirectoryMatcher::MatchPathName(Data*, char *pathname, int)
{
    u_long flags= 0;
    
    if (strcmp(pathname, ".") == 0 || strcmp(pathname, "..") == 0)
	return TRUE;
    
    if (gSystem->GetPathInfo(pathname, 0, 0, &flags, 0) == 0) {
	if (flags & ePiDir)
	    return TRUE;
    }
    return FALSE;
}

//---- ExecMatcher -------------------------------------------------------------

class ExecMatcher: public TypeMatcher {
public:
    ExecMatcher() : TypeMatcher(cDocTypeExec, 0, 0, FALSE, 10001)
	{ }
    bool MatchNameAndContents(Data*, char *pathname, char*, int);
    Bitmap *GetBitmap();
};
inline Bitmap *ExecMatcher::GetBitmap()
	{ return gExecIcon; }
	
static ExecMatcher execmatcher;

bool ExecMatcher::MatchNameAndContents(Data*, char *pathname, char*, int)
{
    u_long flags= 0;
    
    if (gSystem->GetPathInfo(pathname, 0, 0, &flags, 0) == 0) {
	if (flags & ePiExec)
	    return TRUE;
    }
    return FALSE;
}


void RefStandardMatchers()
{
    TypeMatcher *tm;
    tm = &cmatcher;
    tm = &omatcher;
    tm = &directorymatcher;
    tm = &execmatcher;
}
