#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma implementation
#endif

#include "CacheUtils.h"
#include "System.h"
#include "ETMath.h"
#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>

#include "ETString.h"
#include "Object.h"
#include "Class.h"
#include "OrdColl.h"

extern "C" {
	int cached_stat(char *, struct stat *);
#if !defined(OS_LINUX)
	int lstat (const char *, struct stat *);
#endif
}

#define LIST_SIZE 10003

statCacheEntry *cache=0;
bool gCacheValid=FALSE;
unsigned numOfFiles=LIST_SIZE;
u_long collisions=0,accessFromCache=0, accessToStat=0;
bool gPrintDebugMsg=FALSE;
bool gUseCache=FALSE;

statCacheEntry::statCacheEntry():Object()
{
    num=0;	
    listSize=0;
    valid=TRUE;
    next=NULL;
    refCount=0;
    isLink=FALSE;
}

statCacheEntry::~statCacheEntry()
{
    if(num!=0) {
		SafeDelete(name);
		if(num>1) {
			statCacheEntry *tmp=next;
			statCacheEntry *tmp1=tmp;
			while(tmp!=NULL) {
				tmp1=tmp->next;
				SafeDelete( tmp);
				tmp=tmp1;
			}
		}
    }
}

void statCacheEntry::SetName(char *path)
{
    SafeDelete(name);
    name=new char[strlen(path)+1];
    strcpy(name,path);
}

void statCacheEntry::CopyTo(struct stat *statptr, statCacheEntry *tmp)
{
	statptr->st_mode=tmp->st_mode;
	statptr->st_uid=tmp->st_uid;
	statptr->st_gid=tmp->st_gid;
	statptr->st_size=tmp->st_size;
	statptr->st_mtime=tmp->st_mtimer;
	statptr->st_ino=tmp->st_ino;
	statptr->st_dev=tmp->st_dev;
}

void statCacheEntry::CopyFrom(statCacheEntry *tmp,const struct stat statStruct)
{
	tmp->st_mode=statStruct.st_mode;
	tmp->st_uid=statStruct.st_uid;
	tmp->st_gid=statStruct.st_gid;
	tmp->st_size=statStruct.st_size;
	tmp->st_mtimer = statStruct.st_mtime;
	tmp->st_ino=statStruct.st_ino;
	tmp->st_dev=statStruct.st_dev;
}

void statCacheEntry::CopyFrom(statCacheEntry *tmp,statCacheEntry *tmp1)
{
	tmp->st_mode=tmp1->st_mode;
	tmp->st_uid=tmp1->st_uid;
	tmp->st_gid=tmp1->st_gid;
	tmp->st_size=tmp1->st_size;
	tmp->st_mtimer = tmp1->st_mtimer;
	tmp->st_ino=tmp1->st_ino;
	tmp->st_dev=tmp1->st_dev;
}
short statCacheEntry::Add(statCacheEntry *entry, short *pos)
{
	statCacheEntry *save=0;
	short validPos=0;
	short isLnk=entry->isLink;
	
	if(num >= 1) {
		if(!strcmp(name,entry->name)) {
		    CopyFrom(this,entry); // already in list
		    if(!valid){
			refCount=1;
			valid=TRUE;
		    }
		    else refCount++;
		    isLink=isLnk;
		    *pos=0;
		    return 1;
		}
		if(!valid) {
		    save=this;
		    validPos=0;
		}
		// find free entry
		statCacheEntry *tmp=this;
		short index=0;
		while(tmp->next!=NULL){
			tmp=tmp->next;
			index++;

			if(!tmp->valid && save == 0) {
			    save=tmp;
			    validPos=index;
			}
			if(!strcmp(tmp->name,entry->name)) {
				CopyFrom(tmp,entry); // already in list
				if(!tmp->valid){
				    tmp->refCount=1;
				    tmp->valid=TRUE;
				}
				else tmp->refCount++;
				tmp->isLink=isLnk;
				*pos=index;
				return 1;
			}
		}
		if(save !=0) {
			SafeDelete(save->name);
			save->name=new char[strlen(entry->name)+1];
			strcpy(save->name,entry->name);
			CopyFrom(save,entry);
			save->refCount=1;
			save->isLink=isLnk;
			*pos=validPos;
			save->valid=TRUE;
			return 1;
		}
		collisions++;
		num++;
		listSize++;
		tmp->next=entry;
		entry->refCount=1; // valid = TRUE is set in constructor
				    // isLink is already set in entry
		*pos=listSize;
		return 0;
	}
	else {// first entry
		num=1;
		listSize=0;
		name=new char[strlen(entry->name)+1];
		strcpy(name,entry->name);
		CopyFrom(this,entry);		
		refCount=1;
		isLink=isLnk;
		*pos=listSize;
		valid=TRUE;
		return 1;
	}
}

short statCacheEntry::Remove(int pos)
{	
	statCacheEntry *tmp=GetAt(pos);
	if(tmp!=0) {
	    tmp->refCount--;
	    if(tmp->refCount==0)
			tmp->valid=FALSE;
	}
	return 0;
}

short statCacheEntry::Add(const char *pathname, const struct stat statStruct, bool isLnk, short *pos)
{
	statCacheEntry *save=0;
	short validPos=0;
	if(num >= 1) {
		if(!strcmp(name,pathname)) {
		    CopyFrom(this,statStruct); // already in list
		    if(!valid) {
			refCount=1;
			valid=TRUE;
		    }
		    else refCount++;
		    isLink=isLnk;
		    *pos=0;
		    return 1;
		}
		else {
		    if(!valid) {
				save=this;
				validPos=0;
			}
			// find free entry
			statCacheEntry *tmp=this;
			short index=0;
			while(tmp->next!=NULL){
				tmp=tmp->next;
				index++;
				if(!tmp->valid && save ==0 ) {
					save=tmp;
					validPos=index;
				}
				if(!strcmp(tmp->name,pathname)) {
					CopyFrom(tmp,statStruct); // already in list
					if(!tmp->valid) {
					    tmp->refCount=1;
					    tmp->valid=TRUE;
					}
					else tmp->refCount++;
					tmp->isLink=isLnk;
					*pos=index;
					return 1;
				}    
			}
			   // no match found -> use first ! valid
			if(save != 0) {
				SafeDelete(save->name);
				save->name=new char[strlen(pathname)+1];
				strcpy(save->name,pathname);
				CopyFrom(save,statStruct);
				save->refCount=1;
				save->isLink=isLnk;
				*pos=validPos;
				save->valid=TRUE;
				return 1;
			}
			// else add at end
			collisions++;
			num++;
			listSize++;
			statCacheEntry *newEntry=new statCacheEntry;
			newEntry->name=new char[strlen(pathname)+1];
			strcpy(newEntry->name,pathname);
			CopyFrom(newEntry,statStruct);
			newEntry->num=1;
			newEntry->refCount=1;
			newEntry->isLink=isLnk; // valid=TRUE is set in constructor
			tmp->next=newEntry;
			*pos=listSize;
		}
	}
	else { // first entry
		num=1;
		name=new char[strlen(pathname)+1];
		strcpy(name,pathname);
		CopyFrom(this,statStruct);
		listSize=0;
		refCount=1;
		isLink=isLnk;
		*pos=0;
		valid=TRUE;
	}
	return 1;
}

short statCacheEntry::Replace(const char *pathname, const struct stat statStruct, short *pos)
{	

	if(num > 1) {
		if(!strcmp(pathname,name)) {
			if(!valid)
			    return 0;
			CopyFrom(this,statStruct);
			*pos=0;
			return 1;
		}
		else {	
		    int index=0;
		    statCacheEntry *tmp;
			tmp=next;
		    while(tmp!=NULL) {
			index++;
			if(!strcmp(pathname,tmp->name)) {
				if(!tmp->valid)
				    return 0;
				CopyFrom(tmp,statStruct);
				*pos=index;
				return 1;
			}
			tmp=tmp->next;
		    }
		}
	}
	else {
		if(num != 0) {
			if(!strcmp(pathname,name)) {
				if(!valid)
				    return 0;
				CopyFrom(this,statStruct);
				*pos=0;
				return 1;
			}
		}
	}
	return 0;
}

statCacheEntry *statCacheEntry::GetAt(short pos)
{
    if(pos==0 && num != 0 && valid) 
		return this;
    if(num > 1) {
		statCacheEntry *tmp=this;
		short x=0;
		while(x<pos && tmp->next!=NULL) {
			tmp=tmp->next;
			x++;
		}
		if(tmp->valid)
			return tmp;
    }
    return 0;    
}

statCacheEntry *statCacheEntry::ForceGetAt(short pos)
{
    if(pos==0 && num != 0) 
		return this;
    if(num > 1) {
		statCacheEntry *tmp=this;
		int x=0;
		while(x<pos && tmp->next!=NULL) {
			tmp=tmp->next;
			x++;
		}
		return tmp;
    }
    return 0;    
}

void statCacheEntry::Invalidate()
{
    valid=FALSE;
}

statCacheEntry *statCacheEntry::Get(const char *pathname, struct stat *statptr)
{

	if(num > 1) {
		if(!strcmp(pathname,name)) {
			if(!valid)
			    return 0;
			CopyTo(statptr,this);
			return this;
		}
		else {	
			statCacheEntry *tmp;
			tmp=next;
			while(tmp!=NULL) {
				if(!strcmp(pathname,tmp->name)) {
					if(!tmp->valid)
					    return 0;
					CopyTo(statptr,tmp);
					return tmp;
				}
				tmp=tmp->next;
			}
		}
	}
	else {
		if(num != 0) {
			if(!strcmp(pathname,name)) {
				if(!valid)
				    return 0;
				CopyTo(statptr,this);
				return this;
			}
		}
	}
	return 0;
}
 
statCacheEntry *statCacheEntry::ForceGet(const char *pathname)
{
 
	if(num > 1) {
		if(!strcmp(pathname,name)) {
			return this;
		}
		else {	
			statCacheEntry *tmp;
			tmp=next;
			while(tmp!=NULL) {
				if(!strcmp(pathname,tmp->name)) {
					return tmp;
				}
				tmp=tmp->next;
			}
		}
	}
	else {
		if(num != 0) {
			if(!strcmp(pathname,name)) {
				return this;
			}
		}
	}
	return 0;
}

void statCacheEntry::Print(short pos, bool deep, bool onlyValid)
{
    if(num==0) {
	if(deep) 
	    fprintf(stderr,"%d\t#\tslot empty\n",pos);
    }
    else {
		if(onlyValid) {
			if(valid)
				fprintf(stderr,"%d\t#\tnum=%d\tpath=%s\n\tlistSize=%d\trefCount=%d\tvalid=%d\n",pos,num,name,listSize,refCount,valid);
		}
		else fprintf(stderr,"%d\t#\tnum=%d\tpath=%s\n\tlistSize=%d\trefCount=%d\tvalid=%d\n",pos,num,name,listSize,refCount,valid);

		statCacheEntry *tmp=this;
		while(tmp->next!=NULL) {
			tmp=tmp->next;
			if(onlyValid) {
				if(tmp->valid)
					fprintf(stderr,"%d\t#\tnum=%d\tpath=%s\n\tlistSize=%d\trefCount=%d\tvalid=%d\n",pos,tmp->num,tmp->name,tmp->listSize,tmp->refCount,tmp->valid);
			}
			else fprintf(stderr,"\t\tnum=%d\tpath=%s\n\tlistSize=%d\trefCount=%d\tvalid=%d\n",tmp->num,tmp->name,tmp->listSize,tmp->refCount,tmp->valid);
		}
    }
}

//----------------------------------------------------------------------

superTreeEntry::superTreeEntry(Collection *col,char *str)
{
    dumpPath=strsave(str);
    trees=col;
}

superTreeEntry::~superTreeEntry()
{
    SafeDelete(dumpPath);
    if(trees) {
	trees->FreeAll();
	delete trees;
    }
}

void superTreeEntry::SetEntries(Collection *newCol)
{
    if(trees) {
	trees->FreeAll();
	delete trees;
    }
    trees=newCol;
}

//----------------------------------------------------------------------


statTreeEntry::statTreeEntry(unsigned nr,short listPos,short isDir, char *path):Object()
{
    hashNum=nr;
    entries=0;
    dir=isDir;
    pos=listPos;
    if(path)
		rootDir=strsave(path);
    else rootDir=0;
}

statTreeEntry::~statTreeEntry()
{
    if(entries) {
		entries->FreeAll();
		SafeDelete(entries);
    }
    SafeDelete (rootDir);
}

void statTreeEntry::DeleteSubs()
{
    if(entries) {
		entries->FreeAll();
    }
}	

void statTreeEntry::SetName(char *path)
{
    SafeDelete(rootDir);
    rootDir=strsave(path);
}

statTreeEntry *statTreeEntry::FindDir(unsigned nr, short itsPos)
{
    statTreeEntry *found=0;
    
    if(hashNum==nr && pos==itsPos) {
		found=this;
    }
    else {
		if(entries) {
			Iter next(entries);
			statTreeEntry *tmp;
			while(tmp=(statTreeEntry *)next()) {
				if(tmp->IsDir()) {
					if ( found=tmp->FindDir(nr,itsPos) )
					break;
				}
			}
		}
    }
    return found;
}

statTreeEntry *statTreeEntry::FindDir(char *path)
{    
    statTreeEntry *found=0;
    statCacheEntry *entry=cache[hashNum].GetAt(pos);
    if(entry && !strcmp(path,entry->name))
		found=this;
    else {
		if(entries) {
			Iter next(entries);
			statTreeEntry *tmp;
			while(tmp=(statTreeEntry *)next()) {
				if(tmp->IsDir()) {
					if(found= tmp->FindDir(path))
					break;
				}
			}
		}
    }
    return found;
}

int statTreeEntry::Size()
{
    if(entries) {
		return entries->Size();
    }
    return 0;
}

void statTreeEntry::Print()
{
    if(dir) {
		statCacheEntry *entry=cache[hashNum].ForceGetAt(pos);
		if(entry)
			fprintf(stderr,"dir %s\n",entry->name);
		if(entries) {
			Iter next(entries);
			statTreeEntry *tmp;
			while(tmp=(statTreeEntry *)next()) 
				tmp->Print();
		}
    }
    else {
		statCacheEntry *entry=cache[hashNum].ForceGetAt(pos);
		if(entry)
			fprintf(stderr,"file %s\n",entry->name);
    }
}

void statTreeEntry::Add(statTreeEntry *item)
{
    if(entries==NULL) 
		entries=new OrdCollection();
    entries->Add(item);
}

void statTreeEntry::InvalidateAll()
{

    statCacheEntry *entry=cache[hashNum].ForceGetAt(pos);
    if(entry)
		entry->Invalidate();    
    if(entries) {
		Iter next(entries);
		statTreeEntry *tmp;
		while(tmp=(statTreeEntry *)next()) {
			tmp->InvalidateAll();
		}
    }
}

void statTreeEntry::RemoveAll()
{
//    fprintf(stderr,"Remove from cache %s\n",cache[hashNum].GetAt(pos)->name);

    cache[hashNum].Remove(pos);
    if(entries) {
		Iter next(entries);
		statTreeEntry *tmp;
		while(tmp=(statTreeEntry *)next()) {
			tmp->RemoveAll();
		}
    }
}

void statTreeEntry::InvalidateDir()
{
    statCacheEntry *entry=cache[hashNum].ForceGetAt(pos);
    if(entry)
		entry->Invalidate();
    if(entries) {
		Iter next(entries);
		statTreeEntry *tmp;
		while(tmp=(statTreeEntry *)next()) {
			unsigned i=tmp->Num();
			short itsPos=tmp->Pos();
			statCacheEntry *entry=cache[i].ForceGetAt(itsPos);
			if(entry)
				entry->Invalidate();
		}
    }
}

//---------------------------------------------------------------

int getStat(const char *pathname, struct stat *statptr)
{   
	if(!gCacheValid) {
	    return cached_stat((char *)pathname,statptr);
	}
	if(gUseCache) {
	    unsigned h = strhash((const byte *)pathname)%numOfFiles;
	    if(!cache[h].Get(pathname,statptr)) {
	        if(gPrintDebugMsg)
		    printf("STAT:: file=%s h=%ld\n",pathname,h);
	        accessToStat++;
		return cached_stat((char *)pathname,statptr);
	    }
	    else {
		accessFromCache++;
//		if(gPrintDebugMsg)
//		    printf("GET:: file=%s h=%ld\n",pathname,h);
	        return 0;
	    }
	}
	else return cached_stat((char *)pathname,statptr);
}

int statEntryExists(const char *pathname)
{
	// if pathname not in the cache, it doesnt exists !
	if(!gCacheValid) {
	    return !gSystem->AccessPathName((char *)pathname, 0);
	}
	if(gUseCache) {
	    unsigned h = strhash((const byte *)pathname)%numOfFiles;
	    struct stat statStruct;
	    if(!cache[h].Get(pathname,&statStruct)) { // not in cache
		return 0; // dont search with stat -> SPEEDUP
	    }
	    return 1;
	}
	else return !gSystem->AccessPathName((char *)pathname, 0);
}

void initCache()
{
    numOfFiles=Math::NextPrime(numOfFiles);
    if(!cache) {
	cache=new statCacheEntry[numOfFiles];
	gCacheValid=TRUE;
    }
}

void clearCache()
{
    if(cache) {
	delete cache;
	gCacheValid=FALSE;
    }
}
 
short isInvalidated(char *pathname) //  is directory pathname already invalidated
{
    if(!gCacheValid) {
	return 1;
    }
    if(!pathname || !*pathname)
	return 1;
    unsigned h = strhash((const byte *)pathname)%numOfFiles;
    statCacheEntry *entry;
    if(entry=cache[h].ForceGet(pathname)) {
	return !entry->valid;
    }
    return 1;
}

short forceFileIsLink(char *absname)
{
	int retval;
	struct stat sbuf;
	retval = lstat(absname, &sbuf);
	if( retval > 0 )
		{
		// lstat fails -> no link ???
		return 0;
		}

	if ((sbuf.st_mode & S_IFMT) == S_IFLNK)
		return 1;
	else
		return 0;
}

short fileIsLink(char *pathname)
{
    if(!gCacheValid) {
	return forceFileIsLink(pathname);
    }
    if(gUseCache) {
	unsigned h = strhash((const byte *)pathname)%numOfFiles;
	struct stat statStruct;
	statCacheEntry *entry;
	if(entry=cache[h].Get(pathname,&statStruct)) { 
	    if(entry->isLink)
		return 1;
	    else return 0;
	}
	else return forceFileIsLink(pathname);
    }
    else return forceFileIsLink(pathname);
}

void SetCacheUsage(short usage)
{
    gUseCache=usage;
}
