#ifndef _CacheUtils_h_
#define _CacheUtils_h_

#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma interface
#endif

#include "Object.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

class Collection;

class  statCacheEntry: public Object{
public:
	int num;
	bool valid;
	short refCount;
	short listSize;
	bool isLink; // only for quick decision if it is a link -> the statbuf values are NOT from the link
	char 	*name;
	long 	st_mode; // to become HW independent use the greated possible 
	long	st_uid;		
	long	st_gid;
	long	st_mtimer;
	long	st_size;
	long	st_ino;
	long	st_dev;
protected:
	struct statCacheEntry	*next;
public:
	statCacheEntry();
	~statCacheEntry();
	
	void SetName(char *path);	
	void Invalidate();

	short Add(const char *pathname, const struct stat, bool ,short *);
	short Add(statCacheEntry *,short *);
	short Replace(const char *pathname, const struct stat, short *);
	short Remove(int);

	statCacheEntry *Get(const char *pathname, struct stat *);
	statCacheEntry *GetAt(short pos);
	statCacheEntry *ForceGetAt(short pos); // dont care about valid
	statCacheEntry *ForceGet(const char *pathname); // dont care about valid
	void Print(short pos,bool deep=FALSE, bool onlyValid=FALSE);

protected:
	void CopyTo(struct stat *, statCacheEntry *);
	void CopyFrom(statCacheEntry *,const struct stat);
	void CopyFrom(statCacheEntry *,statCacheEntry *);

};

class superTreeEntry : public Object{
    Collection *trees; // collection of statTreeEntries
    char *dumpPath;
public:
    superTreeEntry(Collection *,char *);
    ~superTreeEntry();
    char *Dump() {return dumpPath;}
    Collection *Entries() {return trees;}
    void SetEntries(Collection *);
};

class statTreeEntry : public Object{
	unsigned hashNum;
	short pos;
	Collection *entries;
	short dir;
	char *rootDir;
public:
	statTreeEntry(unsigned,short,short,char *path=0);
	~statTreeEntry();
	void DeleteSubs();
	void Print();
	void Add(statTreeEntry *);
	statTreeEntry *FindDir(unsigned,short);
	statTreeEntry *FindDir(char *);
	int Size();
	Collection *Entries() { return entries;}
	unsigned Num() { return hashNum;}
	short Pos() { return pos;}
	char *Name() { return rootDir;}
	short IsDir() { return dir;}
	void SetName(char *);
	void InvalidateAll();
	void InvalidateDir();
	void RemoveAll();
};

extern "C" {
    int getStat(const char *pathname, struct stat *statptr);
    int statEntryExists(const char *pathname);
    void initCache();
    void clearCache();
    short isInvalidated(char *pathname);
    short fileIsLink(char *pathname);
    void SetCacheUsage(short);
}

#endif
