#ifndef UnixSystem_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define UnixSystem_First

#include "System.h"
#include "Directory.h"

struct fd_set;
class Collection;
class Directory;
class PttyConnection;

//---- UnixDirectory ------------------------------------------------------------

class UnixDirectory : public Directory {
    void *dirp;
public:
    UnixDirectory(char *name);
    ~UnixDirectory();
    char *operator()();
};

//---- UnixSystem --------------------------------------------------------------

class UnixSystem : public System {
    char * work_dir;	// current working directory
    int maxfd;
public:
    UnixSystem();
    ~UnixSystem();
    bool Init();
    void InnerLoop(int timeout= 200);

//---- directory-information ---------------------------
    Directory *MakeDirectory(char *name);			// opens directory stream
    bool ChangeDirectory(char *path);				// changes the working directory
    char *WorkingDirectory();						// returns the working directory
    char *HomeDirectory();							// returns the home directory
	
//---- directory-manipulation ---------------------------
    bool CreateDirectory(char *path, int mode);		// creates a new directory
    bool CreateDirectoryWithParents(char *path, int mode); // creates a new directory and the missing parent directories
    bool RemoveEmptyDirectory(char *path);		// removes a directory (if empty)
    bool RemoveAllFilesInDirectory(char *path);		// removes all files in a directory (not rcursive)

//---- path-information --------------------------------
    int GetPathInfo(char *path, u_long *id, u_long *size, u_long *flags, u_long *modtime);
    bool AccessPathName(char *path, int mode= 0);	// checks accessibility
    char *BaseName(char *pathname);					// "/dir1/dir2/filename" -> "filename"
    char *DirName(char *pathname);					// "/dir1/dir2/filename" -> "/dir1/dir2"
    bool IsAbsolutePath(char *);					// TRUE if <dir> starts w/ pathseperator
    bool ExpandPathName(char *patbuf, int buflen);	// expand the metacharacters as in the shell    
    char *ConcatPathNames(const char *, ...);		// concats 0 or more pathname(s)
    const char *GetServerPath();			// returns the system dependent SERVERPATH

//---- stat chaching -----------------------------------
    void StatCache(bool sw);						// switch on/off stat caching

    char *GetError();
    
    bool CanRead(int fd, int timeout);
    bool CanWrite(int fd, int timeout);
    void Wait(unsigned int duration);
    int GetPid();
    void SetUpMask(fd_set*, Collection*, SysEventCodes c);
    void DispatchReadWrite(fd_set*, Collection*, SysEventCodes c);
    void DispatchSignals(bool asynch);
    void AddHandler(SeqCollection *s, SysEvtHandler *h);
    PttyConnection *MakePttyConnection(char *prog, char **args);
    
    //---- running programs -------------------------------------
    u_int Sleep(u_int seconds);
    int  RunProgSynch(const char *cmdline);
    void RunProgAsynch(const char *cmdline);
 
    //---- environment manipulation -----------------------------
    char *Getenv(char *name);
    void Setenv(char *name, char *value);

    char *Which(char *search, char *file, int mode);

    //---- RPC --------------------------------------------------
    int OpenConnection(char *server, char *service);
    int AnnounceTcpService(char *servicename);
    int AnnounceUnixService(char *servername);
    int AcceptConnection(int sock);
    int ConnectTcpService(char *hostname, char *service);
    int ConnectUnixService(char *service);
    char *GetSessionName(int argc, char **argv, char *server);
    void InitSession(int argc, char **argv);
};

#endif
