
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include "ptty.h"

extern char * sys_errlist[];
extern char * gProgname;
 
extern int unix_set_nonblocking_mode();
 
/* ----- configuration --------------- */
 
#if defined (OS_SOLARIS_2) || defined (OS_UNIXWARE) || defined (OS_SINIX)
#  define HAVE_SYSV_PTY
#  define POSIX_TERMIOS
#endif

#if defined (OS_HPUX_8) || defined (OS_AIX_3_2)
#  define POSIX_TERMIOS
#endif

#if defined (OS_SCO)
#  define _IBCS2
#endif
 
#if !defined (USE_SGTTYB) && !defined (USE_TERMIO)
#  ifdef OS_BSD
#    define USE_SGTTYB
#  else
#    define USE_TERMIO
#  endif
#endif


#if defined (HAVE_SYSV_PTY)
# include <stropts.h>
  extern int    grantpt  _PROTO_(( int fd ));
  extern int    unlockpt _PROTO_(( int fd ));
  extern char * ptsname  _PROTO_(( int fd ));
#endif


#ifdef USE_SGTTYB

#  include <sgtty.h>

   static struct sgttyb  tty_sgttyb;
   static struct tchars  tty_tchars;
   static struct ltchars tty_ltchars;
   static int            tty_localmode;
   static int            tty_ldisc;

#else	/* USE_TERMIO */

#  ifdef POSIX_TERMIOS
#    ifdef OS_AIX_3_2
#      ifndef _POSIX_SOURCE
#        define _POSIX_SOURCE 1
#      endif
#    endif
#    include <termios.h>
#  else
#    include <termio.h>
#  endif

#endif	/* USE_TERMIO */


/* define control characters */

#define CC_CTRL(c) ((c) & 0x1F)

#define CC_INTR		CC_CTRL('c')
#define CC_QUIT		034
#define CC_ERASE	CC_CTRL('h')
#define CC_KILL		CC_CTRL('u')
#define CC_EOF		CC_CTRL('d')
#define CC_EOL		'\n'
#define CC_EOL2		0
#define CC_SWTCH	CC_CTRL('z')
#define CC_START	CC_CTRL('q')
#define CC_STOP		CC_CTRL('s')
#define CC_SUSP		CC_CTRL('z')
#define CC_DSUSP	CC_CTRL('y')
#define CC_RPRNT	CC_CTRL('r')
#define CC_FLUSH	CC_CTRL('o')
#define CC_WERASE	CC_CTRL('w')
#define CC_LNEXT	CC_CTRL('v')


#ifdef USE_TERMIO

# ifdef POSIX_TERMIOS
  static struct termios tty_termio;
# else
  static struct termio tty_termio;
# endif

#define TERMIO_CC_SIZE (sizeof (tty_termio.c_cc) / sizeof (tty_termio.c_cc[0]))

static struct
  {
    int index;
# ifdef POSIX_TERMIOS
    cc_t cc;
# else
    unsigned char cc;
# endif
  }
init_termio_cc[] =
  {
# ifdef VINTR
    { VINTR,  CC_INTR },
# endif
# ifdef VQUIT
    { VQUIT,  CC_QUIT },
# endif
# ifdef VERASE
    { VERASE, CC_ERASE },
# endif
# ifdef VKILL
    { VKILL,  CC_KILL },
# endif
# ifdef VEOF
    { VEOF,   CC_EOF },
# endif
# ifdef VEOL
    { VEOL,   CC_EOL },
# endif
# ifdef VEOL2
    { VEOL2,  CC_EOL2 },
# endif
# ifdef VSWTCH
    { VSWTCH, CC_SWTCH },
# endif
# ifdef VSTART
    { VSTART, CC_START },
# endif
# ifdef VSTOP
    { VSTOP,  CC_STOP },
# endif
# ifdef VSUSP
    { VSUSP,  CC_SUSP },
# endif
# ifdef VDSUSP
    { VDSUSP, CC_DSUSP },
# endif
# ifdef VREPRINT
    { VREPRINT, CC_RPRNT },
# endif
# ifdef VDISCARD
    { VDISCARD, CC_FLUSH },
# endif
# ifdef VWERASE
    { VWERASE, CC_WERASE },
# endif
# ifdef VLNEXT
    { VLNEXT, CC_LNEXT },
# endif
    { -1, 0 }
  };

#endif /* USE_TERMIO */


/*---- local functions ---------------------------------------------*/

static int
ptty_open_master ( pty )
  ptty_t * pty;
{
  int ret;
  int mfd;
  char * pty_slave_name;
  struct stat stat_buf;
 
#if defined (HAVE_SYSV_PTY)
 
  mfd = open ("/dev/ptmx", O_RDWR);
  if ( mfd == -1 )
    return -1;
 
  {
    sigset_t sig_prev, sig_block;
    sigemptyset (&sig_block);
    sigaddset (&sig_block, SIGCLD);
    sigprocmask (SIG_BLOCK, &sig_block, &sig_prev);
 
    ret = grantpt (mfd);
 
    sigprocmask (SIG_SETMASK, &sig_prev, 0);
  }
 
  if ( ret == -1 || unlockpt (mfd) == -1
      || (pty_slave_name = ptsname (mfd)) == 0 )
    {
      close (mfd);
      return -1;
    }
  pty->master_fd = mfd;
  strcpy (pty->slave_dev_name, pty_slave_name);
  return 0;
 
#elif defined (OS_IRIX)
 
  {
    sigset_t sig_prev, sig_block;
    sigemptyset (&sig_block);
    sigaddset (&sig_block, SIGCLD);
    sigprocmask (SIG_BLOCK, &sig_block, &sig_prev);
 
    pty_slave_name = _getpty (&mfd, O_RDWR, 0600, 0);
 
    sigprocmask (SIG_SETMASK, &sig_prev, 0);
  }
 
  if ( pty_slave_name == 0
      || mfd < 0
      || fstat (mfd, &stat_buf) < 0 )
    return -1;
  pty->master_fd = mfd;
  strcpy (pty->slave_dev_name, pty_slave_name);
  return 0;
 
#elif defined (OS_AIX_3_2)
 
  mfd = open ("/dev/ptc", O_RDWR);
  if ( mfd < 0 )
    return -1;
  pty->master_fd = mfd;
  strcpy (pty->slave_dev_name, ttyname (mfd));
  return 0;
 
#else
 
# ifdef OS_HPUX_8
  char * pty_master_dir = "/dev/ptym";
  char * pty_slave_dir = "/dev/pty";
# else
  char * pty_master_dir = "/dev";
  char * pty_slave_dir = pty_master_dir;
# endif
 
  DIR * dirp;
  ret = -1;
  dirp = opendir (pty_master_dir);
  if ( dirp != 0)
    {
      struct dirent * dir_entry;
      while ( (dir_entry = readdir (dirp)) != 0 )
	{
	  if ( strncmp (dir_entry->d_name, "pty", 3) == 0 )
	    {
	      char pty_device [64];
	      sprintf (pty_device, "%s/%s", pty_master_dir, dir_entry->d_name);
	      mfd = open (pty_device, O_RDWR);
	      if ( mfd >= 0 )
		{
		  sprintf (pty_device, "%s/t%s", pty_slave_dir, &dir_entry->d_name[1]);
		  if ( access (pty_device, F_OK | R_OK | W_OK) == 0 )
		    {
		      pty->master_fd = mfd;
		      strcpy (pty->slave_dev_name, pty_device);
		      ret = 0;
		      break;
		    }
		  else
		    {
		      close (mfd);
		    }
		}
	    }
	}
      closedir (dirp);
    }
  return ret;
 
#endif
}


static void
ptty_get_tty_mode ( pty )
  ptty_t * pty;
{
  static int have_terminal_settings = 0;

#ifdef USE_SGTTYB
  if ( ! have_terminal_settings )
    {
      tty_sgttyb.sg_flags  = ANYP | ECHO | CRMOD | PASS8 | CTLECH | DECCTQ;
      tty_sgttyb.sg_ispeed = B9600;
      tty_sgttyb.sg_ospeed = B9600;
      tty_sgttyb.sg_erase  = CC_ERASE;
      tty_sgttyb.sg_kill   = CC_KILL;
 
      tty_tchars.t_intrc   = CC_INTR;
      tty_tchars.t_quitc   = CC_QUIT;
      tty_tchars.t_startc  = CC_START;
      tty_tchars.t_stopc   = CC_STOP;
      tty_tchars.t_eofc    = CC_EOF;
      tty_tchars.t_brkc    = 0;
 
      tty_ltchars.t_suspc  = CC_SUSP;
      tty_ltchars.t_dsuspc = CC_DSUSP;
      tty_ltchars.t_rprntc = CC_RPRNT;
      tty_ltchars.t_flushc = CC_FLUSH;
      tty_ltchars.t_werasc = CC_WERASE;
      tty_ltchars.t_lnextc = CC_LNEXT;
 
      tty_localmode = LCRTBS | LCRTERA | LCRTKIL | LCTLECH | LPASS8;
      tty_ldisc = NTTYDISC;
 
      have_terminal_settings = 1;
    }
 
  pty->cc_erase = tty_sgttyb.sg_erase;
  pty->cc_kill  = tty_sgttyb.sg_kill;
  pty->cc_rprnt = tty_ltchars.t_rprntc;
  pty->cc_susp  = tty_ltchars.t_suspc;
  pty->cc_intr  = tty_tchars.t_intrc;
  pty->cc_quit  = tty_tchars.t_quitc;
  pty->cc_eof   = tty_tchars.t_eofc;

#else	/* USE_TERMIO */

  if ( ! have_terminal_settings )
    {
      int i;
      for ( i = 0; init_termio_cc[i].index >= 0; i++ )
	if ( init_termio_cc[i].index < TERMIO_CC_SIZE )
	  tty_termio.c_cc[init_termio_cc[i].index] = init_termio_cc[i].cc;
 
      tty_termio.c_iflag = IGNBRK | IGNPAR | ICRNL | IXON;
      tty_termio.c_oflag = 0;
      tty_termio.c_cflag = CS8 | CREAD | CLOCAL;
# ifdef POSIX_TERMIOS
      cfsetispeed (&tty_termio, B9600);
      cfsetospeed (&tty_termio, B9600);
# else
      tty_termio.c_cflag |= B9600;
# endif
      tty_termio.c_lflag = ISIG | ICANON | ECHO | ECHOE | ECHOK;
# ifdef ECHOCTL
      tty_termio.c_lflag |= ECHOCTL;
# endif
 
      have_terminal_settings = 1;
    }

# ifdef VINTR
  pty->cc_intr  = tty_termio.c_cc [VINTR];
# endif
# ifdef VQUIT
  pty->cc_quit  = tty_termio.c_cc [VQUIT];
# endif
# ifdef VERASE
  pty->cc_erase = tty_termio.c_cc [VERASE];
# endif
# ifdef VKILL
  pty->cc_kill  = tty_termio.c_cc [VKILL];
# endif
# ifdef VEOF
  pty->cc_eof   = tty_termio.c_cc [VEOF];
# endif
# ifdef VSUSP
  pty->cc_susp  = tty_termio.c_cc [VSUSP];
# endif
# ifdef VREPRINT
  pty->cc_rprnt = tty_termio.c_cc [VREPRINT];
# endif

#endif	/* USE_TERMIO */
}

static void
ptty_init ( pty )
  ptty_t * pty;
{
  pty->master_fd = -1;
  pty->slave_pid = -1;
  pty->slave_dev_name[0] = 0;
  pty->cc_erase = 0;
  pty->cc_kill = 0;
  pty->cc_rprnt = 0;
  pty->cc_susp = 0;
  pty->cc_intr = 0;
  pty->cc_quit = 0;
  pty->cc_eof = 0;
}


/*---- external functions ------------------------------------------*/

void
ptty_cleanup ( pty )
  ptty_t * pty;
{
  if ( pty->master_fd >= 0 )
    {
      close (pty->master_fd);
    }
  ptty_init (pty);
}

void
ptty_kill_child ( pty )
  ptty_t * pty;
{
  if ( pty->slave_pid >= 0 )
    {
#if defined (OS_AIX_3_2) || defined (OS_OSF1)
      kill ( - pty->slave_pid, SIGKILL );
#else
      kill ( - pty->slave_pid, SIGTERM );
#endif
      pty->slave_pid = -1;
    }
}

int
ptty_spawn_slave ( pty, name, args )
  ptty_t *  pty;
  char   *  name;
  char   ** args;
{
  ptty_init (pty);
 
  if ( ptty_open_master (pty) == -1 )
    {
      Error ("ptty_spawn_slave", "cannot get a new pseudo tty\n");
      return -1;
    }
 
  ptty_get_tty_mode (pty);
 
  pty->slave_pid = fork();
  if ( pty->slave_pid < 0 )
    {
      SysError ("ptty_spawn_slave", "cannot spawn \"%s\"", name);
      ptty_cleanup (pty);
      return -1;
    }
  else if ( pty->slave_pid == 0 )	/* slave context */
    {
      int i;
      int slave_fd;
      int fd_table_size;
      int code;
 
      /* Reset default action for signals. */
      for ( i = 1; i < NSIG; i++ )
	signal (i, SIG_DFL);
 
      /* Close all inherited file descriptors. */

#ifdef OS_BSD
      fd_table_size = getdtablesize();
#else
      fd_table_size = sysconf (_SC_OPEN_MAX);
#endif
      for ( i = 0; i < fd_table_size; i++ )
	close (i);
 
      /* Create new process group and
	 disconnect from controlling terminal. */
#ifdef OS_BSD
      setpgrp (0, getpid());
      {
	int fd = open ("/dev/tty", O_RDWR);
	if ( fd >= 0 )
	  {
	    ioctl (fd, TIOCNOTTY, 0);
	    close (fd);
	  }
      }
#else
      setsid();
#endif

      /* Open slave pseudo tty device.
	 On BSD systems this automatically becomes the
	 controlling terminal.
	 On certain SYSV systems this must be done explicitly. */
 
      slave_fd = open (pty->slave_dev_name, O_RDWR);

#ifdef OS_SYSV
#  ifdef TIOCSCTTY
      ioctl (slave_fd, TIOCSCTTY, 0);
#  endif
#endif

#ifdef HAVE_SYSV_PTY
      ioctl (slave_fd, I_PUSH, "ptem");
      ioctl (slave_fd, I_PUSH, "ldterm");
#endif

      /* Set terminal modes and control characters. */

#ifdef USE_SGTTYB
      ioctl (slave_fd, TIOCSETP, &tty_sgttyb);
      ioctl (slave_fd, TIOCSETC, &tty_tchars);
      ioctl (slave_fd, TIOCSLTC, &tty_ltchars);
      ioctl (slave_fd, TIOCLSET, &tty_localmode);
      ioctl (slave_fd, TIOCSETD, &tty_ldisc);
#else
# ifdef POSIX_TERMIOS
      tcsetattr (slave_fd, TCSANOW, &tty_termio);
# else
      ioctl (slave_fd, TCSETA, &tty_termio);
# endif
#endif

      /* Connect stdin, stdout and stderr to slave device. */
      for ( i = 0; i <= 2; i++ )
	if ( i != slave_fd )
	  dup (slave_fd);
      if ( slave_fd > 2 )
	close (slave_fd);
 
      /* Execute program */
      execvp (name, args);
      code = errno;
      fprintf (stderr, "%s: exec(\"%s\") failed: %s\n",
	       gProgname, name, sys_errlist [errno]);
      _exit (code);
    }

  unix_set_nonblocking_mode (pty->master_fd);
 
  return 0;
}


void
ptty_set_size ( pty, lines, columns)
  ptty_t * pty;
  int      lines;
  int      columns;
{
#if !defined (OS_SOLARIS) && !defined (OS_AIX)
# ifdef TIOCSWINSZ
  if ( pty->master_fd >= 0 )
    {
      struct winsize ws;
      ws.ws_row = lines;
      ws.ws_col = columns;
      ws.ws_xpixel = 0;
      ws.ws_ypixel = 0;
      ioctl (pty->master_fd, TIOCSWINSZ, &ws);    
    }
# endif
#endif
}
