
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <pwd.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifndef OS_SCO
# include <sys/un.h>
#endif
#include <netdb.h>
#include <fcntl.h>

#include <memory.h>

#ifdef OS_NEXT_3
# include <sys/dir.h>
# define dirent direct
#else
# include <dirent.h>
#endif

# include <unistd.h>

#ifdef OS_AIX
#  include <sys/select.h>
#  include <grp.h>
#else
#  include <limits.h>
#endif

#ifdef OS_HPUX
# include <grp.h>
#endif

/* configuration */

#if (defined (OS_SYSV) && OS_SYSV == 4) || defined (OS_AIX) || defined (OS_IRIX) \
     || defined (OS_LINUX) || defined (OS_OSF1) || defined (OS_SCO) || defined (OS_USL)
# define HAVE_sigaction
#endif

#ifndef MAXPATHLEN
# define MAXPATHLEN 1024
#endif

#undef TRUE
#undef FALSE
#define TRUE  1
#define FALSE 0

static char server_path [] = "/tmp/.sniffdir2";
static char tcp_protocol [] = "tcp";

enum ePathInfoFlags {
    ePiExec     =   1,
    ePiDir      =   2,
    ePiSpecial  =   4
};

char *cuserid();

typedef unsigned int bool;
typedef void (*sig_handler_t)();

void unix_signal ();

extern int errno, select();
extern char *sys_errlist[], *getenv();

extern char *gProgname;

#define MAXSIGNALS (sizeof (signal_map) / sizeof (signal_map[0]))

static struct
  {
    int		  code;
    sig_handler_t handler;
  }
signal_map [] =
  {
    { SIGINT, 0 },
    { SIGQUIT, 0 },
    { SIGILL, 0 },
    { SIGBUS, 0 },
    { SIGSEGV, 0 },
    { SIGPIPE, 0 },
    { SIGALRM, 0 },
#ifdef SIGCHLD
    { SIGCHLD, 0 }
#else
    { SIGCLD, 0 }
#endif
  };

 

/* simple hash routine used by cached_stat() */
#define STAT_CACHE_SIZE 5000

static unsigned int
strhash(const char *str)
{
    register unsigned int h = 0;
    register const char *p= str;

    while (*p != 0)
    	h += *p++;
    return h % STAT_CACHE_SIZE;
}


bool	gDebugStat = FALSE;		/* debug stat() */
bool	statCacheOn = FALSE;		/* switch on/off stat() caching */

typedef struct {
	char	file[256];
	struct  stat sbuf;
	int	retval;
	int	errno;
	bool	isValid;
} cachetype;
static 	cachetype cache[STAT_CACHE_SIZE];


static int cached_count = 0;
static int stat_count = 0;

/*
 * Switch on/off stat caching
 */
void
statCache(bool sw)
{
	int i;
	if(sw == statCacheOn)
		return;

	if(!sw)
	{	
		for (i = 0; i < STAT_CACHE_SIZE; i++)
			cache[i].isValid = FALSE;
		
		if (gDebugStat)
			printf("********* CACHED stat OFF (%d%% cashed, called: stat=%d cached=%d) *********\n\n", (100*cached_count)/((cached_count ? cached_count : 1)+stat_count), stat_count, cached_count);
		stat_count = cached_count = 0;
	}
	else if (gDebugStat)
		{	
		printf("\n********* CACHED stat ON  (called: %d stat and access so far) *********\n", stat_count);
		stat_count = cached_count = 0;
		}
	statCacheOn = sw;	
}

/* Cache routine for stat() system calls
   Loading a projet in SNiFF+, stat() will called several times for the same file.
   (about 18000 stat() calls for a project with 1000 files -> performace problem)
   cached_stat() caches the result of stat() system calls if statCacheOn == TRUE
*/
int 
cached_stat (char * file, struct stat *sbuf)
{
	int 	i,retval;
	unsigned int h;

	if(statCacheOn)
		{
		h = strhash(file);
		
		if(cache[h].isValid && strcmp(file, cache[h].file) == 0)
			{
			memcpy(sbuf,&cache[h].sbuf,sizeof(struct stat));
			cached_count++;
			errno = cache[h].errno;
			
			if (gDebugStat)
			    printf("stat from cache    (%s)\n", file);
 
			return cache[h].retval;
			}
		}

	if (gDebugStat)
		printf("stat from disk    (%s)\n", file);
 
	stat_count++;
	retval = stat(file, sbuf);
		
/*	if(statCacheOn && cache[h].retval >= 0) */
		/* stat is saved only if retval >=0 because of mkdir, create file etc. */
	if(statCacheOn)
		{
		strncpy(cache[h].file,file,sizeof(cache[0].file)-1);
		memcpy(&cache[h].sbuf,sbuf,sizeof(struct stat));
		cache[h].retval = retval;
		cache[h].errno = errno;
		cache[h].isValid = TRUE;
		}
	return retval;		
}


void
unix_initialize()
{
  signal (SIGHUP, SIG_IGN);
  signal (SIGPIPE, SIG_IGN);
#ifdef SIGTTOU
  signal (SIGTTOU, SIG_IGN); /* background write attempted to control terminal */
#endif
}

int unix_select ( nfds, readready, writeready, timeout )
    unsigned int nfds;
    fd_set *     readready;
    fd_set *     writeready;
    unsigned int timeout;
{
    int retcode;
    struct timeval tv;
    
    tv.tv_sec= timeout / 1000;
    tv.tv_usec= (timeout % 1000) * 1000;
    retcode= select(nfds, readready, writeready, 0, &tv);
    if (retcode == -1) {
	if (errno == EINTR)
	    return -2;
	if (errno == EBADF)
	    return -3;
	return -1;
    }
    return retcode;
}

static void
sighandler (sig
#if defined (hpux) || defined (__hpux__)
		, code, scp
#endif
	    )
  int sig;
#if defined (hpux) || defined (__hpux__)
  int code;
  struct sigcontext * scp;
#endif
{
  int i;

#if defined (hpux) || defined (__hpux__)
  /* restart interrupted system calls */
  if ( scp->sc_syscall == SYS_WRITE || scp->sc_syscall == SYS_READ )
    scp->sc_syscall_action = SIG_RESTART;
#endif

  for ( i = 0; i < MAXSIGNALS; i++ )
    if ( signal_map[i].code == sig && signal_map[i].handler != 0 )
      {
	(*signal_map[i].handler)(i);
	return;
      }
}

void
unix_signal (sig, handler)
  int sig;
  sig_handler_t handler;
{
  sig_handler_t the_real_handler = sighandler;

  if ( signal_map[sig].handler != handler )
    {
#ifdef HAVE_sigaction

      struct sigaction sa;
      sa.sa_handler = the_real_handler;
      sigemptyset (&sa.sa_mask);
      sa.sa_flags = 0
# ifdef SA_RESTART
		  | SA_RESTART		/* restart interrupted system calls */
# endif
# ifdef SA_NOCLDSTOP
		  | SA_NOCLDSTOP	/* don't care if child is stopped/continued */
# endif
		  ;

      sigaction (signal_map[sig].code, &sa, 0);
#else
      struct sigvec sv;
      sv.sv_handler = the_real_handler;
      sv.sv_mask = 0;
      sv.sv_flags = 0;

# if defined(hpux) || defined(__hpux__)
      sigvector (signal_map[sig].code, &sv, 0);
# else
      sigvec (signal_map[sig].code, &sv, 0);
# endif

#endif
      signal_map[sig].handler = handler;
    }
}

void
unix_resetsignal()
{
  int i;

#ifdef HAVE_sigaction
  struct sigaction sa;
  sa.sa_handler = SIG_DFL;
  sigemptyset (&sa.sa_mask);
  sa.sa_flags = 0;
#else
  struct sigvec sv;
  sv.sv_handler = SIG_DFL;
  sv.sv_mask = 0;
  sv.sv_flags = 0;
#endif

  for ( i = 0; i < MAXSIGNALS; i++ )
    if ( signal_map[i].handler != 0 )
      {
	signal_map[i].handler = 0;
#ifdef HAVE_sigaction
	sigaction (signal_map[i].code, &sa, 0);
#else
# if defined (hpux) || defined (__hpux__)
	sigvector (signal_map[i].code, &sv, 0);
# else
	sigvec (signal_map[i].code, &sv, 0);
# endif
#endif
      }
}

char *
unix_signame (sig)
  int sig;
{
  static char sig_name [20];
  sprintf (sig_name, "%d", signal_map[sig].code);
  return sig_name;
}

static struct passwd *
get_passwd_entry()
{
    static struct passwd *pw = NULL;
    if (pw == NULL)
	pw = getpwuid(getuid());
    return pw;
}

char *unix_homedirectory (name)
    char *name;
{
    static char path[MAXPATHLEN],
		home_dir[MAXPATHLEN];
    struct passwd *pw;

    if (name)
    {
	pw = getpwnam(name);
	if (pw)
	{
	    strncpy(path, pw->pw_dir, MAXPATHLEN);
	    return path;
	}
    }
    else
    {
	if (home_dir[0] == 0)
	{
	    char *home = getenv("HOME");
	    if (home && *home)
		strcpy(home_dir, home);
	    else
	    {
		pw = get_passwd_entry();
		if (pw)
		    strncpy(home_dir, pw->pw_dir, MAXPATHLEN);
		else
		    strcpy(home_dir, "/");
	    }
	}
	return home_dir;
    }
    return 0;
}

#ifndef OS_BSD
char *getwd(path)
    char *path;
{
    if ( getcwd(path, MAXPATHLEN) == NULL ) {
	strcpy(path, sys_errlist[errno]);
	return NULL; 
    }
    return path;
}
#endif

char *unix_initialworkingdirectory()
{
    static char start_msg[] = "starting from \"%s\"\n";
    static char workdir[MAXPATHLEN];
    int workdir_valid;

    /* Try to start from the
       1) current working directory (or $PWD)
       2) home directory
       3) root directory
     */

    workdir_valid = 1;
    if ( getwd(workdir) ) {
	if ( unix_fileaccess(workdir, R_OK) != 0 ) {	/* cannot read "./" */
	    SysError("", "cannot read \"./\"");
	    workdir_valid = 0;
	}
    } else {	/* cannot get cwd */
	Warning("", "%s\n", workdir);
	workdir_valid = 0;
    }

    if ( ! workdir_valid ) {
	strcpy(workdir, unix_homedirectory(NULL));
	if ( chdir(workdir) == 0 )
	    Warning("", start_msg, workdir);
	else {
	    strcpy(workdir, "/");
	    chdir(workdir);
	    Warning("", start_msg, workdir);
	}
    }
#ifdef S_ISLNK
    /* If the system supports symbolic links, then try to use $PWD.
       $PWD is set by some shells. It would give us "/home/joe" whereas
       'getcwd' would return "/tmp_mnt/net/brutus/home/joe" assuming
       that "/home" is a symbolic link to "/tmp_mnt/net/brutus/home".
       Unfortunately, $PWD is not reliable. So we must ensure
       that it really points to the current working directory. */
    else {
	char *pwd = getenv("PWD");
	if ( pwd != NULL && pwd[0] != '\0' ) {
	    struct stat cwd_stat, pwd_stat;
	    if ( stat( pwd, &pwd_stat ) == 0
		&& S_ISDIR( pwd_stat.st_mode )
		&& stat( workdir, &cwd_stat ) == 0
		&& pwd_stat.st_dev == cwd_stat.st_dev
		&& pwd_stat.st_ino == cwd_stat.st_ino )
		strcpy(workdir, pwd);
	}
    }
#endif

    return workdir;
}

char *unix_getdirentry (dirp)
    DIR *dirp;
{
    struct dirent *dp;

    if (dirp && (dp= readdir(dirp)))
	return dp->d_name;
    return 0;
}

int unix_filestat (path, id, size, flags, modtime)
    char *path;
    u_long *size, *modtime, *id, *flags;
{
    struct stat statbuf;
    
    if (path != 0 && getStat(path, &statbuf) >= 0) {
	if (id)
	    *id= (statbuf.st_dev << 24) + statbuf.st_ino;
	if (size)
	    *size= statbuf.st_size;
	if (modtime)
	    *modtime= statbuf.st_mtime;
	if (flags) {
	    if (statbuf.st_mode & ((S_IEXEC)|(S_IEXEC>>3)|(S_IEXEC>>6)))
		*flags|= ePiExec;
	    if ((statbuf.st_mode & S_IFMT) == S_IFDIR)
		*flags|= ePiDir;
	    if ((statbuf.st_mode & S_IFMT) != S_IFREG
			    && ((statbuf.st_mode & S_IFMT) != S_IFDIR))
		*flags|= ePiSpecial;
	}
	return 0;
    }
    return 1;
}

#ifdef OS_HPUX

gid_t getgrpid(char *name)
{
	struct group *grp;
	if ((grp = getgrnam(name)) == NULL)
	    return -1;
	return grp->gr_gid;
}

int
getgroups (int ngroups, gid_t gidset[])
{
    FILE	*fp;
    int		lineLen = 2000;
    char	line[lineLen + 1];
    int	len;
    char *grp;
    char *p;
    int	i;
    gid_t	gid;
    char	buff[300];

    sprintf(buff, "groups %s", cuserid(NULL));
    
    if ((fp = popen(buff, "r")) == NULL)
	    return 0;

    line[0] = '\0';

    fgets(line, lineLen, fp);

    pclose(fp);

    for (i = 0, p = grp = line; *p && i < ngroups; p++)
	{
	if (*p == ' ' || *p == '\n')
	    {
	    *p = '\0';
	    gid = getgrpid(grp);
	    if (gid != -1)
		{
		gidset[i] = gid;
		i++;
		}
	    grp = p+1;
	    }
	}

    if (*grp && i < ngroups)
	{
	gid = getgrpid(grp);
	if (gid != -1)
	    {
	    gidset[i] = gid;
	    i++;
	    }
	}

    return i;
}
#endif

static int
fileaccess (statbuf, mode)
  struct stat * statbuf;
    int mode;
{
  static int userid = -1;
  static int groupSize = -1;

#ifdef OS_SUNOS
  static int grouplist[NGROUPS_MAX + 1];
#else
  static gid_t grouplist[NGROUPS_MAX + 1];
#endif

    int i;
  int check_bit;
   
/* 
S_IRUSR 00400           read permission: owner
S_IWUSR 00200           write permission: owner
S_IXUSR 00100           execute permission: owner
S_IRGRP 00040           read permission: group
S_IWGRP 00020           write permission: group
S_IXGRP 00010           execute permission: group
S_IROTH 00004           read permission: other
S_IWOTH 00002           write permission: other
S_IXOTH 00001           execute permission: other

R_OK    4	test for read permission
W_OK    2	test for write permission
X_OK    1	test for execute (search) permission
F_OK    0	test for presence of file
*/
 
  if ( mode == F_OK )		/* F_OK == 0 */
    return 0;
 
  if ( userid == -1 )
    	{
      userid = geteuid();
      grouplist[0] = getegid();
      groupSize = getgroups (NGROUPS_MAX, &grouplist[1]);
      if ( groupSize == -1 )
	groupSize = 0;
      groupSize++;
    }
    	
  errno = EPERM;
	
  if ( mode & R_OK )
	    {
      check_bit = S_IROTH;
      if ( statbuf->st_uid == userid ) /* user's file */
		{
	  check_bit = S_IRUSR;
		}
      else
		{
	  for ( i = 0; i < groupSize; i++ )
	    if ( statbuf->st_gid == grouplist[i] )
	      {
		check_bit = S_IRGRP;
		break;
		}
	}
      if ( ! (statbuf->st_mode & check_bit) )
	return 1;
	    }
	    
  if ( mode & W_OK )
	    {
      check_bit = S_IWOTH;
      if ( statbuf->st_uid == userid ) /* user's file */
		{
	  check_bit = S_IWUSR;
		}
      else
		{
	  for ( i = 0; i < groupSize; i++ )
	    if ( statbuf->st_gid == grouplist[i] )
	      {
		check_bit = S_IWGRP;
		break;
		}
	}
      if ( ! (statbuf->st_mode & check_bit) )
	return 1;
	    }
	    
  if ( mode & X_OK )
	    {
      check_bit = S_IXOTH;
      if ( statbuf->st_uid == userid ) /* user's file */
		{
	  check_bit = S_IXUSR;
		}
      else
	{
	  for ( i = 0; i < groupSize; i++ )
	    if ( statbuf->st_gid == grouplist[i] )
		{
		check_bit = S_IXGRP;
		break;
	      }
	}
      if ( ! (statbuf->st_mode & check_bit) )
	return 1;
    }
		
		    return 0;
}
 
int unix_fileaccess (path, mode)
    char *path;
    int mode;
{
    struct stat statbuf;
 
/*    if (!statCacheOn)
    	{
	if (gDebugStat)
		printf("access  (%s)\n", path);

	stat_count++;
    	if (access(path, mode) == 0)
	    	return 0;
    	return 1;
	    }
 */
    if (path != 0 && getStat(path, &statbuf) >= 0) 	/* cached_stat sets errno */
    	return fileaccess(&statbuf, mode);
    else {
	if (gDebugStat)
		printf("access  (%s)\n", path);

	stat_count++;
    	if (access(path, mode) == 0)
	    	return 0;
    	return 1;
    }
    return 1;
}

int unix_waitchild()
{
#ifdef OS_NEXT_3
    return wait3 (NULL, WNOHANG, NULL);
#else
    int status;
    return waitpid(-1, &status, WNOHANG);
#endif
}

char *
cuserid (str_return)
    char *str_return;
{
    static char user_name[64];
    struct passwd *pw;

    if (user_name[0] == 0)
    {
	pw = get_passwd_entry();
	if (pw)
	    {
	    strncpy(user_name, pw->pw_name, sizeof(user_name));
	    user_name[sizeof(user_name) - 1] = '\0';
	    }
	else
	    sprintf(user_name, "%d", getuid());
    }

    if (str_return)
	strcpy(str_return, user_name);

    return user_name;
}

char *get_long_user_name()
{
    struct passwd *pw = get_passwd_entry();
    return pw ? pw->pw_gecos : "";
}


#ifdef OS_NEXT_3

/* NEXTSTEP lacks 'putenv' */

int
putenv (string)
  char *string;
{
  return 0;
}

#endif	/* OS_NEXT_3 */


/*---- RPC ---------------------------------------------------------------*/

int
unix_set_nonblocking_mode (fd)
  int fd;
{
  int flags = fcntl (fd, F_GETFL, 0);
  if ( flags == -1 )
    return -1;
  return fcntl (fd, F_SETFL, flags | O_NONBLOCK);
}

int unix_accept (sock)
    int sock;
{
    int fd = accept (sock, 0, 0);
    if ( fd >= 0 )
      unix_set_nonblocking_mode (fd);
    return fd;
}


/* TCP sockets */

int unix_TcpConnect (hostname, servicename)
    char *hostname, *servicename;
{
    int saved_errno;
    int sock;
    struct sockaddr_in server;
    struct hostent *host_ptr;
    struct servent *sp;
    
    if ((sp= getservbyname(servicename, tcp_protocol)) == NULL) {
	Error("TcpConnect", "no service \"%s\" with protocol \"%s\"\n",
						servicename, tcp_protocol);
	return -1;
    }
    
    if ((host_ptr= gethostbyname(hostname)) == 0) {
	Error("TcpConnect", "unknown host \"%s\"\n", hostname);
	return -1;
    }
    
    /* Check the address type for an internet host */
    if (host_ptr->h_addrtype != AF_INET) {
	Error("TcpConnect", "%s is not an internet host\n", hostname);
	return -1;
    }
    
    memset(&server, 0, sizeof(server));
    memcpy(&server.sin_addr, host_ptr->h_addr, host_ptr->h_length);
    server.sin_family = host_ptr->h_addrtype;
    server.sin_port= htons(sp->s_port);
    
    /* Create socket */
    if ((sock= socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	return -1;
    }
	
    if (connect(sock, (struct sockaddr*) &server, sizeof(server)) < 0) {
	saved_errno = errno;
	close(sock);
	errno = saved_errno;
	return -1;
    }

    unix_set_nonblocking_mode (sock);

    return sock;
}

int unix_TcpService (servicename)
    char *servicename;
{
    int saved_errno;
    int sock;
    struct sockaddr_in inserver;
    struct servent *sp;
    
    if ((sp= getservbyname(servicename, tcp_protocol)) == NULL) {
	Error("TcpService", "no service \"%s\" with protocol \"%s\"\n",
						    servicename, tcp_protocol);
	return -1;
    }
    
    /* Create tcp socket */
    if ((sock= socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	return -1;
    }
	
    memset(&inserver, 0, sizeof(inserver));
    inserver.sin_family = AF_INET;
    inserver.sin_addr.s_addr = htonl(INADDR_ANY);
    inserver.sin_port = htonl(sp->s_port);
    
    /* bind socket */
    if ( bind(sock, (struct sockaddr*) &inserver, sizeof(inserver)) < 0 ) {
	saved_errno = errno;
	close(sock);
	errno = saved_errno;
	return -1;
    }

    /* Start accepting connections */
    if (listen (sock, 5)) {
	saved_errno = errno;
	close(sock);
	errno = saved_errno;
	return -1;
    }
    
    return sock;
}


/* Unix domain sockets */

/* SCO Unix lacks Unix domain sockets.
   We simulate them by creating TCP sockets and
   publish the port in the "socket file". Clients willing to connect
   read the port number from that file. */

#ifndef INADDR_LOOPBACK
#define INADDR_LOOPBACK ((u_long) 0x7F000001)
#endif

int
unix_UnixService (server_name)
  char * server_name;
{
  int ret;
  int sock;
#ifdef OS_SCO
  int fd;
  char port_str [20];
  struct sockaddr_in srv_addr;
#else
  struct sockaddr_un srv_addr;
#endif
  int srv_addr_len;
  int saved_errno;
  char socket_file [64];

  /* assure that socket directory exists */
  mkdir (server_path, 0777);
  chmod (server_path, 0777);

  sprintf (socket_file, "%s/%s_%s", server_path, server_name, cuserid (0));
  unlink (socket_file);

  srv_addr_len = sizeof (srv_addr);
  memset (&srv_addr, 0, srv_addr_len);

#ifdef OS_SCO

  srv_addr.sin_family      = AF_INET;
  srv_addr.sin_addr.s_addr = htonl (INADDR_ANY);
  srv_addr.sin_port        = 0;		/* let the system choose a port */

  sock = socket (AF_INET, SOCK_STREAM, 0);

#else	/* ! OS_SCO */

  srv_addr.sun_family      = AF_UNIX;
  strcpy (srv_addr.sun_path, socket_file);

# ifdef OS_AIX
  srv_addr.sun_len =
# endif
  srv_addr_len = (sizeof (srv_addr)
		  - sizeof (srv_addr.sun_path)
		  + strlen (srv_addr.sun_path));

  sock = socket (AF_UNIX, SOCK_STREAM, 0);

#endif	/* ! OS_SCO */

  if ( sock == -1 )
    return -1;

  /* bind address to socket */
  ret = bind (sock, (struct sockaddr *) &srv_addr, srv_addr_len);
  if ( ret == -1 )
    {
      saved_errno = errno;
      close (sock);
      errno = saved_errno;
      return -1;
    }

  /* start accepting connections */
  ret  = listen (sock, 5);
  if ( ret == -1 )
    {
      saved_errno = errno;
      close (sock);
      unlink (socket_file);
      errno = saved_errno;
      return -1;
    }

#ifdef OS_SCO

  /* find out which port number has been assigned by the system */
  ret = getsockname (sock, (struct sockaddr *) &srv_addr, &srv_addr_len);
  if ( ret == -1 )
    {
      saved_errno = errno;
      close (sock);
      unlink (socket_file);
      errno = saved_errno;
      return -1;
    }

  /* write the port number into the socket file for use by clients */
  fd = open (socket_file, O_CREAT | O_TRUNC | O_WRONLY, 0644);
  if ( fd == -1 )
    {
      saved_errno = errno;
      close (sock);
      unlink (socket_file);
      errno = saved_errno;
      return -1;
    }
  sprintf (port_str, "%d\n", ntohs (srv_addr.sin_port));
  write (fd, port_str, strlen (port_str));
  close (fd);
#endif	/* OS_SCO */
  return sock;
}
int
unix_UnixConnect (server_name)
  char * server_name;
{
  int ret;
  int sock;
#ifdef OS_SCO
  int fd;
  int port;
  char port_str [20];
  struct sockaddr_in srv_addr;
#else
  struct sockaddr_un srv_addr;
#endif
  int srv_addr_len;
  int saved_errno;
  char socket_file [64];
  sprintf (socket_file, "%s/%s_%s", server_path, server_name, cuserid (0));
  srv_addr_len = sizeof (srv_addr);
  memset (&srv_addr, 0, srv_addr_len);

#ifdef OS_SCO

  fd = open (socket_file, O_RDONLY, 0);
  if ( fd == -1 )
    {
      return -1;
    }
  port = 0;
  ret = read (fd, port_str, sizeof (port_str) - 1);
  close (fd);
  if ( ret <= 0 )
    {
      errno = ENXIO;
      return -1;
    }
  port_str [ret] = 0;
  port = atoi (port_str);

  srv_addr.sin_family      = AF_INET;
  srv_addr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);
  srv_addr.sin_port        = htons (port);

  sock = socket (AF_INET, SOCK_STREAM, 0);

#else	/* ! OS_SCO */

  srv_addr.sun_family      = AF_UNIX;
  strcpy (srv_addr.sun_path, socket_file);
# ifdef OS_AIX
  srv_addr.sun_len =
# endif
  srv_addr_len = (sizeof (srv_addr)
		  - sizeof (srv_addr.sun_path)
		  + strlen (srv_addr.sun_path));

  sock = socket (AF_UNIX, SOCK_STREAM, 0);
#endif	/* ! OS_SCO */

  if ( sock == -1 )
    return -1;

  ret = connect (sock, (struct sockaddr *) &srv_addr, srv_addr_len);
  if ( ret == -1 )
    {
      saved_errno = errno;
      close (sock);
      errno = saved_errno;
      return -1;
    }

  unix_set_nonblocking_mode (sock);

  return sock;
}
