#ifdef __GNUG__
#pragma implementation
#endif

#include "View.h"

#include "Class.h"
#include "Error.h"
#include "Document.h"
#include "Menu.h"
#include "CmdNo.h"
#include "ClipBoard.h"
#include "PrintManager.h"
#include "Alert_e.h"
#include "Clipper.h"
#include "OrdColl.h"
#include "ETString.h"
#include "Data.h"
#include "FileDialog.h"
#include "Progress.h"

//---------------------------- Controler -----------------------------------------
NewAbstractMetaImpl0(Controler,RefCounted);
 
//---- View --------------------------------------------------------------------
NewAbstractMetaImpl(View, VObject, (TP(clippers), TP(nexthandler), TP(controler), TP(focus)));

View::View(EvtHandler *eh, Rectangle itsExtent, int id) : VObject(itsExtent, id)
{
    clippers= 0;
    ResetFlag(eVObjOpen|eVObjHFixed|eVObjVFixed);
    nexthandler= eh;
    controler=0; // Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
}

View::~View()
{
    if (clippers) {
	SeqCollection *tmp= clippers;
	clippers= 0;
	tmp->ForEach(VObject,Remove)(this);
	delete tmp;
    }
    SetControler(0); // Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
}
 
void View::DoObserve(int i, int c, void *v, Object *op)
{
    if (op == (PrintManager*) gPrintManager) {
	if (Width() != gPrintManager->GetViewRect().Width())
	    SetWidth(gPrintManager->GetViewRect().Width());
    } else
	VObject::DoObserve(i, c, v, op);
}

Document *View::GetDocument()
{
    return (Document*) FindNextHandlerOfClass(Meta(Document));
}

void View::SetExtent(Point newExtent)
{
    if (GetExtent() != newExtent) {
	VObject::SetExtent(newExtent);
	if (clippers) {
	    clippers->ForEach(EvtHandler,Control)(GetId(),cPartExtentChanged,this);
	}
    }
}

void View::ExtentChanged(VObject*)
{
    if (clippers) {
	clippers->ForEach(VObject,ExtentChanged)(this);
    }
}

void View::SetContainer(VObject *v)
{
    VObject::SetContainer(v);
    
    if (clippers == 0)
	clippers= new OrdCollection;
    if (!clippers->FindPtr(v))
	clippers->Add(v);
}

void View::ClearContainer(VObject *clipper)
{
    VObject::ClearContainer(clipper);
    if (clippers) {
	clippers->RemovePtr(clipper);
	if (clippers->Size() <= 0)
	    SafeDelete(clippers);
    }
}

void View::CheckOpen()
{
    ResetFlag(eVObjOpen);
    if (clippers) {
	Iter next(clippers);
	VObject *fp;
    
	while (fp= (VObject*) next()) {
	    if (fp->IsOpen()) {
		SetFlag(eVObjOpen);
		break;
	    }
	}
    }
    if (IsOpen())
	Update();  // hack
}

void View::Open(bool)
{
    //if (mode)
    //    VObject::Open(mode);
    //else
	CheckOpen();
}

//---- drawing

void View::DrawAll(Rectangle r, bool)
{
    if (!gPrinting)
	Update();
    Draw(r);
    if (!TestFlag(eViewNoPrint))
	gPrintManager->ShowPageGrid(r, this);
}

void View::InvalidateRect(Rectangle r)
{
    if (clippers) {
	clippers->ForEach(VObject,InvalidateViewRect)(r);
    }
}

void View::Update()
{
}

Rectangle View::GetViewedRect()
{
    if (focus)
	return focus->GetViewedRect();
    return gRect0;
}

//---- Printing

Rectangle View::NextPageBreak(int, Rectangle r)
{
    return r;
}

void View::PrintAdorn(Rectangle, int gPageNo)
{
    Rectangle r(gPrintManager->GetViewRect());
    Point o;
    Font *f= new_Font(eFontTimes, 12, eFaceBold);
    r.origin.y-= 10;

    byte * p = 0;
    Manager * mgr = (Manager *) FindNextHandlerOfClass (Meta(Manager));
    if ( mgr != 0 )
      p = (byte *) mgr->GetName();
    if ( p == 0 )
      p = (byte *) "Untitled";

    o= f->AdjustString(p, r.NW(), eAdjVBottom, eAdjHLeft);
    GrShowString(f, gInkBlack, o, p);

    p= (byte*) form("Page: %d", gPageNo);
    o= f->AdjustString(p, r.NE(), eAdjVBottom, eAdjHRight);
    GrShowString(f, gInkBlack, o, p);
}

void View::PrintingWillBegin (const Point& paperSize)
{
}

void View::PrintingDidEnd()
{
}

//---- Selection

void View::ShowSelection(bool, bool)
{
}

void View::ClearSelection(bool)
{
}

//---- scrolling

void View::ConstrainScroll(Point*)
{
}

void View::RevealRect(Rectangle revealRect, Point minToSee)
{
    //if (focus)
    //    focus->RevealRect(revealRect, minToSee);
    //else
    if (clippers) {
	clippers->ForEach(VObject,RevealRect)(revealRect, minToSee);
    }
}

void View::RevealAlign(Rectangle revealRect, VObjAlign al)
{
    //if (focus)
    //    focus->RevealAlign(revealRect, al);
    //else
    if (clippers) {
	clippers->ForEach(VObject,RevealAlign)(revealRect, al);
    }
}

void View::Scroll(int mode, Point scroll, bool redraw)
{
    //if (focus)
    //    focus->Scroll(mode, scroll, redraw);
    //else
    if (clippers) {
	clippers->ForEach(VObject,Scroll)(mode, scroll, redraw);
    }
}

//---- event handling

EvtHandler *View::GetNextHandler()
{
    return nexthandler;
}

void View::SetNextHandler(EvtHandler *eh)
{
    nexthandler= eh;
}

//----Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995: renamed Input to DoInput
Command *View::DoInput(Point lp, Token &t, Clipper *vf) //
{
    focus= vf;
    if (t.IsMouseButton() && !(t.Flags & eFlgButDown) && TestFlag(eVObjKbdFocus))
	SetFirstHandler(this);
    //Command *cmd= VObject::Input(lp, t, vf);
    Command *cmd= DispatchEvents(lp, t, vf);
    focus= 0;
    return cmd;
}

//---- BEGIN Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
void View::SetControler(Controler*c)
{
    if(controler)
	controler->RemovedFromView(this);
    AssignRef(&controler,c);
    if(controler)
	controler->AddedToView(this);
}
 
Controler* View::GetControler()
{
    return controler;
}
 
 
Command *View::Input(Point lp, Token &t, Clipper *vf)
{
    focus= vf;
    if (t.IsMouseButton() && !(t.Flags & eFlgButDown) && TestFlag(eVObjKbdFocus))
	SetFirstHandler(this);
    Command *cmd=0;
    if(controler)
	cmd=controler->Input(this, lp, t, vf);
    if(!cmd)
    	//cmd= VObject::Input(lp, t, vf);
	cmd= DispatchEvents(lp, t, vf);
    focus= 0;
    return cmd;
}
//----- END Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
 
Command *View::DoCursorKeyCommand(EvtCursorDir d, Token t)
{
    if (focus)
	focus->Scroll(cPartScrollStep, t.CursorPoint());
    return VObject::DoCursorKeyCommand(d, t);
}

Command *View::DoFunctionKeyCommand(int key, Token t)
{
    int cmd = -1;
    switch (key) {	// this is for Sun L? function keys
    case 12: cmd = cREDO;   break;
    case 14: cmd = cUNDO;   break;
    case 16: cmd = cCOPY;   break;
    case 18: cmd = cPASTE;  break;
    case 20: cmd = cCUT;    break;
    case 19: cmd = cFIND;   break;
    }
    if (cmd > 0)
	return DoMenuCommand(cmd);

    return VObject::DoFunctionKeyCommand(key, t);
}

Command *View::DoOtherFunctionKeyCommand(int key, Token t)
{
    if (focus) {
	switch (key) {
	case eEvtKey_PageUp:
	    focus->Scroll(cPartScrollPage, Point(0,-1));
	    return gNoChanges;
	case eEvtKey_PageDown:
	    focus->Scroll(cPartScrollPage, Point(0,1));
	    return gNoChanges;
	case eEvtKey_Home:
	    focus->Scroll(cPartScrollAbs, gPoint0);
	    return gNoChanges;
	case eEvtKey_End:
	    focus->Scroll(cPartScrollAbs, Point(0,cMaxInt));
	    return gNoChanges;
	}
    }

    int cmd = -1;
    switch (key) {
    case eEvtKey_Redo:  cmd = cREDO;   break;
    case eEvtKey_Undo:  cmd = cUNDO;   break;
    case eEvtKey_Copy:  cmd = cCOPY;   break;
    case eEvtKey_Paste: cmd = cPASTE;  break;
    case eEvtKey_Cut:   cmd = cCUT;    break;
    case eEvtKey_Find:  cmd = cFIND;   break;
    }
    if (cmd > 0)
	return DoMenuCommand(cmd);

    return VObject::DoOtherFunctionKeyCommand(key, t);
}
 
//---- clipboard

Command *View::PasteData(Data*)
{
    return gNoChanges;
}

extern void PrepareTypes();
extern bool EndTypes(Data *data);

bool View::CanPaste(Data *data)
{
    if (data) {
	PrepareTypes();
	Command *cmd = PasteData(data);
	if (cmd && cmd != gNoChanges)
	    delete cmd;
	return EndTypes(data);
    }
    return FALSE;
}

bool View::HasSelection()
{
    return FALSE;
}

bool View::IsReadOnly()
{
    return FALSE;
}

static FileDialog *importDialog;

Command *View::DoMenuCommand(int cmd)
{
    Command *cm;
//---- BEGIN Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
    if(controler) {
	cm=controler->DoMenuCommand(this,cmd);
	if(cm)
	    return cm;
    }
//----- END Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
    
    switch (cmd) {
    case cIMPORT:
	if (importDialog == 0)
	    importDialog= new FileDialog;
    
	if (importDialog->ShowInWindow(eFDImport, GetWindow(), this) == cIdOk) {
	    Data *data= importDialog->GetData();
	    if (CanPaste(data)) {
		gProgress->Start(form("Pasting file: %s", data->ShortName()),
							(int)data->SizeHint());
		cm= PasteData(data);
		gProgress->Stop();
		return cm;
	    }
	    ShowAlert(eAlertCaution, "Cannot paste file (%s)", data->Type().AsString());
	}
	return gNoChanges;

    case cCOPY:
    case cCUT:
	return gNoChanges;

    case cPASTE:
	if (gClipBoard->CanPaste(this))
	{
	    cm = gClipBoard->PasteClipboard(this);
	    if (cm == gNoChanges)
		ShowAlert(eAlertCaution, "Cannot paste clipboard (%s)",
				gClipBoard->GetType()->Type().AsString());
	    return cm;
	}
	return gNoChanges;

    case cPRINT:
	Print();
	return gNoChanges;

    default:
	break;
    }
    return EvtHandler::DoMenuCommand(cmd);
}

void View::DoSetupMenu(Menu *menu)
{
    EvtHandler::DoSetupMenu(menu);
    if (menu->GetId() == cEDITMENU) {
	bool hasSelection = HasSelection();
	if (!IsReadOnly()) {
	    VObject *paste= menu->FindItem(cPASTE);
	    if (paste && gClipBoard->CanPaste(this))
		paste->Enable(TRUE, FALSE);
	    if (hasSelection)
		menu->EnableItem(cCUT);
	    menu->EnableItem(cIMPORT);
	}
	if (hasSelection)
	    menu->EnableItem(cCOPY);
    }
//---- BEGIN Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
    if(controler) {
	controler->DoSetupMenu(this,menu);
    }
//----- END Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
}

void View::InspectorId(char *buf, int sz)
{
    Window *w= GetWindow();
    if (w && w->GetNextHandler()) {
	w->GetNextHandler()->InspectorId(buf, sz-6);
	strn0cpy(buf, form("in %s", buf), sz);
    } else
	VObject::InspectorId(buf, sz);
}
 
