#ifndef _View_h_
#define _View_h_

#ifdef __GNUG__
#  pragma interface
#endif

#include "VObject.h"
#include "Symbol.h"
#include "RefCounted.h"  // Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
class Controler;         // Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995

class SeqCollection;;
class EvtHandler;
class Clipper;
class Document;
class Data;
class Menu;
 
extern bool gPrinting;

enum ViewFlags {
    eViewNoPrint=   BIT(eVObjLast+1),
    eViewLast   =   eVObjLast + 1
};

enum ViewSizeDet {
    eViewHFit   = BIT(0),
    eViewVFit   = BIT(1),
    eViewDefault= 0
};

class View: public VObject {
protected:
    SeqCollection *clippers;
    EvtHandler *nexthandler;
    Clipper *focus;
private:
    Controler *controler;// Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
public:
    MetaDef(View);

    View(EvtHandler *eh, Rectangle itsExtent, int id= -1);
    ~View();

    void SetExtent(Point);
    SeqCollection *GetClipperList()
	{ return clippers; }
    void CheckOpen();
    void Open(bool mode);
    EvtHandler *GetNextHandler();
    void SetNextHandler(EvtHandler *);
    Document *GetDocument();
    Rectangle GetViewedRect();
    void ExtentChanged(VObject *what);

    //---- Drawing ---------------------------------
    void DrawAll(Rectangle, bool);
    void InvalidateRect(Rectangle);

    virtual void Update();

    //---- Printing ---------------------------------
    virtual Rectangle NextPageBreak(int page, Rectangle pr);
    virtual void PrintAdorn(Rectangle, int);
    // paperSize in view coords
    virtual void PrintingWillBegin (const Point& paperSize);
    virtual void PrintingDidEnd();

    void DoObserve(int i, int c, void *v, Object *op);

    //---- Selection --------------------------------
    virtual void ShowSelection(bool redraw= TRUE, bool show= TRUE);
    void HideSelection(bool redraw= TRUE)
	{ ShowSelection(redraw, FALSE); }
    virtual void ClearSelection(bool redraw= TRUE);

    //---- Scrolling --------------------------------
    void SetContainer(VObject *v);
    void ClearContainer(VObject *v);
    virtual void ConstrainScroll(Point*);
    void RevealRect(Rectangle r, Point e);
    void RevealAlign(Rectangle, VObjAlign al= (VObjAlign)(eVObjHLeft+eVObjVTop));
    void Scroll(int mode, Point scroll, bool redraw= TRUE);

    //---- Clipboard --------------------------------
    virtual Command *PasteData(Data*);
    virtual bool HasSelection();
    virtual bool IsReadOnly();
    bool CanPaste(Data *data);

    //---- Menu related methods ---------------------
    Command *DoMenuCommand(int cmd);
    void DoSetupMenu(Menu*);

    //---- Events -----------------------------------
    Command *Input(Point lp, Token &t, Clipper *cl);
    Command *DoCursorKeyCommand(EvtCursorDir, Token);
    Command *DoFunctionKeyCommand(int, Token);
    Command *DoOtherFunctionKeyCommand(int, Token);

    //---- Inspecting -------------------------------
    void InspectorId(char *buf, int sz);
 
//---- BEGIN Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
    // A controler is an event handler that which may handle the Input() of a view.
    // If the controler is 0 or the controler returns 0 on Input() the view will
    // handle its input as before.
    //
    void SetControler(Controler *controler);
    Controler* GetControler();
    // DoInput does the same as Input but does not call the controler.
    // This can be used by controlers to avoid infinte recursions.
    Command *DoInput(Point lp, Token &t, Clipper *cl);
};
 
//---------------------------- Controler -----------------------------------------
 
class Controler : public RefCounted {
public:
    MetaDef(Controler);
    Controler()
	{ };
    virtual Command *Input(View*v, Point lp, Token &t, Clipper *cl)
	{ return 0; }
    virtual void DoSetupMenu(View*v, Menu*m)
	{ }
    virtual Command* DoMenuCommand(View*v, int cmd)
	{ return 0; }
    
protected:
    friend class View;
    virtual void AddedToView(View*v)
	{ }
    virtual void RemovedFromView(View*v)
	{ }
protected:
    ~Controler()// use Ref/Unref insread!
	{ };
 
//----- END Michael Scharf Tue Dec 5 2:44:13 GMT+0100 1995
};

#endif

