#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma implementation
#endif

#include "WindowSystem.h"

#include "Class.h"
#include "CLib.h"
#include "Error.h"
#include "WindowPort.h"
#include "OrdColl.h"
#include "Env.h"
#include "Look.h"
#include "TextItem.h"
#include "CompVObject.h"
#include "HelpHandler.h"
 
extern int gDoubleClickTime;
 
WindowSystem    *gWindowSystem;
int             gDepth= 1;
bool            gColor;
 
static int	gDelayUpdate=0;
 
//---- MenuBorderItem ----------------------------------------------------------
#ifdef __GNUG__
#pragma interface
#endif

class MenuBorderItem: public CompositeVObject {
    Layout *l;
public:
    MenuBorderItem(VObject *inner, VObject *title= 0)
				: CompositeVObject(cIdNone, inner, title, 0)
	{ l= title ? gLook->PopUpMenuLayout() : gLook->PullDownMenuLayout(); }
    ~MenuBorderItem();
    void Draw(Rectangle r);
    Metric GetMinSize();
    void SetOrigin(Point at);
    void SetExtent(Point e);
};
 
inline MenuBorderItem::~MenuBorderItem()
	{ Remove(At(0));  /* avoid deletion of inner */ }
inline Metric MenuBorderItem::GetMinSize()
	{ return l->GetMinSize(this); }	
inline	void MenuBorderItem::SetOrigin(Point at)
	{ VObject::SetOrigin(at); l->SetOrigin(this, at); }
inline      void MenuBorderItem::SetExtent(Point e)
	{ VObject::SetExtent(e); l->SetExtent(this, e); }
inline    void MenuBorderItem::Draw(Rectangle r)
	{ l->Adorn(this, r);  }	
//---- WSInit ------------------------------------------------------------------

extern void RefStandardMatchers();
extern void RefAsciiConverter();
extern void RefEtConverter();
extern void RefXPMConverter();
extern void RefEpsfConverter();
extern void RefMacPictConverter();
extern void RefPbmConverter();
extern void RefGIFConverter();
extern void RefRtfConverter();
extern void RefSunIconConverter();
extern void RefSunRasterConverter();
extern void RefXIconConverter();


void WindowSystem::WSInit()
{
    if (gWindowSystem)
	return;

    InitSystem(0);
    HelpHandler::Init();

    if (! gInMain) {
	fprintf(stderr, "--- InitWindowSystem called before ETInit\n");
	CLib::FastExit(1);
    }

    RefStandardMatchers();
    RefAsciiConverter();
    RefEtConverter();
    RefXPMConverter();
    RefEpsfConverter();
    RefMacPictConverter();
    RefPbmConverter();
    RefGIFConverter();
    RefRtfConverter();
    RefSunIconConverter();
    RefSunRasterConverter();
    RefXIconConverter();

    batch= Env::GetValue("WindowSystem.DoubleBuffer", TRUE);

    gWindowSystem= MakeWindowSystem();

    if (gWindowSystem == 0 || !gWindowSystem->Init()) {
//	fprintf(stderr, "%s: cannot initialize window system\n", gProgname);
	::Fatal("WindowSystem::WSInit", "%s: cannot initialize window system\n", gProgname);
	CLib::FastExit(1);
    }
}

//---- WindowSystem ------------------------------------------------------------
NewMetaImpl(WindowSystem, SysEvtHandler, (TP(allwins),T(set_wait_depth),
    T(lasttime), TB(fullscreen), TP(grabport), T(Clicks), T(screenRect), T(batch)));

bool WindowSystem::batch;
u_long WindowSystem::lasttime;
bool WindowSystem::fullscreen;
WindowPort *WindowSystem::grabport;
Rectangle WindowSystem::screenRect= Rectangle(0, 0, 1152, 900);
Token WindowSystem::lastClick;
int WindowSystem::Clicks;
OrdCollection *WindowSystem::allwins;
WindowColorMap *WindowSystem::wcmap;
int WindowSystem::set_wait_depth;


WindowSystem::WindowSystem() : SysEvtHandler(0)
{
    allwins= new OrdCollection;
    set_wait_depth = 0;
}

WindowSystem::~WindowSystem()
{
    if (gWindowSystem == this)
	gWindowSystem= 0;
    SafeDelete(gFontManager);
    SafeDelete(gInkManager);
    SafeDelete(allwins);
    //SafeDelete(wcmap);
}

bool WindowSystem::Init()
{
    gColor = (gDepth > 1);

    gFontManager= MakeFontManager();
    if (gFontManager == 0) {
	Fatal("Init", "no font manager");
	return FALSE;
    }
    
    if (gFontManager->Init()) {
	Fatal("Init", "initialization of font manager failed");
	return FALSE;
    }

    gInkManager= new InkManager();

    if (gDepth == 1)
	gHighlightColor= gInkXor;
    else {
	if (gDepth == 2)
	    gHighlightColor= new_Grey(0.6);		// light grey
	else
	    gHighlightColor= new_RGBColor(255, 255, 0);	// yellow
	Env::Bind(gHighlightColor, "WindowSystem.HighlightColor");
    }

    MakeLook(Env::GetValue("WindowSystem.Motif", 1));
 
    gDoubleClickTime = Env::GetValue ("WindowSystem.DoubleClickTime", 500);
 
    return TRUE;
}

VObject *WindowSystem::MakeWindowBorder(WindowType wt, VObject *in,
							char *title, bool, bool)
{
    switch (wt) {
    case eWinPulldownMenu:
	return new MenuBorderItem(in);
    case eWinPopupMenu:
	return new MenuBorderItem(in, new TextItem(title));
    default:
	return in;
    }
}

void WindowSystem::SetWait()
{
    SetWait(eCrsHourglass);
}

void WindowSystem::SetWait(GrCursor cursor)
{
    set_wait_depth++;
    if (set_wait_depth == 1) {
	allwins->ForEach(WindowPort,DevSetCursor)(cursor);
    }
}

void WindowSystem::ResetWait()
{
    if (set_wait_depth > 0)
    {
	set_wait_depth--;
	if (set_wait_depth == 0)
	{
	    Iter next(allwins);
	    WindowPort *wp;
	    while (wp= (WindowPort*) next())
		wp->DevSetCursor(wp->GetCursor());
	}
    }
}

void WindowSystem::Update()
{
    if(gDelayUpdate==0) {
    	allwins->ForEach(WindowPort,DevUpdate)();
    }
}

void WindowSystem::BeginDelayUpdate()
{
    gDelayUpdate++;
}

void WindowSystem::EndDelayUpdate()
{
    gDelayUpdate--;
    if(gDelayUpdate<0)
    	Object::Error("EndDelayUpdate","Unbalanced Calls");
    if(gDelayUpdate==0)
    	Update();
}

bool WindowSystem::InDelayUpdate()
{
    return bool(gDelayUpdate!=0);
}

WindowPort *WindowSystem::MakeWindow(Window*, WindowType, GrCursor) 
{
    return 0;
}

void WindowSystem::RemoveWindow(WindowPort *wp)
{
    if (allwins)
	allwins->RemovePtr(wp);
}

void WindowSystem::AddWindow(WindowPort *wp)
{
    if (allwins)
	allwins->Add(wp);
}

FontManager *WindowSystem::MakeFontManager()
{
    return 0;
}

DevBitmap *WindowSystem::MakeDevBitmap(Point, u_short*, u_short, bool, u_short)
{
    return 0;
}

void WindowSystem::graphicDelay(u_int duration)
{
    gSystem->Wait(duration);
}

void WindowSystem::DevBell(long)
{
}

int WindowSystem::DevHaveSelection(Data*)
{
    return 0;
}

void WindowSystem::DevCheckSelection()
{
}
//------------------- GrShowWaitCursor --------------------------
 
void GrShowWaitCursor::Done()
{
}
 
//------------------- GrDelayUpdate --------------------------
 
void GrDelayUpdate::Done()
{
}
