#ifndef _WindowSystem_h_included_
#define _WindowSystem_h_included_

#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma interface
#endif

#include "Port.h"
#include "System.h"
#include "Symbol.h"

enum WindowType {
    eWinPlain,
    eWinStandard,
    eWinDialog,
    eWinPulldownMenu,
    eWinPopupMenu,
    eWinIcon
};

enum WindowFlags {
    eWinDefault     =   0,
    eWinBlock       =   1,
    eWinFixed       =   2,
    eWinCanClose    =   4
};

class VObject;
class Window;
class WindowPort;
class Data;
class WindowColorMap;

//---- WindowSystem ------------------------------------------------------------

class WindowSystem : public SysEvtHandler {
public:
    static bool batch;
    static u_long lasttime;
    static bool fullscreen;
    static WindowPort *grabport;
    static Token lastClick;
    static int Clicks;
    static Rectangle screenRect;
    static OrdCollection *allwins;
    static WindowColorMap *wcmap;
    static int set_wait_depth;

    static void WSInit();

public:
    MetaDef(WindowSystem);
    WindowSystem();
    ~WindowSystem();
    
    void Update();
    
    // can be called nested
    void BeginDelayUpdate();
    void EndDelayUpdate();
    bool InDelayUpdate();
    
    virtual bool Init();
    virtual WindowPort *MakeWindow(Window *bw, WindowType wt, GrCursor curs);
    virtual void RemoveWindow(WindowPort*);
    void AddWindow(WindowPort*);
    virtual FontManager *MakeFontManager();
    virtual DevBitmap *MakeDevBitmap(Point sz, u_short *data, u_short depth,
					bool freedata= FALSE, u_short pad= 2);
    virtual void graphicDelay(unsigned int);
    virtual void DevBell(long duration);
    virtual int DevHaveSelection(Data *data);
    virtual VObject *MakeWindowBorder(WindowType wt,
			VObject *in, char *title, bool block, bool fixed);
    void SetWait();
    void SetWait(GrCursor);
    void ResetWait();
    virtual void DevCheckSelection();
};

extern WindowSystem *gWindowSystem;
extern bool         gColor;
extern int          gDepth;

extern WindowSystem *MakeWindowSystem();

inline void GraphicDelay(unsigned int duration)
    { gWindowSystem->graphicDelay(duration); }

//---- Show wait cursor --------------------------------------------------

class GrShowWaitCursor {
    int dummy;
public:
    GrShowWaitCursor()
	{ gWindowSystem->SetWait(); }
    ~GrShowWaitCursor()
	{ gWindowSystem->ResetWait(); }
    void Done(); // this is a dummy function
};

//---- Delay Update of graphic -------------------------------------------

class GrDelayUpdate {
    int dummy;
public:
    GrDelayUpdate()
	{ gWindowSystem->BeginDelayUpdate(); }
    ~GrDelayUpdate()
	{ gWindowSystem->EndDelayUpdate(); }
    void Done(); // this is a dummy function
};
#endif
