#if defined (__GNUG__) && __GNUG__ >= 2
#  pragma implementation
#endif

#include "XFont.h"
#include "Class.h"
#include "XWinSystem.h"
#include "XBitmap.h"
#include "Env.h"

static struct {
    char *et_font;
    char *x_font;
    char *medium;
    char *bold;
    char *roman;
    char *italic;
    bool  is_iso8859_1;
} fontMap[] =
{
  { "times",  "times",			"medium", "bold",   "r", "i", TRUE },
  { "helvet", "helvetica",		"medium", "bold",   "r", "o", TRUE },
  { "courie", "courier",		"medium", "bold",   "r", "o", TRUE },
  { "symbol", "symbol",			"medium", "medium", "r", "r", FALSE },
  { "chicag", "lucida",			"medium", "bold",   "r", "i", TRUE },
  { "avgard", "itc avant garde",	"medium", "demi",   "r", "o", TRUE },
  { "bookmn", "bookman",		"light",  "demi",   "r", "i", TRUE },
  { "ceslbk", "new century schoolbook",	"medium", "bold",   "r", "i", TRUE },
  { "nhelve", "helvetica",		"medium", "bold",   "r", "o", TRUE },
  { "palati", "palatino",		"medium", "bold",   "r", "i", TRUE },
  { "zchanc", "zapfchancery",		"medium", "medium", "i", "i", TRUE },
  { "zdingb", "zapfdingbats",		"medium", "medium", "r", "r", FALSE }
};

//---- XFontManager ----------------------------------------------------------
Font *
XFontManager::MakeFont(FontFamily *ff, int ps, GrFace fc, bool dflt)
{
    int fid= ff->Fid();
    char xname[500];
    char *nm;
    char *weight;
    char *slant;
    char *encoding;
    XFontStruct *xf;
    static char no_font_msg[] = "cannot load font \"%s\"\n";

    // try ET++ fonts
    sprintf ( xname, "%s%02d%s", fontMap[fid].et_font, ps,
			gFontManager->StyleString(fc) );
    if ( xf = XLoadQueryFont (XWinSystem::display, xname) )
	return new XServerFont (ff, ps, fc, xf);

    // try standard X fonts

    nm = fontMap[fid].x_font;
    weight = (fc & eFaceBold) ? fontMap[fid].bold : fontMap[fid].medium;
    slant = (fc & eFaceItalic) ? fontMap[fid].italic : fontMap[fid].roman;
    encoding = fontMap[fid].is_iso8859_1 ? "iso8859-1" : "*-*";

    // try point size (pixel size * 10)
    sprintf ( xname, "-*-%s-%s-%s-*-*-*-%d0-*-*-*-*-%s",
			nm, weight, slant, ps, encoding );
    if ( xf = XLoadQueryFont (XWinSystem::display, xname) )
	return new XServerFont(ff, ps, fc, xf);

    // try exact pixel size
    sprintf ( xname, "-*-%s-%s-%s-*-*-%d-*-*-*-*-*-%s",
			nm, weight, slant, ps, encoding );
    if ( xf = XLoadQueryFont (XWinSystem::display, xname) )
	return new XServerFont (ff, ps, fc, xf);

    if (dflt) {
	char *otherfont;
	Warning("MakeFont", no_font_msg, xname);

	// try some default fonts
	if (fid != eFontCourier) {
	    if (fid == eFontChicago)
		otherfont = "variable";
	    else
		otherfont = "8x13";
	    if ((xf= XLoadQueryFont(XWinSystem::display, otherfont)))
		return new XServerFont(ff, ps, fc, xf);
	    Warning("MakeFont", no_font_msg, otherfont);
	}

	// last resort
	otherfont = "fixed";
	if ((xf= XLoadQueryFont(XWinSystem::display, otherfont)))
	    return new XServerFont(ff, ps, fc, xf);
	Warning("MakeFont", no_font_msg, otherfont);
    }
    return NULL;
}

//---- XServerFont -------------------------------------------------------------
//NewMetaImpl(XServerFont,Font,(T(fUnderlinePos),T(fUnderlineWidth)));

XServerFont::XServerFont(FontFamily *ff, int ps, GrFace fc, XFontStruct *xFont)
			: Font(ff, ps, fc)
{
    if (xFont) {
	register int i, j;
	register byte w;
	
	if (size == 0) {
	    u_long val;
	    if (XGetFontProperty(xFont, XA_POINT_SIZE, &val))
		size= (short) val/10;
	    else
		size= 12;
	}
		
	xfid = xFont->fid;
	ht = xFont->max_bounds.ascent;
	bs = xFont->max_bounds.descent;
	ils = ht+bs;
	int env_ils = Env::GetValue("*.LineSpacing", 1);
	if (env_ils < 0)
	    env_ils = 0;
	else if (env_ils > 10)
	    env_ils = 10;
	ils += env_ils;
	
	if (xFont->per_char && xFont->default_char >= 0 && xFont->default_char < 256)
	    w= (byte) xFont->per_char[xFont->default_char].width;
	else
	    w= xFont->max_bounds.width;
    
	for (i= 0; i < 256; i++)
	    cw[i]= w;
	
	if (xFont->per_char) {
	    if (xFont->min_byte1 == 0 && xFont->max_byte1 == 0)
		for (i= xFont->min_char_or_byte2, j= 0; i <= xFont->max_char_or_byte2; i++, j++)
		    cw[i]= (byte) xFont->per_char[j].width;
	    else
		fprintf(stderr, "%s: cannot handle two-byte fonts\n", gProgname);
	}
	if (!XGetFontProperty(xFont, XA_UNDERLINE_POSITION, &fUnderlinePos)) {
	    fUnderlinePos = bs/2;
	}
	if (!XGetFontProperty(xFont, XA_UNDERLINE_THICKNESS, &fUnderlineWidth)) {
	    fUnderlineWidth = 1;
	}
	XFreeFontInfo(NULL, xFont, 1);
    }
}

DevBitmap *XServerFont::CharAsBitmap(byte c, Point *offset)
{ 
    XGCValues gcv;
    XCharStruct cc;
    Pixmap pm;
    GC gc;
    Point e;
    int dir, dd, aa;
    
    XQueryTextExtents(XWinSystem::display, GetId(), (char*)&c, 1, &dir, &aa, &dd, &cc);
    e.x= cc.rbearing - cc.lbearing;
    e.y= cc.ascent + cc.descent;
    if (e.x <= 0)
	e.x= 1;
    if (e.y <= 0)
	e.y= 1;
    offset->x= cc.lbearing;
    offset->y= cc.ascent;
    pm= XWinSystem::CreatePixmap(e, 1);
    gcv.background= 0;
    gcv.foreground= 0;
    gc= XWinSystem::CreateGC(pm, GCForeground|GCBackground, &gcv);
    XFillRectangle(XWinSystem::display, pm, gc, 0, 0, e.x+1, e.y+1);
    XSetForeground(XWinSystem::display, gc, 1);
    XSetFont(XWinSystem::display, gc, GetId());
    XDrawString(XWinSystem::display, pm, gc, -cc.lbearing, cc.ascent, (const char*) &c, 1);
    XFreeGC(XWinSystem::display, gc);
    
    XImage *im= XGetImage(XWinSystem::display, pm, 0, 0, e.x, e.y,
						    (u_long) AllPlanes, ZPixmap);
    XFreePixmap(XWinSystem::display, pm);

    if (im)
	return new XBitmap(e, im);
    return 0;
}
