#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma implementation
#endif

#include "XWinSystem.h"
#include "XWindowPort.h"
#include "XFont.h"
#include "XBitmap.h"
#include "XColorMap.h"

#include <X11/cursorfont.h>

#include "Class.h"
#include "System.h"
#include "ETString.h"
#include "Text.h"
#include "MemBuf.h"
#include "OrdColl.h"
#include "Data.h"
#include "ClipBoard.h"
#include "Symbol.h"
#include "Env.h"
#include "ETCType.h"


XAtom
	cDeleteWindow ("WM_DELETE_WINDOW", TRUE),
	cOpenLookDecorAdd ("_OL_DECOR_ADD"),
	cOpenLookDecorDel ("_OL_DECOR_DEL"),
	cOpenLookResize ("_OL_DECOR_RESIZE"),
	cOpenLookWinAttr ("_OL_WIN_ATTR"),
	cOpenLookWinCmd ("_OL_WT_CMD"),
	cOpenLookWinBase ("_OL_WT_BASE"),
	cMotifWmHints ("_MOTIF_WM_HINTS"),
	cDecWmMoved ("WM_MOVED"),
	cAtomWmState ("WM_STATE", TRUE);

static XAtom
	cPRIMARY (XA_PRIMARY),
	cCLIPBOARD ("CLIPBOARD", TRUE),
	cTIMESTAMP ("TIMESTAMP", TRUE),
	cLENGTH ("LENGTH", TRUE),
	cTEXT ("TEXT", TRUE),
	cTARGETS ("TARGETS", TRUE),
	cET_CLIPBOARD ("_ET_CLIPBOARD", TRUE);


//---- XAtom -------------------------------------------------------

XAtom::XAtom(char *nm, bool cr)
{
    name = nm;
    create = cr;
    valid = FALSE;
    atom = None;
}

XAtom::XAtom(Atom a)
{
    name = NULL;
    valid = TRUE;
    atom = a;
}

XAtom::operator Atom()
{
    if (!valid) {
	atom = XInternAtom(XWinSystem::display, name, ! create);
	valid = TRUE;
    }
    return atom;
}

char *XAtom::Name()
{
    if (name == NULL)
	name = XGetAtomName(XWinSystem::display, atom);
    return name;
}


//---- XSelectionData ----------------------------------------------------

class XSelectionData : public StreamBufData {
    char *buf;
    int len;
    XTime time_stamp;
public:
    XSelectionData(char *buf, int len, XTime time_stamp);
    ~XSelectionData();

    XTime TimeStamp() const
	{ return time_stamp; }
    int Length() const
	{ return len; }

    StreamBuf *GetStreamBuf(bool forwriting);
};

XSelectionData::XSelectionData(char *b, int l, XTime ts)
		: StreamBufData(NULL, cDocTypeAscii)
{
    buf = b;
    len = l;
    time_stamp = ts;
}

XSelectionData::~XSelectionData()
{
    delete buf;
}

StreamBuf *XSelectionData::GetStreamBuf(bool)
{
    return sb = new MemBuf(len, buf, len, FALSE);
}

//-----------------------------------------------------------------------


XSelectionData *XWinSystem::ReadSelection(XAtom& selection, XTime current_time)
{
    XEvent ev;
    XSelectionEvent *se = &ev.xselection;
    XWindow owner;
    int status;
    XTime selection_time = 0;
    Atom type_return;
    int actual_format_return;
    unsigned char *prop_return = NULL;
    unsigned long nitems_return, bytes_after_return;

    owner = XGetSelectionOwner(display, selection);
    if (owner == None || owner == selection_window)
	return NULL;

    /* first get timestamp of selection */

    XConvertSelection (display, selection, cTIMESTAMP,
		       cET_CLIPBOARD, selection_window, current_time);
    XFlush (display);
    // wait at most 5 seconds for a SelectionNotify event
    if ( ! XCheckTypedEvent (display, SelectionNotify, &ev) )
      {
	gSystem->CanRead (ConnectionNumber (display), 5000);
	if ( ! XCheckTypedEvent (display, SelectionNotify, &ev) )
	  return 0;
      }

    if (se->property == None)	/* no timestamp ? */
	return NULL;

    status = XGetWindowProperty(display,
		    selection_window, se->property,
		    0, 1, XTrue,
		    AnyPropertyType,
		    &type_return, &actual_format_return,
		    &nitems_return, &bytes_after_return,
		    &prop_return);

    if (status == Success)
    {
	if ((type_return == cTIMESTAMP || type_return == XA_INTEGER)
		&& actual_format_return == 32
		&& nitems_return >= 1)
	    selection_time = *(XTime *)prop_return;
	XFree((char*)prop_return);
    }

    if (selection_time <= last_selection_time)
	return NULL;

    // We've got the timestamp of the selection and it is newer than the last
    // selection. Now fetch the actual contents of the selection.
    XConvertSelection (display, selection, XA_STRING,
		       cET_CLIPBOARD, selection_window, current_time);
    XFlush (display);
    // wait at most 5 seconds for a SelectionNotify event
    if ( ! XCheckTypedEvent (display, SelectionNotify, &ev) )
      {
	gSystem->CanRead (ConnectionNumber (display), 5000);
	if ( ! XCheckTypedEvent (display, SelectionNotify, &ev) )
	  return 0;
      }

    if (se->property == None)
	return NULL;

    long offset = 0;
    long readsize = 1024;	// read in 4096 byte chunks
    char *buf = NULL;
    char *ptr;

    prop_return = NULL;
    do
    {
	if (prop_return)
	{
	    XFree((char*)prop_return);
	    prop_return = NULL;
	}
	status = XGetWindowProperty(display,
			selection_window, se->property,
			offset, readsize, XTrue,
			AnyPropertyType,
			&type_return, &actual_format_return,
			&nitems_return, &bytes_after_return,
			&prop_return);

	if (status == Success)
	{
	    if (type_return == XA_STRING || type_return == cTEXT)
	    {
		if (buf == NULL)
		{
		    if ((nitems_return + bytes_after_return) == 0)
			break;

		    buf = new char[nitems_return + bytes_after_return];
		    ptr = buf;
		}
		memcpy(ptr, prop_return, nitems_return);
		offset += readsize;
		ptr += nitems_return;
	    }
	    else if ( type_return != None )
	    {
		char *nm = XGetAtomName (XWinSystem::display, type_return);
		Warning ("ReadSelection", "cannot handle selection of type \"%s\"\n", nm);
		XFree (nm);
		break;
	    }
	}
	else
	{
	    Warning("ReadSelection", "cannot read \"%s\" selection\n",
			selection.Name());
	    break;
	}

    } while (bytes_after_return > 0);

    if (prop_return)
	XFree((char*)prop_return);

    if (buf)
	return new XSelectionData(buf, ptr - buf, selection_time);

    return NULL;
}

XSelectionData *XWinSystem::ReadCutBuffer0()
{
    XSelectionData *sel = NULL;
    if (cut_buffer0_change_time > 0 &&
	cut_buffer0_change_time > last_selection_time)
    {
	int length;
	char *buffer = XFetchBuffer(display, &length, 0);
	if (length > 0 && buffer && (length = strlen(buffer)) > 0)
	{
	    char *newbuf = new char[length];
	    memcpy(newbuf, buffer, length);
	    sel = new XSelectionData(newbuf, length, cut_buffer0_change_time);
	}
	if (buffer)
	    XFree(buffer);
    }

    return sel;
}

void XWinSystem::DevCheckSelection()
{
    if (own_selection)
	return;

    XTime current_time = GetCurrentServerTime();
    XSelectionData *sel[3];
    XSelectionData *the_selection = NULL;

    sel[0] = ReadCutBuffer0();
    sel[1] = ReadSelection(cPRIMARY, current_time);
    sel[2] = ReadSelection(cCLIPBOARD, current_time);

    int i;
    for (i = 0; i < 3; i++)
	if (sel[i] && sel[i]->TimeStamp() >= last_selection_time
	    && (the_selection == NULL
		|| sel[i]->TimeStamp() > the_selection->TimeStamp()))
	the_selection = sel[i];

    if (the_selection)
    {
	if (the_selection->Length() > 0)
	{
	    gClipBoard->SetData(the_selection);
	    last_selection_time = the_selection->TimeStamp();
	}
	else
	    the_selection = NULL;
    }

    for (i = 0; i < 3; i++)
	if (sel[i] && sel[i] != the_selection)
	    delete sel[i];
}

void XWinSystem::OwnSelection(XAtom& selection, XTime time_stamp)
{
    XWindow owner;
    XSetSelectionOwner(display, selection, selection_window, time_stamp);
    owner = XGetSelectionOwner(display, selection);
    if (owner != selection_window)
	Warning("OwnSelection", "cannot become owner of \"%s\" selection\n",
		selection.Name());
}

void XWinSystem::LostSelection()
{
    own_selection = FALSE;
    last_selection_time = None;
}

int XWinSystem::DevHaveSelection(Data *data)
{
    XTime current_time = GetCurrentServerTime();
    OwnSelection(cPRIMARY, current_time);
    OwnSelection(cCLIPBOARD, current_time);
    own_selection = TRUE;
    last_selection_time = current_time;
    return 0;
}

void XWinSystem::SendClipBoard(XSelectionRequestEvent *sre)
{
    XEvent ev;
    XSelectionEvent *sev = &ev.xselection;
    XTime current_time = GetCurrentServerTime();

    Data *data = own_selection ? gClipBoard->GetType() : NULL;
    Object *text = data ? data->AsObject(Meta(Text)) : NULL;
    char *buf = text ? text->AsString() : NULL;
    int length = buf ? strlen(buf) : 0;
    bool have_data = (length > 0);

    SafeDelete(text);

    sev->type	   = SelectionNotify;
    sev->display   = sre->display;
    sev->requestor = sre->requestor;
    sev->selection = sre->selection;
    sev->target	   = sre->target;
    sev->property  = sre->property;
    sev->time	   = sre->time;

    if (sev->property == None)
	sev->property = sev->target;

    if (sev->time != CurrentTime
	&& (sev->time < last_selection_time || sev->time > current_time))
	have_data = FALSE;

    if (have_data)
    {
	if (sev->target == XA_STRING || sev->target == cTEXT)
	{
	    XChangeProperty(display, sev->requestor, sev->property,
				XA_STRING, 8*sizeof(char),
				PropModeReplace,
				(unsigned char *) buf, length);
	    XSendEvent(display, sev->requestor, XFalse, 0, &ev);
	    XFlush(display);
	}
	else if (sev->target == cTIMESTAMP)
	{
	    XChangeProperty(display, sev->requestor, sev->property,
				cTIMESTAMP, 8*sizeof(XTime),
				PropModeReplace,
				(unsigned char *) &last_selection_time, 1);
	    XSendEvent(display, sev->requestor, XFalse, 0, &ev);
	    XFlush(display);
	}
	else if (sev->target == cLENGTH)
	{
	    XChangeProperty(display, sev->requestor, sev->property,
				XA_INTEGER, 8*sizeof(length),
				PropModeReplace,
				(unsigned char *) &length, 1);
	    XSendEvent(display, sev->requestor, XFalse, 0, &ev);
	    XFlush(display);
	}
	else if (sev->target == cTARGETS)
	{
	    Atom targets[5];
	    targets[0] = XA_STRING;
	    targets[1] = cTEXT;
	    targets[2] = cTIMESTAMP;
	    targets[3] = cLENGTH;
	    targets[4] = cTARGETS;
	    XChangeProperty(display, sev->requestor, sev->property,
				XA_ATOM, 8*sizeof(Atom),
				PropModeReplace,
				(unsigned char *) targets, 5);
	    XSendEvent(display, sev->requestor, XFalse, 0, &ev);
	    XFlush(display);
	}
	else
	{
	    have_data = FALSE;
	}
    }

    if (!have_data)
    {
	sev->property = None;
	XSendEvent(display, sev->requestor, XFalse, 0, &ev);
	XFlush(display);
    }
	
    SafeDelete(buf);
}

XTime XWinSystem::GetCurrentServerTime()
{
    XEvent ev;
    XChangeProperty(display, selection_window,
				XA_WM_NAME, XA_STRING, 8,
				PropModeAppend, (unsigned char *)"", 0);
    XWindowEvent(display, selection_window, PropertyChangeMask, &ev);
    lasttime = ev.xproperty.time;
    return ev.xproperty.time;
}

// --- X error handler ------------------------------------

static int
NonFatalErrorHandler (XDisplay * display, XErrorEvent * ev)
{
  char error_text [256];
  char request_code_text [80];
  char request_code [10];

  error_text[0] = 0;
  XGetErrorText (display, ev->error_code, error_text, sizeof (error_text));

  sprintf (request_code, "%d", ev->request_code);
  request_code_text[0] = 0;
  XGetErrorDatabaseText (display, "XRequest", request_code, "",
			 request_code_text, sizeof (request_code_text));

  fprintf (stderr,
	   "%s: X protocol error\n"
	   "  Error code:      %d, %s\n"
	   "  Request serial:  %lu\n"
	   "  Request opcode:  %d, %s\n"
	   "  Minor opcode:    %d\n"
	   "  Resource id:     0x%lx\n",
	   gProgname,
	   ev->error_code, error_text,
	   ev->serial,
	   ev->request_code, request_code_text,
	   ev->minor_code,
	   ev->resourceid);

  return 0;
}

//---- XWinSystem -----------------------------------------------------------

NewMetaImpl0(XWinSystem,WindowSystem);

XCursor XWinSystem::CurMap[cMaxCursors];
XDisplay *XWinSystem::display;
XWindow XWinSystem::rootwin;
int XWinSystem::screen;
u_long XWinSystem::black, XWinSystem::white;
Visual *XWinSystem::dfltvisual;
Pixmap XWinSystem::batchPixmap;

unsigned int XWinSystem::modMask;
bool XWinSystem::useDeleteKey;

XrmDatabase XWinSystem::resource_db;
bool XWinSystem::explicit_input_focus;

void
WindowSystem_ReadEnvVars()
{
  XWinSystem::useDeleteKey = Env::GetValue ("WindowSystem.UseDeleteKey", FALSE);

  switch ( Env::GetValue ("*.UseModifierKey", 1) )
    {
      default:
      case 1:  XWinSystem::modMask = Mod1Mask; break;
      case 2:  XWinSystem::modMask = Mod2Mask; break;
      case 3:  XWinSystem::modMask = Mod3Mask; break;
      case 4:  XWinSystem::modMask = Mod4Mask; break;
      case 5:  XWinSystem::modMask = Mod5Mask; break;
    }
} 

WindowSystem *MakeWindowSystem()
{
    return new XWinSystem;
}

XWinSystem::XWinSystem()
{
}

bool XWinSystem::Init()
{
    display = XOpenDisplay(NULL);
    if (display == NULL) {
	fprintf(stderr, "%s: cannot open display \"%s\"\n",
			gProgname, XDisplayName(NULL));
	return FALSE;
    }

    XSetErrorHandler (NonFatalErrorHandler);

    XForceScreenSaver (display, ScreenSaverReset);

    screen= DefaultScreen(display);
    rootwin= RootWindow(display, screen);
    dfltvisual= DefaultVisual(display, screen);
    gDepth= DefaultDepth(display, screen);
    
    black= BlackPixel(display, screen);
    white= WhitePixel(display, screen);
    screenRect.extent.x= DisplayWidth(display, screen);
    screenRect.extent.y= DisplayHeight(display, screen);

    switch (dfltvisual->c_class)
    {
	case StaticGray:
	case StaticColor:
	case GrayScale:
	case PseudoColor:
	    wcmap = new XColorIndexMapper(DefaultColormap(display, screen),
					  dfltvisual->map_entries);
	    break;

	case DirectColor:
	case TrueColor:
	    wcmap = new XColorDirectMapper(dfltvisual);
	    break;
    }

    SetResourceId(ConnectionNumber(display));
    gSystem->AddFileInputHandler(this);
    bellduration= -1;
    
    own_selection = FALSE;
    last_selection_time = cut_buffer0_change_time = 0;
    selection_window = XCreateWindow(display, rootwin, 0, 0, 1, 1, 0,
					gDepth, InputOutput,
					(Visual*) CopyFromParent, 0, 0);
    XSelectInput(display, rootwin, PropertyChangeMask);
    XSelectInput(display, selection_window, PropertyChangeMask);

    WindowSystem_ReadEnvVars();
    InitResourceDatabase();

    return WindowSystem::Init();
}

XWinSystem::~XWinSystem()
{
  SafeDelete (gFontManager);
  SafeDelete (gInkManager);

  if ( selection_window )
    XDestroyWindow (display, selection_window);

  if ( batchPixmap )
    {
      XFreePixmap (display, batchPixmap);
      batchPixmap = 0;
    }

  if ( resource_db != 0 )
    {
      XrmDestroyDatabase (resource_db);
    }

  if ( display )
    {
      XCloseDisplay (display);
      display = 0;
    }
}

void
XWinSystem::InitResourceDatabase()
{
  char * res_string;

  XrmInitialize();

  res_string = XResourceManagerString (display);
  if ( res_string != 0 )
    {
      XrmDatabase dpy_db = XrmGetStringDatabase (res_string);
      if ( dpy_db != 0 )
	XrmMergeDatabases (dpy_db, &resource_db);
    }

#if defined (XlibSpecificationRelease) && XlibSpecificationRelease >=5
  res_string = XScreenResourceString (DefaultScreenOfDisplay (display));
  if ( res_string != 0 )
    {
      XrmDatabase screen_db = XrmGetStringDatabase (res_string);
      if ( screen_db != 0 )
	XrmMergeDatabases (screen_db, &resource_db);
      XFree (res_string);
    }
#endif

  explicit_input_focus = FALSE;

  if ( resource_db != 0 )
    {
      XrmValue res_value;
      char * res_type;

      if ( cMotifWmHints )
	{
	  if ( XrmGetResource (resource_db, "Mwm.keyboardFocusPolicy", "KeyboardFocusPolicy",
			       &res_type, &res_value) )
	    {
	      if ( StrCmp ((byte *) res_type, (byte *) "String", -1, upcase) == 0
		  && StrCmp ((byte *) res_value.addr, (byte *) "explicit", -1, upcase) == 0 )
		explicit_input_focus = TRUE;
	    }
	}
      else if ( cOpenLookWinAttr )
	{
	  if ( XrmGetResource (resource_db, "OpenWindows.SetInput", "SetInput",
			       &res_type, &res_value) )
	    {
	      if ( StrCmp ((byte *) res_type, (byte *) "String", -1, upcase) == 0
		  && StrCmp ((byte *) res_value.addr, (byte *) "select", -1, upcase) == 0 )
		explicit_input_focus = TRUE;
	    }
	}
    }
}

bool
XWinSystem::HasInterest (SysEventCodes code)
{
  {
    WindowSystem::allwins->ForEach (XWindowPort,TestInput)();
  }

  Update();
  XFlush (display);
  while ( XEventsQueued (display, QueuedAlready) > 0 )
    {
      Token t;
      XWindowPort * port;
      XEvent xe;
      XNextEvent (display, &xe);
      if ( port = ProcessEvent (xe, &t) )
	port->InputNotify (&t);
    }

  return TRUE;
}

void
XWinSystem::Notify (SysEventCodes, int)
{
  XEventsQueued (display, QueuedAfterReading);
  while ( XEventsQueued (display, QueuedAlready) > 0 )
    {
      Token t;
      XWindowPort * port;
      XEvent xe;
      XNextEvent (display, &xe);
      if ( port = ProcessEvent (xe, &t) )
	port->InputNotify (&t);
    }
  Update();
  XFlush (display);
}

XWindowPort *
XWinSystem::NextToken (Token * t, int timeout, XWindowPort * port)
{
  int events_queued = XEventsQueued (display, QueuedAfterReading);
  if ( events_queued <= 0 )
    {
      Update();
      events_queued = XEventsQueued (display, QueuedAfterFlush);
    }

  if ( events_queued <= 0 )
    {
      if ( timeout >= 0 )	// block
	{
	  timeout -= 20;
	  if ( timeout < 0 )
	    timeout = 0;
	  if ( gSystem->CanRead (ConnectionNumber (display), timeout) )
	    {
	      if ( t )
		{
		  t->Code = eEvtNone;
		  t->Flags = 0;
		}
	      return port;
	    }
	}
    }

  XEvent xe;
  XNextEvent (display, &xe);
  return ProcessEvent (xe, t);
}

XWindowPort *
XWinSystem::ProcessEvent (XEvent & xe, Token * t)
{
  if ( xe.type == MappingNotify )
    {
      XRefreshKeyboardMapping (&xe.xmapping);
    }
  else if ( xe.xany.window == rootwin )
    {
      if ( xe.type == PropertyNotify && xe.xproperty.atom == XA_CUT_BUFFER0 )
	{
	  if ( xe.xproperty.state == PropertyNewValue )
	    cut_buffer0_change_time = xe.xproperty.time;
	  else	/* property deleted */
	    cut_buffer0_change_time = None;
	}
    }
  else if ( xe.xany.window == selection_window )
    {
      switch ( xe.type )
	{
	case SelectionClear:
	  LostSelection();
	  break;
	case SelectionRequest:
	  SendClipBoard (&xe.xselectionrequest);
	  break;
	}
    }
  else
    {
      XWindowPort * port = XWinSystem::Find (xe);
      if ( port != 0 && port->MapEvent (xe, t) )
	return port;
    }

  return 0;
}

XWindowPort *
XWinSystem::Find (XEvent & ev)
{
  XWindowPort * port;
  Iter next (allwins);
  while ( port = (XWindowPort *) next() )
    if ( port->GetWinId() == ev.xany.window )
      return port;
  return 0;
}

WindowPort *XWinSystem::MakeWindow(Window *bw, WindowType wt, GrCursor curs)
{
    return new XWindowPort(bw, wt, curs);
}

VObject *XWinSystem::MakeWindowBorder(WindowType wt, VObject *in,
					    char *title, bool block, bool fixed)
{
    return WindowSystem::MakeWindowBorder(wt, in, title, block, fixed);
}

XCursor XWinSystem::UseCursor(GrCursor c)
{
    if (CurMap[c] == 0) {
	unsigned int xc;

	switch (c) {
	case eCrsNone:		xc= XC_xterm; break;
	case eCrsArrow:		xc= XC_left_ptr; break;
	case eCrsBoldArrow:	xc= XC_left_ptr; break;
	case eCrsCross:		xc= XC_crosshair; break;
	case eCrsBoldCross:	xc= XC_cross; break;
	case eCrsUpArrow:	xc= XC_sb_up_arrow; break;
	case eCrsDownArrow:	xc= XC_sb_down_arrow; break;
	case eCrsLeftArrow:	xc= XC_sb_left_arrow; break;
	case eCrsRightArrow:	xc= XC_sb_right_arrow; break;
	case eCrsHourglass:	xc= XC_watch; break;
	case eCrsPen:		xc= XC_pencil; break;
	case eCrsMoveHand:	xc= XC_hand2; break;
	case eCrsHand:		xc= XC_hand2; break;
	case eCrsCrossHair:	xc= XC_crosshair; break;
	case eCrsMoveStretch:	xc= XC_hand2; break;
	case eCrsIBeam:		xc= XC_xterm; break;
	case eCrsLeftRightArrow: xc= XC_sb_h_double_arrow; break;
	case eCrsUpDownArrow:	xc= XC_sb_v_double_arrow; break;
	case eCrsHelp:		xc= XC_question_arrow; break;
	default:		xc= XC_left_ptr; break;
	}
	CurMap[c]= XCreateFontCursor(display, xc);
    }
    return CurMap[c];
}

FontManager *XWinSystem::MakeFontManager()
{
    return new XFontManager();
}

DevBitmap *XWinSystem::MakeDevBitmap(Point sz, u_short *data,
				u_short depth, bool freedata, u_short pad)
{
    return new XBitmap(sz, data, depth, freedata, pad);
}

void XWinSystem::DevBell(long millisec)
{
    XBell(display, 0);
}

void XWinSystem::graphicDelay(u_int millisec)
{
    XSync(display, XFalse);
    gSystem->Wait(millisec);
}
