#ifndef _XWinSystem_h_
#define _XWinSystem_h_

#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma interface
#endif

#include "X11.h"
#include "WindowSystem.h"

class XWindowPort;
class XSelectionData;

const int cMaxCursors= 30;

//---- XAtom -------------------------------------------------------

class XAtom {
    char *name;
    bool valid, create;
    Atom atom;
public:
    XAtom(char *nm, bool create= FALSE);
    XAtom(Atom atom);
    operator Atom();
    char *Name();
};

extern XAtom
    cDeleteWindow,
    cOpenLookDecorAdd,
    cOpenLookDecorDel,
    cOpenLookResize,
    cOpenLookWinAttr,
    cOpenLookWinCmd,
    cOpenLookWinBase,
    cMotifWmHints,
    cDecWmMoved,
    cAtomWmState;

//---- XWinSystem -----------------------------------------------------

class XWinSystem: public WindowSystem {
public:
    long bellduration;

private:
    XWindow selection_window;
    XTime last_selection_time,
	  cut_buffer0_change_time;
    bool own_selection;			// whether we currently own the selection

public:
    static XCursor CurMap[cMaxCursors];
    static XDisplay *display;
    static XWindow rootwin;
    static int screen;
    static u_long black, white;
    static Visual *dfltvisual;
    static Pixmap batchPixmap;

    static unsigned int modMask; 	// Mask for used Modifier
    static bool useDeleteKey;

    static XrmDatabase resource_db;
    static bool explicit_input_focus;

public:
    MetaDef(XWinSystem);
    XWinSystem();
    ~XWinSystem();

    bool Init();

    // --- selection handling
    void DevCheckSelection();
    int  DevHaveSelection(Data *);

    // --- other
    static unsigned int GetModMask()
      { return modMask; }
    static bool UseDeleteKey()
      { return useDeleteKey; }
    static bool ExplicitInputFocus()
      { return explicit_input_focus; }

private:
    XSelectionData *ReadSelection(XAtom& selection, XTime current_time);
    XSelectionData *ReadCutBuffer0();
    void OwnSelection(XAtom& selection, XTime when);
    void LostSelection();
    void SendClipBoard(XSelectionRequestEvent *);
    XTime GetCurrentServerTime();
    void InitResourceDatabase();

public:
    bool HasInterest(SysEventCodes code);
    void Notify(SysEventCodes code, int val);
    XWindowPort *NextToken(Token *t, int timeout= -1, XWindowPort *p= 0);
    XWindowPort *ProcessEvent(XEvent &, Token *);

    static XWindowPort *Find(XEvent &ev);
    WindowPort *MakeWindow(Window *bw, WindowType wt, GrCursor curs);
    VObject *MakeWindowBorder(WindowType wt, VObject *in,
				char *title, bool block, bool fixed);
    FontManager *MakeFontManager();
    DevBitmap *MakeDevBitmap(Point sz, u_short *data, u_short depth,
				bool freedata, u_short pad);
    void graphicDelay(unsigned int);
    void DevBell(long duration);

    static Pixmap CreatePixmap(Point e, int d)
	{ return XCreatePixmap(display, rootwin, e.x, e.y, d); }
    static GC CreateGC(Drawable d, u_long valuemask, XGCValues *values)
	{ return XCreateGC(display, d, valuemask, values); }
    static XCursor UseCursor(GrCursor c);
};

extern void WindowSystem_ReadEnvVars (void);

#endif
