#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma implementation
#endif

#include "XWindowPort.h"
#include "XWinSystem.h"
#include "XFont.h"
#include "XBitmap.h"

#include "WindowColorMap.h"
#include "Window.h"
#include "ETString.h"
#include "Env.h"
#include "Application.h"
#include "Class.h"

#include "MwmUtil.h"
#include "Sunkeysym.h"

#include <X11/keysym.h>

extern char *gProgname;
extern int gArgc;
extern char **gArgv;

static XWindow mainwindow;
static bool down;

const u_long Msk1 =	ButtonPressMask | ButtonReleaseMask |
			PointerMotionMask | PointerMotionHintMask |
			EnterWindowMask | LeaveWindowMask;

const u_long dfltMsk =	Msk1 |
			StructureNotifyMask | ExposureMask |
			KeyPressMask | FocusChangeMask;

const u_long ptrMsk =	Msk1;

static Pixmap MakePixmap(Pixmap &pm, Point &oldsize, Point newsize)
{
    if (pm == None || newsize.x > oldsize.x || newsize.y > oldsize.y) {
	if (pm)
	    XFreePixmap(XWinSystem::display, pm);
	oldsize= newsize;
	pm= XWinSystem::CreatePixmap(oldsize, gDepth);
    }
    return pm;
}

#define SetTS() if (solid != FillSolid) XSetTSOrigin(XWinSystem::display, gc, origin.x, origin.y)

//---- window creation/destruction ---------------------------------------------

NewMetaImpl0(XWindowPort,WindowPort);

XWindowPort::XWindowPort (Window *bw, WindowType wt, GrCursor curs)
    : WindowPort (bw, wt, curs)
{
    XSetWindowAttributes setWAttr;
    long flags= 0;
    WindowFlags wflags = win->GetWindowFlags();

    bool transient= FALSE;
    bool resize= !bool(wflags & eWinFixed);
    bool block= bool(wflags & eWinBlock);
    bool dialog= FALSE;
    overlay= FALSE;

    switch (wt) {
    case eWinPlain:
    case eWinStandard:
	//resize= TRUE;
	break;
    case eWinDialog:
	transient= block;
	dialog= TRUE;
	break;
    case eWinPulldownMenu:
    case eWinPopupMenu:
	overlay= TRUE;
	break;
    case eWinIcon:
	break;
    }

    wm_state = saved_wm_state = WithdrawnState;
    wmhints = 0;
    sizehints = 0;

    func= -1;
    currpsz= -1;
    solid= FillSolid;
    depth= XDefaultDepth(XWinSystem::display, XWinSystem::screen);
    
    if (wcmap == 0) {
	wcmap= WindowSystem::wcmap;
	wcmap->Ref();
    }

    flags|= CWBackPixmap;
    setWAttr.background_pixmap= None;

    flags|= CWBorderPixmap;
    setWAttr.border_pixmap= None;

    flags |= CWCursor;
    setWAttr.cursor =
	XWinSystem::UseCursor(XWinSystem::set_wait_depth > 0 ? eCrsHourglass : cursor);

    if ( overlay )
      {
	flags |= CWSaveUnder;
	setWAttr.save_under = XTrue;
	flags |= CWOverrideRedirect;
	setWAttr.override_redirect = XTrue;
      }

    flags |= CWEventMask;
    setWAttr.event_mask = dfltMsk;
    if ( ! overlay )
      setWAttr.event_mask |= PropertyChangeMask;	// to trace WM_STATE changes

    id= winid= XCreateWindow(XWinSystem::display, XWinSystem::rootwin,
				0, 0, 1, 1, 0, GetDepth(), InputOutput,
				(Visual*) CopyFromParent, flags, &setWAttr);

    if ( mainwindow == 0 ) {
	mainwindow = winid;
	gApplication->MakeWindows();
	Window *w = gApplication->GetWindow();
	if ( w ) {
	    XWindowPort *xw = (XWindowPort *) w->GetPortDesc();
	    if ( xw )
		mainwindow = xw->GetWinId();
	}
    }

    XGCValues gcv;
    gcv.graphics_exposures = XFalse;
    gc = XWinSystem::CreateGC (winid, GCGraphicsExposures, &gcv);

    if ( ! overlay )
      {
	int wmprot = 0;
	Atom WMProtocols[3];
	if ( cDeleteWindow )
	    WMProtocols[wmprot++]= cDeleteWindow;
	if ( wmprot > 0 )
	    XSetWMProtocols (XWinSystem::display, winid, WMProtocols, wmprot);

	wmhints = XAllocWMHints();

	wmhints->flags |= InputHint;
	wmhints->input = XTrue;

	wmhints->flags |= StateHint;
	wmhints->initial_state = NormalState;

	// get icon bitmap and set icon pixmap hint
	if ( win )
	  {
	    Manager *mgr = (Manager *) win->FindNextHandlerOfClass(Meta(Manager));
	    Bitmap *icon_bm = 0;
	    if ( mgr )
		icon_bm = mgr->GetIcon();
	    if ( icon_bm == 0 )
		icon_bm = gApplication->GetIcon();
	    if ( icon_bm )
	      {
		XBitmap *xbm = (XBitmap *) icon_bm->GetDevBitmap();
		if ( xbm )
		  {
		    wmhints->flags |= IconPixmapHint;
		    wmhints->icon_pixmap = xbm->GetPixmap();
		  }
	      }
	  }

	XClassHint *classhints= XAllocClassHint();
	classhints->res_name= gProgname;
	classhints->res_class= "ET++";
	
	XSetWMProperties(XWinSystem::display, winid, 0, 0,
		(winid == mainwindow) ? gArgv : NULL,
		(winid == mainwindow) ? gArgc : 0,
		0, wmhints, classhints);
	XFree((char*)classhints);

	// set default icon title
	DevSetIconTitle(gProgname);

	if (cOpenLookWinAttr)
	  {
	    long props[5];
	    props[0]= props[1]= props[2]= props[3]= props[4]= 0;
	    props[0] |= 1;
	    if (dialog)
	      {
		props[0] |= 4;  // pin state
		props[1]= (Atom) cOpenLookWinCmd;
		props[3]= 1;
	      }
	    else
	      props[1]= (Atom) cOpenLookWinBase;
	    XChangeProperty(XWinSystem::display, winid, cOpenLookWinAttr,
		cOpenLookWinAttr, 32, PropModeReplace, (byte*) &props, 5);
	  }
	if (cOpenLookDecorAdd && resize)
	  {
	    Atom props[5];
	    props[0]= cOpenLookResize;
	    XChangeProperty(XWinSystem::display, winid, cOpenLookDecorAdd,
		XA_ATOM, 32, PropModeReplace, (byte*) &props, 1);
	  } 
	if (cOpenLookDecorDel && !resize)
	  {
	    Atom props[5];
	    props[0]= cOpenLookResize;
	    XChangeProperty(XWinSystem::display, winid, cOpenLookDecorDel,
		XA_ATOM, 32, PropModeReplace, (byte*) &props, 1);
	  }

	// set Motif Window Manager hints
	if (cMotifWmHints)
	  {
	    MwmHints mwm_hints;
	    mwm_hints.flags =	MWM_HINTS_FUNCTIONS
				| MWM_HINTS_DECORATIONS
				| MWM_HINTS_INPUT_MODE;
	    mwm_hints.functions = MWM_FUNC_MOVE | MWM_FUNC_CLOSE;
	    mwm_hints.decorations = MWM_DECOR_BORDER
				    | MWM_DECOR_TITLE
				    | MWM_DECOR_MENU;
	    mwm_hints.input_mode = block ? MWM_INPUT_FULL_APPLICATION_MODAL
					 : MWM_INPUT_MODELESS;
	    if (resize)
	      {
		mwm_hints.functions |= MWM_FUNC_RESIZE;
		mwm_hints.decorations |= MWM_DECOR_RESIZEH;
		if (!dialog)
		  {
		    mwm_hints.functions |= MWM_FUNC_MAXIMIZE;
		    mwm_hints.decorations |= MWM_DECOR_MAXIMIZE;
		  }
	      }
	    if (!dialog)
	      {
		mwm_hints.functions |= MWM_FUNC_MINIMIZE;
		mwm_hints.decorations |= MWM_DECOR_MINIMIZE;
	      }

	    XChangeProperty(XWinSystem::display, winid,
			cMotifWmHints, cMotifWmHints, 32, PropModeReplace,
			(unsigned char *) &mwm_hints, PROP_MWM_HINTS_ELEMENTS);
	  }
      }
}

XWindowPort::~XWindowPort()
{
  if ( winid != 0 )
    {
      XFreeGC (XWinSystem::display, gc);
      XDestroyWindow (XWinSystem::display, winid);
      if ( sizehints != 0 )
	{
	  XFree ((char *) sizehints);
	  sizehints = 0;
	}
      if ( wmhints != 0 )
	{
	  XFree ((char *) wmhints);
	  wmhints = 0;
	}
      winid = 0;
    }
}

//---- Patterns & Rasterops ----------------------------------------------------

void XWindowPort::DevSetPattern(Bitmap *b)
{
    DevBitmap *bm= b->GetDevBitmap();
    XBitmap *xbm= (XBitmap*) bm->PrepareBitmap(this, bm->Size(), b->GetColorMapper());
    SetFunction(GXcopy);
    if (b->MaxPixelValue() > 1) {
	SetFillStyle(FillTiled);
	XSetTile(XWinSystem::display, gc, xbm->GetPixmap());
    } else {
	SetFillStyle(FillOpaqueStippled);
	XSetStipple(XWinSystem::display, gc, xbm->GetPixmap());
	XSetForeground(XWinSystem::display, gc, XWinSystem::black);
	XSetBackground(XWinSystem::display, gc, XWinSystem::white);
    }
}

void XWindowPort::DevSetColor2(u_long id)
{
    FlushColorMap();
    SetFunction(GXcopy);
    SetFillStyle(FillSolid);
    XSetForeground(XWinSystem::display, gc, id);
}

void XWindowPort::DevSetOther(int pid)
{
    SetFillStyle(FillSolid);
    switch (pid) {
    case -1:
	SetFunction(GXinvert);
	break;
    case 1:
	SetFunction(GXcopy);
	XSetForeground(XWinSystem::display, gc, XWinSystem::white);
	break;
    case 2:
	SetFunction(GXcopy);
	XSetForeground(XWinSystem::display, gc, XWinSystem::black);
	break;
    }
}

void XWindowPort::setFunction(int f)
{
    func= f;
    if (func == GXinvert) {
	XSetPlaneMask(XWinSystem::display, gc, XWinSystem::white ^ XWinSystem::black);
	XSetFunction(XWinSystem::display, gc, GXinvert);
    } else {
	XSetPlaneMask(XWinSystem::display, gc, AllPlanes);
	XSetFunction(XWinSystem::display, gc, func);
    }
}

void XWindowPort::setLine(int psz)
{
    currpsz= psz;
    if (psz == 1 && solid == FillSolid)
	psz= 0;
    XSetLineAttributes(XWinSystem::display, gc, psz, LineSolid, CapButt, JoinMiter);
}

//---- Cursors -----------------------------------------------------------------

void XWindowPort::DevSetCursor(GrCursor c)
{
    if (winid) {
	XDefineCursor(XWinSystem::display, winid, XWinSystem::UseCursor(c));
	XFlush(XWinSystem::display);
    }
}

//---- Event mapping -----------------------------------------------------------

void XWindowPort::StateToFlags(Token *t, unsigned int state)
{
    if (state & ShiftMask)
	t->Flags|= eFlgShiftKey;
    if (state & ControlMask)
	t->Flags|= eFlgCntlKey;
    if (state & XWinSystem::GetModMask())
	t->Flags|= eFlgMetaKey;
}

void
XWindowPort::GetOrigins (Point & pos_me, Point & pos_parent)
{
  XWindow a_win;
  XGrabServer (XWinSystem::display);
  XTranslateCoordinates (XWinSystem::display, winid, XWinSystem::rootwin,
			 0, 0, &pos_me.x, &pos_me.y, &a_win);
  pos_parent = pos_me;

  // Walk up the window hierarchy starting with `this->winid'.
  // Search the (grand)parent of `this->winid' which is a direct child
  // of the root window.
  // Put the offset of that window into `parent'.
  a_win = winid;
  while ( TRUE )
    {
      XWindow root;
      XWindow parent;
      XWindow * children;
      unsigned int n_children;
      Status status;

      children = 0;
      status = XQueryTree (XWinSystem::display, a_win,
			   &root, &parent, &children, &n_children);
      if ( status == 0 )
	break;

      // Destroy list of children windows. It is not used here.
      if ( children != 0 )
	XFree ((char *) children);

      if ( parent == root )
	{
	  if ( a_win != winid )
	    {
	      XTranslateCoordinates (XWinSystem::display, a_win, XWinSystem::rootwin,
				     0, 0, &pos_parent.x, &pos_parent.y, &a_win);
	    }
	  break;
	}

      a_win = parent;
    }

  XUngrabServer (XWinSystem::display);
}

bool XWindowPort::MapEvent(XEvent &xe, Token *tp)
{
    Rectangle newrect;
    Token t(eEvtNone);
    char buffer[40];
    int keysym;
    KeySym _keysym;

    FlushColorMap();
//printf("XWindowPort::MapEvent type:%d\n", xe.type);
    switch (xe.type) {

    case NoExpose:
	return FALSE;

    case GraphicsExpose:
    case Expose:
	newrect = Rectangle(xe.xexpose.x, xe.xexpose.y,
			    xe.xexpose.width, xe.xexpose.height);
	if (xe.xexpose.count > 0) {
	    InvalidateRect(newrect);
	} else {  // last expose
	    if (configure) {
		InvalidateRect (Rectangle (gPoint0, portrect.extent));
		OpenNotify(portrect);
		configure= FALSE;
	    } else
		InvalidateRect(newrect, TRUE);      // last exposure in sequence
	}

	win->Send(cTOP, 0, win);
	return FALSE;

    case ConfigureNotify:
	if ( portrect.extent.x != xe.xconfigure.width
	     || portrect.extent.y != xe.xconfigure.height )
	    configure = TRUE;
	GetOrigins (myOrigin, portrect.origin);
	portrect.extent.x = xe.xconfigure.width;
	portrect.extent.y = xe.xconfigure.height;
	return FALSE;

    case MapNotify:
	configure = TRUE;
	GetOrigins (myOrigin, portrect.origin);
	win->Send(cSHOW_WIN, 0, win);
	return FALSE;

    case UnmapNotify:
	state = eWsHidden;
	win->Send(cHIDE_WIN, 0, win);
	return FALSE;

    case PropertyNotify:
	if ( xe.xproperty.atom == cAtomWmState && xe.xproperty.state == PropertyNewValue )
	  GetWmState();
	return FALSE;

    case KeyPress:
	// Lookup for Meta characters
	// Used by crazy HP/IBM keyboards :-(
	// 3 column in modmap -> Meta ; 4 column -> Shift + Meta
	int mod, index;
	mod = xe.xkey.state & (Mod1Mask|Mod2Mask|Mod3Mask|Mod4Mask|Mod5Mask);
	_keysym = NoSymbol;
	if(mod)
	    {
	    if (mod & Mod1Mask)
		{
		if (!(xe.xkey.state & ShiftMask))
		    index = 2;
		else
		    index = 3;
		}
	    else if (mod & Mod2Mask) index = 4; 
		else if (mod & Mod3Mask) index = 5; 
		    else if (mod & Mod4Mask) index = 6; 
			else if (mod & Mod5Mask) index = 7;
	    
	    _keysym = XLookupKeysym(&xe.xkey, index);
	    }
	if(_keysym == NoSymbol)
	    XLookupString(&xe.xkey, buffer, sizeof(buffer), &_keysym, 0);
	keysym= (int) _keysym;

	if (keysym >= XK_KP_0 && keysym <= XK_KP_9)
	    keysym-= (XK_KP_0 - XK_0);

	switch (keysym) {
	case XK_KP_Space:   t.Code= ' ';            break;
	case XK_KP_Tab:     t.Code= '\t';           break;
	case XK_KP_Enter:   t.Code= '\n';           break;
	case XK_KP_Equal:   t.Code= '=';            break;
	case XK_KP_Multiply:t.Code= '*';            break;
	case XK_KP_Add:     t.Code= '+';            break;
	case XK_KP_Separator:t.Code= ',';           break;
	case XK_KP_Subtract:t.Code= '-';            break;
	case XK_KP_Decimal: t.Code= '.';            break;
	case XK_KP_Divide:  t.Code= '/';            break;

	case XK_Left:       t.Code= eEvtCursorLeft; break;
	case XK_Up:         t.Code= eEvtCursorUp;   break;
	case XK_Right:      t.Code= eEvtCursorRight;break;
	case XK_Down:       t.Code= eEvtCursorDown; break;
	case XK_Tab:        t.Code= '\t';           break;
	case XK_Delete:
	  t.Code = XWinSystem::UseDeleteKey() ? 0x7F : '\b';
	  break;
	case XK_BackSpace:  t.Code= '\b';           break;
	case XK_Linefeed:   t.Code= '\n';           break;
	case XK_Return:     t.Code= '\r';           break;
	case XK_Escape:     t.Code= '\033';         break;

	case XK_Help:       t.Code = eEvtHelpKey;   break;

	case XK_Print:      t.Code = eEvtKey_Print;    break;
	case XK_Prior:      t.Code = eEvtKey_PageUp;   break;
	case XK_Next:       t.Code = eEvtKey_PageDown; break;
	case XK_Home:       t.Code = eEvtKey_Home;     break;
	case XK_End:        t.Code = eEvtKey_End;      break;
	case XK_Undo:       t.Code = eEvtKey_Undo;     break;
	case XK_Redo:       t.Code = eEvtKey_Redo;     break;
	case XK_Find:       t.Code = eEvtKey_Find;     break;
	case XK_Cancel:     t.Code = eEvtKey_Cancel;   break;
	case SunXK_Props:   t.Code = eEvtKey_Props;    break;
	case SunXK_Front:   t.Code = eEvtKey_Front;    break;
	case SunXK_Open:    t.Code = eEvtKey_Open;     break;
	case SunXK_Cut:     t.Code = eEvtKey_Cut;      break;
	case SunXK_Copy:    t.Code = eEvtKey_Copy;     break;
	case SunXK_Paste:   t.Code = eEvtKey_Paste;    break;

	default:
	    if (keysym > 0 && keysym < 256)
		t.Code= keysym;
	    else if (keysym >= XK_F1 && keysym <= XK_F35)
		t.Code= (short) (eEvtPfkFirst + (keysym - XK_F1) + 1);
	    else
		return FALSE;
	    break; 
	}
	StateToFlags(&t, xe.xkey.state);
/*
	if ( t.IsAscii() ) {
	    if (t.Flags & eFlgCntlKey) {
		if (t.Flags & eFlgShiftKey)
		    t.Code-= 0x40;
		else
		    t.Code-= 0x60;
	    }
	    if (t.Flags & eFlgMetaKey)
		t.Code+= 0x80;
	}
*/
	t.Pos= Point(xe.xkey.x, xe.xkey.y);
	t.At= xe.xkey.time;
	break;

    case ButtonPress:
	t.Code= eEvtButtons + xe.xbutton.button - 1;
	StateToFlags(&t, xe.xbutton.state);
	t.At= xe.xbutton.time;
	t.Pos= Point(xe.xbutton.x, xe.xbutton.y);
	down= TRUE;
	break;

    case ButtonRelease:
	t.Code= eEvtButtons + xe.xbutton.button - 1;
	StateToFlags(&t, xe.xbutton.state);
	t.At= xe.xbutton.time;
	t.Pos= Point(xe.xbutton.x, xe.xbutton.y);
	t.Flags|= eFlgButDown;
	XUngrabPointer(XWinSystem::display, CurrentTime);
	down= FALSE;
	break;

    case MotionNotify:
	if (xe.xmotion.is_hint) {
	    while(XCheckTypedEvent(XWinSystem::display, MotionNotify, &xe))
		;   // overread MotionNotifies
	    t.Pos= getMousePos(xe.xmotion.state);
	    if (t.Pos == lastpos)
		return FALSE;
	} else
	    t.Pos= Point(xe.xmotion.x, xe.xmotion.y);
	t.Code= eEvtLocMove;
	t.At= xe.xmotion.time;
	StateToFlags(&t, xe.xmotion.state);
	break;

    case LeaveNotify:
    case EnterNotify:
	if (xe.xcrossing.focus
	    && (xe.xcrossing.detail == NotifyAncestor
		|| xe.xcrossing.detail == NotifyInferior
		|| xe.xcrossing.detail == NotifyNonlinear) ) {
	    active= (xe.type == EnterNotify);
	    if (win)
		win->FocusNotify(active || focus);
	}
	return FALSE;

    case FocusIn:
    case FocusOut:
	if (xe.xfocus.detail == NotifyAncestor
	    || xe.xfocus.detail == NotifyInferior
	    || xe.xfocus.detail == NotifyNonlinear) {
	    focus= (xe.type == FocusIn);
	    if (win)
		win->FocusNotify(active || focus);
	}
	return FALSE;

    case ClientMessage:
	if (cDeleteWindow && xe.xclient.data.l[0] == cDeleteWindow) {
	    if (win)
		win->Control(cCLOSE, cPartAction, 0);
	} else if (cDecWmMoved && xe.xclient.message_type == cDecWmMoved) {
	    // Window has been moved by DECWindows window manager
	    // get new window origin in screen coordinates
	    GetOrigins (myOrigin, portrect.origin);
	} else {
	    if ( gDebug && xe.xclient.message_type != None )
	    {
		char *p = XGetAtomName(XWinSystem::display,
					xe.xclient.message_type);
		fprintf(stderr, "%s: unknown ClientMessage: %s\n", gProgname, p);
		XFree(p);
	    }
	}
	return FALSE;

    default:
	return FALSE;
    }

    if (tp)
	*tp= t;
    return TRUE;
}

//---- window management -------------------------------------------------------

// Read the WM_STATE property

static int
x_ignore_error_handler (XDisplay * display, XErrorEvent * ev)
{
  return 0;
}

void
XWindowPort::GetWmState (void)
{
  int status;
  Atom actual_type_return;
  int actual_format_return;
  unsigned long nitems_return;
  unsigned long bytes_after_return;
  unsigned char * prop_return = 0;
  int (*old_error_handler) (XDisplay *, XErrorEvent *);

  old_error_handler = XSetErrorHandler (x_ignore_error_handler);

  status = XGetWindowProperty (XWinSystem::display, winid,
				cAtomWmState,
				0L,			/* long_offset */
				2,			/* long_length */
				False,		/* delete property ? */
				cAtomWmState,		/* req_type */
				&actual_type_return,
				&actual_format_return,
				&nitems_return,
				&bytes_after_return,
				&prop_return);
  if ( status == 0 )
    wm_state = ((unsigned int *) prop_return)[0];
  if ( prop_return != 0 )
    XFree ((char *) prop_return);

  XSetErrorHandler (old_error_handler);
}

void XWindowPort::UpdateHints(const Point &o, const Point &e)
{
    if (!overlay) {
	Point ee;
	if (sizehints == 0)
	    sizehints= XAllocSizeHints();
	sizehints->flags= 0;
	
	if (win)
	    ee= win->GetMinSize().extent;
	else
	    ee= e;
	if (sizehints->x != o.x || sizehints->y != o.y
	    || sizehints->width != e.x || sizehints->height != e.y
		|| sizehints->min_width != ee.x || sizehints->min_height != ee.y) {
		
	    sizehints->flags |= USPosition;
	    sizehints->x = o.x;
	    sizehints->y = o.y;

	    sizehints->flags |= USSize;
	    sizehints->width = e.x;
	    sizehints->height = e.y;

	    sizehints->flags |= PMinSize;
	    sizehints->min_width = ee.x;
	    sizehints->min_height = ee.y;
	    XSetWMNormalHints(XWinSystem::display, winid, sizehints);
	}
    } 
}

void
XWindowPort::DevShow (WindowPort * fp, Rectangle r)
{
  if ( r.origin.x == 0 )
    r.origin.x = 1;
  if ( r.origin.y == 0 )
    r.origin.y = 1;

  father = (XWindowPort *) fp;
  if ( father )
    {
      Point off (father->myOrigin - father->portrect.origin);
      r.origin += off;
      if ( win->GetWindowType() == eWinDialog
	  && (win->GetWindowFlags() & eWinBlock)
	  && father->GetWinId() != 0 )
	{
	  XSetTransientForHint (XWinSystem::display, winid, father->GetWinId());
	}
    }
 
  if ( overlay )
    r.origin += r.AmountToTranslateWithin (WindowSystem::screenRect);
  else
    {
      UpdateHints (r.origin, r.extent);
      wmhints->flags |= StateHint;
      wmhints->initial_state = NormalState;
      XSetWMHints (XWinSystem::display, winid, wmhints);
    }

  if ( portrect != r )
    {
      XMoveResizeWindow (XWinSystem::display, winid,
			 r.origin.x, r.origin.y, r.extent.x, r.extent.y);
    }

  XForceScreenSaver (XWinSystem::display, ScreenSaverReset);

  XMapRaised (XWinSystem::display, winid);

  while ( state != eWsShown )
    {
      XEvent xe;
      XWindowEvent (XWinSystem::display, winid,
		    StructureNotifyMask | ExposureMask, &xe);
      MapEvent (xe);
    }

  if ( ! overlay && XWinSystem::ExplicitInputFocus() )
    XSetInputFocus (XWinSystem::display, winid, RevertToPointerRoot, CurrentTime);

  if ( father && down )
    {
      XGrabPointer (XWinSystem::display, winid, XFalse, ptrMsk,
		    GrabModeAsync, GrabModeAsync, None, None,
		    WindowSystem::lasttime);
    }
}

void XWindowPort::DevHide()
{     
  if ( overlay )
    XUnmapWindow (XWinSystem::display, winid);
  else
    XWithdrawWindow (XWinSystem::display, winid, XWinSystem::screen);
  wm_state = WithdrawnState;

  XFlush (XWinSystem::display);

  XEvent xe;
  while ( XCheckMaskEvent (XWinSystem::display,
			   StructureNotifyMask | ExposureMask, &xe) )
    {
      XWindowPort *port = XWinSystem::Find (xe);
      if ( port )
	port->MapEvent (xe);
    }

  if ( father && down )
    {
      XGrabPointer (XWinSystem::display, father->GetWinId(), XFalse, ptrMsk,
		    GrabModeAsync, GrabModeAsync, None, None,
		    WindowSystem::lasttime);
    }
}

void
XWindowPort::DevIconize()
{
  XIconifyWindow (XWinSystem::display, winid, XWinSystem::screen);
}

void
XWindowPort::DevTop (bool top)
{
  if ( top )
    {
      if ( ! overlay )
	{
	  wmhints->flags |= StateHint;
	  wmhints->initial_state = NormalState;
	  XSetWMHints (XWinSystem::display, winid, wmhints);
	}
      XMapRaised (XWinSystem::display, winid);
      if ( ! overlay && XWinSystem::ExplicitInputFocus() )
	XSetInputFocus (XWinSystem::display, winid, RevertToPointerRoot, CurrentTime);
    }
  else
    XLowerWindow (XWinSystem::display, winid);
}

void XWindowPort::DevSetExtent(Point e)
{
    UpdateHints(portrect.origin, e);
    if (portrect.extent != e)
	XResizeWindow(XWinSystem::display, winid, e.x, e.y);
}

void XWindowPort::DevSetOrigin(Point o)
{
    UpdateHints(o, portrect.extent);
    if (portrect.origin != o)
	XMoveWindow(XWinSystem::display, winid, o.x, o.y);
}

//---- graphical primitives ----------------------------------------------------

void XWindowPort::CopyArea(XWindow from, XWindow to, const Point &o, const Point &e, const Point &d)
{
    SetFunction(GXcopy);
    XCopyArea(XWinSystem::display, from, to, gc, o.x, o.y, e.x, e.y, d.x, d.y);
}

void XWindowPort::DevClip(Rectangle *r)
{
    if (WindowSystem::fullscreen) {
	XSetClipMask(XWinSystem::display, gc, None);
    } else {
	XRectangle xcr;
	xcr.x= r->origin.x;
	xcr.y= r->origin.y;
	xcr.width= r->extent.x;
	xcr.height= r->extent.y;
	XSetClipRectangles(XWinSystem::display, gc, 0, 0, &xcr, 1, YXBanded);
    }
} 

void XWindowPort::DevStrokeLine2(int psz, Rectangle*, GrLineCap, Point start, Point end)
{
    if (psz < 0)
	return;
    SetLine(psz);
    SetTS();
    XDrawLine(XWinSystem::display, id, gc, start.x, start.y, end.x, end.y);
}

static void AdjustRect(int psz, Rectangle *r)
{
    if (psz > 1) {
	r->origin+= psz/2;
	r->extent-= psz;
    } else
	r->extent-= 1;
}

void XWindowPort::DevStrokeRect2(int psz, Rectangle *r)
{
    if (psz < 0)
	return;
    SetLine(psz);
    AdjustRect(psz, r);
    SetTS();
    XDrawRectangle(XWinSystem::display, id, gc, r->origin.x, r->origin.y, r->extent.x, r->extent.y);
}

void XWindowPort::DevFillRect(Rectangle *r)
{
    SetTS();
    XFillRectangle(XWinSystem::display, id, gc, r->origin.x, r->origin.y, r->extent.x, r->extent.y);
}

void XWindowPort::DevStrokeRRect2(int psz, Rectangle *r, Point dia)
{
    XArc a[4];
    XSegment s[4];
    
    if (psz < 0)
	return;

    AdjustRect(psz, r);
    Point o= r->origin;
    Point e= r->extent-dia;
    
    for (int i= 0; i < 4; i++) {
	a[i].angle1= (i*90) << 6;
	a[i].angle2= 90 << 6;
	a[i].width= dia.x;
	a[i].height= dia.y;
	a[i].x= o.x;
	a[i].y= o.y;
    }
    a[0].x+= e.x;
    a[3].x+= e.x;
    a[2].y+= e.y;
    a[3].y+= e.y;
    
    s[0].x1= s[3].x1= o.x+dia.x/2;
    s[0].x2= s[3].x2= o.x+r->extent.x-dia.x/2;
    s[1].x1= s[1].x2= o.x;
    s[2].x1= s[2].x2= o.x+r->extent.x;
    
    s[0].y1= s[0].y2= o.y;
    s[1].y1= s[2].y1= o.y+dia.y/2;
    s[1].y2= s[2].y2= o.y+r->extent.y-dia.y/2;
    s[3].y1= s[3].y2= o.y+r->extent.y;
    
    SetLine(psz);
    SetTS();
    XDrawArcs(XWinSystem::display, id, gc, a, 4);
    XDrawSegments(XWinSystem::display, id, gc, s, 4);
}

void XWindowPort::DevFillRRect2(Rectangle *r, Point dia)
{
    XArc a[4];
    XRectangle rr[3];
    Point o= r->origin;
    Point e= r->extent-dia;
    
    for (int i= 0; i < 4; i++) {
	a[i].angle1= (i*90) << 6;
	a[i].angle2= 90 << 6;
	a[i].width= dia.x;
	a[i].height= dia.y;
	a[i].x= o.x;
	a[i].y= o.y;
    }
    a[0].x+= e.x;
    a[3].x+= e.x;
    a[2].y+= e.y;
    a[3].y+= e.y;
    
    rr[0].x= rr[2].x= o.x + dia.x/2;
    rr[1].x= o.x;
    
    rr[0].y= o.y;
    rr[1].y= o.y+dia.y/2;
    rr[2].y= o.y+r->extent.y-dia.y/2;
    
    rr[0].width= rr[2].width= e.x;
    rr[1].width= r->extent.x;
    
    rr[0].height= rr[2].height= dia.y/2;
    rr[1].height= e.y;
    
    SetTS();
    XFillArcs(XWinSystem::display, id, gc, a, 4);
    XFillRectangles(XWinSystem::display, id, gc, rr, 3);
}

void XWindowPort::DevStrokeOval2(int psz, Rectangle *r)
{
    if (psz < 0)
	return;
    SetLine(psz);
    AdjustRect(psz, r);
    SetTS();
    XDrawArc(XWinSystem::display, id, gc, r->origin.x, r->origin.y,
				r->extent.x, r->extent.y, 0, 360*64);
}

void XWindowPort::DevFillOval2(Rectangle *r)
{
    SetTS();
    XFillArc(XWinSystem::display, id, gc, r->origin.x, r->origin.y,
					r->extent.x, r->extent.y, 0, 360*64);
}

void XWindowPort::DevStrokeWedge2(int psz, GrLineCap, Rectangle *r, int s, int len)
{
    if (psz < 0)
	return;
    SetLine(psz);
    AdjustRect(psz, r);
    SetTS();
    XDrawArc(XWinSystem::display, id, gc, r->origin.x, r->origin.y,
		r->extent.x, r->extent.y, ((450-s) % 360)*64, -len*64);
}

void XWindowPort::DevFillWedge2(Rectangle *r, int s, int len)
{
    SetTS();
    XFillArc(XWinSystem::display, id, gc, r->origin.x, r->origin.y,
			r->extent.x, r->extent.y, ((450-s) % 360)*64, -len*64);
}

void XWindowPort::DevStrokePolygon2(Rectangle *r,
			Point *pts, int npts, GrPolyType, int psz, GrLineCap)
{
    if (psz < 0)
	return;
    register int i, x= r->origin.x, y= r->origin.y;
    XPoint *xp, *xpp;
    xp= xpp= (XPoint*) Alloca(npts * sizeof(XPoint));
    
    SetLine(psz);
    for (i= 0; i < npts; i++, xpp++, pts++) {
	xpp->x= pts->x + x;
	xpp->y= pts->y + y;
    }
    SetTS();
    XDrawLines(XWinSystem::display, id, gc, xp, npts, CoordModeOrigin);
    Freea(xp);
}

void XWindowPort::DevFillPolygon2(Rectangle *r, Point *pts, int npts, GrPolyType)
{
    register int i, x= r->origin.x, y= r->origin.y;
    XPoint *xp, *xpp;
    xp= xpp= (XPoint*) Alloca(npts * sizeof(XPoint));
    
    for (i= 0; i < npts; i++, xpp++, pts++) {
	xpp->x= pts->x + x;
	xpp->y= pts->y + y;
    }
    SetTS();
    XFillPolygon(XWinSystem::display, id, gc, xp, npts,
				(npts <= 3) ? Convex : Complex, CoordModeOrigin);
    Freea(xp);
}

void XWindowPort::DevShowBitmap(Rectangle *r, Bitmap *bm)
{
    DevBitmap *dbm= bm->GetDevBitmap();
    XBitmap *x= (XBitmap*) dbm->PrepareBitmap(this, r->extent, bm->GetColorMapper());
    Pixmap pm= x->GetPixmap();
    XBitmap *mask= (XBitmap*) dbm->GetMask(r->extent, bm->GetColorMapper());
    
    if (mask) {
	XSetForeground(XWinSystem::display, gc, 0x00000000);
	XSetBackground(XWinSystem::display, gc, 0xffffffff);
	SetFunction(GXand);
	XCopyPlane(XWinSystem::display, mask->GetPixmap(), id, gc, 0, 0,
		r->extent.x, r->extent.y, r->origin.x, r->origin.y, 0x00000001);
	SetFunction(GXor);
	XCopyArea(XWinSystem::display, pm, id, gc, 0, 0,
		r->extent.x, r->extent.y, r->origin.x, r->origin.y);
   } else {
	if (dbm->depth > 1) {
	    CopyArea(pm, id, gPoint0, r->extent, r->origin);
	} else {
	    XSetStipple(XWinSystem::display, gc, pm);
	    XSetTSOrigin(XWinSystem::display, gc, r->origin.x, r->origin.y);
	    SetFillStyle(FillStippled);
	    XFillRectangle(XWinSystem::display, id, gc, r->origin.x, r->origin.y,
						    r->extent.x, r->extent.y);
	}
    }
}

//---- text batching -----------------------------------------------------------

static XTextItem ti[100], *tip= ti;
static byte tbc[400], *tbp= tbc;
static Font *lastfont;

struct Ulines {
    XPoint start;
    XPoint end;
};
static XRectangle ul[100];
static int ullast=-1;

void XWindowPort::DevShowChar(Font *fdp, Point gap, byte c, Rectangle *rr)
{
    if (gap.y != 0 || tip-ti >= 100-1 || tbp-tbc >= MaxTextBatchCnt) // X cannot handle multiple lines !!
	FlushAnyText();
    if (gap.x > 24)     //XXX BUG on LINUX???
	FlushAnyText(); //REMOVE if not needed...
 
    bool isul=( fdp->Face() & eFaceUnderline );
    if (tbp == tbc) {   // new batch
	ullast=-1;
	if(isul) {
	    ullast++;
	    ul[ullast].x=rr->origin.x;
	    ul[ullast].y=rr->SW().y-((XServerFont*)fdp)->fUnderlinePos;
	}
	tip= ti;
	tip->delta= 0;
	tip->chars= (char*)tbp;
	tip->nchars= 0;
	tip->font= ((XServerFont*)fdp)->GetId();
	lastfont= fdp;
    } else if (fdp != lastfont || gap.x != 0) {   // switch to next
	if(isul) {
	    ullast++;
	    ul[ullast].x=rr->origin.x;
	    ul[ullast].y=rr->SW().y-((XServerFont*)fdp)->fUnderlinePos;
	}
	tip++;
	tip->delta= gap.x;
	tip->chars= (char*)tbp;
	tip->nchars= 0;
	if (fdp != lastfont)
		tip->font= ((XServerFont*)fdp)->GetId();
	else
		tip->font= (XFont) None;
	lastfont= fdp;
    }
    if ( isul ) {
	ul[ullast].width=rr->E().x-ul[ullast].x+1;
	ul[ullast].height=((XServerFont*)fdp)->fUnderlineWidth;
    }
    *tbp++= c;
    tip->nchars++;
}

void XWindowPort::DevShowTextBatch(Rectangle*r, Point pos)
{
    if (tbp > tbc) {
	if (tip-ti == 0) {
	    XSetFont(XWinSystem::display, gc, tip->font);
	    XDrawString(XWinSystem::display, id, gc, pos.x, pos.y,
						    tip->chars, tip->nchars);
	} else
	    XDrawText(XWinSystem::display, id, gc, pos.x, pos.y, ti, tip-ti+1);
	tbp= tbc;
    }
    if(ullast>=0) {
	XFillRectangles(XWinSystem::display, id, gc, ul, ullast+1);
    }
} 

//---- scrolling ---------------------------------------------------------------

void
XWindowPort::DevScrollRect (Rectangle dst, Point delta)
{
  XEvent xe;
  bool need_update;
  bool finished;
  Rectangle src (dst.origin - delta, dst.extent);
  Rectangle expose;

  XSetGraphicsExposures (XWinSystem::display, gc, XTrue);
  CopyArea (winid, winid, src.origin, src.extent, dst.origin);
  XSetGraphicsExposures (XWinSystem::display, gc, XFalse);

  // Possible event sequence:
  // [ Expose ... ] { GraphicsExpose ... | NoExpose }

  need_update = FALSE;
  finished = FALSE;
  do
    {
      XWindowEvent (XWinSystem::display, winid, ExposureMask, &xe);

      switch ( xe.type )
	{
	case NoExpose:
	  finished = TRUE;
	  break;

	case Expose:
	  // Normal expose events were generated BEFORE scrolling took place.
	  // Thus we have to compute the intersection of the exposed rectangle
	  // and the scrolled rectangle. If the intersection is not empty
	  // we must "scroll" that rectangle (move its origin by `delta') and
	  // invalidate it as well.
	  expose = Rectangle (xe.xexpose.x, xe.xexpose.y,
			      xe.xexpose.width, xe.xexpose.height);
	  if ( expose.Clip (src) )
	    {
	      expose.Moveby (delta);
	      InvalidateRect (expose);
	    }
	  expose = Rectangle (xe.xexpose.x, xe.xexpose.y,
			      xe.xexpose.width, xe.xexpose.height);
	  InvalidateRect (expose);
	  need_update = TRUE;
	  break;

	case GraphicsExpose:
	  expose = Rectangle (xe.xgraphicsexpose.x, xe.xgraphicsexpose.y,
			      xe.xgraphicsexpose.width, xe.xgraphicsexpose.height);
	  InvalidateRect (expose);
	  need_update = TRUE;
	  finished = (xe.xgraphicsexpose.count == 0);
	  break;
	}
    }
  while ( ! finished );

  if ( need_update )
    DevUpdate();
}

//---- input -------------------------------------------------------------------

void XWindowPort::TestInput()
{
    if (havepushbacktoken) {
	InputNotify(&pushbacktoken);
	havepushbacktoken= FALSE;
    }
}

void XWindowPort::DevGetEvent(Token *t, int timeout, bool)
{
    while (((XWinSystem*)gWindowSystem)->NextToken(t, timeout, this) == 0)
	;
}

Point XWindowPort::getMousePos(u_int &state)
{
    XWindow rootW, childW;
    int x0, y0, wx, wy;
    
    XQueryPointer(XWinSystem::display, winid, &rootW, &childW, &x0, &y0, &wx, &wy, &state);
    return Point(wx, wy);
}

void XWindowPort::DevSetMousePos(Point p, bool m)
{
    XWarpPointer(XWinSystem::display, None, m ? (XWindow)None : winid, 0, 0, 0, 0, p.x, p.y);
}

void XWindowPort::DevGrab(bool mode, bool fs)
{
    if (fs) {
	if (mode) {
	    id= XWinSystem::rootwin;
	    origin+= portrect.origin;
	    XGrabServer(XWinSystem::display);
	    XSetSubwindowMode(XWinSystem::display, gc, IncludeInferiors);
	    XSetClipMask(XWinSystem::display, gc, None);
	} else {
	    origin-= portrect.origin;
	    id= winid;
	    XSetSubwindowMode(XWinSystem::display, gc, ClipByChildren);
	    XUngrabServer(XWinSystem::display);
	    XFlush(XWinSystem::display);
	}
    }
}

//---- double buffering --------------------------------------------------------

static Point batchSize;
static Rectangle batchRect; 
static bool inBatch;

void XWindowPort::DevGiveHint(int code, int, void *vp)
{
    switch (code) {    
    case eHintBatch:
	if (!WindowSystem::batch || overlay || (id != winid))
	    return;
	inBatch= TRUE;
	batchRect= *((Rectangle*)vp);
	id= MakePixmap(XWinSystem::batchPixmap, batchSize, batchRect.origin+batchRect.extent);
	break;

    case eHintUnbatch:
	if (!WindowSystem::batch || !inBatch)
	    return;
	inBatch= FALSE;
	id= winid;
	CopyArea(XWinSystem::batchPixmap, winid, batchRect.origin,
					    batchRect.extent, batchRect.origin);
	break;

    case eHintCacheStart:
	{
	    batchInfo *bi= (batchInfo*) vp;
	    ((XBitmap*)bi->b)->GetPixmap();
	    CopyArea(((XBitmap*)bi->b)->pm, id, bi->p, bi->r.extent, bi->r.origin);
	}
	break;

    case eHintCacheEnd:
	{
	    batchInfo *bi= (batchInfo*) vp;
	    MakePixmap(((XBitmap*)bi->b)->pm, ((XBitmap*)bi->b)->size, bi->r.extent);
	    CopyArea(id, ((XBitmap*)bi->b)->pm, bi->r.origin, bi->r.extent, gPoint0);
	}
	break;
    }
}

void XWindowPort::DevSetTitle(char *title)
{
    if (winid && !overlay) {
	XTextProperty tp;
	if (XStringListToTextProperty(&title, 1, &tp)) {
	    XSetWMName(XWinSystem::display, winid, &tp);
	    XFree((char*)tp.value);
	}
    }
}

void XWindowPort::DevSetIconTitle(char *title)
{
    if (winid && !overlay) {
	XTextProperty tp;
	if (XStringListToTextProperty(&title, 1, &tp)) {
	    XSetWMIconName(XWinSystem::display, winid, &tp);
	    XFree((char*)tp.value);
	}
    }
}

void
XWindowPort::DevHideAndRemember()
{
  saved_wm_state = wm_state;
  XWithdrawWindow (XWinSystem::display, winid, XWinSystem::screen);
  wm_state = WithdrawnState;
}

void
XWindowPort::DevUnhide()
{
  if ( wm_state == saved_wm_state || wm_state != WithdrawnState )
    return;

  switch (saved_wm_state)
    {
      case NormalState:
      case IconicState:
	// Withdrawn -> {Normal,Iconic}
	wmhints->flags |= StateHint;
	wmhints->initial_state = saved_wm_state;
	XSetWMHints (XWinSystem::display, winid, wmhints);
	XMapWindow (XWinSystem::display, winid);
	break;
    }
}
