#ifndef _XWindowPort_h_
#define _XWindowPort_h_

#if defined(__GNUG__) && __GNUG__ >= 2
#  pragma interface
#endif

#include "X11.h"
#include "XWinSystem.h"
#include "WindowPort.h"

class XWindowPort : public WindowPort {
protected:
    XWindow	winid;  // window id
    XWindow	id;     // current drawable id
    GC		gc;     // window graphic context
    XWindowPort *father;
    int		currpsz;
    int		solid;
    int		func;
    Point	myOrigin;
    bool	active : 1;
    bool	focus : 1;
    bool	configure : 1;
    // window management
    XSizeHints *sizehints;
    XWMHints *	wmhints;
    unsigned int wm_state;
    unsigned int saved_wm_state;

protected:
    void setLine(int psz);
    void SetLine(int psz)
	{ if (psz != currpsz) setLine(psz); }
    Point getMousePos(u_int&);
    void setFunction(int f);
    void StateToFlags(Token *t, unsigned int state);
    void SetFunction(int f)
	{ if (func != f) setFunction(f); }
    void SetFillStyle(int fs)
	{ if (solid != fs) { XSetFillStyle(XWinSystem::display, gc, solid= fs); } }
    void CopyArea(XWindow from, XWindow to, const Point &o, const Point &e, const Point &d);
    void UpdateHints(const Point &o, const Point &e);
    void GetOrigins(Point &me, Point &parent);
    void GetWmState (void);

public:
    MetaDef (XWindowPort);

    XWindowPort(Window *bw, WindowType wt, GrCursor curs);
    ~XWindowPort();

    XWindow GetWinId()
	{ return winid; }
    void TestInput();

    bool MapEvent(XEvent &xe, Token *t= 0);

    //---- window management
    void DevShow(WindowPort *father, Rectangle);
    void DevHide();
    void DevIconize();
    void DevHideAndRemember();
    void DevUnhide();
    void DevTop(bool);
    void DevSetExtent(Point);
    void DevSetOrigin(Point);

    //---- clipping
    void DevClip(Rectangle*);

    //---- graphic
    void DevStrokeLine2(int, Rectangle*, GrLineCap, Point, Point);
    void DevStrokeRect2(int, Rectangle*);
    void DevFillRect(Rectangle*);
    void DevStrokeRRect2(int, Rectangle*, Point);
    void DevFillRRect2(Rectangle*, Point);
    void DevStrokeOval2(int, Rectangle*);
    void DevFillOval2(Rectangle*);
    void DevStrokeWedge2(int, GrLineCap, Rectangle*, int, int);
    void DevFillWedge2(Rectangle*, int, int);
    void DevStrokePolygon2(Rectangle*, Point*, int, GrPolyType t, int psz, GrLineCap cap);
    void DevFillPolygon2(Rectangle*, Point*, int, GrPolyType);
    void DevShowBitmap(Rectangle*, struct Bitmap*);
    void DevShowChar(Font *fdp, Point pos, byte c, Rectangle*);
    void DevShowTextBatch(Rectangle*, Point);
    void DevSetPattern(Bitmap*);
    void DevSetColor2(u_long id);
    void DevSetOther(int);

    //---- scrolling
    void DevScrollRect(Rectangle, Point);

    //---- input
    void DevGrab(bool, bool);
    void DevGetEvent(Token *t, int timeout, bool);    

    //---- cursor
    void DevSetCursor(GrCursor);

    //---- mouse
    void DevSetMousePos(Point, bool);

    //---- misc
    void DevGiveHint(int, int, void*);
    void DevSetTitle(char *);
    void DevSetIconTitle(char *);
};

#endif
