#ifdef __GNUG__
#pragma implementation
#endif

#include "CLib.h"
#include "ETString.h"
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
bool gPrintMem=FALSE;
#ifdef OS_WINDOWS
	#include "Error.h"
	int symlink(const char *from, const char *to)
	{
		::Warning("in CLib.c","symlink is not supported");
		return 0;
	}
#else
#if !defined(OS_AIX)
  extern "C" int symlink(const char *from, const char *to);
#endif
#endif
#if defined(OS_IRIX_5_3)
extern "C" void gnuqsort(void *pbase, size_t total_elems,size_t size,int (*cmp) (const void *, const void *));
#endif
 
#if defined(ultrix) || defined(__ultrix__) || defined(_AIX)
  extern "C" int setitimer (int, const struct itimerval *, struct itimerval *);
#endif
 
#if defined(__GNUG__) || defined(OS_WINDOWS) || defined(OS_SINIX)
#  include <fcntl.h>
#  include <errno.h>
#else
    #if defined(OS_USL)
    #  include <errno.h>
    #endif
    extern "C" {
	#if !defined(OS_USL) && !defined(OS_IRIX) && !defined(OS_OSF1) && !defined(OS_SINIX)
	    int creat (const char *, int);
	#endif
	int open (const char *, int, ...);
#if defined(OS_OSF1) || defined(OS_USL) || defined(OS_WINDOWS) 
	void *bsearch (const void *, const void *, size_t, size_t, 
			int (*)(const void*, const void*));
#else
#if !defined(hpux) &&  !defined(__hpux__) &&!defined(OS_IRIX_5_3) && !defined(OS_SINIX)
	char *bsearch (const char *, char *, unsigned, int, 
			int (*)(const void*, const void*));
#endif
#endif 
	int rename (const char *, const char *);
    }
#endif	/* __GNUG__ */

#ifdef __DECCXX
#  include <errno.h>
#endif

#ifdef OS_IRIX_5_3
#  include <errno.h>
#endif

#ifdef OS_NEXT_3
  extern "C" int putenv(const char *string);
#endif

extern "C" char **environ;
extern int gDebug;
extern char *gProgname;

//#define	MEMSTAT 1
#ifdef MEMSTAT

// =================== MemStat =====================
#define MAXMEM	1024*200
#define MAXSTAT	1024*100
class MemSlot
{
public:
	void		*mem;
	unsigned	size;
	MemSlot() { mem = 0; size = 0; }
};

class StatSlot
{
public:
	unsigned	size;
	unsigned	currCnt;
	unsigned	lastCnt;
	unsigned	currRealCnt;
	unsigned	lastRealCnt;
	StatSlot() { size = currCnt = lastCnt = currRealCnt = lastRealCnt =0; }
};

class MemStatCounters 
{
public:
	unsigned 	mallocCount,
			reallocCount,
			freeCount;
	int		sumMem, sumGap;
	char		*heapStart, *heapEnd;
	MemStatCounters() { mallocCount  = reallocCount  = freeCount = 0; 
			    heapStart = heapEnd = 0;}
};


class	MemStat
{
	MemSlot		slots[MAXMEM];	
	StatSlot	stats[MAXSTAT];
	MemStatCounters ctr;
	MemStatCounters ctrLast;
	unsigned	lastSlot, lastMStat, lastGStat, lastStat;	
	int		slotsSinceLastCompact;
	bool		memProtEnabled;
	int		breakIfSize;
public:
	MemStat() { lastSlot = lastMStat = lastGStat = lastStat = breakIfSize = 0; memProtEnabled = TRUE;}
	~MemStat() {}
	void	AllocateSlot(char *, unsigned);
	void	FreeSlot(char *);
	void	DoMalloc(char *addr, unsigned sz);
	void	DoFree(char *addr);
	void	DoRealloc(char *oldaddr, char *addr, unsigned sz);
	void	BuildStatistic();
	void	AddStatSlot(unsigned sz, StatSlot (&stats)[], unsigned &lastStat);
	void	AddRealStatSlot(unsigned sz, StatSlot (&stats)[], unsigned &lastStat);
	void	ResetRealStatSlot(StatSlot (&st)[], unsigned &lStat);
	void	Sort(int (*compare)(MemSlot*,MemSlot*));
	void 	Compress();

	// the following functions can be called from gdb
	// eg.: print memStat.Print()
	// or  p memStat.Reset()
	
	void	Print(int p = 0);	// print statistic. values in () are count since last Print()
					// p == 1: print changes only
	void	Check();		// check malloc overlapping
	void	Reset();		// reset counters
	void	Switch(bool o) { memProtEnabled = o; } // switch on=1/off=0 malloc/free protocolling (size statistic is always switched on)
	void	Fatal();		// seeting a breakpoint to Fatal()
					// allows to figure out where the failure has happend
	void	Break();
	void	BreakSize(int sz) { breakIfSize = sz; }
};

void
MemStat::Break()
{
	fprintf(stderr,"MemStat: size %d allocated.\n",breakIfSize);
}

void
MemStat::Reset()
{
    for(int i = 0; i < lastSlot; i++)
    {
	slots[i].mem = 0;
	slots[i].size = 0;
    }
    lastSlot = 0;
    for(i = 0; i < lastStat; i++)
    {
	stats[i].size = 0;
	stats[i].currCnt = 0;
	stats[i].lastCnt = 0;
	stats[i].currRealCnt = 0;
	stats[i].lastRealCnt = 0;
    }
    lastStat = 0;
}
void
MemStat::Fatal()
{
    // this function is only needed to be able to set a breakpoint on it
    fprintf(stderr,"FATAL !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
}
static int
compareMem(MemSlot *p1, MemSlot *p2)
{
	return (int)(p1->mem - p2->mem);
}
static int
compareStat(StatSlot *p1, StatSlot *p2)
{
    if(p1->currRealCnt*p1->size != p2->currRealCnt*p2->size)
	return (int)(p2->currRealCnt*p2->size - p1->currRealCnt*p1->size);
    else
	return (int)(p2->size - p1->size);
}
static int
compareSize(MemSlot *p1, MemSlot *p2)
{
	if(p1->size != p2->size)
	    return (int)(p1->size - p2->size);
	else
	    return (int)(p1->mem - p2->mem);
}

void
MemStat::DoMalloc(char *addr, unsigned sz)
{
    if(breakIfSize && breakIfSize == sz)
	Break();

    AddStatSlot(sz, stats, lastStat);
    
    if(memProtEnabled)
    {
	if(addr == NULL)
	{
		fprintf(stderr,"MemStat: FATAL: malloc NULL pointer\n");
		Fatal();
		return;
	}
	if(sz == 0)
	{
		fprintf(stderr,"MemStat: INFO: malloc ZERO size\n");
		return;	
	}
	ctr.mallocCount++;
	AllocateSlot(addr,sz);
	if(addr+sz > ctr.heapEnd)
	    ctr.heapEnd = addr + sz;
	if(addr < ctr.heapStart || ctr.heapStart == 0)
	    ctr.heapStart = addr;
    }
}
void
MemStat::DoFree(char *addr)
{
    if(memProtEnabled)
    {
	if(addr == NULL)
	{
		fprintf(stderr,"MemStat: FATAL: free NULL pointer\n");
		Fatal();
		return;
	}
	ctr.freeCount++;
	FreeSlot(addr);
    }
}
void
MemStat::DoRealloc(char *oldaddr, char *addr, unsigned sz)
{
    if(memProtEnabled)
    {
	if(addr == NULL)
	{
		fprintf(stderr,"MemStat: FATAL: realloc NULL pointer\n");
		Fatal();
		return;
	}
	if(oldaddr)
	{
		
	    DoFree(oldaddr);
	    ctr.freeCount--;
	}
	else
	{			// realloc with ZERO addr is a malloc 
	    ctr.reallocCount--;
	    ctr.mallocCount++;
	}
	
	ctr.reallocCount++;
	DoMalloc(addr,sz);
	ctr.mallocCount--;
    }
    else
	DoMalloc(addr,sz);
}
void
MemStat::AllocateSlot(char *addr, unsigned sz)
{
/*
    for(int i = 0; i < MAXMEM; i++)
	if(slots[i].size == 0)
	{
	    slots[i].size = sz;
	    slots[i].mem = addr;
	    if(i > lastSlot)
		lastSlot = i;
	    return;
	}
*/
    if(lastSlot >= MAXMEM-1)
    {
	fprintf(stderr,"MemStat: FATAL: malloc table overflow (%d slots)\n",MAXMEM);
	Fatal();
	return;
    }

    slots[lastSlot].mem = addr;
    slots[lastSlot++].size = sz;
    
//Check
//   for(int i = lastSlot-2; i >= 0 ; i--)
//        if(slots[i].mem == addr)
//        {
//		fprintf(stderr,"MemSat: FATAL: double allocated memory\n");
//		Fatal();
//        }

}
void
MemStat::Compress()
{
fprintf(stderr,"Compressing Memory.... ");
fflush(stderr);
    int compressed = 0;
    int nextNon0 = lastSlot-1;
    for(int curr = 0; curr < lastSlot; curr++)
    {
	if(slots[curr].size)
	    continue;
	    
// zero size slot found, look for the next nonZero slot
//	for(int nextNon0 = curr+1; nextNon0 < lastSlot && slots[nextNon0].size == 0; nextNon0++)
//	    ;

	// zero size slot found, look for the next nonZero slot
	for(; nextNon0 > curr && slots[nextNon0].size == 0; nextNon0--)
	    ;
//fprintf(stderr," gap=%d\n",nextNon0-curr);
	if(nextNon0 <= curr) // ready, all zero slots at the end
	{
	    for(int j=curr; j < lastSlot; j++)
		compressed++;
fprintf(stderr,"%d\n",compressed);
	    lastSlot = curr;	// reduce the current length
	    return;
	}
	else	// copy the next non zero slot into the curr slot
	{
	    slots[curr].mem = slots[nextNon0].mem;
	    slots[curr].size = slots[nextNon0].size;
	    slots[nextNon0].size = 0;
	    slots[nextNon0].mem = 0;
	}
    }
    // returning here: no zero slots found 
fprintf(stderr,"0\n");
}
void
MemStat::FreeSlot(char *addr)
{
    for(int i = lastSlot-1; i >= 0 ; i--)
	if(slots[i].mem == addr)
	{
	    slots[i].size = 0;
	    slots[i].mem = 0;
	    if(i == lastSlot-1)
		lastSlot--;
	    
	    if(slotsSinceLastCompact++ > 1000)
	    {
		Compress();
		slotsSinceLastCompact = 0;
	    }
	    return;
	}

//    fprintf(stderr,"MemStat: FATAL: freeing unallocated memory at %x\n",addr);
//   Fatal();
}
void
MemStat::BuildStatistic()
{
fprintf(stderr,"Building Statistics.... ");
fflush(stderr);
    int	prev = -1, gap = 0;
    ctr.sumMem = 0;
    ctr.sumGap = 0;
int cnt = 0;
    ResetRealStatSlot(stats, lastStat);
    for(int i = 0; i < lastSlot ; i++)
    {
	if(slots[i].size == 0)
	    continue;
	    
	cnt++;
	if(prev >= 0)	
	{
	    if(slots[i].mem == slots[prev].mem)
	    {
		fprintf(stderr,"MemStat: FATAL: Dupliated malloc on addr %d\n",slots[i].mem);
	    }
	    gap = (int)(slots[i].mem - slots[prev].mem) - slots[prev].size;
//fprintf(stderr,"%d - %d - %d = %d\n",slots[i].mem,slots[prev].mem,slots[prev].size,gap);
	}
	else
	    gap = 0;
	    
	ctr.sumMem  += slots[i].size;
	ctr.sumGap  += gap;

	AddRealStatSlot(slots[i].size, stats, lastStat);
	    
	prev = i;
    }
fprintf(stderr,"slots: %d\n",i);
}
void
MemStat::AddStatSlot(unsigned sz, StatSlot (&st)[], unsigned &lStat)
{
    if(sz == 0)
	return;
	
    for(int i = 0; i < lStat; i++)
    {
	if(st[i].size == sz)
	{
	    st[i].currCnt++;
	    return;
	}
    }
    if(i == MAXSTAT)
    {
	fprintf(stderr,"MemStat: FATAL: statistic table overflow (%d slots)\n",MAXSTAT);
	Fatal();
    }
    else
    {
	st[i].size = sz;
	st[i].currCnt = 1;
	lStat++;
    }
}
void
MemStat::ResetRealStatSlot(StatSlot (&st)[], unsigned &lStat)
{
    for(int i = 0; i < lStat; i++)
	st[i].currRealCnt = 0;
}
void
MemStat::AddRealStatSlot(unsigned sz, StatSlot (&st)[], unsigned &lStat)
{
    if(sz == 0)
	return;
	
    for(int i = 0; i < lStat; i++)
    {
	if(st[i].size == sz)
	{
	    st[i].currRealCnt++;
	    return;
	}
    }
    if(i == MAXSTAT)
    {
	fprintf(stderr,"MemStat: FATAL: statistic table overflow (%d slots)\n",MAXSTAT);
	Fatal();
    }
    else
    {
	st[i].size = sz;
	st[i].currRealCnt = 1;
	lStat++;
    }
}
void
MemStat::Check()
{
    Sort(compareMem);
    fprintf(stderr,"Checking for overlapping memory .... ");
    fflush(stderr);
    int	prev = -1, gap = 0;
    for(int i = 0; i < lastSlot ; i++)
    {
	if(slots[i].size == 0)
	    continue;
	    
	if(prev >= 0)	
	{
	    if(slots[i].mem < slots[prev].mem + slots[prev].size)
	    {
		fprintf(stderr,"MemStat: FATAL: allocated memory at %d (size=%d) overlaps %d (size=%d)\n",
		    slots[prev].mem,slots[prev].size,slots[i].mem,slots[i].size);
	    }
	}
	prev = i;
    }
    fprintf(stderr,"\n");
}

void
MemStat::Print(int flag)
{
    Sort(compareMem);
    BuildStatistic();

    float frg,d;
    fprintf(stderr,"========== Memory Statistic ===========\n");
    fprintf(stderr,"Allocated heap:     %-8d (%d) (start: %d  end: %d)\n",
	    ctr.heapEnd - ctr.heapStart, (ctr.heapEnd - ctr.heapStart)-(ctrLast.heapEnd - ctrLast.heapStart),
	    ctr.heapStart, ctr.heapEnd);
    fprintf(stderr,"Used bytes:         %-8d   (%d)\n",ctr.sumMem, ctr.sumMem - ctrLast.sumMem);
    fprintf(stderr,"Memory gap bytes:   %-8d   (%d)\n",ctr.sumGap, ctr.sumGap - ctrLast.sumGap);
    d=((float)(ctr.heapEnd - ctr.heapStart))*100;
    d=d?1/d:0;
    fprintf(stderr,"Memory fragm:       %4.2f%%    (%4.2f%%)\n",
	frg = d*ctr.sumGap,
	frg - d*ctrLast.sumGap);
    fprintf(stderr,"malloc() calls:     %-8d   (%d)\n",ctr.mallocCount, ctr.mallocCount - ctrLast.mallocCount);
    fprintf(stderr,"realloc() calls:    %-8d   (%d)\n",ctr.reallocCount, ctr.reallocCount - ctrLast.reallocCount);
    fprintf(stderr,"free() calls:       %-8d   (%d)\n",ctr.freeCount, ctr.freeCount - ctrLast.freeCount);

    ctrLast = ctr;

    if(flag == 2)
	return;
	
    fprintf(stderr,"----------- malloc() realloc() sizes -------------\n");

    qsort(stats,lastStat,sizeof(StatSlot),compareStat);

    int sumCurrCnt = 0, sumLastCnt = 0, sumCurrSize = 0, sumLastSize = 0;
    int sumCurrRealCnt = 0, sumLastRealCnt = 0, sumCurrRealSize = 0, sumLastRealSize = 0;
    float sumProcSize = 0;
	
    for(int i = 0; i < lastStat; i++)
    {
	if(flag == 1 && stats[i].currRealCnt == stats[i].lastRealCnt)
	    continue;
	sumCurrRealSize += stats[i].currRealCnt * stats[i].size;
    }
    fprintf(stderr,"Size   Curr alloc       Curr size (bytes)      %%           Sum  alloc       Sum size (bytes)\n");
    for(i = 0; i < lastStat; i++)
    {
	if(flag == 1 && stats[i].currRealCnt == stats[i].lastRealCnt)
	    continue;
	d=(float)sumCurrRealSize;
	d=d?1/d:0;
	sumProcSize += (float)(stats[i].currRealCnt*stats[i].size * 100)*d;
	    
	fprintf(stderr,"%-6d %6d(%-6d)   %8d(%-8d) %4.2f%% %5.2f%%   %6d(%-6d)   %8d(%-8d)\n",
	    stats[i].size,
	    stats[i].currRealCnt,stats[i].currRealCnt - stats[i].lastRealCnt,
	    stats[i].currRealCnt*stats[i].size, (stats[i].currRealCnt - stats[i].lastRealCnt)*stats[i].size,
	    ((float)(stats[i].currRealCnt*stats[i].size * 100)*d),
	    sumProcSize,
	    stats[i].currCnt,stats[i].currCnt - stats[i].lastCnt,
	    stats[i].currCnt*stats[i].size, (stats[i].currCnt - stats[i].lastCnt)*stats[i].size
	    );

	    
	sumCurrCnt += stats[i].currCnt;
	sumLastCnt += stats[i].currCnt - stats[i].lastCnt;
	sumCurrSize += stats[i].currCnt*stats[i].size;
	sumLastSize += (stats[i].currCnt - stats[i].lastCnt)*stats[i].size;

	sumCurrRealCnt += stats[i].currRealCnt;
	sumLastRealCnt += stats[i].currRealCnt - stats[i].lastRealCnt;
//	sumCurrRealSize += stats[i].currRealCnt*stats[i].size;
	sumLastRealSize += (stats[i].currRealCnt - stats[i].lastRealCnt)*stats[i].size;
	    
	stats[i].lastCnt = stats[i].currCnt;
	stats[i].lastRealCnt = stats[i].currRealCnt;
	
	    
   }
   fprintf(stderr,"\nSum:   %-6d(%6d)   %-8d(%8d)                %-6d(%6d)   %-8d(%8d)\n",
	    sumCurrRealCnt,sumLastRealCnt,
	    sumCurrRealSize, sumLastRealSize,
	    sumCurrCnt,sumLastCnt,
	    sumCurrSize, sumLastSize);
}
void
MemStat::Sort(int (*compare)(MemSlot *,MemSlot *))
{
    Compress();
    fprintf(stderr,"Sorting memory ...");
    fflush(stderr);
    qsort(slots,lastSlot,sizeof(MemSlot),compare);
    fprintf(stderr,"\n");
}
MemStat	memStat;
void PrintMemStat(){memStat.Print();}
void ResetMemStat(){memStat.Reset();}
#else /* MEMSTAT */
void PrintMemStat(){}
void ResetMemStat(){}
#endif /* MEMSTAT */


// ----- CLib --------------------------------------------------------

char **CLib::Environ()
{
    return environ;
}

int CLib::EnvironLength()
{
    int len = 0;
    for (int i = 0; environ[i]; i++)
	len += strlen (environ[i]) + 1;
    return len;
}

void CLib::QSort(
    void *base, int nel, int width, int (*cmp)(const void*, const void*)
)
{
#if defined(OS_IRIX_5_3)
    gnuqsort(base, (size_t)nel, (size_t)width, cmp);
#else
    qsort(base, nel, width, cmp);   
#endif
}
    
char *CLib::BSearch(
    const char *key, char *base, unsigned nel, int size, 
    int (*cmp)(const void*, const void*)
)
{
#if defined(OS_OSF1) ||  defined(OS_IRIX_5_3)
    return (char*) bsearch((const void *)key, (const void *)base, (size_t)nel, (size_t)size, cmp); 
#else
    return (char*) bsearch(key, base, nel, size, cmp);
#endif
}
   
char* CLib::Getenv(const char *name)
{
    return getenv(name);
}
    
int CLib::Putenv(const char *entry)
{
    return putenv((char *)entry);
}
    
//replaced by System::RunProgSynch()/RunProgAsynch()
//int CLib::System(const char *cmd)
//{
//    return system(cmd);
//}
    
void CLib::Abort()
{
    _exit(1);
}

#if 1
	// If nobody uses it we will remove it from next release
	// setitimer isn't supported on WindowsNT
#else
unsigned int CLib::Alarm(unsigned int seconds)
{
    struct itimerval val;
    struct itimerval oldval;

    val.it_value.tv_sec = seconds;
    val.it_value.tv_usec = 0;
    val.it_interval.tv_sec = seconds;
    val.it_interval.tv_usec = 0;
    setitimer(ITIMER_REAL, &val, &oldval);
    return oldval.it_value.tv_sec;
}
#endif

void CLib::Exit(int code)
{
    exit(code);
}

void CLib::FastExit(int code)
{
    _exit(code);
}

int CLib::Open(const char *name, int mode)
{
#ifndef OS_WINDOWS
    return open(name, mode);
#else
    return open(name, mode|O_BINARY);
#endif
}

int CLib::Close(int fd)
{
    return close(fd);
}

int CLib::Creat(const char *name, int mode)
{
#ifndef OS_WINDOWS
    return creat(name, mode);
#else
    //to keep low-level-interface small
    //maybe use the following also for (POSIX-)UNIX...
    //and/or change CLib::Open() also to POSIX-style...
    return open(name, O_WRONLY|O_CREAT|O_TRUNC | O_BINARY, mode);
#endif
}

#ifdef OS_WINDOWS_WINSOCK
#   include "Debug.h"
#   include <sys/socket.h>
    //"replacement" for CLib:Read()
    int CLib::Recv(int sock, char *buf, int sz, int flags)
    {
	int rv= recv(sock, buf, sz, flags);
	if (rv==0 || rv==SOCKET_ERROR) {
	    return -1;	//error
	} else {
	    ASSERT(rv==sz);//is my understanding of recv() correct
	    return rv;
	}
    }
    //"replacement" for CLib:Write()
    int CLib::Send(int sock, char *buf, int sz, int flags)
    {
	int rv= send(sock, buf, sz, flags);
	if (rv==0 || rv==SOCKET_ERROR) {
	    return -1;	//error
	} else {
	    ASSERT(rv==sz);//is my understanding of send() correct
	    return rv;
	}
    }
    //"replacement" for CLib:Close()
    int CLib::Closesocket(int sock)
    {
	return closesocket(sock);
    }
#endif
 
int CLib::Read(int fd, char *buf, int sz)
{
    int ret;

    do
	ret = read(fd, buf, sz);
    while (ret == -1 && errno == EINTR);

    if (ret < 0) {
#ifdef OS_WINDOWS
	if (errno == EAGAIN)
#else
	if (errno == EWOULDBLOCK || errno == EAGAIN)
#endif
	    return -2;
	return -1;      // fatal error
    }
    return ret;
}

int CLib::Write(int fd, const char *buf, unsigned int sz)
//used for any kind of writes! (also eg. sockets)
{
    int ret;
    do
	ret= write(fd, buf, sz);
    while (ret == -1 && errno == EINTR); 

    if (ret < 0) {
#ifdef OS_WINDOWS
	if (errno == EAGAIN)
#else
	if (errno == EWOULDBLOCK || errno == EAGAIN)
#endif
	    return -2;
	return -1;      // fatal error
    }
    return ret;
}

long CLib::Lseek(int fd, long off, int mode)
{
    return lseek(fd, off, mode);
}

int CLib::Unlink(const char *name)
{
    return unlink(name);    
}

#if 1
	// If nobody uses it we will remove it from next release
	// link isn't supported on WindowsNT
#else
int CLib::Link(const char *from, const char *to)
{
    return link(from, to);
}
#endif

int CLib::Rename(const char *from, const char *to)
{
    return rename(from, to);
}

int CLib::Symlink(const char *from, const char *to)
{
    return symlink(from, to);
}

static void
malloc_failed (void)
{
    static char msg[] = ": virtual memory exhausted\n";
    write (2, gProgname, strlen(gProgname));
    write (2, msg, sizeof(msg) - 1);
    exit (1);
}
 
static void (*malloc_failed_hook) (void) = malloc_failed;
 
void CLib::SetMallocFailedHook ( void (*hook) (void) )
{
    malloc_failed_hook = hook;
}

#ifdef OS_WINDOWS
#define MEMCLEAR_RUEDI(p, len)  if ((len) > 0) memset((void*)(p), 0, (len))
int CLib::Free(char *p)
{
    free(p);
    return 0;
}
 
char *CLib::ReAlloc (char *p, int sz)
{
    //void *vp= ::realloc(p, sz);
    //if (vp == 0)  Fatal("CLib::ReAlloc", "storage exhausted");
    //MEMCLEAR_RUEDI(vp, size);	//ruedi: we should only clear the new stuff!!
    //return (char*)vp;
    
    char *cp;
    if ( sz <= 0 )	sz = 1;
    while ( (cp=(char*)::realloc(p,sz)) == 0 )
	malloc_failed_hook ();
    //MEMCLEAR_RUEDI(vp, size);	//ruedi: we should only clear the new stuff!!
    return cp;
}
 
char *CLib::CAlloc (int n, int sz)
{
    //void *vp= ::calloc(n, sz);
    //if (vp == 0) Fatal("CLib::Calloc", "storage exhausted");
    //MEMCLEAR_RUEDI(vp, n*sz);	//ruedi: calloc already sets to 0. but bbsts
    //return (char*)vp;
 
    char *cp;
    if ( sz <= 0 )	sz = 1;
    if ( n <= 0 )	n = 1;
    while ( (cp=(char*)::calloc(n,sz)) == 0 )
	malloc_failed_hook ();
    MEMCLEAR_RUEDI(cp, n*sz);	//ruedi: calloc already sets to 0. but bbsts
    return cp;
}
 
#else
 
int CLib::Free(char *p)
{
#ifdef MEMSTAT
    memStat.DoFree(p);
#endif
    free(p);
    return 0;
}

char *CLib::ReAlloc (char *p, int sz)
{
    char *cp;
    if ( sz <= 0 )
	sz = 1;

    while ( (cp = (char *) realloc (p, sz)) == 0 )
	malloc_failed_hook ();

#ifdef MEMSTAT
    memStat.DoRealloc(p,cp,sz);
#endif
    return cp;
}

char *CLib::CAlloc (int n, int sz)
{
    if(gPrintMem)
	fprintf(stderr,"n=%d sz=%d\n",n,sz);	
    char *cp;
    if ( sz <= 0 )
	sz = 1;
    if ( n <= 0 )
	n = 1;
    while ( (cp = (char *) calloc (n, sz)) == 0 )
	malloc_failed_hook ();
#ifdef MEMSTAT
    memStat.DoMalloc(cp,sz*n);
#endif
    if(gPrintMem)
	fprintf(stderr,"n=%d sz=%d\n",n,sz); 
    return cp;
}
#endif //OS_WINDOWS

// --- Glob --- file name pattern matching ----------------------------------

int
Glob::HasPattern (char *pattern)
{
  register char *p = pattern;
  register char c;
  int	open = 0;

  while ((c = *p++) != '\0')
    {
    switch (c)
      {
      case '?':
      case '*':
	return 1;

      case '[':		// Only accept an open brace if there is a close
	open++;		// brace to match it.  Bracket expressions must be
	break;		// complete, according to Posix.2

      case ']':
	if (open)
	  return 1;
	break;      

      case '\\':
	if (*p++ == '\0')
	  return 0;
	break;      
      }
    }
  return 0;
}

int
Glob::Match (char *pattern, char *text, int dot_special)
{
  register char *p = pattern, *t = text;
  register char c;

  while ((c = *p++) != '\0')
    switch (c)
      {
      case '?':
	if (*t == '\0' || (dot_special && t == text && *t == '.'))
	  return 0;
	else
	++t;
	break;

      case '\\':
	if (*p++ != *t++)
	  return 0;
	break;

      case '*':
	if (dot_special && t == text && *t == '.')
	  return 0;
	return Glob::MatchAfterStar(p, t);

      case '[':
	{
	register char c1 = *t++;
	int invert;

	if (!c1)
	  return (0);
	invert = ((*p == '!') || (*p == '^'));
	if (invert)
	  p++;

	c = *p++;
	while (1)
	  {
	  register char cstart = c, cend = c;

	  if (c == '\\')
	    {
	    cstart = *p++;
	    cend = cstart;
	    }

	  if (c == '\0')
	    return 0;

	  c = *p++;
	  if (c == '-' && *p != ']')
	    {
	    cend = *p++;
	    if (cend == '\\')
	      cend = *p++;
	    if (cend == '\0')
	      return 0;
	    c = *p++;
	    }
	  if (c1 >= cstart && c1 <= cend)
	    goto match;
	  if (c == ']')
	    break;
	  }
	if (!invert)
	  return 0;
	break;

	match:		/* Skip the rest of the [...] construct that already matched. */
	while (c != ']')
	  { 
	  if (c == '\0')
	    return 0;
	  c = *p++;
	  if (c == '\0')
	    return 0;
	  else if (c == '\\')
	    ++p;
	  }
	if (invert)
	  return 0;
	break;
	}

        default:
	  if (c != *t++)
	    return 0;
      }
  return *t == '\0';
}

int
Glob::MatchAfterStar (char *pattern, char *text)
{
  register char *p = pattern, *t = text;
  register char c, c1;

  while ((c = *p++) == '?' || c == '*')
    if (c == '?' && *t++ == '\0')
      return 0;

  if (c == '\0')
    return 1;

  if (c == '\\')
    c1 = *p;
  else
    c1 = c;

  while (1)
    {
      if ((c == '[' || *t == c1) && Glob::Match(p - 1, t, 0))
	return 1;
      if (*t++ == '\0')
	return 0;
    }
}
