#ifdef __GNUG__
#pragma implementation
#endif

#include "ClassManager.h"

#include "Class.h"
#include "ETString.h"
#include "OrdColl.h"
#include "SortedOList.h"
#include "ObjectTable.h"

ClassManager *gClassManager;

//---- ClassManager ------------------------------------------------------------

NewMetaImpl(ClassManager,Object, (TP(classTable), TB(sorted)));

ClassManager::ClassManager()
{
    classTable= 0;
    sorted= FALSE;
}

ClassManager::~ClassManager()
{
    SafeDelete(classTable);
}

//---- class table management --------------------------------------------------

OrdCollection *ClassManager::sortClassTable()
{
    classTable->Sort();
    sorted= TRUE;
    return classTable;
}

void ClassManager::Add(Class *cl)
{
    if (classTable == 0)
	classTable= new OrdCollection(300);
    if (classTable->FindPtr(cl) == 0)
	classTable->Add(cl);
    sorted= FALSE;
    //Send(cIdNone, 8888, cl);
}

void ClassManager::Remove(Class *cl)
{
    if (classTable)
	classTable->RemovePtr(cl);
}

Class *ClassManager::Find(char *name, double version)
{
    Class lookFor(name, version);  
    int at= GetClassTable()->BinarySearch(&lookFor);
    if (at != -1)
	return (Class*) GetClassTable()->At(at);
    return 0;
}

Class *ClassManager::FindOrLoad(char *name, double version)
{
    Class *clp = Find(name, version);
    if (clp == 0)
	Error("FindOrLoad", "class %s not linked to application", name);
    return clp;
}

Object *ClassManager::Load(char *module, char *name)
{
    Class *clp = gClassManager->Find(name);
    if (clp == 0) {
	Error("Load", "class %s not linked to application", name);
	return 0;
    }
    return clp->Proto()->New();
}

void ClassManager::SetupSubclasses()
{
    OrdCollectionIter next(GetClassTable());
    register Class *clp, *super;

    while (clp= (Class*) next()) 
	if (clp->SubClasses()) 
	    clp->SubClasses()->Empty();

    next.Reset();
    while (clp= (Class*) next()) 
	if (super= clp->Super())
	    super->AddSubclass(clp);
}

//---- sort list according to number of instances of a class -------------------

class CumSortedObjList : public SortedObjList {
public:
    CumSortedObjList()
	{ }
    int CompareObjects(Object*, Object*);
};
    
int CumSortedObjList::CompareObjects(Object *op1, Object *op2)
{
    return Guard(op2, Class)->GetInstanceCount() - Guard(op1, Class)->GetInstanceCount();
}

void ClassManager::InstanceStatistics(bool cumsorted)
{
    Class *clp;
    int cum= 0;
    Collection *col, *csl= 0;
    
    ObjectTable::UpdateInstCount(FALSE);
    if (cumsorted) {
	col= csl= new CumSortedObjList;
	csl->AddAll(classTable);
    } else
	col= GetClassTable();
    Iter next(col);

    fprintf(stdout, "Instances per class on heap\n");
    fprintf(stdout, "=============================\n");
    
    int	sum = 0, sum1;
    while (clp= (Class*) next()) 
    	{
		if (clp->GetInstanceCount() > 0) 
			{
			sum1 = clp->GetInstanceCount() * clp->Size();
			sum += sum1;
		    fprintf(stdout, "%-20s%8d%8d%9d%9d\n", clp->Name(),
						clp->GetInstanceCount(), clp->Size(), sum1, sum);
		    cum+= clp->GetInstanceCount();
			}
    	}
    fprintf(stdout, "-----------------------------\n");
    fprintf(stdout, "Total %d\n", cum);
    fprintf(stdout, "=============================\n");
    
    SafeDelete(csl);
}

