#ifdef __GNUG__
#pragma implementation
#endif

#include "Error.h"
#include "Debug.h"	//ruedi
#include "ETString.h"
#include "System.h"
#include "Env.h"

static ErrorHandlerFunc errorhandler= DefaultErrorHandler;

ErrorHandlerFunc SetErrorHandler(ErrorHandlerFunc newhandler)
{
    ErrorHandlerFunc oldhandler= errorhandler;
    errorhandler= newhandler;
    return oldhandler;
}

ErrorHandlerFunc GetErrorHandler()
{
    return errorhandler;
}  

void AbstractMethod(char *s)
{
    Warning(s, "abstract method called");	//ruedi: removed comments
}

void MayNotUse(char *s)
{
    Warning(s, "may not use");			//ruedi: removed comments
}

void DefaultErrorHandler(int level, bool abort, char *location, char *msg)
{
    char *type= "Warning";
    if (level < Env::GetValue("System.IgnoreLevel", cError))
	return;
    if (level >= cFatal)
	type= "Fatal";
    else if (level >= cSysError)
	type= "SysError";
    else if (level >= cError)
	type= "Error";
	  
#ifdef OS_WINDOWS    
    TPR("%s in <%s>: %s\n", type, location, msg);		  
#else
    fprintf (stderr, "%s in <%s>: %s\n", type, location, msg);
    fflush (stderr);
#endif					  
    if (abort) {
	fprintf (stderr, "aborting\n");
	fflush (stderr);
	gSystem->Abort();
    }
}

void ErrorHandler(int level, char *location, char *fmt, va_list ap)
{
    char buf[1024], *bp;
    
    vsprintf(buf, fmt, ap);
    if (level >= cSysError && level < cFatal)
	bp= form("%s (%s)", buf, gSystem->GetError());
    else
	bp= buf;
    errorhandler(level, level >= Env::GetValue("System.AbortLevel", cSysError), location, bp);
}

void Error(char *location, char *va_(fmt), ...)
{
    va_list ap;
    va_start(ap,va_(fmt));
    ErrorHandler(cError, location, va_(fmt), ap);
    va_end(ap);
}

void SysError(char *location,  char *va_(fmt), ...)
{
    va_list ap;
    va_start(ap, va_(fmt));
    ErrorHandler(cSysError, location, va_(fmt), ap);
    va_end(ap);
}

void Warning(char *location, char *va_(fmt), ...)
{
    va_list ap;
    va_start(ap,va_(fmt));
    ErrorHandler(cWarning, location, va_(fmt), ap);
    va_end(ap);
}

void Fatal(char *location, char *va_(fmt), ...)
{
    va_list ap;
    va_start(ap,va_(fmt));
    ErrorHandler(cFatal, location, va_(fmt), ap);
    va_end(ap);
}

static int gAssertLevel= -1;

int GetAssertLevel()
{
    if (gAssertLevel == -1 && gSystem) 
	gAssertLevel= Env::GetValue("System.AssertLevel", 0);
    return gAssertLevel;
}

int SetAssertLevel(int l)
{
    int old= gAssertLevel;
    gAssertLevel= l;
    return old;
}

