#ifdef __GNUG__
#pragma implementation
#endif

//#define USE_MMAP

#include "FileBuf.h"

#include "CLib.h"
#include "System.h"
#include "Env.h"
#include "Error.h"
#include "Debug.h"
#include <string.h>

#if defined(OS_SUNOS)
#	include <unistd.h>
#endif

extern "C" char *mmap(char*, size_t, int, int, int, size_t);
extern "C" int munmap(char*, size_t);
extern "C" int madvise(char*, size_t, int);

Filebuf::Filebuf() : StreamBuf(0)
{
    fd= -1;
    opened= mapped= FALSE;
    itsFile=0;
}

Filebuf::Filebuf(int nfd) : StreamBuf(0)
{
    fd= nfd;
    opened= TRUE;
    mapped= FALSE;
    itsFile=0;
}

Filebuf::Filebuf(int nfd, char* p, int l) : StreamBuf(l, p)
{
    fd= nfd;
    opened= TRUE;
    mapped= FALSE;
    if(p && *p) {
   	itsFile=new char[strlen(p)+1];
    	strcpy(itsFile,p);
    }
    else itsFile=0;
}

Filebuf::Filebuf(char* name, open_mode om) : StreamBuf(0)
{
    opened= FALSE;
    open(name, om);
    if(name && *name) {
    	itsFile=new char[strlen(name)+1];
    	strcpy(itsFile,name);
    }	
    else itsFile=0;
}

Filebuf::~Filebuf()
{
    close();
    SafeDelete(itsFile);
}

Filebuf* Filebuf::open (char *name, open_mode om)
{
    close();
    
    fpos= 0;
    switch (om) {
    case input:     
	fd= CLib::Open(name, O_RDONLY);
#ifdef USE_MMAP
	if (gSystem && Env::GetValue("System.MMAP", 0)) {
	    int id, flags;
	    long size, modtime;
	    gSystem->GetPathInfo(name, &id, &size, &flags, &modtime);
	    base= gptr= mmap(0, (size_t)size, 1, 1, fd, 0);
	    if (base == (char*) -1) {
		perror("mmap");
	    } else {
		madvise(base, (size_t)size, 2);
		fprintf(stderr, "%s s:%d a:%08x fd:%d\n", name, size, base, fd);
		ptr= base;
		gptr= eptr= base + size;
		fprintf(stderr, "mmap: %d %d\n", fpos, fpos+(gptr-base));
		alloc= FALSE;
		CLib::Close(fd);
		mapped= TRUE;
	    }
	}
#endif  
	break;
    case output:    
	fd= CLib::Creat(name, 0664);
	break;
    case append:    
	fd= CLib::Open(name, O_WRONLY);
	if (fd < 0)
	    fd= CLib::Creat(name, 0664);
	if (fd >= 0)
	    fpos= CLib::Lseek(fd, 0, SEEK_END);
	break;
    }

    if (fd < 0) {
	opened= FALSE;
	return 0;
    }
    
    if (base == 0)
	setbuf(0, BUFSIZE);

    opened= TRUE;
    return this;
}

bool Filebuf::close()
{
    if (opened) {
	if (mapped) {
#ifdef USE_MMAP
	    fprintf(stderr, "unmap 0x%08x\n", base);
	    munmap(base, eptr-base);
	    mapped= FALSE;
#endif            
	} else {
	    if (fd > 2)
		CLib::Close(fd);
	}
	opened= FALSE;
	return TRUE;
    }
    return FALSE;
}

int Filebuf::Overflow(u_char *bp, int size)
{
    if (opened)
	return CLib::Write(fd, (char*)bp, size);
    return EOF;
}

int Filebuf::Underflow(u_char *p, int l)
{
    if (mapped || !opened)
	 return EOF;
    return CLib::Read(fd, (char*)p, l);
}

long Filebuf::Seek(long pos, int whence)
{
    return CLib::Lseek(fd, pos, whence);
}



//---- FilebufAscii ---------------------------------------------------------------


//---- FilebufAsciiEncoder --------------------------------------------------------------
FilebufAsciiEncoder::FilebufAsciiEncoder(char *path, ELineSeparatorStyle lss) 
    : FilebufAscii(path, output)
{
    fLineSeparatorStyle= lss;
}

int FilebufAsciiEncoder::Overflow(u_char *bp, int size)
{
    //return Filebuf::Overflow(bp, size);
    int countLF= 0;
    int countCR= 0;
    int i;
    for (i=0; i<size; i++) {	//count the LFs
	if (bp[i]=='\n') countLF++;
	if (bp[i]=='\r') countCR++;
    }
    if (countCR!=0) {
	Warning("FilebufAsciiEncoder::Overflow", 
		"There should be no '\r' in stream - probably used wrong StreamBuf!" );
    }
    
    if (fLineSeparatorStyle==cLineSeparatorStyleDos) {
	int i,j= 0;
	u_char *bp2= (u_char*)Storage::ChunkAlloc(size+countLF);	
	for (i=0,j=0; i<size; i++) {
	    if (bp[i]=='\n') {
		bp2[j++]= '\r'; bp2[j++]= '\n'; 
	    } else {
		bp2[j++]= bp[i];
	    }
	}
	ASSERT(j == size+countLF);
	int rv= Filebuf::Overflow(bp2, size+countLF);
	Storage::ChunkFree(bp2);
	ASSERT(rv==size+countLF);
	//return rv; // caller wants NOT to know how many effectively written, 
	return size; // BUT how many of the logical stream got written
    } else if (fLineSeparatorStyle==cLineSeparatorStyleMac) {
	u_char *bp2= (u_char*)Storage::ChunkAlloc(size);	
	for (int i=0; i<size; i++) {
	    if (bp[i]=='\n') {
		bp2[i]='\r';
	    } else {
		bp2[i]= bp[i];
	    }
	}
	int rv= Filebuf::Overflow(bp2, size);
	Storage::ChunkFree(bp2);
	return rv;
    } else {	//UNIX, rest...
	//nothing to do
	return Filebuf::Overflow(bp, size);
    }
}


//---- FilebufAsciiDecoder --------------------------------------------------------------
FilebufAsciiDecoder::FilebufAsciiDecoder(char *path) 
    : FilebufAscii(path, input)
{
}

int FilebufAsciiDecoder::Underflow(u_char *bp, int size)
{
    //if (mapped || !opened)
    //   return EOF;
    //return CLib::Read(fd, (char*)p, l);

    if (mapped || !opened)
	 return EOF;

    char ch;
    //problem: was the char just before our first char a CR?
    //general policy: CR is immediately considered as lineseparator,
    //NL is only considered as lineseparator if not after a CR
    bool lastCharIsCR= FALSE;
    int lpos= CLib::Lseek(fd, -1, SEEK_CUR);
    if (lpos>=0) {
	if (CLib::Read(fd, &ch, 1)==1 ) {
	    if (ch=='\r') lastCharIsCR= TRUE;
	} else {
	    Error("FilebufAsciiDecoder::Underflow", "impossible!?!");
	}
    }

    int i=0;
    while (i<size) {
	int r;
	if ((r=CLib::Read(fd, &ch, 1)) < 1) {
	    if (r==0) {	//EOF
		if(i>0) return i;
		else	return EOF;
	    } else return r;
	} else {
	    //make canonical lineseparators (i.e. only \n internally)
	    // UNIX:    LF   -> LF
	    // WIN/DOS: CRLF -> LF
	    // MAC:     CR   -> LF
	    if (ch=='\r') {
		bp[i++]= '\n';
		lastCharIsCR= TRUE;
	    } else if (ch=='\n') {
		if (lastCharIsCR) {
		    //do nothing
		} else {
		    bp[i++]= '\n';
		}
		lastCharIsCR= FALSE;
	    } else {
		bp[i++]= ch;
		lastCharIsCR= FALSE;
	    }
        }
    }
    return i;
}


