#ifndef ObjectTable_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define ObjectTable_First

#include "Object.h"
#include "Iterator.h"

//---- Object Table ----------------------------------------------------------
// this class registers all instances of Object or its subclasses
// in a hash table

class ObjectTable : public Object {
friend class ObjectTableIter;
    static Object *lastobject;
    static bool died;
    static Class *tab[1000];
    static Object **table;
    static int size, tally, iterlevel;
    static class OrdCollection *delayed; // insertions while an iterator is active are delayed
    static OrdCollection *rdelayed;
    //---- statistics
    static int convoi, seed;
    static long cumconvoi;

    static bool HighWaterMark()
	{ return (bool) (tally >= (3 * size /4)); }
    static void Expand(int newsize);
    static int FindElement (Object*);
    static void FixCollisions (int index);
    static void HashStatistics();

    ObjectTable();

public:
    MetaDef(ObjectTable);

    ~ObjectTable();
    static void Add(Object *op);
    static void Remove(Object *op);
    static void UpdateInstCount(bool heaponly= FALSE);
    static bool PtrIsValid(Object *op)
	{ return table ? (table[FindElement(op)] != 0) : TRUE; }
    static void *CheckPtrAndWarn(char *msg, void *vp);
    static void Remember(OrdCollection *ct);
    static int Instances()          // return total number of instances
	{ return tally; }
    static void InstanceStatistics();
    static Object *SomeInstance(Class *);
    static Object *SomeMember(Class *);
};

extern "C" void gIsA(Object *op);

class ObjectTableIter : public Iterator {
    int cursor;              // for iterating
    class Class *fromClass;  // looking for instances of class in operator()()
    bool members;            // return only direct members of a clas in "   "
public:
    ObjectTableIter(Class *cl= 0, bool members= TRUE);
    ~ObjectTableIter();
    Object *operator()();
};

#endif

