#ifndef Rectangle_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define Rectangle_First

#include "Metric.h"

class Rectangle {
public:
    Point origin;
    Point extent;
      
public:
    
    Rectangle();

    Rectangle(int x, int y, int w, int h);

    Rectangle(const Point &o, const Point &e);
	
    Rectangle(const Point &e);

    Rectangle(int w, int h);
    
    Rectangle(const Rectangle&);
    
    int Width() const
	{ return extent.x; }

    int Height() const
	{ return extent.y; }

    int Left() const
	{ return origin.x; }

    int Top() const
	{ return origin.y; }
	
    Point Center() const
	{ return Point(origin.x+extent.x/2, origin.y+extent.y/2); }

    Point NW() const
	{ return origin; }
	
    Point N() const
	{ return Point(origin.x+extent.x/2, origin.y); }

    Point NE() const
	{ return Point(origin.x+extent.x-1, origin.y); }

    Point E() const
	{ return Point(origin.x+extent.x-1, origin.y+extent.y/2); }
	
    Point SE() const
	{ return origin + extent - 1; }
	
    Point S() const
	{ return Point(origin.x+extent.x/2, origin.y+extent.y-1); }
	
    Point SW() const
	{ return Point(origin.x, origin.y+extent.y-1); }

    Point W() const
	{ return Point(origin.x, origin.y+extent.y/2); }
	    
    friend Rectangle NormRect(const Point &p1, const Point &p2);
	
    friend Rectangle BoundingBox(int, Point *pts, Point *npts= 0);

    int Area() const
	{ return extent.x * extent.y; }

    Rectangle &Moveby(const Point &p)
	{ origin+= p; return *this; }
	
    friend Rectangle Moveby(Rectangle r, const Point &p)
	{ r.origin+= p; return r; }
	
    Rectangle operator+= (const Point &p)
	{ origin+= p; return *this; }
	
    Rectangle operator+ (const Point &p)
	{ return Rectangle(origin+p, extent); }
	
    Rectangle operator- (const Point &p)
	{ return Rectangle(origin-p, extent); }
	
    Rectangle operator-= (const Point &p)
	{ origin-= p; return *this; }

    void Moveto(const Point &p)
	{ origin= p; }

    void Scaleby(const Point &p)
	{ extent*= p; }

    friend bool operator== (const Rectangle &r1, const Rectangle &r2)
	{ return (bool) (r1.origin == r2.origin && r1.extent == r2.extent); }

    friend bool operator!= (const Rectangle &r1, const Rectangle &r2)
	{ return (bool) (r1.origin != r2.origin || r1.extent != r2.extent); }
	
    friend Rectangle Union(const Rectangle &r1, const Rectangle &r2);

    friend Rectangle Inter(Rectangle r1, const Rectangle &r2);

    friend int Difference(Rectangle *rp, const Rectangle &r1, const Rectangle &r2);
    
    Rectangle Intersect(const Rectangle &r);

    bool Clip(const Rectangle &r);
    
    bool ContainsRect(const Rectangle &r) const;
    bool OvalContainsRect(const Rectangle &r) const;
    
    Rectangle Inset(const Point &p) const;

    Rectangle Expand(const Point &p) const;
	
    Rectangle Merge(const Rectangle &r);

    bool ContainsPoint(const Point &p) const
	{ return (bool) (p >= origin && p < origin + extent); }

    bool Intersects(const Rectangle &r) const;
		  
    bool IsEmpty() const
	{ return (bool) (extent.x <= 0 || extent.y <= 0); }
	    
    bool IsNotEmpty() const
	{ return (bool) (extent.x > 0 && extent.y > 0); }

    Point Constrain(const Point &p);
	
    friend Point ConstrainMoveRect(const Rectangle &r1, const Rectangle &r2, Point delta);

    Point AmountToTranslateWithin(const Rectangle &r);

    int PointToAngle(const Point &) const;
    
    Point AngleToPoint(int) const;
    Point OvalAngleToPoint(int) const;
	    
    Rectangle WedgeBBox(int s, int e);
    
    //         7|0|1
    //        -------
    //         6| |2
    //        -------
    //         5|4|3

    int PointToCorner(const Point &p) const;

    Point CornerToPoint(int n) const;
    
    friend OStream& operator<< (OStream &s, const Rectangle &r);
		   
    friend IStream& operator>> (IStream &s, Rectangle &r);
    
    char *AsString() const;

    static int Diff(Rectangle *rp, const Rectangle &r1, const Rectangle &r2);
};

SimpleMetaDef(Rectangle);

extern const Rectangle gRect0;

#endif
