#ifndef Region_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define Region_First

#include "Rectangle.h"
#include "Storage.h"
#include "ETString.h"

class Region;
class RegionRectIter;
class RegionImpl;
//--- Region -------------------------------------------------------------------

class Region {
public:
    enum EOpcode {
	eUnion           = 0,
	eDifference      = 1,
	eRevDifference   = 2,
	eIntersection    = 3,
	eCopy            = 4
    };
 
    Region();
    Region(const Coord *s, long len);
    Region(const Rectangle &r);
    Region(const Region &r);
    ~Region();

    Region &operator=(const Region &r);
    const Rectangle &BBox() const;
    long GetRects(Rectangle *rp) const;
    long Length() const;
    Coord *GetBuf() const;
    bool IsNotEmpty() const;
    bool IsEmpty() const;
    bool ContainsPoint(const Point &p) const;
    bool IsEqual(const Region &r2) const;
    friend inline bool operator==(const Region &r1, const Region &r2);
    friend inline bool operator!=(const Region &r1, const Region &r2);
 
    friend Region RegionOp(EOpcode code, const Region &r1, const Region &r2);
    friend Region SymDifference(const Region &r1, const Region &r2);
    friend inline Region Union(const Region &r1, const Region &r2);
    friend inline Region Intersection(const Region &r1, const Region &r2);
    friend inline Region Difference(const Region &r1, const Region &r2);

    OStream &PrintOn(OStream &os) const;
    IStream &ReadFrom(IStream &is);
    friend inline OStream& operator<< (OStream &os, const Region &r);
    friend inline IStream& operator>> (IStream &is, Region &r);
protected:
    RegionImpl *fImpl;
};

SimpleMetaDef(Region);

//---- RegionRectIter ----------------------------------------------------------
 
class RegionRectIter {
    Region fRegion;
    Coord *fBufPtr;
    long fNext;
public:
    RegionRectIter(const Region &r);
    bool operator()(Rectangle &r);
};
inline bool operator==(const Region &r1, const Region &r2)
    { return r1.IsEqual(r2); }
inline bool operator!=(const Region &r1, const Region &r2)
    { return !r1.IsEqual(r2); }

inline Region Union(const Region &r1, const Region &r2)
    { return RegionOp(Region::eUnion, r1, r2); }
inline Region Intersection(const Region &r1, const Region &r2)
    { return RegionOp(Region::eIntersection, r1, r2); }
inline Region Difference(const Region &r1, const Region &r2)
    { return RegionOp(Region::eDifference, r1, r2); }

inline OStream& operator<< (OStream &os, const Region &r)
    { return r.PrintOn(os); }
inline IStream& operator>> (IStream &is, Region &r)
    { return r.ReadFrom(is); }

//---- OldRegion ---------------------------------------------------------------

class OldRegion {
public:
    int size;
    Rectangle *rl, bbox;
public:
    OldRegion();
    OldRegion(Rectangle &r);
    ~OldRegion();
    void Merge(Rectangle);
    Rectangle *GetAt(int ix);
    void operator=(OldRegion &r);
    void Empty();

#if !defined(OS_WINDOWS)
    //---- memory allocation
    void *operator new (size_t sz)
	{ return Storage::ChunkAlloc(sz); }
    void operator delete(void *vp)
	{ Storage::ChunkFree(vp); }
#endif
};

#endif
