#ifdef __GNUG__
#pragma implementation
#endif

#include "StreamConnection.h"

#include "Class.h"
#include "Error.h"
#include "ETString.h"
#include "MemBuf.h"
#include "OrdColl.h"
#include "Set.h"
#include "CLib.h"
#include "ETMath.h"
#include "Env.h"
#include "CmdNo.h"

const int cHeaderSize = 45;

bool gIACDebug;
StreamConnection *gRpc;

OrdCollection *gServers;

static ProxyReference ref(0);

//---- ConnectionStack -------------------------------------------------

class ConnectionStack {
public:
    static void Push(ServiceConnection *s);
    static void Pop();
    static ServiceConnection *Top();
private:
    static void Init();
    static OrdCollection *cons;
};

OrdCollection *ConnectionStack::cons= 0;

void ConnectionStack::Push(ServiceConnection *s)
{
    Init();
    cons->AddLast(s);
}

void ConnectionStack::Pop()
{
    if (!cons || !cons->Size()) {
    	fprintf(stderr,"ConnectionStack::Pop call with empty stack\n");
    	return;
    }
    cons->RemoveLast();
}

ServiceConnection * ConnectionStack::Top()
{
    if (!cons || !cons->Size()) {
    	fprintf(stderr,"ConnectionStack::Top call with empty stack\n");
    	return 0;
    }
    return Guard(cons->Last(),ServiceConnection);
}

void ConnectionStack::Init()
{
    if (!cons)
    	cons= new OrdCollection;
}

//---- Message -----------------------------------------------------------------

block::block()
{
    len= cHeaderSize;
    pos= 0;
    buf= 0;
}

block::~block()
{
    SafeDelete(buf);
}

NewMetaImpl0(Message, Object);

Message::Message()
{
    to= from= 0;
    serial= 0;
}
	
Message::Message(int t, int f, int reply, char *buf, int l, bool makecopy)
{
    to= t;
    from= f;
    serial= reply;
    if (l < 0)
	l= strlen(buf)+1;
    data[1].len= l;
    if (makecopy) {
	data[1].buf= new char[l];
	strncpy(data[1].buf, buf, l);
    } else
	data[1].buf= buf;
}

int Message::Read(int fd)
{
    for (int i= 0; i < 2; i++) {
	block *b= &data[i];
	
//	if (b->pos < b->len) {
	while (b->pos < b->len) {
	    if (b->buf == 0)
		b->buf= new char[b->len];
#ifdef OS_WINDOWS_WINSOCK
	    int cnt= CLib::Recv(fd, b->buf+b->pos, b->len-b->pos, 0);
#else
	    int cnt= CLib::Read(fd, b->buf+b->pos, b->len-b->pos);
#endif
	    if (cnt == 0 || cnt == -1)
		return -1;
	    if (cnt == -2)
		return -2;
    
	    b->pos+= cnt;
	    
	    if (b->pos >= b->len) {
		if (i == 1) {
		    return 0;
		}
		if (sscanf(b->buf, "%d %d %d %d",
				    &to, &from, &data[1].len, &serial) != 4)
		    return -1;
	    }
//	    return -2;
	}
    }
    return 0;
}

int Message::Write(int fd)
{
    for (int i= 0; i < 2; i++) {
	block *b= &data[i];
	
	while (b->pos < b->len) {
#ifdef OS_WINDOWS_WINSOCK
	    int wl= CLib::Send(fd, b->buf+b->pos, b->len-b->pos, 0);
#else
	    int wl= CLib::Write(fd, b->buf+b->pos, b->len-b->pos);
#endif
	    if (wl == -1)   // fatal error
		return -1;
	    if (wl == -2)   // try again
		return -2;
	    b->pos+= wl;
	}
    }
    Print("Write", 0);
    return 0;                   // completely written
}

void Message::Print(char *msg, char*)
{
    if (gIACDebug) {
	int l= data[1].len;
	
	fprintf(stderr, "%s fr:%d to:%d ser:%d l:%d <", msg, from, to, serial, l);
	if (l < 60) {
	    
	    fprintf(stderr, "%s>\n", data[1].buf);
	} else {
	    fwrite(data[1].buf, 1, 60, stderr);
	    fprintf(stderr, "...>\n");
	}
    }
}

IStream &Message::ReadFrom(IStream &is)
{
    is >> from >> to >> serial;
    fprintf(stderr,"from %d to %d serial %d\n",from,to,serial);
    is.ReadString(&data[0].buf);
    fprintf(stderr,"data[0].buf %s\n",data[0].buf);
    data[0].len= strlen(data[0].buf);
    is.ReadString(&data[1].buf);
    fprintf(stderr,"data[1].buf %s\n",data[1].buf);
    data[1].len= strlen(data[1].buf);
    return is;
}

OStream &Message::PrintOn(OStream &os)
{
    os << from SP << to SP << serial SP;
    os.PrintString(data[0].buf) NL;
    os.PrintString(data[1].buf) NL;
    return os;
}

//---- StreamConnection --------------------------------------------------------

NewMetaImpl0(StreamConnection, SysEvtHandler); 

int StreamConnection::timeout = 60000;
 
StreamConnection::StreamConnection(int f) : SysEvtHandler(f)
{
    gIACDebug= Env::GetValue("IAC.Debug", FALSE);
    serial= 100;
    wqueue= 0;
    queue= 0;
    rm= 0;
    sid= 0;
    name= 0;
    //SetName(gProgname);
    error= !Open(f);
    gSystem->AddWorkHandler(this);
}

StreamConnection::~StreamConnection()
{
    error= TRUE;
#ifdef OS_WINDOWS_WINSOCK
    CLib::Closesocket(GetResourceId());
#else
    CLib::Close(GetResourceId());
#endif
    SafeDelete(name);
}

void StreamConnection::SetName(char *n)
{
    SafeDelete(name);
    name= strsave(n);
}
    
bool StreamConnection::Open(int fd)
{
    if (fd >= 0) {
	SetResourceId(fd);
	gSystem->AddFileInputHandler(this);
	gSystem->AddFileOutputHandler(this);
	error= FALSE;
	return TRUE;
    }
    error= TRUE;
    return FALSE;
}

int StreamConnection::Error(int code)
{
    error= TRUE;
    if (gIACDebug)
	fprintf(stderr, "StreamConnection::Error: disconnecting from server; reason: %d\n", code);
    return 0;
}

bool StreamConnection::Reconnect()
{
/*
    if (error) {
	int sock = gSystem->OpenConnection(SERVERNAME, SERVICENAME);
	if (sock >= 0)
	    error = !Open(sock);
    }
*/
    return error;
}

void StreamConnection::SetTimeout(int to)
{
    timeout= to;
}

void StreamConnection::SendMessage(Message *m)
{
    if (m) {
	SafeDelete(m->data[0].buf);
	m->data[0].len= cHeaderSize;
	m->data[0].pos= 0;
	m->data[0].buf= new char[cHeaderSize];
	m->data[1].pos= 0;
	sprintf(m->data[0].buf, "%d %d %d %d", m->to, m->from, m->data[1].len, m->serial);
	if (wqueue == 0)
	    wqueue= new OrdCollection;
	wqueue->Add(m);
    }
}

int StreamConnection::DrainOutput(int timeout)
{
    if (Reconnect())
	return -1;
    // wait for output to drain
    while (wqueue) {
	if (gSystem->CanWrite(GetResourceId(), timeout)) {
	    if (gIACDebug)
		fprintf(stderr, "DrainOutput: timeout\n");
	    return -1;
	}
	Notify(eSysEvtWrite, GetResourceId());
    }
    return 0;
}

int StreamConnection::WaitForMessage(int timeout)
{
    if (Reconnect())
	return -1;
    if (gSystem->CanRead(GetResourceId(), timeout))
	return -1;
    Notify(eSysEvtRead, GetResourceId());
    return 0;
}

Message *StreamConnection::WaitForReply(int seq, int timeout)
{
    if (Reconnect())
	return 0;
    for (;;) {
	if (queue && queue->Size() > 0) {
	    register Message *m;
	    Iter next(queue);
	    
	    while (m= (Message*) next()) {
		if (m->IsReply()) {    // a reply
		    int s= m->serial;
		    queue->RemovePtr(m);
		    if (s == -seq) {
			return m;
		    } else {
			delete m;
			if (gIACDebug)
			    fprintf(stderr, "lost sync\n");
			return 0;
		    }
		}
	    }
	}
	
	DrainOutput(4000);

	if (WaitForMessage(timeout) == -1) {
	    if (gIACDebug)
		fprintf(stderr, "CanRead: timeout\n");
	    break;
	}
    }

    return 0;
}

bool StreamConnection::HasInterest(SysEventCodes code)
{
    if (error)
	return FALSE;
    if ( ! SysEvtHandler::HasInterest (code) )
	return FALSE;
    if ( code == eSysEvtWrite )
	return (wqueue != 0 && wqueue->Size() > 0);
    return TRUE;
}

void StreamConnection::Notify(SysEventCodes sec, int)
{
    if (error)
	return;
	
    switch (sec) {
    case eSysEvtWork:
	if (queue && queue->Size() > 0) {
	    Message *m= (Message*) queue->RemoveFirst();
	    if (m) {
		if (Dispatch2(m))
		    delete m;
	    }
	}
	break;
	
    case eSysEvtRead:
	if (rm == 0)
	    rm= new Message;
	    
	switch (rm->Read(GetResourceId())) {
	case 0:     // got one
	    if (queue == 0)
		queue= new OrdCollection;
	    if (rm->from == 0)
		rm->from= sid;
	    rm->Print("Received", 0);
	    queue->Add(rm);
	    rm= 0;
	    break;
	case -1:    // fatal error
	    Error(2);
	    break;
	case -2:    // try again later
	    break;
	}
	break;

    case eSysEvtWrite:
	if (wqueue && wqueue->Size() > 0) {
	    Message *m= (Message*) wqueue->First();
	    if (m) {
		switch (m->Write(GetResourceId())) {
		case 0:     // completely written
		    if ((Message*)wqueue->RemovePtr(m) == m) {
			SafeDelete(m);
			if (wqueue->Size() <= 0)
			    SafeDelete(wqueue);
		    } else  // inconsistent queue
			Error(5);
		    break;
		case -1:    // fatal error
		    Error(4);
		    break;
		case -2:    // try again
		    break;
		}
	    }
	}
	break;
    }
}

int StreamConnection::SendReply(Message *m, char *buf, int len, bool dodelete)
{
    int rc= TRUE;
    if (buf)
	rc= IntSendTo(m->from, -m->serial, 0, buf, len);
    if (dodelete)
	SafeDelete(buf);
    return rc;
}

bool StreamConnection::Dispatch2(Message *m)
{
    bool eventHanldedByService= FALSE;
    if (m->IsReply()) {     
	if (gIACDebug)
	    fprintf(stderr, "StreamConnection::Dispatch2: oops: got a reply!!\n");
    } else {        // not a reply
	char req[200], *bp= req, *buf= m->data[1].buf, *cp;
	int len= m->data[1].len;
    
	cp= buf;
	while (*cp && *cp != ' ' && bp < &req[199])
	    *bp++= *cp++;
	*bp= 0;
	if (*cp == ' ')
	    cp++;
	len-= (cp-buf);
	buf= cp;
    
	if (gServers) {
	    Iter next(gServers);
	    Service *sv;
	    char servicename[200];
	    
	    strcpy(servicename, req);
	    char *cp= strchr(servicename, ':');
	    if (cp)
		*cp++= 0;
	    int servicenumber;
	    if (sscanf(servicename,"%d",&servicenumber) == 1) {
	    	// dispatching by service number -> used by proxies
		if (gIACDebug)
	    	    fprintf(stderr, "StreamConnection::Dispatch2: dispatching by service number %d\n",
	    	    	servicenumber);
	    	Guard(gServers->At(servicenumber),Service)->
	    	    ExtCommand(m, this, cp, buf, len);
	    }
	    while (sv= (Service*) next()) {
		if (cp) {   // syntax:  service:request
		    if (strcmp(servicename, sv->GetName()) == 0) { // found
			sv->ExtCommand(m, this, cp, buf, len);
			eventHanldedByService= TRUE;
		    }
		} else {    // syntax:  request
		    sv->ExtCommand(m, this, servicename, buf, len);
		}
	    }
	}
	if (!eventHanldedByService) {
	    char *retbuf= 0;
	    int retlen= -1;
	    Dispatch(m->from, req, buf, len, retbuf, retlen);
	    SendReply(m, retbuf, retlen);
	}
    }
    return TRUE;
}

//---- old interface -----------------------------------------------------------

int StreamConnection::SendTo(int to, char *req, char *buf, int l)
{
    return IntSendTo(to, serial++, req, buf, l);
}

int StreamConnection::Talk(int to, char *req, char *buf, int len, 
						  char **retbuf, int *retlen)
{
    if (Reconnect())
	return -1;

    int seq= serial++;

    if (!IntSendTo(to, seq, req, buf, len))
	return FALSE;
	
    if (to != 0) {

	Message *m= WaitForReply(seq, timeout);
	
	if (m) {
	    if (retbuf) {
		*retbuf= m->data[1].buf;
		m->data[1].buf= 0;
	    }
	    if (retlen)
		*retlen= m->data[1].len;
	    delete m;
	}
    }
    
    return TRUE;
}

void StreamConnection::Talk(char *prog, char *va_(fmt), ...)
{
    va_list ap;
    char *rbuf = 0;
    int to;

    va_start(ap, va_(fmt));

    Talk(1, "start", prog, -1, &rbuf);
    if ( rbuf ) {
	if ( sscanf(rbuf, "%d", &to) == 1 ) {
	    char buf[2000];
	    strcpy(buf, prog);
	    strcat(buf, ":");
	    vsprintf(&buf[strlen(buf)], va_(fmt), ap);
	    if (gIACDebug)
		fprintf(stderr, "[%s]\n", buf);
	    Talk(to, buf);
	}
	delete rbuf;
    }

    va_end(ap);
}

int StreamConnection::IntSendTo(int to, int reply, char *req, char *buf, int l)
{
    if (Reconnect())
	return FALSE;
	
    int rl= 0;
    
    if (req)
	rl= strlen(req);
    
    if (buf) {
	if (l < 0)
	    l= strlen(buf);
    } else
	l= 0;
	
    int tl= l + rl + 1;
    if (l > 0 && rl > 0)
	tl++;
	
    char *bp= new char[tl];
    
    if (rl > 0)
	strcpy(bp, req);
    if (l > 0 && rl > 0)
	strcat(bp, " ");
    if (l > 0)
	strcat(bp, buf);
	
    SendMessage(new Message(to, 0, reply, bp, tl, FALSE));

    return TRUE;
}

void StreamConnection::Dispatch(int, char*, char*, int, char *&, int&)
{
}

//---- Service -----------------------------------------------------------------

NewAbstractMetaImpl(Service,Dispatchable, (TP(name), TP(svcs)));

Service::Service(const char *nm) : Dispatchable(0)
{
    name= strsave((char *)nm);
    if (gServers == 0)
	gServers= new OrdCollection;
    SetId(gServers->Size());
    gServers->Add(this);
    svcs= 0;
    isDeleted= FALSE;
}

Service::~Service()
{
    SafeDelete(name);
    if (gServers) {
	gServers->RemovePtr(this);
	if (gServers->Size() <= 0)
	    SafeDelete(gServers);
    }
}

void Service::ExtCommand(Message *m, StreamConnection *sc, 
					    char *req, char *args, int arglen)
{
    MemBuf *mb= new MemBuf;
    OStream os(mb);

    MemBuf *mb1= new MemBuf;
    mb1->sputn(args, arglen);
    mb1->SwitchToRead();
    IStream is(mb1);

    Dispatch(m->from, req, is, os);
    
    os.flush();
    mb->SwitchToRead();
    if (mb->ContentsSize() > 0)
	sc->SendReply(m, mb->Base(), (int) mb->ContentsSize());
}

void Service::Dispatch(int, char *request, IStream &is, OStream &os)
{
	if (strcmp(request, "enroll") == 0 || strcmp(request, "remove") == 0) {
	}
	else
	  fprintf(stderr,"Service::Dispatch unknown request %s\n", request);
}

void Service::AddServiceConnection(ServiceConnection *s)
{
    if (svcs == 0)
    	svcs= new OrdCollection();
    svcs->Add(s);
}

Collection *Service::Services()
{
    return svcs;
}

void Service::RemoveServiceConnection(ServiceConnection *s)
{
    svcs->RemovePtr(s);
}

//---- ServiceConenction -----------------------------------------------------------------

class  KeyValueAssoc : public Object {
public:
    MetaDef(KeyValueAssoc);
    KeyValueAssoc(Object *k, ObjectSvc *op);
    u_long Hash();
    bool IsEqual (Object *o);

    Object *key;
    ObjectSvc *proxy;
};

inline u_long KeyValueAssoc::Hash() { return key->Hash(); };
inline   bool KeyValueAssoc::IsEqual (Object *o) { return key == o; }
 
 
NewMetaImpl(KeyValueAssoc,Object, (TP(key), TP(proxy)));

KeyValueAssoc::KeyValueAssoc(Object *k, ObjectSvc *op)
{
    key= k;
    proxy= op;
}

NewAbstractMetaImpl(ServiceConnection,Service, (T(connectionId), T(connectionPort)));

ServiceConnection::ServiceConnection(int cport, int cid) : Service("")
{
    connectionPort= cport;
    connectionId= cid;
    nextid= 0;
    proxymap= 0;
}
ServiceConnection::~ServiceConnection()
{
     // Do not forget: remove reference in Service
     if (proxymap) {
         proxymap->FreeAll();
         delete proxymap;
     }
}

void ServiceConnection::ExtCommand(Message *m, StreamConnection *sc, 
					    char *req, char *args, int arglen)
{
    MemBuf *mb= new MemBuf;
    OStream os(mb);

    MemBuf *mb1= new MemBuf;
    mb1->sputn(args, arglen);
    mb1->SwitchToRead();
    IStream is(mb1);
    
    int proxynumber;
    if (sscanf(req,"%d:",&proxynumber) == 1) {
        // dispatching to proxies
	if (gIACDebug)
    	    fprintf(stderr, "Service::ExtCommand: dispatching by service number %d\n",
    	    	proxynumber);
        char *c= strchr(req,':');
        Proxy p(proxynumber);
        Guard(svcs->Find(&p),Dispatchable)
        	->Dispatch(m->from, ++c, is, os);
    }
    else
    	Dispatch(m->from, req, is, os);
    
    os.flush();
    mb->SwitchToRead();
    if (mb->ContentsSize() > 0)
	sc->SendReply(m, mb->Base(), (int) mb->ContentsSize());
}

const char *ServiceConnection::GetName()
{
    static char buf[10];
    sprintf(buf,"%d",ConnectionId());
    return buf;
}

Collection *ServiceConnection::Proxies()
{
    return svcs;
}

int  ServiceConnection::NextProxyId()
{
    return nextid++;
}

void ServiceConnection::AddProxy(Proxy *s)
{
    if (svcs == 0) {
	svcs= new Set();
	proxymap= new Set();
    }
    svcs->Add(s);
    if (s->IsKindOf(ObjectSvc))
    	proxymap->Add(new KeyValueAssoc(Guard(s,ObjectSvc)->GetObject(),
    		(ObjectSvc*)s));
}

void ServiceConnection::RemoveProxy(Proxy *p)
{
    svcs->Remove(p);
    if (p->IsKindOf(ObjectSvc)) {
    	Object *kva= proxymap->Remove(Guard(p,ObjectSvc)->GetObject());
    	SafeDelete(kva);
    }
}

void ServiceConnection::DeleteProxy(Proxy *p)
{
    svcs->Remove(p);
    if (p->IsKindOf(ObjectSvc)) {
    	Object *kva= proxymap->Remove(Guard(p,ObjectSvc)->GetObject());
    	SafeDelete(kva);
    }
    delete p;
}

Proxy *ServiceConnection::FindObjectProxy(Object *o)
{
    KeyValueAssoc k(o,0);
    KeyValueAssoc *kva= (KeyValueAssoc*)proxymap->Find(&k);
    if (kva)
    	return kva->proxy;
    else
    	return 0;
}

Proxy *ServiceConnection::GetRefForObjectProxy(Object *o)
{
    KeyValueAssoc k(o,0);
    KeyValueAssoc *kva= (KeyValueAssoc*)proxymap->Find(&k);
    if (kva == 0)
    	return 0;
    ref.SetId(kva->proxy->Id());
    return &ref;
}

void ServiceConnection::PrepareRequest()
{
    SetupStreams();
    ConnectionStack::Push(this);
}

void ServiceConnection::FinishRequest()
{
    ConnectionStack::Pop();
}

IStream &ServiceConnection::ReadFrom(IStream &is)
{
    is >> connectionPort >> connectionId;
    return is;
}

OStream &ServiceConnection::PrintOn(OStream &os)
{
    os << GetPort() SP << Id() NL;
    return os;
}

//---- ServiceConnectionSvc ----------------------------------------------

NewMetaImpl0(ServiceConnectionSvc, ServiceConnection);

ServiceConnectionSvc::ServiceConnectionSvc(int cport, int cid)
    : ServiceConnection(cport,cid)
{
}

ServiceConnectionSvc::~ServiceConnectionSvc()
{
    if (svcs) {
        svcs->FreeAll();
    	delete svcs;
    }
    
}

void ServiceConnectionSvc::Dispatch(int from, char *req, IStream &is, OStream &os)
{
    if (strcmp(req,"ProxyDeleted") == 0) {
    	gSystem->AddCleanupObject(this);
    	os.PrintString("Ok") NL;
    }
    else
    	ServiceConnection::Dispatch(from,req,is,os);
};

//---- ServiceConnectionProxy --------------------------------------------

NewMetaImpl0(ServiceConnectionProxy, ServiceConnection);

ServiceConnectionProxy::ServiceConnectionProxy(int cport, int cid)
    : ServiceConnection(cport,cid)
{
}

ServiceConnectionProxy::~ServiceConnectionProxy()
{
    Deleted();
    PrepareRequest();
    Talk(ConnectionPort(),"ProxyDeleted");
    char buf[10];
    is->ReadString((unsigned char**)&buf);
    FinishRequest();
    if (svcs) {
    	svcs->FreeAll();
    	delete svcs;
    }
}

//---- ProxyReference -----------------------------------------------------------------

NewMetaImpl(ProxyReference, Proxy, (T(id)));

ProxyReference::ProxyReference(int i)
{
    id= i;
}

void ProxyReference::SetId(int i)
{
    id= i;
}

Object *ProxyReference::ReadAndMap(IStream &is)
{
    int i;
    is >> i;
    Proxy p(i);
    return ConnectionStack::Top()->Proxies()->Find(&p);
}

OStream& ProxyReference::PrintOn(OStream &os)
{
    os << id NL;
    return os;
}

IStream& ProxyReference::ReadFrom(IStream &is)
{
    is >> id;
    return is;
}   

//---- Dispatchable -----------------------------------------------------------------

NewMetaImpl(Dispatchable,SysEvtHandler, (T(id)));

int Dispatchable::port= -1;

const char *Dispatchable::GetName()
{
    AbstractMethod("GetName");
    return 0;
}

int Dispatchable::GetPort()
{
    if (port == -1) {
	char *retbuf, *dummy;
	int retlen;
	gRpc->Talk(1, "myport", dummy, 0, &retbuf, &retlen);
	if (retbuf != 0)
	    sscanf(retbuf, "%d", &port);
	else {
	    fprintf(stderr,"Cannot obtzin my own port number from appcomm.");
	    return -1;
	}
    }
    return port;
}

Dispatchable::Dispatchable(int ident) : SysEvtHandler(0)
{
    id= ident;
    os= 0;
    is= 0;
}

Dispatchable::~Dispatchable()
{
    SafeDelete(is);
    SafeDelete(os);
}

int Dispatchable::Talk(int to, char *request)
{
    char *buf= 0;
    int len= 0;

    os->flush();
    omb->SwitchToRead();

    gRpc->Talk(to, form("%s:%s", GetName(), request),
				omb->Base(), (int)omb->ContentsSize(), &buf, &len);

    SafeDelete(os);
    omb= new MemBuf(10000);
    os= new OStream(omb);
	
    SafeDelete(is);
    imb= new MemBuf;
    is= new IStream(imb);
	
    if (buf) {
	imb->sputn(buf, len);
	imb->SwitchToRead();
	SafeDelete(buf);
    }
    
    return 0;
}

int Dispatchable::SendTo(int to, char *request)
{
    os->flush();
    omb->SwitchToRead();

    gRpc->SendTo(to, form("%s:%s", GetName(), request), omb->Base(),
						(int)omb->ContentsSize());

    SafeDelete(os);
    omb= new MemBuf;
    os= new OStream(omb);
	
    return 0;
}

void Dispatchable::SetupStreams()
{
    if (os == 0) {
    	omb= new MemBuf();
    	os= new OStream(omb);
    }
}

void Dispatchable::Dispatch(int, char*, IStream&, OStream&)
{
    AbstractMethod("Dispatch");
}

OStream &Dispatchable::PrintOn(OStream &os)
{
    os << Id();
    return os;
}

IStream &Dispatchable::ReadFrom(IStream &is)
{
    is >> id;
    return is;
}


//---- Proxy -----------------------------------------------------------------

NewAbstractMetaImpl(Proxy, Dispatchable, (TP(service)));

Proxy::Proxy(int oid) : Dispatchable(oid)
{
}

Proxy::~Proxy()
{ 
}

const char *Proxy::GetName()
{
	static char buf[50];
	sprintf(buf,"%d:%d",ConnectionServiceId(),Id());
	return buf;
}

int Proxy::ConnectionPort()
{
    return GetService()->ConnectionPort();
}

int Proxy::ConnectionServiceId()
{
    return GetService()->ConnectionId();
}

void Proxy::PrepareRequest()
{
    SetupStreams();
    ConnectionStack::Push(GetService());
}

void Proxy::FinishRequest()
{
    ConnectionStack::Pop();
}

int  Proxy::NextProxyId()
{
    return GetService()->NextProxyId();
}

void Proxy::AddProxy(Proxy *s)
{
    GetService()->AddProxy(s);
}

void Proxy::RemoveProxy(Proxy *s)
{
    GetService()->RemoveProxy(s);
}

void Proxy::DeleteProxy(Proxy *s)
{
    GetService()->DeleteProxy(s);
}

u_long Proxy::Hash () 
{
    return Id();
}

OStream &Proxy::PrintOn(OStream &os)
{
    return os;
}

IStream &Proxy::ReadFrom(IStream &is)
{
    return is;
}

//---- ObjectSvc -------------------------------------------------------

NewAbstractMetaImpl0(ObjectProxy, Proxy);

ObjectProxy::ObjectProxy(int oid) : Proxy(oid)
{
    Proxy::InitNew();
    InitNew();
}

ObjectProxy::~ObjectProxy()
{
    if (!GetService()->IsDeleted()) {
    	GetService()->RemoveProxy(this);
        PrepareRequest();
        SendTo(ConnectionPort(),"ProxyDeleted");
        FinishRequest();
    }
}

void ObjectProxy::InitNew()
{
    SetService(ConnectionStack::Top());
    SetId(NextProxyId());
    AddProxy(this);
}

void ObjectProxy::ResetCaches()
{
}

void ObjectProxy::Dispatch(int, char *request, IStream &is, OStream &os)
{
    if (strcmp(request, "ProxyDeleted") == 0) {
        GetService()->DeleteProxy(this);
    }
    else if (strcmp(request, "ProxyChanged") == 0) {
        ResetCaches();
        Send(cIdNone,cPartProxyChanged);
    }
    else {
        fprintf(stderr,"ObjectProxy::Dispatch %s", request);
        os << "error";
    }
}

//---- ObjectSvc -------------------------------------------------------

NewAbstractMetaImpl(ObjectSvc, Proxy, (TP(obj)));

ObjectSvc::ObjectSvc(int oid, ServiceConnection *s, Object *o) : Proxy(oid)
{
    SetService(s);
    obj= o;
    if (obj)
    	obj->AddObserver(this);
}

ObjectSvc::~ObjectSvc()
{
    if (obj)
        obj->RemoveObserver(this);
}

void ObjectSvc::Dispatch(int, char *request, IStream &is, OStream &os)
{
    if (strcmp(request, "ProxyDeleted") == 0) {
        GetService()->DeleteProxy(this);
    }
    else {
        fprintf(stderr,"ObjectSvc::Dispatch %s", request);
        os << "error";
    }
}

void ObjectSvc::DoObserve(int id, int part, void *what, Object *op)
{
    if (id == cIdNone && part == cPartSenderDied) {
        SendTo(ConnectionPort(),"ProxyDeleted");
        GetService()->DeleteProxy(this);
    } else
    	SendTo(ConnectionPort(),"ProxyChanged");
}
