#ifndef StreamConnection_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define StreamConnection_First

#include "System.h"
#include "EvtHandler.h"

#define SERVERNAME "sniffappcomm"
#define SERVICENAME "sniffappcomm"

extern bool gIACDebug;
extern OrdCollection *gServers;

class MemBuf;
class SeqCollection;
class Set;
class StreamConnection;
class ServiceConnection;
class Dictionary;
class Proxy;
class ObjectSvc;

//---- Message -----------------------------------------------------------------

struct block {
    int len, pos;
    char *buf;
    
    block();
    ~block();
};

class Message : public Object {
public:
    int from, to, serial;
    block data[2];

public:
    MetaDef(Message);
    Message();
    Message(int to, int from, int reply, char *buf, int l= -1, bool makecopy= TRUE);
    int Read(int fd);
    int Write(int fd);
    void Print(char *from, char *to);
    bool IsReply()
	{ return serial < 0; }

    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);
};

//---- StreamConnection --------------------------------------------------------

class StreamConnection : public SysEvtHandler {
protected:
    Message *rm;
    SeqCollection *wqueue;
    OrdCollection *queue;
    bool error;
    int serial, sid;
    char *name;
    static int timeout;

    bool Open(int fd);

public:
    int IntSendTo(int to, int reply, char *req, char *buf, int l);
    int SendTo(int to, char *req, char *buf= 0, int l= -1);
public:
    MetaDef(StreamConnection);
    StreamConnection(int);
    ~StreamConnection();
    void SetName(char*);
    
    void SendMessage(Message *m);
    int SendReply(Message *m, char *retbuf, int retlen= -1, bool dodelete= FALSE);
    Message *WaitForReply(int seq, int timeout);
    int DrainOutput(int timeout);
    int WaitForMessage(int timeout);
    bool Reconnect();
    static void SetTimeout(int to);
    static int GetTimeout();

    // obsolete interface
    int Talk(int client, char *req, char *buf= 0, int len= -1,
					    char **retbuf= 0, int *retlen= 0);
    void Talk(char *prog, char *va_(fmt), ...);

    virtual void Dispatch(int from, char *req, char *buf, int len,
						    char *&retbuf, int &retlen);
    virtual bool Dispatch2(Message *m);
    int Error(int code);

protected:
    bool HasInterest(SysEventCodes code);
    void Notify(SysEventCodes, int);
};

inline int StreamConnection::GetTimeout()
{
    return timeout;
}

class Dispatchable : public SysEvtHandler {
public:
    MetaDef(Dispatchable);
    Dispatchable(int ident);
    ~Dispatchable();
    virtual const char *GetName();
    int Talk(int to, char *request);
    int SendTo(int to, char *request);
    void SetupStreams();
    virtual void Dispatch(int from, char *req, IStream &is, OStream &os);
    int Id() { return id; }
    static int GetPort();

    OStream& PrintOn(OStream&);
    IStream& ReadFrom(IStream&);
protected:
    void SetId(int i) { id= i; }
protected:
    MemBuf *omb, *imb;
    OStream *os;
    IStream *is;
    int id;
    static int port;
};

//---- Service -----------------------------------------------------------------

class Service : public Dispatchable {
public:
    MetaDef(Service);
    Service(const char *name);
    ~Service();
    const char *GetName() { return name; }
    virtual void ExtCommand(Message *m, StreamConnection *sc,
					    char *req, char *args, int arglen);
    void Dispatch(int from, char *req, IStream &is, OStream &os);
    
    //---- cleanup handling
    void Deleted() { isDeleted= TRUE; }
    bool IsDeleted() { return isDeleted; }
    
    //---- service connection handling
    Collection *Services();
    void AddServiceConnection(ServiceConnection *s);
    void RemoveServiceConnection(ServiceConnection *s);
protected:
    char *name;
    Collection *svcs;
    bool isDeleted;
};

//---- ServiceConnection -------------------------------------------------

class ServiceConnection : public Service {
public:
    MetaDef(ServiceConnection);
    ServiceConnection(int cport= -1, int cid= -1);
    ~ServiceConnection();
    void ExtCommand(Message *m, StreamConnection *sc,
			char *req, char *args, int arglen);
    int ConnectionPort() { return connectionPort; }
    int ConnectionId() { return connectionId; }
     const char*GetName(); // connection name
   
    //---- proxy handling
    Collection	*Proxies();
    int		NextProxyId();
    void	AddProxy(Proxy *s);
    void	RemoveProxy(Proxy *p);
    void	DeleteProxy(Proxy *s);
    Proxy	*FindObjectProxy(Object *o);
    Proxy	*GetRefForObjectProxy(Object *o);
    
    //---- request handling
    void PrepareRequest();
    void FinishRequest();
    
    //---- input/output
    virtual OStream& PrintOn(OStream&);
    virtual IStream& ReadFrom(IStream&);
    
protected:
    int connectionPort;
    int connectionId;
    int nextid;
    Set *proxymap;
};

//---- ServiceConnectionSvc ----------------------------------------------

class ServiceConnectionSvc : public ServiceConnection {
public:
    MetaDef(ServiceConnectionSvc);
    ServiceConnectionSvc(int cport= -1, int cid= -1);
protected:
    ~ServiceConnectionSvc();
    void Dispatch(int from, char *req, IStream &is, OStream &os);
};

//---- ServiceConnectionProxy --------------------------------------------

class ServiceConnectionProxy : public ServiceConnection {
public:
    MetaDef(ServiceConnectionProxy);
    ServiceConnectionProxy(int cport= -1, int cid= -1);
    ~ServiceConnectionProxy();
};

//---- Proxy -------------------------------------------------------------

class Proxy: public Dispatchable {
public:
    MetaDef(Proxy);
    Proxy(int oid= -1);
    ~Proxy();
	
    //---- access
    const char*GetName(); // connection name
    int	  ConnectionPort();
    int	  ConnectionServiceId();
    
    //---- request handling
    void PrepareRequest();
    void FinishRequest();
	
    //---- service/subservice handling
    ServiceConnection	*GetService() { return service; }
    int  NextProxyId();
    void AddProxy(Proxy *s);
    void RemoveProxy(Proxy *s);
    void DeleteProxy(Proxy *s);

    //---- comparing -------------------------------------------------------
    u_long Hash();
    
    //---- input/output
    virtual OStream& PrintOn(OStream&);
    virtual IStream& ReadFrom(IStream&);

protected:
    void SetService(ServiceConnection *s) { service= s; }
protected:
    ServiceConnection *service;
};

//---- ProxyReference ----------------------------------------------------

class ProxyReference : public Proxy {
public:
    MetaDef(ProxyReference);
    ProxyReference(int i);
    void SetId(int i);
    Object *ReadAndMap(IStream&);
    virtual OStream& PrintOn(OStream&);
    virtual IStream& ReadFrom(IStream&);    
protected:
    int id;
};

//---- ObjectProxy ---------------------------------------------------------

class ObjectProxy: public Proxy {
public:
    MetaDef(ObjectProxy);
    ObjectProxy(int oid);
    ~ObjectProxy();
    void Dispatch(int from, char *req, IStream &is, OStream &os);

    //---- change propagation
    virtual void ResetCaches();
protected:
    void InitNew();
};

//---- ObjectSvc ---------------------------------------------------------

class ObjectSvc: public Proxy {
public:
    MetaDef(ObjectSvc);
    ObjectSvc(int oid, ServiceConnection *s, Object *o);
    ~ObjectSvc();
    Object *GetObject() { return obj;}
    void Dispatch(int from, char *req, IStream &is, OStream &os);
    void DoObserve(int id, int part, void *what, Object *op);
protected:
    Object *obj;
};

#endif

