#ifdef __GNUG__
#pragma implementation
#endif

#include "System.h"

#include "Class.h"
#include "CLib.h"
#include "ETString.h"
#include "Error.h"
#include "Directory.h"  
#include "OrdColl.h"
#include "Env.h"
#include "Point.h"

char    *gEtDir;
System  *gSystem;
char    *gProgname= "???";
char    *gProgPath;
bool     gInMain;
int      gDebug;
bool     gQuitApp;

static bool cleanuplock;

 
void ETInit(int argc, char **argv)
{
    if (gEtDir == 0)
	gEtDir = CLib::Getenv("ET_DIR");

    char *arg;
    for (int i = 1; i < argc; i++) {
	arg = argv[i];
	if (arg[0] == '-' && arg[1] == '-' && arg[2])
	    Env::SetValue(&arg[2], eEnvCommand);
    }

    gDebug = Env::GetValue("System.Debug", 0);

    if (Env::GetValue("System.MemStat", 0))
	Storage::EnableStatistics();
    int msize = Env::GetValue("System.MemStat.size", -1);
    int mcnt = Env::GetValue("System.MemStat.cnt", -1);
    if (msize != -1 || mcnt != -1)
	Storage::EnableStatistics(msize, mcnt);
}

void ETCleanup()
{
    SafeDelete(gSystem);
}
void ETInitAll(int argc, char **argv)
{
    // Initializes et++ in the standard way
    gEtDir = CLib::Getenv("ET_DIR");
    gInMain= TRUE;
    InitSystem(argv[0]);
    if (gEtDir == NULL || *gEtDir == 0)
    {
	char *gSniffDir;
	gSniffDir = CLib::Getenv("SNIFF_DIR");
	if (gSniffDir == NULL || *gSniffDir == 0) 
	{
	    fprintf(stderr,"%s: SNIFF_DIR environment variable not set\n",
		    gSystem->BaseName(argv[0]));
	    CLib::FastExit(1);
	}
	gEtDir = new char [strlen(gSniffDir) + 1];
	strcpy(gEtDir, gSniffDir);
    }
 
    Env::Init();
    gSystem->InitSession(argc,argv);
 
    ETInit( argc, argv );
}
 
//---- SysEvtHandler -----------------------------------------------------------

NewAbstractMetaImpl(SysEvtHandler, Object, (T(resourceId), TP(owner)));

SysEvtHandler::SysEvtHandler(int id)
{
    owner= 0;
    resourceId= id;
}

SysEvtHandler::~SysEvtHandler()
{
    if (!cleanuplock && owner) {
	owner->ForEach(SeqCollection,RemovePtr)(this);
    }
    SafeDelete(owner);
}

void SysEvtHandler::Notify(SysEventCodes, int)
{
}

bool SysEvtHandler::HasInterest(SysEventCodes)
{
    return owner != 0;
}

void SysEvtHandler::AddOwner(SeqCollection *c)
{
    if (owner == 0)
	owner= new OrdCollection;
    if (owner->FindPtr(c) == 0)
	owner->Add(c);
}

bool SysEvtHandler::IsEqual(Object *op)
{
    return resourceId == ((SysEvtHandler*)op)->resourceId;
}

void SysEvtHandler::Remove()
{
    if (owner) {
	owner->ForEach(SeqCollection,RemovePtr)(this);
	SafeDelete(owner);
    }
    gSystem->AddCleanupObject(this);
}

int SysEvtHandler::Error(int)
{
    return 0;
}

//---- System ------------------------------------------------------------------

NewMetaImpl(System, Object, (TS(errorstr), TS(osid), TB(done), TP(fileInputHandler),
	TP(fileOutputHandler), TP(zombieHandler), TP(signalHandler),
	TP(asyncSignalHandler), TP(timeoutHandler), TP(cleanupList),
	TP(workHandler),TP(backgroundWorkHandler),
	TS(gEtDir), TS(gProgname), TB(gInMain), TB(gDebug), TB(gQuitApp)
));

System::System(char *name)
{
    osid= name;
}

bool System::Init()
{
    fileInputHandler    = new OrdCollection;
    fileOutputHandler   = new OrdCollection;
    zombieHandler       = new OrdCollection;
    signalHandler       = new OrdCollection;
    asyncSignalHandler  = new OrdCollection;
    timeoutHandler      = new OrdCollection;
    workHandler         = new OrdCollection;
    backgroundWorkHandler= new OrdCollection;
    cleanupList= 0;

    return FALSE;
}

static void remove(SeqCollection *&oc)
{
    if (oc) {
	Iter next(oc);
	register SysEvtHandler *seh;
	while (seh= (SysEvtHandler*) next())
	    gSystem->AddCleanupObject(seh);
	SafeDelete(oc);
    }
}

System::~System()
{
    remove(fileInputHandler);
    remove(fileOutputHandler);
    remove(zombieHandler);
    remove(signalHandler);
    remove(asyncSignalHandler);
    remove(timeoutHandler);
    remove(workHandler);
    remove(backgroundWorkHandler);

    if (gSystem == this)
	gSystem= 0;
	
    if (cleanupList) {
	cleanuplock= TRUE;
	cleanupList->FreeAll();
	cleanuplock= FALSE;
	SafeDelete(cleanupList);
    }
}

void System::Control()
{
    done= FALSE;
    while (! done) {
	InnerLoop(100);
	if (cleanupList)
	    Remove();
    }
}

void System::ExitControl()
{
    gQuitApp= done= TRUE;
}

void System::InnerLoop(int)
{
}

Directory *System::MakeDirectory(char *name)
{
    return 0;
}

bool 
System::RemoveDirectory(char *path)
{
    if (RemoveAllFilesInDirectory(path))
		return RemoveEmptyDirectory(path);
    else
		return FALSE;
}

bool
System::RemoveEmptyDirectory(char *path)
{
	return FALSE;
}

bool
System::RemoveAllFilesInDirectory(char *path)
{
	return FALSE;
}

int System::GetPathInfo(char *, u_long *, u_long *, u_long *, u_long *)
{   
    return -1;
}

void System::StatCache(bool)
{
}

bool System::AccessPathName(char*, int)
{
    return FALSE;
}

bool System::ExpandPathName(char*, int)
{
    return FALSE;
}
    
bool System::CreateDirectory(char *, int)
{
    return FALSE;
}
    
bool System::CreateDirectoryWithParents(char *, int)
{
    return FALSE;
}

bool System::ChangeDirectory(char *)
{
    return FALSE;
}
    
char *System::WorkingDirectory()
{
    return 0;
}

char *System::HomeDirectory()
{
    return 0;
}

void System::Wait(unsigned int)
{
}

int System::GetPid()
{
    return -1;
}

bool System::CanRead(int, int)
{
    return FALSE;
}

bool System::CanWrite(int, int)
{
    return FALSE;
}

void System::Remove()
{
    while (cleanupList != 0) {
	SeqCollection *tmp= cleanupList;
	cleanupList= 0;
	tmp->FreeAll();
	delete tmp;
    }
}
bool System::Work()
{
    bool retval=FALSE;
    if (backgroundWorkHandler && backgroundWorkHandler->Size() > 0) {
	Iter next(backgroundWorkHandler);
	SysEvtHandler *eh;
	while(eh=(SysEvtHandler*)next()) {
	    if(eh->HasInterest(eSysEvtWork)){
		eh->Notify(eSysEvtWork, 0);
		retval= TRUE;
	    }
	}
    }
    return retval;
}
 
 
void System::AddHandler(SeqCollection *s, SysEvtHandler *h)
{
    s->Add(h);
    h->AddOwner(s);
}

SysEvtHandler *System::RemoveHandler(SeqCollection *s, SysEvtHandler *h)
{
    return (SysEvtHandler*) s->RemovePtr(h);
}

void System::AddCleanupObject(Object *op)
{
    if (cleanupList == 0)
	cleanupList= new OrdCollection;
    if (cleanupList->FindPtr(op) == 0)
	cleanupList->Add(op);
}

void System::Abort(int)
{
    CLib::Abort();
}

void System::Exit(int code, bool mode)
{    
    if (mode)
	CLib::Exit(code);
    CLib::FastExit(code);
}

PttyConnection *System::MakePttyConnection(char *, char **)
{
    return 0;
}

//---- running programs -------------------------------------
u_int System::Sleep(u_int seconds)
{
    return 0;
}
int System::RunProgSynch(const char *cmdline)
{
    return 0;
}
 
void System::RunProgAsynch(const char *cmdline)
{
}
 
//---- environment manipulation ------------------------------------------------

char *System::Getenv(char *)
{
    return 0;
}

void System::Setenv(char *, char *)
{
}

void System::Unsetenv(char *)
{
}

char *System::GetError()
{
    return NULL;
}

char *System::BaseName(char *)
{
    return NULL;
}

char *System::DirName(char *)
{
    return ".";
}

char *System::RemExecutableExt(char *name)
{
    return name;
}
 
char *System::ConcatPathNames(const char *, ...)
{
    return NULL;
}
 
const char *
System::GetServerPath()
{
    return NULL;
}

bool System::IsAbsolutePath(char *)
{
    return FALSE;
}
 
char *System::Which(char *, char *, int)
{
    return 0;
}
 
bool System::GetSystemInfo (char* name, char* buf, size_t size) {
    return FALSE;
}
 
//---- RPC ---------------------------------------------------------------------

int System::OpenConnection(char*, char*)
{
    return -1;
}

int System::AnnounceTcpService(char*)
{
    return -1;
}

int System::AnnounceUnixService(char*)
{
    return -1;
}

int System::AcceptConnection(int)
{
    return -1;
}

int System::ConnectUnixService(char*)
{
    return -1;
}

int System::ConnectTcpService(char*, char *)
{
    return -1;
}
 
char *System::GetSessionName(int, char **, char *server)
{
    return server;
}
 
void System::InitSession(int, char **)
{
}
 
