#ifndef System_First
#ifdef __GNUG__
//pragma once
#pragma interface
#endif
#define System_First

#include "Object.h"
#include "Main.h"

class Directory;
class PttyConnection; 
class SeqCollection;
 
//---- class SysEvtHandler ---------------------------------------
// abstract class for handling events from system resources

enum SysEventCodes {
    eSysEvtRead,
    eSysEvtWrite,
    eSysEvtZombie,
    eSysEvtSignal,
    eSysEvtAsyncSignal,
    eSysEvtTimeout,
    eSysEvtWork
};

class SysEvtHandler : public Object {
    SeqCollection *owner;
    int resourceId;
protected:
    ~SysEvtHandler();
public:
    MetaDef(SysEvtHandler);
    SysEvtHandler(int resource);
    int GetResourceId()
	{ return resourceId; }
    void SetResourceId(int id)
	{ resourceId= id; }
    virtual void Notify(SysEventCodes code, int val);
    virtual bool HasInterest(SysEventCodes code);
    void Remove();
    bool IsEqual(Object *op);
    void AddOwner(SeqCollection *o);
    virtual int Error(int code);
};

//---- Signals -----------------------------------------------------------------

enum Signals {
    eSigInterrupt,
    eSigQuit,
    eSigIllegalInstruction,
    eSigBus,
    eSigSegmentationViolation,
    eSigPipe,
    eSigAlarm,    
    eSigChild
};

//---- System ------------------------------------------------------------------

typedef void (*CallFunc)(void*, void*, void*, void*);

enum ePathInfoFlags {
    ePiExec     =   1,
    ePiDir      =   2,
    ePiSpecial  =   4
};

class System : public Object {
protected:
    bool done;
    char *errorstr, *osid;

    SeqCollection *fileInputHandler;
    SeqCollection *fileOutputHandler;
    SeqCollection *zombieHandler;    
    SeqCollection *signalHandler;    
    SeqCollection *asyncSignalHandler;    
    SeqCollection *timeoutHandler;
    SeqCollection *workHandler;
    SeqCollection *backgroundWorkHandler;
    SeqCollection *cleanupList;

public:  
    static bool anyremoved;

public:
    MetaDef(System);
    System(char *name);
    virtual ~System();

    virtual bool Init();

    char *GetErrorStr()
	{ return errorstr; }
    virtual char *GetError();
    char *GetSystemId()
	{ return osid; }
    virtual void Control();
    void ExitControl();
    virtual void InnerLoop(int timeout= 200);
    virtual bool CanRead(int fd, int timeout);
    virtual bool CanWrite(int fd, int timeout);

//---- directory-information ---------------------------
    virtual Directory *MakeDirectory(char *name);			// opens directory stream
    virtual bool ChangeDirectory(char *path);				// changes the working directory
    virtual char *WorkingDirectory();						// returns the working directory
    virtual char *HomeDirectory();							// returns the home directory
	
//---- directory-manipulation ---------------------------
    virtual bool CreateDirectory(char *path, int mode);		// creates a new directory
    virtual bool CreateDirectoryWithParents(char *path, int mode); // creates a new directory and the missing parent directories
    virtual bool RemoveDirectory(char *path);			// removes a directory
    virtual bool RemoveEmptyDirectory(char *path);		// removes a directory (if empty)
    virtual bool RemoveAllFilesInDirectory(char *path);		// removes all files in a directory (not rcursive)

//---- path-information --------------------------------
    virtual int GetPathInfo(char *path, u_long *id, u_long *size, u_long *flags, u_long *modtime);
    virtual bool AccessPathName(char *path, int mode= 0);	// checks accessibility
    virtual char *BaseName(char *pathname);					// "/dir1/dir2/filename" -> "filename"
    virtual char *DirName(char *pathname);					// "/dir1/dir2/filename" -> "/dir1/dir2"
    virtual char *RemExecutableExt(char *name);			// eg: "program.exe" -> "program"
    virtual bool IsAbsolutePath(char *);					// TRUE if <dir> starts w/ pathseperator
    virtual bool ExpandPathName(char *patbuf, int buflen);	// expand the metacharacters as in the shell    
    virtual char *ConcatPathNames(const char *, ...);		// concats 0 or more pathname(s)
    virtual const char *GetServerPath();			// returns the system dependent SERVERPATH
    //---- stat caching -----------------------------------
    virtual void StatCache(bool sw);						// switch on/off stat caching

    virtual void Exit(int code, bool mode= TRUE);
    virtual void Wait(unsigned int);
    virtual void Abort(int code= 0);
    virtual int GetPid();

    //---- running programs / sleep ----------------------------
    virtual u_int Sleep(u_int seconds);
    virtual int  RunProgSynch(const char *cmdline);
    virtual void RunProgAsynch(const char *cmdline);
 
    //---- pseudo tty connection --------------------------------
    virtual PttyConnection *MakePttyConnection(char *prog, char **args);

    //---- environment manipulation -----------------------------
    virtual void Setenv(char *name, char *value);
	// set environment variable name to value   
    virtual void Unsetenv(char *name);
	// remove environment variable
    virtual char *Getenv(char *);

    virtual char *Which(char *search, char *file, int mode = 0);
    virtual bool GetSystemInfo (char* name, char* buf, size_t size);

    //---- handling of system events ----------------------------
    virtual void AddHandler(SeqCollection *s, SysEvtHandler *h);
    virtual SysEvtHandler *RemoveHandler(SeqCollection *s, SysEvtHandler *h);

    void AddFileInputHandler(SysEvtHandler *h) 
	{ AddHandler(fileInputHandler, h); }
    void AddFileOutputHandler(SysEvtHandler *h)
	{ AddHandler(fileOutputHandler, h); }
    void AddZombieHandler(SysEvtHandler *h)
	{ AddHandler(zombieHandler, h); }
    void AddSignalHandler(SysEvtHandler *h)    // synch
	{ AddHandler(signalHandler, h); }
    void AddAsyncSignalHandler(SysEvtHandler *h)
	{ AddHandler(asyncSignalHandler, h); }
    void AddTimeoutHandler(SysEvtHandler *h)
	{ AddHandler(timeoutHandler, h); }
    void AddWorkHandler(SysEvtHandler *h)
	{ AddHandler(workHandler, h); }
    void AddBackgroundWorkHandler(SysEvtHandler *h)
	{ AddHandler(backgroundWorkHandler, h); }
    void *RemoveTimeoutHandler(SysEvtHandler *h)
	{ return RemoveHandler(timeoutHandler, h); }
    void AddCleanupObject(Object*);
    void Remove();
    bool Work(); // give the background workhandlers a little bit work...
		 // Return TRUE if anybody did some work
 
    //---- RPC --------------------------------------------------
    virtual int OpenConnection(char *server, char *service);
    virtual int AnnounceTcpService(char *service);
    virtual int AnnounceUnixService(char *server);
    virtual int AcceptConnection(int sock);
    virtual int ConnectUnixService(char *service);
    virtual int ConnectTcpService(char *hostname, char *service);
    virtual char *GetSessionName(int argc, char **argv, char *server);
    virtual void InitSession(int argc, char **argv);};

extern System  *gSystem;
extern char    *gEtDir;
extern int      gDebug;
extern bool     gInMain;
extern char    *gProgname;
extern char    *gProgPath;
extern bool     gQuitApp;
extern void ETInitAll(int argc, char **argv);
extern void ETInit(int argc, char **argv);
extern void InitSystem(char *progname);
extern const char *path_separator;
extern const char *path_parentdir;
#endif
