//-*-Mode: C++;-*-
#ifndef _Token_h_
#define _Token_h_

#ifdef __GNUG__
#  pragma interface
#endif

#include "Rectangle.h"

enum EvtCursorDir {
    eCrsUp,
    eCrsDown,
    eCrsLeft,
    eCrsRight
}; 

enum EventCodes {
    eEvtNone        = -1,

    eEvtAscFirst    = 0,
    eEvtAscLast     = 127,
    eEvtMetaFirst   = 128,   // returned if key depressed with 'Left' or 'Right' key
    eEvtMetaLast    = 255,

    eEvtPseudo      = 0x7f << 8,			// mouse move events
    eEvtLocMove     = eEvtPseudo + 0,
    eEvtLocStill    = eEvtPseudo + 1,
    eEvtEnter       = eEvtPseudo + 2,
    eEvtExit        = eEvtPseudo + 3,
    eEvtLocMoveBut  = eEvtPseudo + 4,
    eEvtIdle        = eEvtPseudo + 5,

    eEvtButtons     = eEvtPseudo + 32,			// button events
    eEvtLeftButton  = eEvtButtons + 0,
    eEvtMiddleButton= eEvtButtons + 1,
    eEvtRightButton = eEvtButtons + 2,
    eEvtHelpButton  = eEvtButtons + 3,                 // pseudo event
    eEvtLastButton  = eEvtButtons + 10,

    eEvtFirstCursorKey= eEvtLastButton + 1,		// cursor keys
    eEvtCursorUp    = eEvtFirstCursorKey + eCrsUp,
    eEvtCursorDown  = eEvtFirstCursorKey + eCrsDown,
    eEvtCursorLeft  = eEvtFirstCursorKey + eCrsLeft,
    eEvtCursorRight = eEvtFirstCursorKey + eCrsRight,
    eEvtLastCursorKey= eEvtCursorRight,

    eEvtPfkFirst    = eEvtLastCursorKey + 1,		// F1 - F64 function keys
    eEvtPfkLast     = eEvtLastCursorKey + 64,

    eEvtOtherKeyFirst  = eEvtPfkLast + 1,		// other function keys
    eEvtKey_Print      = eEvtOtherKeyFirst + 0,
    eEvtKey_PageUp     = eEvtOtherKeyFirst + 1,
    eEvtKey_PageDown   = eEvtOtherKeyFirst + 2,
    eEvtKey_Home       = eEvtOtherKeyFirst + 3,
    eEvtKey_End        = eEvtOtherKeyFirst + 4,
    eEvtKey_Undo       = eEvtOtherKeyFirst + 5,
    eEvtKey_Redo       = eEvtOtherKeyFirst + 6,
    eEvtKey_Find       = eEvtOtherKeyFirst + 7,
    eEvtKey_Cancel     = eEvtOtherKeyFirst + 8,
    eEvtKey_Props      = eEvtOtherKeyFirst + 9,
    eEvtKey_Front      = eEvtOtherKeyFirst + 10,
    eEvtKey_Open       = eEvtOtherKeyFirst + 11,
    eEvtKey_Cut        = eEvtOtherKeyFirst + 12,
    eEvtKey_Copy       = eEvtOtherKeyFirst + 13,
    eEvtKey_Paste      = eEvtOtherKeyFirst + 14,
    eEvtOtherKeyLast   = eEvtKey_Paste,
    
    eEvtHelpKey       =  eEvtOtherKeyLast + 1
};

enum EventFlags {
    eFlgNone        = 0,
    eFlgButDown     = BIT(0),
    eFlgShiftKey    = BIT(1),
    eFlgCntlKey     = BIT(2),
    eFlgMetaKey     = BIT(3) // left or right keys on SUN-2 and SUN-3 keyboards 
}; 
  
class Token {
public:
    short Code;   // enum EventCodes
    short Flags;  // enum EventFlags
    u_long At;
    Point Pos;
public:
    Token();
    Token(EventCodes, EventFlags f= eFlgNone, Point pos= gPoint0);

    bool DoubleClick(Token &t);
    bool IsAscii()
	{ return bool(Code >= eEvtAscFirst && Code <= eEvtAscLast); }
    bool IsKey()
	{ return bool(Code >= eEvtAscFirst && Code <= eEvtMetaLast); }
    bool IsFunctionKey()
	{ return bool(Code >= eEvtPfkFirst && Code <= eEvtPfkLast); }
    bool IsOtherFunctionKey()
	{ return bool(Code >= eEvtOtherKeyFirst && Code <= eEvtOtherKeyLast); }
    bool IsHelpKey()
	{ return bool(Code == eEvtHelpKey); }
    bool IsCursorKey()
	{ return bool(Code >= eEvtFirstCursorKey && Code <= eEvtLastCursorKey); }

    bool IsShift()
	{ return bool(Flags & eFlgShiftKey); }
    bool IsControl()
	{ return bool(Flags & eFlgCntlKey); }
    bool IsMeta()
	{ return bool(Flags & eFlgMetaKey); }

    bool IsButtonClick(EventCodes ec)
	{ return (Code == ec && !(Flags & eFlgButDown)); }
    bool IsLeftClick()
	{ return IsButtonClick(eEvtLeftButton); }
    bool IsMiddleClick()
	{ return IsButtonClick(eEvtMiddleButton); }
    bool IsRightClick()
	{ return IsButtonClick(eEvtRightButton); }
    bool IsMouseButton()
	{ return bool(Code >= eEvtButtons && Code <= eEvtRightButton); }

    char MapToAscii();
    Point CursorPoint();
    EvtCursorDir CursorDir()
	{ return EvtCursorDir((EvtCursorDir)Code - eEvtFirstCursorKey); }
    int FunctionCode()
	{ return Code - eEvtPfkFirst; }
    void Dump();
};

#endif
