#ifdef __GNUG__
#pragma implementation
#endif
#include "ET++.h"
  
#include "GotoDialog_e.h"
#include "FindChange_e.h"
#include "VObjectText.h"
#include "ShellTView.h"
#include "FileDialog.h"
#include "Alert.h"
#include "System.h"
#include "Filler.h"
#include "WindowSystem.h"
#include "Env.h"
#include "MenuBar.h"
#include "ShellTool.h"
#include "ETMath.h"
#include "Data.h"

#include "BrowserView.h"
#include "BrowserItems.h"
#include "BrowserDoc.h"
#include "Preferences.h"
#include "BrowserCmdNo.h"
#include "SharedDocObjects.h"

static char *errFormats[]= {
	/* cc */            "\"%[^\"]\", line%d",
	/* grep -n, cpp*/   "%[^:]:%d:",
	0
};

static Alert *sharedAlert;

ONEXIT(BrowserDoc)
{
    SafeDelete(sharedAlert);
}

//---- ShellManager ------------------------------------------------------------

class ShellManager : public ShellTool {
    BrowserDocument *receiver;
public:
    MetaDef(ShellManager);
    ShellManager(BrowserDocument *d)
	{ receiver= d; }
    void SetReceiver(BrowserDocument *d)
	{ receiver= d; }
    Command *DoMenuCommand(int cmd);
};

NewMetaImpl(ShellManager,ShellTool, (TP(receiver)));

Command *ShellManager::DoMenuCommand(int cmd)
{
    if (cmd == cSHOWERR) {
	static byte buf[500];
	int start, end;
	view->GetSelection(&start, &end);
	Mark *m= view->MarkAtLine(view->CharToLine(start));
	view->GetText()->CopyInStr(buf, sizeof(buf), m->Pos(), m->End());
	receiver->ShowError((char*)buf);
	return gNoChanges;
    }
    return ShellTool::DoMenuCommand(cmd);
}

//---- BrowserDocument ---------------------------------------------------------

NewMetaImpl(BrowserDocument,Document, (TP(browserView), TP(textView), TP(text), 
		    T(nFileLists), T(isCCode), TP(shellManager), TP(scroller)));

    // my external type is cDocTypeAscii, e.g. ordinary ascii files  
BrowserDocument::BrowserDocument() : Document(cDocTypeAscii)
{   
    nFileLists= Math::Range(1, 6, Env::GetValue("filebrowser.BrowserPanes", 3));
    shellManager= 0;
    doRevert= isCCode= FALSE;
    text= MakeText();
    bv= 0;
	filler = new Filler(gPoint1, VObjAlign(eVObjVFixed));
}

BrowserDocument::~BrowserDocument()
{
    SafeDelete(bv);
    if (SharedDocObjects::RefCount(text) > 1)
	ClearCmdProcessor();
    SafeDelete(textView);
    SharedDocObjects::Delete(text, this);
    SafeDelete(filler);
}

CommandProcessor *BrowserDocument::MakeCmdProcessor()
{
    if (SharedDocObjects::RefCount(text) > 1) {
	Iter next(SharedDocObjects::MakeDocumentIter(text));
	BrowserDocument *dp= (BrowserDocument*)next();
	if (dp)
	    return dp->GetCmdP();
    }
    return Document::MakeCmdProcessor();
}

void BrowserDocument::SetWindowTitle(char *t)
{
	Manager::SetWindowTitle(form("Filebrowser: %s", t));
}

VObject *BrowserDocument::DoMakeContent()
{
	VObject *buttons =
		new HBox(1, (VObjAlign)(eVObjHLeft),
   			new ActionButton(cIdLoad, "Load ..."),
   			new ActionButton(cIdSave, "Save"),
   			new ActionButton(cIdSaveAs, "Save As ..."),
   			new ActionButton(cIdRevert, "Revert"),
			0
		);
    browserView= new BrowserView(this, nFileLists);
    
    textView= new FileBrowserTextView(this, Rectangle(Point(2000, cFit)), text);
    SetFirstHandler(textView);
    
    if (Env::GetValue("CodeText.WordWrap", FALSE)) {
	scroller= new Scroller(textView, Point(500, 350), cIdNone, eScrollRight);
	scroller->GetClipper()->SetFlag(eVObjLayoutCntl);
    } else
	scroller= new Scroller(textView, Point(500, 350));
    scroller->SetBgInk(gInkWhite);
    tvClipper= scroller->GetClipper();
	
    return
	browser= new Expander(cIdNone, eVert, gPoint2,
		buttons,
	    browserView,
	    scroller,
	    0
	);
}

bool BrowserDocument::DoWrite(OStream &s, Data*)
{
    text->PrintOnAsPureText(s);
    return TRUE;
}

void BrowserDocument::Revert()
{
    doRevert= TRUE;
    Document::Revert();
    doRevert= FALSE;
}

bool BrowserDocument::DoReadData(Data *data)
{   
    IStream s(data->GetStreamBuf(FALSE), 0, TRUE);

    Document *dp= IsAlreadyOpen(data);
    if (doRevert) {
	text->ReadFromAsPureText(s, data->SizeHint());
	ShowReverted(text, data);
	return TRUE;
    } 
    SharedDocObjects::DeleteDelayed(text, this);
    if (dp == 0) {        //---- new unshared copy
	text= MakeText();
	text->ReadFromAsPureText(s, data->SizeHint());
    } else {          //---- shared copy
	TextView *tv= Guard(dp, BrowserDocument)->GetTextView();
	text= Guard(tv->GetText(), VObjectText);
	SharedDocObjects::AddRef(text, this);
	SetSharedWindowTitle();
    }
    Symbol dataType = data->Type();
    bool isCCode = dataType == cDocTypeCPlusPlus||dataType == cDocTypeCHeader;
    InstallText(text, isCCode);
    return TRUE;
}

void BrowserDocument::ShowReverted(Text *t, Data *data)
{
    Iter next(SharedDocObjects::MakeDocumentIter(t));
    BrowserDocument *dp; 
    Symbol dataType = data->Type();
    bool isCCode = dataType == cDocTypeCPlusPlus||dataType == cDocTypeCHeader;
    while (dp= (BrowserDocument*)next()) 
	dp->InstallText(t, isCCode);
}

void BrowserDocument::InstallText(Text *t, bool isC)
{
    textView->SetText(t);
    
    if (isCCode= isC) {
	textView->FormatCode();
	textView->CreateFunctionMenu(functionMenu);
	pdm->Enable(functionMenu->GetCollection()->Size() > 0);
    } else {
	textView->SetDefaultStyle();
	textView->ResetFunctionMenu();
	pdm->Disable();
    }
    textView->SetSelection(0, 0);
    textView->RevealSelection();
}

bool BrowserDocument::DoFileIsAlreadyOpen(Document *shown, char *name)
{
    if (shown->UniqueId() == UniqueId())
	return FALSE;
    
    if (sharedAlert == 0)
	sharedAlert= new Alert(eAlertCaution, (byte*) "Caution", gCaution,
					"Open",    cIdYes,
					"Show",    cIdNo,
					"Cancel", cIdCancel, 0);
    switch(sharedAlert->Show("File @B%s@P is already open?", gSystem->BaseName(name))) {
    case cIdYes:
	return TRUE;
    case cIdNo:
	shown->Show();
	return FALSE;
    case cIdCancel:
	return FALSE;
    }
    return FALSE;
}

bool BrowserDocument::SavedChanges()
{
    if (SharedDocObjects::RefCount(text) > 1)
	return TRUE;
    return Document::SavedChanges();
}

void BrowserDocument::SetSharedWindowTitle()
{
    SetWindowTitle(form("%s (%d)", GetBaseName(),
				    SharedDocObjects::RefId(text, this)+1));  
}

VObjectText *BrowserDocument::MakeText()
{
    VObjectText *tp;
    Font *fd= new_Font(gFixedFont->Fid(), Env::GetValue("filebrowser.FontSize", 12), eFacePlain);
	tp= new VObjectText(256, fd);
    tp->SetDefTab(fd->Width(' ') * Env::GetValue("CodeText.TabPos", 8));
    SharedDocObjects::AddRef(tp, this);
    return tp;
}

Document *BrowserDocument::IsAlreadyOpen(Data *data)
{
    Iter next(SharedDocObjects::MakeIterator());
    Document *dp;
    Object *op;
    while (op= next()) {
	Iter next2(SharedDocObjects::MakeDocumentIter(op));
	while(dp= (Document*)next2()) 
	    if (dp->UniqueId() == data->UniqueId() && dp != this)
		return dp;
    }
    return 0;
}

void BrowserDocument::InputKbd(Token &t)
{
    bool selectKey= t.IsAscii() && (t.Code != '\r') && (t.Code != '\n');
    if (selectKey && ShowsFileLists() && browserView->GrabKeyToken(t))
	return;
    Manager::InputKbd(t);
}

void BrowserDocument::Control(int sender, int what, void *data)
{
	switch (sender)
		{
   	case cIdLoad:
		DoMenuCommand(cLOAD);
   		return;
   	case cIdSave:
		DoMenuCommand(cSAVE);
   		return;
   	case cIdSaveAs:
		DoMenuCommand(cSAVEAS);
   		return;
   	case cIdRevert:
		DoMenuCommand(cREVERT);
   		return;
		}

    switch (what) {
    case cPartExtentChanged:
	if (data == (void*) tvClipper) {
	    int w= tvClipper->Width();
	    if (textView->Width() != w) 
		textView->SetWidth(w);
	}
	break;
	
    case cFbLoadFile:
	Load(((FileData*)data)->Name());
	break;
	
    case cFbShowShell:
	ShowShell((char*)data);
	break;
	
    case cFbShellCmd:
	ShellExecute((char*)data);
	break;

    default:
	Document::Control(sender, what, data);
    }
}

void BrowserDocument::ShowError(char *errMsg)
{
    char file[400];
    int line;
    
    for (int i= 0; errFormats[i]; i++) {
	if (sscanf(errMsg, errFormats[i], file, &line) == 2) {
	    FileData *fd= new FileData(file);
	    if (UniqueId() != fd->UniqueId()) {     // file not already loaded ?
		if (gSystem->AccessPathName(file, 4) != 0) {
		    ShowAlert(eAlertNote, "Can't find @B%s@P in\n%s", file, 
						    gSystem->WorkingDirectory());
		    delete fd;
		    return;
		}
		LoadData(fd, TRUE);
	    } else
		delete fd;
	    Mark *lm= textView->MarkAtLine(line-1);
	    textView->SetSelection(lm->Pos(), lm->End());
	    textView->RevealSelection();
	    GraphicDelay(500);
	    textView->SetSelection(lm->Pos(), lm->Pos());
	    return;
	}
    }
    ShowAlert(eAlertNote, "Error message no understood!");
}

void BrowserDocument::ApplyPreferences()
{
    Font *nfp, *ofp= text->GetFont();
    nfp= new_Font(ofp->Fid(), Env::GetValue("CodeText.FontSize", 12), eFacePlain);
    int from, to;
    textView->GetSelection(&from, &to);
    
    text->SetDefTab(nfp->Width(' ') * Env::GetValue("CodeText.TabPos", 8));
    textView->SetAutoIndent(Env::GetValue("CodeText.AutoIndent", TRUE));
    if (nfp != ofp) {
	textView->SetFont(nfp);
	textView->FormatCode();
	textView->SetSelection(from, to);
    }
}

void BrowserDocument::ShellExecute(char* data)
{
    if (gSystem->DirName(GetData()->FullName()))
	gSystem->ChangeDirectory(gSystem->DirName(GetData()->FullName()));
    ShowShell(gSystem->WorkingDirectory());
    shellManager->ShellExecute(data);
}

void BrowserDocument::ExecuteMake()
{
    if (gSystem->DirName(GetData()->FullName()))
	gSystem->ChangeDirectory(gSystem->DirName(GetData()->FullName()));
    ShowShell(gSystem->WorkingDirectory());
    shellManager->ShellExecute(form("%s\n", Env::GetValue("filebrowser.MAKE", "make")));
}

void BrowserDocument::ShowShell(char *data)
{
    if (shellManager == 0) 
		shellManager= new ShellManager(this);
	shellManager->SetName(form("Shell: %s", gSystem->BaseName(data)));
    shellManager->Open();
    shellManager->ShellExecute(form("cd %s\n", data));
    shellManager->SetReceiver(this);
    gSystem->ChangeDirectory(data);
}

//---- Menus --------------------------------------------------------------

MenuBar *BrowserDocument::DoMakeMenuBar()
{
    MenuBar *mb= Document::DoMakeMenuBar();
    
    Menu *m= new Menu("Utilities", FALSE);
    m->AppendItems(
		"Go to Line ...",   cGOTOLINE,
		"Make@M",             cMAKE,
		"Reformat",         cREFORMAT,
		"Preferences ...",  cPREFERENCES,
		"Hide Browser",     cTOGGLEVIEWS,
		0);
    mb->AddMenu(m);
	
    functionMenu= new Menu("Marks", FALSE);
    mb->AddMenu(functionMenu);
    pdm= (PullDownButton*) mb->At(0)->At(3);
    pdm->Disable();
    
    FindChange::InstallChange(mb, textView);
    return mb;
}

void BrowserDocument::DoSetupMenu(Menu *menu)
{
    FindChange::DoSetupMenu(menu);
    if (menu == functionMenu)
	textView->SetupFunctionMenu(menu);
    else {
	if (isCCode && text->IsKindOf(StyledText))
	    menu->EnableItem(cREFORMAT);
	menu->EnableItems(cGOTOLINE, cMAKE, cPREFERENCES, cTOGGLEVIEWS, cFIND, 0);
	menu->ToggleItem(cTOGGLEVIEWS, ShowsFileLists(), "Hide Browser",
							 "Show Browser");
	Document::DoSetupMenu(menu);
    }
}

bool BrowserDocument::ShowsFileLists()
{
    return bv == 0;
}

Command *BrowserDocument::DoMenuCommand(int cmd)
{
    switch (cmd) {

    case cGOTOLINE:
	GotoLine(textView);
	break;
	
    case cFIND:
	FindChange::ShowChangeDialog(textView);
	return gNoChanges;

    case cFINDAGAIN:
	FindChange::FindAgain(textView);
	return gNoChanges;

    case cMAKE:
	ExecuteMake();
	break;
	
    case cREFORMAT:
	textView->FormatCode();
	textView->ForceRedraw();
	break;

    case cPREFERENCES:
	if (PrefDialog::ShowDialog(textView) == cIdOk)
	    ApplyPreferences();
	break;
	
    case cTOGGLEVIEWS:
	if (bv) {
	    browser->SetAt(1, browserView);
	    bv= 0;
	} else {
	    browser->SetAt(1, filler);
	    bv= browserView;
	}
	break;
	
    default:
	if (cmd >= cMARKS && cmd < cMARKS+1000)
	    textView->GotoFunction(cmd-cMARKS);
	else
	    return Document::DoMenuCommand(cmd);
	break;
    }
    return gNoChanges;
}

