#ifdef __GNUG__
#pragma implementation
#endif
#include "BrowserView.h"

#include "ET++.h"
#include "OrdColl.h"
#include "Buttons.h"
#include "Directory.h"
#include "Bitmap.h"
#include "MenuBar.h"
#include "Env.h"
#include "ByteArray.h"
#include "System.h"
#include "SortedOList.h"
#include "FileDialog.h"
#include "ETMath.h"

#include "BrowserCmdNo.h"
#include "ChangeDirDiag.h"
#include "BrowserItems.h"

//---- BrowserView -------------------------------------------------------------

NewMetaImpl(BrowserView,VBox, (TP(path), TP(directories), TP(fileLists), T(left),
						    TP(changeDir), T(nShown)));

BrowserView::BrowserView(EvtHandler*, int n)
	    : VBox(gPoint8, (VObjAlign)(eVObjHCenter|eVObjHExpand|eVObjVExpand))
{ 
    nShown= n; 
    left= 0; 
    path= new OrdCollection;
    directories= new OrdCollection;
    changeDir= 0;

    fileLists= new Expander(cIdNone, eHor, gPoint4, (SeqCollection*)0);
    
    for (int i= 0; i < nShown; i++) 
	fileLists->Add(new FileList(cIdFirstList+i));

    shiftLeft = new ActionButton(cIdShiftLeft, "<<");
    shiftRight= new ActionButton(cIdShiftRight, ">>");
    
    // initialize left most file list
    ShowDirectory(-1, "."); 

    Add(fileLists);
    Add(new HBox(gPoint10, (VObjAlign)(eVObjHEqual|eVObjVBase),
	    shiftLeft, 
	    new ActionButton (cIdChDir, "change directory..."), 
	    shiftRight,
	    0
	));
}

BrowserView::~BrowserView()
{
    if (path) {
	path->FreeAll();
	SafeDelete(path);
    }
    if (directories) {
	directories->FreeAll();
	SafeDelete(directories);
    }
    SafeDelete(changeDir);
}

void BrowserView::Control(int id, int part, void *)
{
    switch (id) {
    case cIdShiftLeft:
	ShiftFileLists(left-1);
	break;

    case cIdShiftRight:
	ShiftFileLists(left+1);
	break;
	
    case cIdChDir:              
	if (changeDir == 0)
	    changeDir= new ChangeDirDiag(this);
	changeDir->ShowOnWindow(GetWindow());
	break;
    
    default: 
	if (id >= cIdFirstList && id < cIdFirstList+nShown) {
	    int at= id-cIdFirstList;
	    if (left + at >= path->Size())
		return;
	    switch (part) {
	    case cFbShowShell:
		Shell(at, PathNameAt(left+at));
		break;
	    case cPartCollSelect:
		LoadFile(at, FileListAt(at));
		break;
	    }
	}
    } 
}

void BrowserView::Shell(int, char *path, char *cmd)
{
    VBox::Control(cIdFileBrowser, cFbShowShell, path);
    if (cmd)
	VBox::Control(cIdFileBrowser, cFbShellCmd, form("%s\n", cmd));
}

void BrowserView::LoadFile(int at, FileList *fl) 
{
    FileItem *fi= fl->GetSelectedItem();
    FileData *data= (FileData*) fi->GetData();
    
    if (data->IsDirectory()) 
	ShowDirectory(at, data->GetBaseName());
    else {
	if (data->IsAscii())
	    // notify next event handler
	    VBox::Control(cIdFileBrowser, cFbLoadFile, data);
    } 
}

//---- directory handling

SeqCollection *BrowserView::ReadDirectory()
{
    Dir next(".");
    char *entry;
    SeqCollection *col= new SortedObjList;
    bool fasticons= Env::GetValue("filebrowser.FastIcons", TRUE);
    
    while (entry= next())                   // do not show "." and ".."
	if (strcmp(entry, ".") != 0 && strcmp(entry, "..")) 
	    col->Add(new FileItem(entry, fasticons));
    return col;
}

void BrowserView::ShowDirectory(int pane, char *name)
{
    if (pane != -1)
	gSystem->ChangeDirectory(gSystem->ConcatPathNames(PathNameAt(left+pane), name, 0));

    SeqCollection *col= ReadDirectory();
    
    // free path to the right of selected pane
    while (path->Size() > left+pane+1) {
	RemoveAndFreeLast(path);
	RemoveAndFreeLast(directories);
    }
    // empty file lists
    for (int i= pane+1; i < nShown; i++)
	FileListAt(i)->Empty();
    
    // shift file lists if at left most one
    if (pane == nShown-1) {
	ShiftFileLists(left+1);
	pane--;
    }
    
    // update path, path names
    path->Add(col);
    directories->Add(new ByteArray(gSystem->WorkingDirectory()));
    FileListAt(pane+1)->SetContents(DirectoryNameAt(path->Size()-1),col);
    DoSetup();
}

void BrowserView::ShowParentDirectory()
{
    gSystem->ChangeDirectory(gSystem->ConcatPathNames(PathNameAt(left), "..", 0));
    path->AddFirst(ReadDirectory());
    directories->AddFirst(new ByteArray(gSystem->WorkingDirectory()));
}

void BrowserView::ShiftFileLists(int to)
{
    if (to < 0 && left == 0)
	ShowParentDirectory();
    left= Math::Range(0, path->Size()-1, to);

    for (int i= 0; i < nShown; i++) {
	FileList *fl= FileListAt(i);
	if (left + i < path->Size())
	    fl->SetContents(DirectoryNameAt(left+i), PathAt(left+i));
	else
	    fl->Empty();
    }
    DoSetup();
}

void BrowserView::DoSetup() 
{
    bool root= path->Size() && strcmp(DirectoryNameAt(left), path_separator) == 0;
    shiftLeft->Enable(!root, TRUE);
    shiftRight->Enable(left != path->Size()-1, TRUE);
}

bool BrowserView::GrabKeyToken(Token &t)
{
    for (int i= 0; i < nShown; i++) {
	FileList *fl= FileListAt(i);
	if (fl->GrabKeyToken(t))
	    return TRUE;
    }
    return FALSE;
}

//---- accessing

char *BrowserView::PathNameAt(int i)
{ 
    return (char*) ((ByteArray*)directories->At(i))->Str(); 
}

char *BrowserView::DirectoryNameAt(int i) 
{
    char *p= PathNameAt(i);
    char *name= gSystem->BaseName(p);
    if (name == 0 || strcmp(p, path_separator) == 0)
	return p;
    return name;    
}

void BrowserView::RemoveAndFreeLast(SeqCollection *col)
{
    Object *op= col->RemoveLast();
    if (op) {
	op->FreeAll();
	SafeDelete(op);
    }
}

//---- FileBrowserTextView -----------------------------------------------------

NewMetaImpl0(FileBrowserTextView,CodeTextView);

FileBrowserTextView::FileBrowserTextView(EvtHandler *eh, Rectangle r, VObjectText *t)
					       : CodeTextView(eh, r, t, TRUE)
{
}
						   
PrettyPrinter *FileBrowserTextView::MakePrettyPrinter(Text *t, CharStyle *cs,
				    CharStyle *fs, CharStyle *cds, CharStyle *ps)
{
    return new FileBrowserPPrinter(this, t, cs, fs, cds, ps);
}

void FileBrowserTextView::ResetFunctionMenu()
{
    GetMarkList()->FreeAll();
}
 
void FileBrowserTextView::SetupFunctionMenu(Menu *functionMenu)
{
    int from, to, at= -1;
    SeqCollection *cp= functionMenu->GetCollection();
    GetSelection(&from, &to);
    Iter next(GetMarkIter());
    
    FunctionMark *fm;
    cp->ForEach(MenuButtonItem,SetValue)(0,FALSE);
    for (int i= 0; fm= (FunctionMark*)next(); i++) {
	if (fm->Pos() > from && at == -1)
	    at= i;
	if (fm->State() & (int)eStateDeleted)
	    functionMenu->DisableItem(cMARKS+i);
    }
    if (at >= 0 && cp->Size() > 0) {
	i= Math::Range(0, cp->Size(), at-1);
	MenuButtonItem *tp= Guard(cp->At(i), MenuButtonItem);
	tp->SetValue(1, FALSE);
    }
}    

void FileBrowserTextView::CreateFunctionMenu(Menu *functionMenu)
{
    OrdCollection *items= new OrdCollection(60);
    Iter next(GetMarkIter());
    FunctionMark *fm;
    
    for (int i= 0; fm= (FunctionMark*)next(); i++)
	items->Add(new MenuButtonItem(cMARKS+i, fm->GetName()));
    functionMenu->SetCollection(items);
}

void FileBrowserTextView::GotoFunction(int i)
{
    MarkList *mlp= GetMarkList();
    if (i != Math::Range(0, mlp->Size(), i))
	return;
    Mark *mp= mlp->At(i);
    if (mp) {
	SetSelection(mp->Pos(), mp->End());
	Rectangle r= SelectionRect();
	r.extent.y+= 200;
	RevealRect(r, r.extent);
    }
}

//---- FileBrowserPPrinter -----------------------------------------------------

FileBrowserPPrinter::FileBrowserPPrinter(FileBrowserTextView *tv, Text *t,
		    CharStyle *cs, CharStyle *fs, CharStyle *cds, CharStyle *ps)
				    : PrettyPrinter(t, cs, fs, cds, ps)
{
    tvp= tv;
    MarkList *mlp= tv->GetMarkList();
    mlp->FreeAll();
}

void FileBrowserPPrinter::Function(int line, int start, int end, char *name, char *classname)
{
    PrettyPrinter::Function(line, start, end, name, classname);
    char *cp= name;
    if (classname)
	cp= form("%s::%s", classname, name);
    tvp->AddMark(new FunctionMark(cp, classname != 0, start, end-start));
}

void FileBrowserPPrinter::ClassDecl(int l, int start, int end, char *name)
{
    PrettyPrinter::ClassDecl(l, start, end, name);
    tvp->AddMark(new FunctionMark(form("class %s", name), FALSE, start, end-start));
}
