#ifdef __GNUG__
#pragma implementation
#endif
#include "ChangeDirDiag.h"

#include "Class.h"
#include "Window.h"
#include "Document.h"
#include "CollView.h"
#include "Scroller.h"
#include "Alert.h"
#include "System.h"
#include "DialogItems.h"
#include "OrdColl.h"
#include "Expander.h"
#include "BorderItems.h"
#include "Fields.h"
#include "Buttons.h"

#include "BrowserView.h"
#include "BrowserDoc.h"
#include "BrowserCmdNo.h"
 
const int cPathBuf2      =   200,
	  cItemMinWidth2 =   300,
	  cMaxDirBuf     =     8;

//---- ChangeDirDiag -----------------------------------------------------------

NewMetaImpl(ChangeDirDiag,Dialog, (TP(bvp), TP(ei1), TP(collview), TP(mylist)));

ChangeDirDiag::ChangeDirDiag(BrowserView *vp) : Dialog("Change Directory")
{
    bvp= vp;
    mylist= new OrdCollection;
    collview= new CollectionView(this, mylist, eCVDontStuckToBorder);
    collview->SetMinExtent(Point(cItemMinWidth2, 0));
    collview->SetId(cIdHistList);
}

ChangeDirDiag::~ChangeDirDiag()
{
    SafeDelete(collview);
}

VObject *ChangeDirDiag::DoMakeContent()
{
    return
	new Matte(
	    new VExpander(20,
		new Scroller(collview, Point(cItemMinWidth2, 16*8), cIdHistList),
		new VExpander(2,
		    new TextItem("New Directory"),
		    ei1= new TextField(cIdNone, 20),
		    0
		),
		new HBox(10, (VObjAlign)(eVObjVBase|eVObjHEqual|eVObjHExpand),
		    new ActionButton(cIdOk, "Change Directory", TRUE),
		    new ActionButton(cIdCancel, "Cancel"),
		    0
		),
		0
	    )
	);
}

void ChangeDirDiag::DoSetDefaults()
{
    BrowserDocument *bd= (BrowserDocument*) FindNextHandlerOfClass(Meta(BrowserDocument));
    
    if (bd) {
	byte sel[30];
	bd->GetTextView()->SelectionAsString(sel, sizeof sel);
	if (sel[0])
	    ei1->SetString((char*)sel);
    }
}

void ChangeDirDiag::DoSetup()
{
    EnableItem(cIdOk, ei1->GetTextSize() > 0);
}

void ChangeDirDiag::Control(int id, int p, void *v)
{
    VObject *gop;
    char *fname;
    
    switch (id) {

    case cIdOk:
	if (ChangeDir())
	    break;
	return;
    
    case cIdHistList:
	switch (p) {
	case cPartCollSelect:
	    gop= (VObject*) collview->GetCollection()->At( (int) v );
	    fname= gop->AsString();
	    if (fname && strlen(fname) > 0) {
		ei1->SetString(fname);
		SafeDelete(fname);
	    }
	    return;
	case cPartCollDoubleSelect:
	    if (ChangeDir()) 
		Dialog::Control(cIdOk, cPartAction, v);
	    else
		return;
	}
    }
    Dialog::Control(id, p, v);
}

bool ChangeDirDiag::ChangeDir()
{
    char pathname[cPathBuf2];
    
    strcpy(pathname, ei1->GetString());
    
    if (gSystem->ExpandPathName(pathname, cPathBuf2-1)) {
	ShowAlert(eAlertNote, "%s (%s)", gSystem->GetErrorStr(), pathname);
	return FALSE;
    }
    if (!gSystem->ChangeDirectory(pathname)) {
	ShowAlert(eAlertNote, "Cannot change directory to @B%s@P", pathname);
	return FALSE;
    }
    UpdateHistory(pathname);
    bvp->ShowDirectory(-1, pathname);
    return TRUE;
}

void ChangeDirDiag::UpdateHistory(char *pathname)
{
    TextItem *ti, *told;

    ti= new TextItem(1,pathname, gSysFont, Point(4,0));
    if (told= (TextItem*)mylist->Find(ti))
	mylist->Remove(told);
    mylist->AddAt(ti, 0);
    if (mylist->Size() > cMaxDirBuf)
	mylist->RemoveAt(cMaxDirBuf);
    collview->SetCollection(mylist,FALSE);
    SafeDelete(told);
}
