#ifdef __GNUG__
#pragma implementation
#endif
#include "SharedDocObjects.h"

#include "Class.h"
#include "Dictionary.h"
#include "OrdColl.h"
#include "Document.h"
#include "Error.h"
#include "System.h"

//---- SharedDocObjects --------------------------------------------------------

IdDictionary *SharedDocObjects::map;

ONEXIT(SharedDocObjects)
{
    SharedDocObjects::CleanUp();
}

void SharedDocObjects::CleanUp()
{
    SafeDelete(map);
}

void SharedDocObjects::AddRef(Object *op, Document *doc)
{
    if (map == 0)
	map= new IdDictionary;

    Collection *refDocs= Guard(map->AtKey(op), Collection);
    if (refDocs == 0) {
	refDocs= new OrdCollection();
	map->PutAtKey(refDocs, op);
    }
    refDocs->Add(doc);
}

void SharedDocObjects::RemoveRef(Object *op, Document *doc)
{
    if (map == 0)
	map= new IdDictionary();
    Collection *refDocs= Guard(map->AtKey(op), Collection);
    if (refDocs == 0) {
	Error("SharedDocObjects::RemoveRef", "object not found");
	return;
    }
    if (refDocs->RemovePtr(doc) == 0) {
	Error("SharedDocObjects::RemoveRef", "Document not found");
	return;
    }
    if (refDocs->IsEmpty()) {
	map->RemoveKey(op);
	SafeDelete(refDocs);
    }
}

int SharedDocObjects::RefCount(Object *op)
{
    if (map == 0)
	map= new IdDictionary();
    Collection *refDocs= Guard(map->AtKey(op), Collection);
    if (refDocs) 
	return refDocs->Size();
    return 0;
}  

void SharedDocObjects::Delete(Object *op, Document *doc)
{
    RemoveRef(op, doc);
    if (RefCount(op) == 0)
	SafeDelete(op);
}

void SharedDocObjects::DeleteDelayed(Object *op, Document *doc)
{
    RemoveRef(op, doc);
    if (RefCount(op) == 0)
	gSystem->AddCleanupObject(op);
}

int SharedDocObjects::RefId(Object *op, Document *dp)
{
    SeqCollection *refDocs= Guard(map->AtKey(op), SeqCollection);
    if (refDocs)
	return refDocs->IndexOfPtr(dp);
    Error("SharedDocObjects::RefId", "object not found");
    return 0;
}  
    
Iterator *SharedDocObjects::MakeDocumentIter(Object *op)
{
    SeqCollection *refDocs= Guard(map->AtKey(op), SeqCollection);
    if (refDocs)
	return refDocs->MakeIterator();
    Error("SharedDocObjects::MakeDocumentIter", "object not found");
    return 0;
}

Iterator *SharedDocObjects::MakeIterator()
{
    return new DictionaryIterator(map);
}

